/*
    Copyright 2008-2009  Rivo Laks <rivolaks@hot.ee>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "gameview.h"

#include "gameworld.h"
#include "ship.h"
#include "asteroid.h"
#include "model.h"
#include "bullet.h"
#include "datastore.h"
#include "shadermanager.h"
#include "renderinfo.h"

#include <kgllib/camera.h>
#include <kgllib/widgetproxy.h>
#include <kgllib/mesh.h>
#include <kgllib/fpscounter.h>
#include <kgllib/texture.h>
#include <kgllib/shapes.h>
#include <kgllib/program.h>

#include <KDebug>
#include <KStandardDirs>
#include <QTimer>

using namespace KGLLib;


GameView::GameView(KGLLib::GLWidget* parent, GameWorld* world) : GLView(parent)
{
    mWorld = world;

    mDetailLevel = 1.0f;
}

GameView::~GameView()
{
}

void GameView::increaseDetailLevel()
{
    mDetailLevel *= 1.2f;
}

void GameView::decreaseDetailLevel()
{
    mDetailLevel /= 1.2f;
}

void GameView::init()
{
    kDebug();

    mSpaceTexture = new KGLLib::Texture(mWorld->dataPath() + "textures/space.png");

    mSpaceMesh = new Mesh();
    Shapes::createCube(mSpaceMesh);
    mSpaceMesh->setTexture(mSpaceTexture);

    mEarthTex = new Texture(mWorld->dataPath() + "textures/earth.png");
    mEarthTex->setWrapMode(GL_CLAMP);

    mEarthMesh = new Mesh();
    Shapes::createSphere(mEarthMesh, 6);
    mEarthMesh->setTexture(mEarthTex);

    mShaderManager = mWorld->data()->shaderManager();
}

void GameView::setupWorldView()
{
    camera()->applyViewport();

    // Setup matrices
    camera()->applyPerspective();

    // Setup camera
    Ship* ship = mWorld->ship();
    camera()->setPosition(ship->position() + ship->up() * 2 + ship->direction() * -5);
    camera()->setDirection(ship->direction());
    camera()->setUp(ship->up());
    camera()->applyView();
}

void GameView::render()
{
    float elapsed = qMin(0.1f, fpsCounter()->timeElapsed());
    mWorld->advance(elapsed);

    renderInfo->startNextFrame();

    // Set up matrices
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glMatrixMode(GL_MODELVIEW);
    setupWorldView();

    // Render space (background). For this, we don't want depth testing, lighting, etc
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    renderSpace();

    // Render objects
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);

    // Ambient + depth pass
    mShaderManager->setRenderMode("ambient");
    mShaderManager->setLight("directional");
    mShaderManager->setLightColor(Vector3f(0.15, 0.15, 0.15));
    renderObjects(mWorld->objects());

    // Light passes
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE, GL_ONE);
    glDepthMask(GL_FALSE);
    glDepthFunc(GL_EQUAL);
    mShaderManager->setRenderMode("light");

    mShaderManager->setLight("directional");
    mShaderManager->setLightColor(Vector3f(0.8, 0.8, 0.8));
    Vector3f lightDir = (camera()->modelviewMatrix() * Vector4f(2.0, 1.0, 0.5, 0.0)).start<3>().normalized();
    mShaderManager->setLightDirection(lightDir);
    renderObjects(mWorld->objects());

    if (mWorld->ship()->lightOn) {
        mShaderManager->setLight("spot");
        mShaderManager->setLightColor(Vector3f(0.8, 0.8, 1.2));
        renderObjects(mWorld->objects());
    }

    /*
    foreach (Bullet* obj, *mWorld->bullets()) {
        mShaderManager->setLight("point");
        mShaderManager->setLightColor(Vector3f(2.0, 1.6, 0.8));
        mShaderManager->setLightRadius(200);
        Vector3f lightPos = (camera()->modelviewMatrix() * obj->position());
        //     kDebug() << "Light pos:" << lightPos.x() << lightPos.y() << lightPos.z() << lightPos.w();
        mShaderManager->setLightPosition(lightPos.start<3>());
        renderObjects(mWorld->objects());
    }
    */

    // Restore states
    glDepthMask(GL_TRUE);
    glDepthFunc(GL_LESS);
    glDisable(GL_CULL_FACE);

    glLineWidth(2.0);
    glPointSize(3.0);
    Vector4f bulletcolor(0.6, 0.5, 0.2, 1.0);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    const QList<Bullet*>* bullets = mWorld->bullets();

    glBegin(GL_LINES);
    for (int i = 0; i < bullets->count(); i++) {
        Vector3f prevpos = bullets->at(i)->position() - bullets->at(i)->velocity() * 0.4f;
        bulletcolor.w() = 1.0;
        glColor4fv(bulletcolor.data());
        glVertex3fv(bullets->at(i)->position().data());
        bulletcolor.w() = 0.6;
        glColor4fv(bulletcolor.data());
        glVertex3fv(prevpos.data());
    }
    glEnd();

    bulletcolor.w() = 1.0;
    glColor4fv(bulletcolor.data());
    glBegin(GL_POINTS);
    for (int i = 0; i < bullets->count(); i++) {
        glVertex3fv(bullets->at(i)->position().data());
    }
    glEnd();
    glDisable(GL_BLEND);

    // Track
    QList<Vector3f> track(mWorld->trackPoints());
    if (!track.isEmpty()) {
        track.prepend(mWorld->ship()->position());
        glColor4f(1, 1, 0, 1);
        glBegin(GL_LINE_STRIP);
        for (int i = 0; i < track.count(); i++) {
            glVertex3fv(track[i].data());
        }
        glEnd();
    }

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    checkGLError("after renderScene()");

    QTimer::singleShot(0, this, SLOT(update()));
}

void GameView::renderObjects(const QList<GameObject*>* objects)
{
    renderInfo->addLight();
    Vector3f cameraPos = mWorld->ship()->position();

    GeometryBuffer* buffer = 0;
    QString material;
    const int maxTextures = 4;
    Texture* texture[maxTextures];
    for (int i = 0; i < maxTextures; i++) {
        texture[i] = 0;
    }
    int activeTextures = 0;

    foreach (const GameObject* obj, *objects) {
        Model* model = obj->model();
        if (!model) {
            continue;
        }

        // Check buffer
        if (model->buffer() != buffer) {
            if (buffer) {
                buffer->unbind();
            }
            model->buffer()->bind();
            buffer = model->buffer();
            renderInfo->addBufferChange();
        }
        // Check material
        if (model->material() != material) {
            material = model->material();
            mShaderManager->setMaterial(material);
            mShaderManager->activateShader();
            renderInfo->addMaterialChange();
        }
        // Check textures
        int newActiveTextures = qMin(maxTextures, model->textureCount());
        for (int i = 0; i < qMax(activeTextures, newActiveTextures); i++) {
            Texture* newTex = model->texture(i);
            if (newTex == texture[i]) {
                // No need to change anything
                continue;
            }
            glActiveTexture(GL_TEXTURE0 + i);
            if (newTex && texture[i]) {
                // Just change the active texture
                newTex->bind();
            } else if (!texture[i]) {
                // Enable (and bind)
                newTex->enable();
            } else {
                // Disable old texture
                texture[i]->disable();
            }

            texture[i] = newTex;
            renderInfo->addTextureChange();
        }
        activeTextures = newActiveTextures;

        // Use pseudo-instancing: instead of modifying the modelview matrix for each object, send
        //  the info about object's position/rotation to the vertex shader in OpenGL persistent
        //  attributes (in this case, unused texture coordinates). This simplifies matrix
        //  calculations and offloads them to the GPU, gaining an impressive performance increase
        //  of about 35% (and more with low-poly object models).
        // Using ODE's code here isn't pretty of course, but it's a bit faster than adding an extra
        //  method into GameObject.
        const dReal* pos = dBodyGetPosition(obj->body);
        const dReal* rot = dBodyGetRotation(obj->body);
        float s = obj->modelScale();
        glMultiTexCoord4f(GL_TEXTURE1, s*rot[0], s*rot[1], s*rot[2], pos[0]);
        glMultiTexCoord4f(GL_TEXTURE2, s*rot[4], s*rot[5], s*rot[6], pos[1]);
        glMultiTexCoord4f(GL_TEXTURE3, s*rot[8], s*rot[9], s*rot[10], pos[2]);

        // Render
        int faces = model->render(obj->radius, obj->position(), cameraPos, mDetailLevel);
        renderInfo->addBatch(faces);
    }

    for (int i = 0; i < activeTextures; i++) {
        if (texture[i]) {
            glActiveTexture(GL_TEXTURE0 + i);
            texture[i]->disable();
            kDebug() << "disable" << i;
        }
    }
    glActiveTexture(GL_TEXTURE0);
    mShaderManager->disableShader();
    if (buffer) {
        buffer->unbind();
    }
    glColor3f(1, 1, 1);
}

void GameView::renderSpace()
{
    glColor3f(1, 1, 1);
    glEnable(GL_CULL_FACE);

    glPushMatrix();
    Vector3f shipPos = mWorld->ship()->position();
    glTranslatef(shipPos.x(), shipPos.y(), shipPos.z());
    glScalef(100, 100, 100);

    mSpaceMesh->render();
    glPopMatrix();
}

