/*
 * Copyright 2008  Rivo Laks <rivolaks@hot.ee>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "gameworld.h"

#include "ship.h"
#include "asteroid.h"
#include "mymath.h"
#include "bullet.h"
#include "model.h"
#include "planet.h"
#include "datastore.h"

#include <kgllib/shapes.h>
#include <kgllib/mesh.h>

#include <KDebug>
#include <KConfig>
#include <KConfigGroup>


int GameWorld::mCollisionCandidatesWithDups = 0;
int GameWorld::mCollisionCandidates = 0;
int GameWorld::mCollisions = 0;


// Check if we have recent enough ODE which contains SAP spaces.
#ifdef dSAP_AXES_XYZ
#define ODE_HAS_SAP_SPACE
#endif


void nearCallback(void *data, dGeomID o0, dGeomID o1)
{
    reinterpret_cast<GameWorld*>(data)->handleCollisionBetween(o0,o1);
}
Vector3f ode2Eigen(const dVector3& v)
{
    return Vector3f(v[0], v[1], v[2]);
}

struct GameWorld::OdeContactFeedback
{
    OdeContactFeedback(GameObject* obj1, GameObject* obj2)
    {
        object1 = obj1;
        object2 = obj2;
    }

    GameObject* object1;
    GameObject* object2;
    dJointFeedback feedback;
};


GameWorld* GameWorld::mInstance = 0;

GameWorld::GameWorld(QObject *parent)
    : QObject(parent)
{
    if (mInstance) {
        kError() << "GameWorld instance has already been created!";
        return;
    }
    mInstance = this;

    mData = DataStore::instance();
    mDataDir = mData->dataDir();

    mShip = 0;

    mGamePaused = false;
    mTime = 0.0f;
    mTrackActive = false;
    mTrackTime = 0.0f;
    mAdvanceTime = 0.0f;
}

GameWorld::~GameWorld()
{
    mInstance = 0;
}

void GameWorld::advance(float elapsed)
{
    if (mGamePaused) {
        return;
    }


    if (mTrackActive) {
        mTrackTime += elapsed;
    }
    applyShipRotation(elapsed);

    mCollisions = 0;
    mCollisionCandidates = 0;

    // ODE doesn't like too big timesteps. If time elapsed is bigger than
    //  maxtimestep then we'll simulate in multiple iterations.
    const float timestep = 0.040;
    mAdvanceTime += elapsed;
    for (; mAdvanceTime >= timestep; mAdvanceTime -= timestep) {
        // Clear previous collision contacts
        dJointGroupEmpty(mOdeContactGroup);
        // Calculate collisions
        dSpaceCollide(mOdeSpace, this, &nearCallback);
        // Advance the simulation
        dWorldQuickStep(mOdeWorld, timestep);

        processContactFeedbacks();

        for (int i = 0; i < mObjects.count(); i++) {
            mObjects[i]->advance(timestep);
        }

        processTrackPoints();
    }
    mTime += elapsed;
}

void GameWorld::handleCollisionBetween(dGeomID o1, dGeomID o2)
{
    mCollisions++;

    const int MAX_CONTACTS = 4;
    dContactGeom contactPoints[MAX_CONTACTS];
    // Check for collisions
    int collisions = dCollide(o1, o2, MAX_CONTACTS, contactPoints, sizeof(dContactGeom));
    mCollisionCandidates += collisions;

    if (collisions > 0) {
        // Create joint for each contact
        for (int i = 0; i < collisions; i++) {
            dBodyID geom1 = dGeomGetBody(o1);
            dBodyID geom2 = dGeomGetBody(o2);
            GameObject* obj1 = reinterpret_cast<GameObject*>(dBodyGetData(geom1));
            GameObject* obj2 = reinterpret_cast<GameObject*>(dBodyGetData(geom2));

            dContact contact;
            contact.surface.mode = dContactBounce;
            contact.surface.mu = dInfinity;
            contact.surface.bounce = 0.5;
            contact.geom = contactPoints[i];

            dJointID c = dJointCreateContact (mOdeWorld, mOdeContactGroup, &contact);
            OdeContactFeedback* fb = new OdeContactFeedback(obj1, obj2);
            dJointSetFeedback(c, &fb->feedback);
            mContactFeedbacks.append(fb);
            dJointAttach (c, geom1, geom2);
        }
    }
}

void GameWorld::processContactFeedbacks()
{
    for (int i = 0; i < mContactFeedbacks.count(); i++) {
        mContactFeedbacks[i]->object1->collision(ode2Eigen(mContactFeedbacks[i]->feedback.f1));
        mContactFeedbacks[i]->object2->collision(ode2Eigen(mContactFeedbacks[i]->feedback.f2));
        delete mContactFeedbacks[i];
    }
    mContactFeedbacks.clear();
}

void GameWorld::processTrackPoints()
{
    for (int i = 0; i < mTrack.count(); i++) {
        const int tpRadius = 60;
        if ((mShip->position() - mTrack[i]).squaredNorm() < tpRadius*tpRadius) {
            kDebug() << "Removing trackpoint" << i;
            mTrack.removeAt(i--);
            if (mTrack.isEmpty()) {
                mTrackActive = false;
            } else {
                mTrackActive = true;
            }
        }
    }
}

float GameWorld::trackTime() const
{
    return mTrackTime;
}

bool GameWorld::init()
{
    dInitODE();
    /*
    QStringList odeExtensions = QString((dGetConfiguration())).split(' ');
    kDebug() << "ODE config:" << odeExtensions;
#ifdef dSINGLE
    if (!odeExtensions.contains("ODE_single_precision")) {
        kError() << "Incompatible version of ODE (AstroDodge was compiled with the single precision version of ODE)";
        emit setLoadingError("Incompatible ODE version");
        return false;
    }
#elif dDOUBLE
    if (!odeExtensions.contains("ODE_double_precision")) {
        kError() << "Incompatible version of ODE (AstroDodge was compiled with the double precision version of ODE)";
        emit setLoadingError("Incompatible ODE version");
        return false;
    }
#endif
    */
    mOdeWorld = dWorldCreate();
    dWorldSetGravity(mOdeWorld, 0.0, -0.0, 0.0);

#ifdef ODE_HAS_SAP_SPACE
    mOdeSpace = dSweepAndPruneSpaceCreate(0, dSAP_AXES_XZY);
#else
    #warning Using newer ODE (=> 0.10) is recommended
    kWarning() << "You're using old version of ODE (< 0.10) and might experience crashes!";
    mOdeSpace = dHashSpaceCreate(0);
#endif

    mOdeContactGroup = dJointGroupCreate(0);
    srand(123456789);

    mNextObjectId = 1;

    return true;
}

void GameWorld::createAsteroids(int count, float mean, float variance, float min, float max,
                                const Vector3f& minpos, const Vector3f& maxpos)
{
    for (int i = 0; i < count; i++) {
        // First the radius
        float radius = clampedGaussFloat(mean, variance, min, max);

        // Select a position so that it won't collide with others
        const int max_tries = 10;
        bool positionFound = false;
        Vector3f position;
        for (int j = 0; j < max_tries; j++) {
            const float posiv = 1000;
            position = Vector3f(randFloat(minpos.x(), maxpos.x()), randFloat(minpos.y(), maxpos.y()),
                                randFloat(minpos.z(), maxpos.z()));
            positionFound = true;
            for (int k = 0; k < mObjects.count(); k++) {
                GameObject* obj = mObjects[k];

                float mindist = radius + obj->radius + 10;
                if ((position - obj->position()).squaredNorm() < mindist*mindist) {
                    positionFound = false;
                    break;
                }
            }
            if (positionFound) {
                //                 kDebug() << i << "found its pos with"  << j << "tries";
                break;
            } else if (j+1 >= max_tries) {
                kWarning() << "Will have bad pos for" << i;
                positionFound = false;
            }
        }
        if (!positionFound) {
            // Start all over again, choose different radius, maybe that helps
            i--;
            continue;
        }

        // Create the new asteroid
        Asteroid* a = new Asteroid(radius);
        a->setMass(a->radius*a->radius);
        a->setPosition(position);
        Vector3f velocity = Vector3f(randFloat(-2, 2), randFloat(-2, 2), randFloat(-2, 2));
        velocity.normalize();
        velocity *= clampedGaussFloat(2, 5, 1, 10);
        velocity *= 0;
        a->setVelocity(velocity);

//         a->rotation = Vector3f(randFloat(0, 360), randFloat(0, 360), randFloat(0, 360));

        addObject(a);
    }
}

void GameWorld::addAsteroids(int count)
{
    Vector3f bounds(500, 500, 500);
    createAsteroids(count/100*1 , 150, 75, 50, 300, -bounds, bounds);
    createAsteroids(count/100*5 , 50, 50, 20, 100, -bounds, bounds);
    createAsteroids(count/100*94, 15, 20, 0.5, 50, -bounds, bounds);
}

void GameWorld::addObject(GameObject* obj)
{
    obj->id = mNextObjectId++;
    obj->init();

    mObjects.append(obj);
}

Planet* GameWorld::addPlanet(const QString& planet)
{
    Mesh* m = new Mesh();
    KGLLib::Shapes::createSphere(m, 4);
//     m->setTexture(objectTexture("earth.png"));

    Planet* p = new Planet(m);
    mPlanets.append(p);
    return p;
}

void GameWorld::setShip(Ship* s)
{
    Q_ASSERT(!mShip);
    mShip = s;
    addObject(mShip);
}

void GameWorld::slotRotateShip(float dx, float dy, float dz)
{
    mShip->rotvelocity += Vector3f(dx, -dy, dz) * 50;
}

void GameWorld::applyShipRotation(float elapsed)
{
    mShip->addRotation(mShip->rotvelocity);
    mShip->rotvelocity = Vector3f(0, 0, 0);
}

void GameWorld::slotShootBullet()
{
    if (mShip->lastshoottime + 0.1 > mTime) {
        return;
    }

    mShip->lastshoottime = mTime;
    Bullet* b = new Bullet;
    b->setPosition(mShip->position() + mShip->direction()*12);
    b->setVelocity(mShip->direction() * 500);
    addObject(b);
    mBullets.append(b);
}

void GameWorld::setShipLight(bool on)
{
    mShip->lightOn = on;
}

GameObject* GameWorld::addTrackPoint(const Vector3f& pos)
{
    mTrack.append(pos);

    GameObject* obj = createObject("checkpoint");
    obj->setPosition(pos);
    addObject(obj);
    return obj;
}

void GameWorld::slotToggleGamePaused()
{
    mGamePaused = !mGamePaused;
}

GameObject* GameWorld::createObject(const QString& name)
{
    KConfig objectconf(mDataDir + "objects/" + name + ".index");
    KConfigGroup modelconf(&objectconf, "Model");
    float radius = modelconf.readEntry("Scale", 1.0);
    GameObject* obj = new GameObject(radius);
    obj->setMass(modelconf.readEntry("Mass", (double)radius*radius*radius));

    addObject(obj);

    return obj;
}
