/*
 * Copyright 2008-2009  Rivo Laks <rivolaks@hot.ee>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "model.h"

#include <kgllib/mesh.h>
#include <kgllib/modelloader.h>

#include <KConfig>
#include <KConfigGroup>
#include <KDebug>

#include <QFileInfo>

using namespace KGLLib;


Model::Model(const QString& filename, const QString& modeldir)
{
    mConfigFile = filename;
    kDebug() << "Loading from" << filename;
    mValid = false;

    QString modelpath = modeldir;
    if (modelpath.isNull()) {
        modelpath = QFileInfo(filename).path() + '/';
    }

    KConfig conf(filename);
    KConfigGroup modelconf(&conf, "Model");

    mMaterial = modelconf.readEntry("Material", "object");

    int lodCount = qMin(5, modelconf.readEntry("LODs", 0));
    if (lodCount == 0) {
        // We have a single model, specified by Model key
        QString modelfile = modelpath + modelconf.readEntry("Model");
        ModelLoader* loader = createModelLoader(modelfile);
        if (!loader) {
            return;
        }

        if (modelconf.hasKey("Scale")) {
            loader->scaleModel(modelconf.readEntry("Scale", 1.0));
        }
        mLoaders.append(loader);
    } else {
        // We have multiple models, each in own LOD-x group
        for (int lod = 0; lod < lodCount; lod++) {
            KConfigGroup modelconf(&conf, QString("LOD-%1").arg(lod));

            QString modelfile = modelpath + modelconf.readEntry("Model");
            kDebug() << "Trying to load from" << modelfile;
            ModelLoader* loader = createModelLoader(modelfile);
            if (!loader) {
                return;
            }
            mLoaders.append(loader);
        }
        kDebug() << "Loaded model from" << filename << "(" << lodCount << "lods)";
    }

    mValid = true;
}

Model::~Model()
{
}

int Model::vertexCount() const
{
    int count = 0;
    foreach (const ModelLoader* loader, mLoaders) {
        count += loader->vertexCount();
    }
    return count;
}

int Model::indexCount() const
{
    int count = 0;
    foreach (const ModelLoader* loader, mLoaders) {
        count += loader->indexCount();
    }
    return count;
}

void Model::setTexture(Texture* texture)
{
    setTexture(0, texture);
}

void Model::setTexture(int index, Texture* texture)
{
    if (mTextures.size() <= index) {
        mTextures.resize(index+1);
    }
    mTextures[index] = texture;
}

void Model::setMaterial(const QString& material)
{
    mMaterial = material;
}

GeometryBuffer* Model::buffer()
{
    if (mLODs.isEmpty()) {
        return 0;
    }

    return mLODs[0]->buffer();
}

void Model::load(GeometryBuffer* buffer, int* vertexoffset, int* indexoffset)
{
    for (int lod = 0; lod < mLoaders.count(); lod++) {
        kDebug() << "Creating mesh for lod" << lod;
        Batch* mesh = new Batch(buffer, *vertexoffset, *indexoffset);

        mLoaders[lod]->loadIntoBatch(mesh);
        mesh->update();
        mLODs.append(mesh);

        *vertexoffset += mesh->vertexCount();
        *indexoffset += mesh->indicesCount();
        delete mLoaders[lod];
    }
    mLoaders.clear();
}

void Model::load()
{
    // TODO: some code duplication here
    for (int lod = 0; lod < mLoaders.count(); lod++) {
        kDebug() << "Creating mesh for lod" << lod;
        Batch* mesh = new Batch();

        mLoaders[lod]->loadIntoBatch(mesh);
        mesh->update();
        mLODs.append(mesh);

        delete mLoaders[lod];
    }
    mLoaders.clear();
}

ModelLoader* Model::createModelLoader(const QString& modelfile)
{
    ModelLoader* loader = new ModelLoader(modelfile);
    if (!loader->isValid()) {
        kError() << "Couldn't load model" << modelfile;
        delete loader;
        return 0;
    }
    return loader;
}

Batch* Model::collisionMesh() const
{
    int lod = qMin(4, mLODs.count()-1);
    return mLODs[lod];
}

int Model::render(float radius, const Vector3f& position, const Vector3f& camerapos, float detail) const
{
    int lod = getLod(radius, position, camerapos, detail);
    mLODs[lod]->renderOnce();
    return mLODs[lod]->indicesCount()/3;
}

