/*
 * Copyright 2008-2009  Rivo Laks <rivolaks@hot.ee>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef MODEL_H
#define MODEL_H

#include <Eigen/Core>

#include <QVector>
#include <QString>

namespace KGLLib
{
class ModelLoader;
class Batch;
class GeometryBuffer;
class Texture;
}

using namespace KGLLib;
using namespace Eigen;

/**
 * Model is used by most GameObjects for rendering.
 **/
class Model
{
public:
    /**
     * Tries to load the model from specified file.
     * The file must be a config file, containing info such as the actual model
     *  file(s) where the model will be loaded from.
     * You will need to call @ref load() to create the actual meshes.
     **/
    Model(const QString& file, const QString& modeldir = QString());
    ~Model();

    QString configFile() const  { return mConfigFile; }

    /**
     * Sets texture of the model.
     * This has to be called after @ref load().
    **/
    void setTexture(Texture* texture);
    void setTexture(int index, Texture* texture);
    int textureCount() const  { return mTextures.count(); }
    Texture* texture(int i)  { return (i < textureCount()) ? mTextures.at(i) : 0; }

    void setMaterial(const QString& material);
    QString material() const  { return mMaterial; }

    GeometryBuffer* buffer();

    /**
     * Loads the model and creates meshes, using the shared GeometryBuffer.
    **/
    void load(GeometryBuffer* buffer, int* vertexoffset, int* indexoffset);
    void load();
    /**
     * @return whether the model is ready to be loaded using @ref load() method
    **/
    bool isValid() const  { return mValid; }

    int vertexCount() const;
    int indexCount() const;


    Batch* lod(int i) const  { return mLODs[i]; }
    Batch* collisionMesh() const;

    int render(float radius, const Vector3f& position, const Vector3f& camerapos, float detail) const;

    inline float getDetailLevel(float radius, const Vector3f& position, const Vector3f& camerapos, float detail) const
    {
        return radius / ((camerapos - position).norm() + 1) * 1e3 * detail;
    }

    inline int getLod(float radius, const Vector3f& position, const Vector3f& camerapos, float detail) const
    {
        static const float lodDists[] = { 20, 30, 50, 70, 150, 250, 500, 1000, 1e30 };
        float d = getDetailLevel(radius, position, camerapos, detail);
        int lodCount = mLODs.count();
        for (int i = 0; i < lodCount; i++) {
            if (d < lodDists[i]) {
                return i;
            }
        }
        return lodCount-1;
    }

protected:
    ModelLoader* createModelLoader(const QString& modelfile);

private:
    QString mConfigFile;
    QVector<ModelLoader*> mLoaders;
    QVector<Batch*> mLODs;
    bool mValid;

    QVector<Texture*> mTextures;
    QString mMaterial;
};

#endif
