/***************************************************************************
 *   Copyright (C) 2009 by Alexander Smirnov                               *
 *   alexander.v.smirnov@gmail.com                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.         *
 ***************************************************************************/

#include "backgammongame.h"
#include "backgammonscene.h"
#include "board.h"
#include "boardrenderer.h"
#include "checker.h"
#include "dice.h"

#include "kdebug.h"

#include <KGamePopupItem>


class BackgammonScene::Private{
public:

    static const int numOfCheckersInOneRow = 6;

    /**
     *  The Game object
     */
    IGameEngine * m_game;
    
    KGamePopupItem* m_infoPopup;
    QList<Checker*> m_proposalCells;

    Private():m_game(NULL)
    {
        m_infoPopup = new KGamePopupItem;
    }

    ~Private()
    {
        delete m_infoPopup;
        deleteGame();
    }

    inline void deleteGame()
    {
        delete m_game;
    }

    inline void setGame(IGameEngine * game)
    {
        m_game = game;
    }
    
    inline void showMessage( const QString& message ) const
    {
        m_infoPopup->setMessageTimeout( 1500 );
        m_infoPopup->showMessage( message, KGamePopupItem::BottomLeft, KGamePopupItem::ReplacePrevious );
    }
};


BackgammonScene::BackgammonScene(BackgammonGame * game)
    : QGraphicsScene(), d(new BackgammonScene::Private)
{
    setGame(game);
    
    addItem( d->m_infoPopup );
}


BackgammonScene::~BackgammonScene()
{
    delete d;
}


void BackgammonScene::drawBackground( QPainter *p, const QRectF&)
{
    BoardRenderer::renderBackground( p, sceneRect() );
    BoardRenderer::renderBoard( p, Board::rect() );

//     if(m_showLabels)
//    BoardRenderer::renderBoardLabels(p, sceneRect() );

}

void BackgammonScene::resizeScene( int width, int height )
{
    kDebug() << "resizeScene" << width << "x" << height;
    setSceneRect( 0, 0, width, height );

    Board::resize(width, height);

    updateObjects();
}

void BackgammonScene::updateObjects()
{
    const QList<QGraphicsItem*> allItems = items();
    bool checkersFound = false;
    bool diceFound = false;
    foreach( QGraphicsItem* item, allItems )
    {
        Checker* checker = qgraphicsitem_cast<Checker*>(item);
        if(checker)
        {
            Board::setCheckerSize(checkerSize());
            setCheckerPosition(checker);
            checkersFound = true;
        }

        Die* die = qgraphicsitem_cast<Die*>(item);
        if(die)
        {
            die->updatePixmap(dieSize());
            die->setPos(
                    Board::cellRect(die->order()+2).center().x(), // on the top of 3rd and 4th cell
                    Board::rect().height()/2); 
            diceFound = true;
        }
    }

    if (!checkersFound) //TODO: add them on scene creation??
    {
        addCheckersToScene(Checker::Black);
        addCheckersToScene(Checker::White);
    }

    if (!diceFound)
    {
        for (int i = 0; i < Board::numberOfDice() ; ++i)
        {
            QGraphicsScene::addItem(Board::die(i));
            kDebug() << "added die, i=" <<i;
        }
    }
}

void BackgammonScene::addCheckersToScene(Checker::Color color )
{
    if (color == Checker::NoColor)
        return;

    for (int c = 1; c <= Board::numberOfCheckers(); ++c)
    {
        Checker * const checker = new Checker(color, checkerSize());

        Board::addChecker(Board::home(color), checker);

        setCheckerPosition(checker);
        QGraphicsScene::addItem(checker);
    }

}


/**
 * return max allowed number of checkers in one row(z-coord)
 * first row contains numOfCheckersInOneRow checkers,
 * second row - first-1
 * third row - second-1 and so on
 * @param row - row, or z-value of a checker. row belongs to [1;N], where N is the most top row
 * @return number of checkers allowed to be placed on #row level
 */
int BackgammonScene::numberOfCheckersInARow(const qint8 row) const
{
    if ( row < d->numOfCheckersInOneRow )
    {
        return d->numOfCheckersInOneRow - row + 1;
    }
    else
    {
        return 1;
    }
    //return ( row < numOfCheckersInOneRow )*( numOfCheckersInOneRow - row ) + 1;
}
/**
 * get checker's row (in a cell)
 * @param order - number of checker in a cell
 * @return row of a checker
 */
int BackgammonScene::checkerRow(const int order) const
{
    int row = 1;

    for ( int sum=numberOfCheckersInARow(row); sum < order; row++, sum+=numberOfCheckersInARow(row) )
    {
        //empty
    }

    return row;
}

/**
 * Get checker position in a row
 * @param order - checker seq number in a cell
 * @return a number from 1 to numberOfCheckersInARow(checker row) indicating checker position in a row
 */
int BackgammonScene::checkerSeq(const int order) const
{
    const int row = checkerRow(order);
    int sum = 0;

    for ( int iRow = 1; iRow < row; iRow++ )
    {
        sum += numberOfCheckersInARow(iRow);
    }

    return order - sum;
}

/**
 * get checker Y coordinate
 * 
 * it depends on cell where checker is supposed to be placed and it's sequental number in frames of this cell.
 * @param cellIdx - number of cell, integer from 0 to 23
 * @param order - sequental number of a checker which is (going to be) placed in this cell
 * @return checker's Y coord 
 */
qreal BackgammonScene::getCheckerY(int cellIdx, int order) const
{
    const QRectF& cell = Board::cellRect(cellIdx);
    const int row = checkerRow(order);
    qreal y;

    if ( numberOfCheckersInARow( row ) == 1 )
    {
        y = cell.center().y() - checkerSize()/2;
    }
    else
    {
        y = cell.y();

        //if cell is in the upper part of the board, then we need to add coords (isBottom=1), to place checker lower
        qint8 isBottom = 1;
        if ( cellIdx > 11 && cellIdx < 24 )
        {
            y += cell.height() - checkerSize();
            //if cell is in the lower part of the board, then we need to substract coords (isBottom=-1), to place checker upper
            isBottom = -1;
        }

        y += isBottom*
                (checkerSize() * ( checkerSeq(order) - 1) +
                checkerSize()/2 * (row - 1) );
    }

    return y;
}

/**
 * Return a position within a board which is supposed for specified cell and checker sequental number
 * 
 * @param cellIdx - number of cell where checker is (supposed to be) located
 * @param order - sequental number of a checker in this cell
 * @return QPointF where checker position is stored
 */
QPointF BackgammonScene::getCheckerPosition(int cellIdx, int order) const
{
    //get Y coord, it depends on number of checkers already placed in the cell and checker row(z-value)
    const qreal y = getCheckerY( cellIdx, order );

    // get X coord - middle of cell
    const qreal x = Board::cellRect(cellIdx).center().x() - checkerSize()/2;

    return QPointF(x,y);
}

/**
 * Overloaded function provided for convinience
 * 
 * Return a position within a board which is supposed for specified checker
 * @param checker is a checker to return position for
 * @return QPointF where checker position is stored
 */
QPointF BackgammonScene::getCheckerPosition(Checker *checker) const
{
    return getCheckerPosition(checker->cell(), checker->order());
}

/**
 * Set checker position within a board
 * @param checker - checker to set position for. It should have specified seq_num in a cell before entering this function
 */
void BackgammonScene::setCheckerPosition(Checker *checker)
{
    setCheckerZValue( checker );
    checker->setPos ( getCheckerPosition (checker) );
}

/**
 * Depending on checker row, set Z value
 */
void BackgammonScene::setCheckerZValue(Checker *checker)
{
    checker->setZValue( checkerRow( checker->order() ) );
}

/**
 * return actual checker size (diameter)
 * @return current checker size
 */
double BackgammonScene::checkerSize() const
{
    return Board::cellRect(0).height()/d->numOfCheckersInOneRow;
}

double BackgammonScene::dieSize() const
{
    return checkerSize() * 0.8;
}

//edit mode, implement it later
// void BackgammonScene::addCheckerToBoard( qint8 cellId, Checker::Color color)
// {
//     if ( d->m_game->canPlaceChecker(color , cellId) )
//     {
//         addCheckersToScene( cellId, color, qAbs(d->m_game->checkersAt(cellId)) );
//         d->m_game->board()[cellId] += color;
//     }
//     else
//     {
//         kDebug() << "Cell already contains checkers of another color, no new checkers have been added.";
//     }
// }

void BackgammonScene::setGame( IGameEngine * game )
{
    if (!game)
        return;

    kDebug() << "Starting new game...";

    stopGame();

    d->setGame(game);

    connect( game,  SIGNAL( setCheckerPositions( const QStringList&) ), 
                    SLOT( arrangeCheckers( const QStringList&) ) );
    connect( game, SIGNAL( diceRolled( qint8,qint8 ) ), SLOT( slotShowDice( qint8, qint8 ) ) );
    connect( game, SIGNAL( showMove( qint8, qint8 ) ), SLOT( moveChecker( qint8, qint8 ) ) );
    
    connect( game, SIGNAL( engineMessage( const QString& ) ), SLOT( showEngineMessage( const QString& ) ) );
    
    connect( game, SIGNAL( proposedMoves( const QList<int>& , Checker::Color ) ),
             SLOT( showProposedMoves( const QList<int>& , Checker::Color ) ) );
             
    connect( game, SIGNAL( hideProposedMoves() ), SLOT( unshowProposedMoves() ) );

    kDebug() << Board::state();
}

void BackgammonScene::arrangeCheckers( const QStringList& newBoardState )
{
    collectCheckers();
    
    kDebug() << newBoardState;

    int cellTo = -1;
    foreach (QString token,newBoardState)
    {
        ++cellTo;
        const int n = token.toInt();

        if (n == 0)
            continue;

        const qint8 pool = Board::home( Board::checkerColor(n) );
        
        //if( cellTo == pool )
        
        for (int j = 0; j < qAbs(n); j++)
        {
            moveChecker(pool, cellTo);
        }
    }
}

void BackgammonScene::collectCheckers()
{
    //move checkers from  old game to a pool
    const QList<QGraphicsItem*> allItems = items();
    foreach( QGraphicsItem* item, allItems )
    {
        Checker* checker = qgraphicsitem_cast<Checker*>(item);
        if(checker)
        {
            moveChecker(checker, Board::home( checker->color() ) );
        }

        Die* die = qgraphicsitem_cast<Die*>(item);
        if (die)
        {
            die->hide();
        }
    }    
}

void BackgammonScene::stopGame()
{
    kDebug() << "Stopping old game...";

    if (!d->m_game)
        return;

    collectCheckers();

    d->deleteGame();

    kDebug() << "game stopped";
}

void BackgammonScene::moveChecker(qint8 cellFrom, qint8 cellTo)
{
    kDebug() << "move checker " << cellFrom << " : " << cellTo;
    
    if (cellFrom == cellTo)
        return;

    Checker * checker = Board::getLastChecker(cellFrom);
    
    //Q_ASSERT(checker != NULL);
    if( checker == 0 )
        kDebug() << "no checker at this index na";
    
    moveChecker(checker, cellTo);
}

void BackgammonScene::moveChecker(Checker * checker, int cellTo)
{
    Board::moveChecker(checker, cellTo);

    checker->animate(getCheckerPosition(checker));
    setCheckerZValue(checker);
}

#include <QGraphicsSceneMouseEvent>
void BackgammonScene::mousePressEvent( QGraphicsSceneMouseEvent* event)
{
    if ( !d->m_game )
    {
        return;
    }
    // TODO: make this call synchronized?
    d->m_game->processUIEvent(event, this);
}

void BackgammonScene::slotShowDice(qint8 d1, qint8 d2)
{
    Board::die(0)->updatePixmap(d1, dieSize());
    Board::die(1)->updatePixmap(d2, dieSize());
    Board::showDice();
}

void BackgammonScene::showEngineMessage( const QString& message )
{
    d->showMessage( message );
}

void BackgammonScene::showProposedMoves( const QList<int>& indexes, Checker::Color color )
{
    foreach( int index, indexes )
    {
        Checker* check = new Checker( color, checkerSize() );
        check->setOpacity( 0.3 );
        check->setCell(index);
        check->setOrder( 1 + Board::numberOfCheckersAt(index) );
        setCheckerPosition(check);
        QGraphicsScene::addItem(check);
        d->m_proposalCells.push_back(check);
    }
}

void BackgammonScene::unshowProposedMoves()
{
    foreach( Checker* check, d->m_proposalCells )
    {
        QGraphicsScene::removeItem( check );
        delete check;
    }
    
    d->m_proposalCells.clear();
}
