/***************************************************************************
 *   Copyright (C) 2009 by Alexander Smirnov                               *
 *   alexander.v.smirnov@gmail.com                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program. If not, see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

#include "checker.h"
#include "boardrenderer.h"
#include "settings.h"

#include <QTimeLine>
#include <QPainter>

#include <kdebug.h>

Checker::Checker(Color color, double size)
 : QObject(),QGraphicsPixmapItem(QPixmap::fromImage(getImage (color, size))), m_color(color), m_animation(0), m_temporal(false)
{
}

Checker::~Checker()
{
    delete m_animation;
}

void Checker::updatePixmap(double size)
{
    QGraphicsPixmapItem::setPixmap(QPixmap::fromImage(getImage(m_color,size)));
}

QImage Checker::getImage (Checker::Color color, double size)
{
    QImage img(size, size, QImage::Format_ARGB32_Premultiplied);
    img.fill(0);

    QPainter p(&img);
    const QRectF rectangle(0,0,size,size);
    BoardRenderer::renderChip( &p, rectangle, color );
    p.end();

    return img;
}

int Checker::cell() const
{
    return m_cell;
}

void Checker::setCell(int cellIdx)
{
    m_cell = cellIdx;
}

Checker::Color Checker::color() const
{
    return m_color;
}

int Checker::order() const
{
    return m_order;
}

void Checker::setOrder(int o)
{
    m_order = o;
}

void Checker::animate(const QPointF& to)
{
    if (!Settings::animationEnabled())
    {
        setPos( to );
        return;
    }

    const double V = 0.1*Settings::animationSpeed();

    slotStopAnimation();

    const QPointF& from = pos();
    m_animationDest = to;

    // S = Vt,=> t=S/V
    const int duration =  qMax(qAbs(to.x() - from.x()),qAbs(to.y() - from.y()))/V;

    QTimeLine *timer = new QTimeLine(duration, this);
    timer->setFrameRange(0, duration);
    timer->setCurveShape(QTimeLine::EaseInOutCurve);

    m_animation = new QGraphicsItemAnimation(this);
    m_animation->setItem(this);
    m_animation->setTimeLine(timer);
    m_animation->setPosAt(0, from);
    m_animation->setPosAt(1, to);

    connect( timer, SIGNAL( finished() ), SLOT( slotStopAnimation() ) );

    timer->start();
}

void Checker::slotStopAnimation()
{
    if ( !m_animation )
        return;

    QGraphicsItemAnimation * const old_animation = m_animation;
    m_animation = 0;
    old_animation->timeLine()->stop();
    old_animation->deleteLater();
    setPos( m_animationDest );
}

QString Checker::toString() const
{
    return QString("cell=").append(QString::number(m_cell)).
            append(", color=").append(QString::number(m_color)).
            append(", order=").append(QString::number(m_order));
}

Checker::Color Checker::getAnotherColor( Color color )
{
    return (Color )( - color );
}
