/***************************************************************************
 *   Copyright (C) 2009 by Alexander Smirnov                               *
 *   alexander.v.smirnov@gmail.com                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.         *
 ***************************************************************************/

#include "backgammongame.h"
#include "iplayer.h"
#include "humanplayer.h"
#include "backgammonscene.h"
#include <kdebug.h>
#include <klocale.h>

#include <QGraphicsSceneMouseEvent>

#include <ctime>

#include <algorithm>

const qint8 BackgammonGame::m_initCheckerPos[] = {  -2, 0, 0, 0, 0, 5, 0, 3, 0, 0, 0, -5, 5, 0, 0, 0, -3, 0, -5,
                                                    0, 0, 0, 0, 2, 0, 0, 0, 0 };
                                                    

BackgammonGame::BackgammonGame( QObject* parent ) : IGameEngine( parent )
{
    qsrand( time(NULL) );
    
    m_players[0] = NULL;
    m_players[1] = NULL;
    
    resetBoard();
}

BackgammonGame::~BackgammonGame()
{}

QStringList BackgammonGame::initialBoardState() const
{
    QString state = "-2:0:0:0:0:5:0:3:0:0:0:-5:5:0:0:0:-3:0:-5:0:0:0:0:2:0:0:0:0";
    return state.split(":");
}

void BackgammonGame::roll()
{
    // In Numerical Recipes in C: The Art of Scientific Computing (William H. Press, Brian P. Flannery, Saul A. Teukolsky,
    // William T. Vetterling; New York: Cambridge University Press, 1992 (2nd ed., p. 277)), the following comments are made:
    //
    // "If you want to generate a random integer between 1 and 10, you should always do it by using high-order bits, as in
    //
    // j = 1 + (int) (10.0 * (rand() / (RAND_MAX + 1.0)));
    //
    // and never by anything resembling
    //
    // j = 1 + (rand() % 10);
    //
    // (which uses lower-order bits)."

    int first_roll = 1 + ( int ) ( 6.0 * ( qrand() / ( RAND_MAX + 1.0 ) ) );
    int second_roll = 1 + ( int ) ( 6.0 * ( qrand() / ( RAND_MAX + 1.0 ) ) );
    
    setDice( qMakePair( first_roll, second_roll ) );
}

/**
 * This function returns the current pip count for specified player
 * @param playerIdx - player index, 0 or 1
 */
int BackgammonGame::pipCount( const int playerIdx ) const
{
    IPlayer * player = m_players[playerIdx];
    int pip = 25 * Board::numberOfCheckersAt(Board::bar (player->color()));

    for (int i = 0; i <= 23; ++i) {
        if (Board::cellColor(i) == player->color())
        {
            pip +=  (player->getDirection() == -1) ?
                    (i+1)*Board::numberOfCheckersAt(i): // (i+1) to avoid multiplication on 0
                    (24 - i)*Board::numberOfCheckersAt(i);
        }
    }
    return pip;
}

void BackgammonGame::processUserAction(qint8 cell)
{}

bool BackgammonGame::playerInput(QDataStream &msg,KPlayer *player)
{       return true; }

// derived from IGameEngine, describes a backgammon rule
bool BackgammonGame::canPlaceChecker(const Checker::Color color, qint8 cellDest) const
{
    Q_ASSERT(Board::indexValid( cellDest ));
    return (color != Checker::NoColor) && (Board::cellColor(cellDest) == color || Board::cellColor(cellDest) == Checker::NoColor);
}

// derived from IGameEngine, describes a backgammon rule
bool BackgammonGame::canMoveChecker(IPlayer* player, qint8 cell1, qint8 cell2) const
{
    if (!Board::indexValid( cell1 ) || !Board::indexValid( cell2 ) )
    {
        kDebug() << "index#"<<cell1<<" or index#"<<cell2<<"  out of range, move is not possible...";
        return false;
    }

    if (Board::cellColor(cell1) != player->color())
    {
        kDebug() << "This is opponent's checker, not yours...";
        return false;
    }

    const Checker::Color cellColor = Board::cellColor(cell1);
    return canPlaceChecker( cellColor, cell2 );
}

void BackgammonGame::updatePip()
{
    emit pipCountChanged(i18n("Pip: ")+
                         m_players[0]->name()+": "+QString::number(pipCount(0)) +
                         "    " +
                         m_players[1]->name()+": "+QString::number(pipCount(1)));
}

bool BackgammonGame::checkGameOver()
{
    if( ( qAbs( m_field[ thrown_ind ] ) >= TOTAL_CHECKERS ) ||
        ( qAbs( m_field[ thrown_ind + 1 ] ) >= TOTAL_CHECKERS ) )
        return true;
    return false;
}

bool BackgammonGame::canThrow( IPlayer::direction dir )
{
    int home_count = 0;

    int index = getHomeIndex(dir);
    
    int start = index - ( HOME_SIZE - 1 )*int( dir );
    int end = index + 1*int( dir );
    
    for( int i = start; i != end; )
    {
        if( m_field[i]*dir < 0 )
            home_count += qAbs( m_field[i] );
        i = i + 1*dir;
    }

    kDebug() << "home count = " << home_count;
  
    int thrown = qAbs( m_field[ getThrownIndex(dir) ] );

    if( ( home_count + thrown ) == TOTAL_CHECKERS )
        return true;
    return false;
}

void BackgammonGame::showBoardChanges( const IPlayer& player, const QVector<Move>& changes )
{
    foreach( Move change, changes )
    {
        qint8 src = change.src;
        qint8 dst = change.dst;

        if( src >= bar_ind )
        {
            kDebug() << "src = "  << src << " going from bar";
            src = Board::bar( player.color() );
        }

        if( dst >= bar_ind )
        {
            kDebug() << "dst = " << dst << " going bar";
            dst = Board::bar( Checker::getAnotherColor( player.color() ) );
        }
        else
        {
            if( dst >= thrown_ind )
            {
                kDebug() << "dst = " << dst << " going home";
                dst =  Board::home( player.color() );
            }
        }

        emit( showMove( src, dst ) );
    }
}

bool BackgammonGame::checkCheckersAtBar( const IPlayer& player )
{
    int direction = player.getDirection();

    if( direction > 0 )
        return m_field[bar_ind] < 0;
    else
        return m_field[bar_ind  + 1] > 0;

    return false;
 }

int BackgammonGame::getBarIndex( IPlayer::direction dir )
{
    if( dir > 0 )
        return bar_ind;
    else
        return bar_ind + 1;
}

int BackgammonGame::getThrownIndex( IPlayer::direction dir )
{
    if( dir > 0 )
        return thrown_ind;
    else
        return thrown_ind + 1;
}

void BackgammonGame::addToBar( IPlayer::direction dir )
{
    int index = getBarIndex(dir);
    addChecker( index, dir );
}

void BackgammonGame::throwChecker( int ind, IPlayer::direction dir )
{
    replaceChecker( ind, dir );
    addChecker( getThrownIndex(dir), dir );
}

int BackgammonGame::getHomeIndex( IPlayer::direction dir )
{
    if( dir > 0 )
        return thrown_ind - 1;  
    else
        return 0;
    
    return 0;
}

bool BackgammonGame::canMakeMove( const IPlayer& player, const Move& move )
{
    IPlayer::direction dir = player.getDirection();
    int src = move.src;
    int dst = move.dst;
    
    if( m_field[src] == 0 )
        return false;

    if( ( src > FIELD_SIZE ) || ( src < 0 ) || 
        ( dst > FIELD_SIZE ) || ( dst < 0 ) )
        return false;
    
    if( cellValueMatchesDirection( src, dir ) )
    {
        kDebug() << "it's opponent's die, not yours " << m_field[src] << " " 
                 << int( dir ) << "src = " << src;
        return false;
    } 
    
    if( dst >= thrown_ind )
    {
        if( canThrow(dir) )
            return true;
        else
        {
            kDebug() << "can't throw";
            return false;
        }
    }
    
    if( cellValuesMatch( src, dst ) )
    {
        // dst is the same plyers cell
        int val = qAbs( m_field[dst] );
        if( val >= MAX_IN_ROW )
            return false;
    }
    else                
    {
        // dst is opponents cell
        if( qAbs( m_field[dst] ) > 1 )
            return false;
    }
    
    if( src < thrown_ind )
    {
        if( !moveMatchesDirection( Move( src, dst ), dir ) )
            return false;
    }

    return true;
}

void BackgammonGame::switchToNextState()
{
    switch( m_currentAction )
    {
        case Roll:
            m_currentAction = Move1;
            break;
        case Move1:
            m_currentAction = Move2;
            break;
        case Move2:
            if (m_dice1 == m_dice2)
            {
                m_currentAction = Move3;
            }
            else
            {
                m_currentAction = Commit;
            }                
            break;
        case Move3:
            m_currentAction = Move4;
            break;
        case Move4:
            m_currentAction = Commit;
            break;
        case Commit:
            m_currentAction = Roll;
            break;
    }
}

void BackgammonGame::switchToState( EGameState state )
{
    m_currentAction = state;
}

IGameEngine::EGameState BackgammonGame::getState() const
{
    return m_currentAction;
}

#include <iostream> 

void BackgammonGame::showField()
{
    for( int i = 0; i < FIELD_SIZE; i++ )
    {
        std::cout << int( m_field[i] ) << " ";
    }
    
    std::cout << std::endl;
}

bool BackgammonGame::canUseFirstDie( const IPlayer& player )
{
    if( firstDieUsed() )
        return false;
        
    return canUseDie( player, m_dice1 );    
}

bool BackgammonGame::canUseSecondDie( const IPlayer& player )
{
    if( secondDieUsed() )
        return false;
    
    return canUseDie( player, m_dice2 );
}

bool BackgammonGame::canUseDie( const IPlayer& player, int die )
{
    IPlayer::direction dir = player.getDirection();
    
    if( checkCheckersAtBar(player) )
    {
        if( canMakeMove( player, Move( getBarIndex(dir), getIndexFromBar( dir, die ) ) ) )
            return true;
        else
            return false;
    }
    
    // Check all player's cells
    for( int i = 0; i < thrown_ind; ++i )
    {
        if( cellMatchesPlayer( player, i ) )
        {
            if( canMakeMove( player, Move( i, getIndexFromSrc( i, dir, die ) ) ) )
                return true;            
        }
    }
    
    return false;
}

int BackgammonGame::getIndexFromBar( IPlayer::direction dir, int die )
{
    /*kDebug() << "get index from bar, home =  " << getHomeIndex( IPlayer::getOpposite(dir) ) 
             << " add = " << int(dir)*die;*/
    return getHomeIndex( IPlayer::getOpposite(dir) ) - 1*int(dir) + int(dir)*die;
}

bool BackgammonGame::moveReady( const IPlayer& player )
{
    if( gotCurrentMoveSrc() && gotCurrentMoveDst() )
        return true;
    
    if( firstDieUsed() || secondDieUsed() )
    {
        if( gotCurrentMoveSrc() )
            return true;
    }
    
    if( gotCurrentMoveSrc() )
    {
        if( !firstDieUsed() )
        {
            if( diceAreEqual() )
                return true;
            
            if( canUseFirstDie( player ) && canUseSecondDie( player ) )
                return false;
            else
                return true;
        }
        
        return true;
    }
    
    return false;
}

bool BackgammonGame::createMove( const IPlayer& player )
{
    qint8 src = m_currentMove.src;
    qint8 dst = m_currentMove.dst;
    
    if( !cellMatchesPlayer( player, src ) )
    {
        kDebug() << "not your checker " << src << ":" << m_field[src];
        emit( engineMessage( i18n( "Select your checker " ) ) );
        
        resetCurrentMove();
        return false;
    }
    
    IPlayer::direction dir = player.getDirection();
    
    if( dst == -1 )
    {
        kDebug() << "no destination yet";
        
        // Get die to be used
        int die = getDieToBeUsed( player );
        
        if( die != -1 )
        {
            // Here we need to compute dst
            if( checkCheckersAtBar( player ) )
            {
                kDebug() << "checker at bar, bar =  " << getBarIndex(player)
                        << " dst  = " << getIndexFromBar( player.getDirection(), die );
                src = getBarIndex( player );
                dst = getIndexFromBar( player.getDirection(), die );
            }
            else
            {
                dst = getIndexFromSrc( src, player.getDirection(), die );
                
                kDebug() << "dst = " << dst;
                
                if( ( dst >= thrown_ind ) || ( dst < 0 ) )
                {
                    dst = getThrownIndex( player.getDirection() );
                
                    if( !canThrowFromSrc( src, dir, die ) )
                    {
                        resetCurrentMove();
                        return false;
                    }
                }
            }
            
            if( canMakeMove( player, Move( src, dst ) ) )
            {
                setDieUsed(die);
                
                m_currentMove.src = src;
                m_currentMove.dst = dst;
                return true;
            }
        }
    }
    else
    {
        kDebug() << "dst != -1 dst = " << m_currentMove.dst;
        kDebug() << "src = " << src;
        
        if( dst >= thrown_ind )
            dst = getThrownIndex(dir);

        int length = qAbs( dst - src );
        
        if( dst >= thrown_ind )
            length = getLengthToThrown( src, dir );
        
        if( src == getBarIndex( player ) )
        {
            if( dst == getIndexFromBar( player.getDirection(), getFirstDie() ) )
                length = m_dice1;
            else if( dst == getIndexFromBar( player.getDirection(), getSecondDie() ) )
                length = m_dice2;
        }
        
        kDebug() << "length = " << length;
        
        int die = getDieToBeUsed( player );
        
        if( ( ( length == getFirstDie() ) && !firstDieUsed() )   || 
            ( ( length == getSecondDie() ) && !secondDieUsed() ) ||
            ( canThrow( player ) && canThrowFromSrc( src, dir, die ) ) )
        {
            // Here we got dst need just check it
            if( canMakeMove( player, m_currentMove ) )
            {
                //setDieUsed( (length < die)? die: length );
                //setDieUsed( die );
                setDieUsed(length);
                return true;
            }
        }
    }
    
    resetCurrentMove();
    return false;
}

const Move BackgammonGame::makeMove( const IPlayer& player, QVector<Move>& changes )
{
    IPlayer::direction dir = player.getDirection();
    qint8 src = m_currentMove.src;
    qint8 dst = m_currentMove.dst;
    
    if( ( dst < 0 ) || ( dst > thrown_ind ) )
    {
        throwChecker( src, dir );
        changes.push_back( Move( src, thrown_ind ) );
        
        resetCurrentMove();
        return Move( src, dst );
    }

    int diff = m_field[src]*m_field[dst];
    
    kDebug() << "src = " << src << " dst = " << dst << " diff " << diff;
    kDebug() << "m_field[src] = " << m_field[src] << " m_field[dst] = " << m_field[dst];

    if( diff > 0 ) 
    {
        addChecker( dst, dir );
        replaceChecker( src, dir );
        changes.push_back( Move( src, dst ) );
    }
    else if( diff == 0 )
    {
        addChecker( dst, dir );
        replaceChecker( src, dir );
        changes.push_back( Move( src, dst ) );
    }
    else
    {
        // Destination cell has different color 
        
        addChecker( dst, dir );
        addToBar( IPlayer::getOpposite(dir) );
        replaceChecker( dst, IPlayer::getOpposite(dir) );
        replaceChecker( src, dir );
        
        changes.push_back( Move( dst, bar_ind ) );
        changes.push_back( Move( src, dst ) );
    }
    
    resetCurrentMove();
    return Move( src, dst );
}

bool BackgammonGame::gotMoves( const IPlayer& player )
{    
    if( firstDieUsed() && secondDieUsed() )
    {
        kDebug() << "both dice used";
        return false;
    }
    
    if( !canUseFirstDie( player ) && !canUseSecondDie( player ) )
    {
        kDebug() << "can't use dice";
        return false;
    }
    
    return true;
}

bool BackgammonGame::cellValueMatchesDirection( int cell, int dir )
{
    return m_field[cell]*dir > 0;
}

int BackgammonGame::getIndexFromSrc( int src, IPlayer::direction dir, int die )
{
    int index = src + die*int(dir);
    
    if( ( index >= thrown_ind ) || ( index < 0 ) )
        index = getThrownIndex(dir);
    
    if( src >= bar_ind )
        index = getIndexFromBar( dir, die );
        
    return index;
}

int BackgammonGame::getDieToBeUsed( const IPlayer& player )
{
    if( canUseFirstDie( player ) )
    {
        kDebug() << "gonna use first die";
        return getFirstDie();
    }
    else if( canUseSecondDie( player ) )
    {
        kDebug() << "gonna use second die";
        return getSecondDie();
    }
    
    return -1;
}

int BackgammonGame::getBarIndex( const IPlayer& player )
{
    return getBarIndex( player.getDirection() );
}

int BackgammonGame::getOutIndex( const IPlayer& player )
{
    return getThrownIndex( player.getDirection() );
}

bool BackgammonGame::cellMatchesPlayer( const IPlayer& player, int src )
{
    //return m_field[src]*player.getDirection() < 0;
    return cellMatchesPlayer( player.getDirection(), src );
}

bool BackgammonGame::cellMatchesPlayer( IPlayer::direction dir, int src )
{
    return m_field[src]*dir < 0;
}

bool BackgammonGame::cellValuesMatch( int src, int dst )
{
    return m_field[src]*m_field[dst] > 0;
}

bool BackgammonGame::moveMatchesDirection( const Move& mv, IPlayer::direction dir )
{
    return ( dir*( mv.dst - mv.src ) ) > 0;
}

bool BackgammonGame::canThrow( const IPlayer& player )
{
    return canThrow( player.getDirection() );
}

int BackgammonGame::getLengthToThrown( int src, IPlayer::direction dir )
{
    if( dir == IPlayer::positive )
        return thrown_ind - src;

    return src + 1;
}

bool BackgammonGame::canThrowFromSrc( int src, IPlayer::direction dir, int die )
{
    int length = getLengthToThrown( src, dir );
    
    kDebug() << "length " << length << " die " << die;
    
    if( length < die )
    {
        int end = getHomeIndex( dir ) - ( HOME_SIZE - 1 )*int( dir );
        
        kDebug() << "end = " << end;
        
        for( int i = src - 1*dir; i != end - 1*dir; i -= 1*dir )
        {
            kDebug() << "i = " << i;
            
            if( cellMatchesPlayer( dir, i ) )
                return false;
        }
    }
    else if( length > die )
        return false;
    
    return true;
}

//-----------------------------------------------------------
// board functions

void BackgammonGame::resetBoard()
{
    resetDice();
    resetField();
    resetCurrentMove();
}

void BackgammonGame::setBoard( const qint8 ( & field ) [FIELD_SIZE] )
{
    memcpy( m_field, field, sizeof(m_field) );
    
    QStringList positions;
    
    for( int i = 0; i < thrown_ind; ++ i )
    {
        positions.push_back( QString::number( m_field[i] ) );
    }
    
    positions.push_back( QString::number( m_field[bar_ind + 1] ) );
    positions.push_back( QString::number( m_field[bar_ind] ) );
    
    positions.push_back( QString::number( m_field[thrown_ind + 1] ) );
    positions.push_back( QString::number( m_field[thrown_ind] ) );
    
    
    emit( setCheckerPositions( positions ) );
}

void BackgammonGame::replaceChecker( int ind, int direction )
{
    if( direction > 0 )
        m_field[ind]++;
    else
        m_field[ind]--;
}

void BackgammonGame::addChecker( int ind, int direction )
{
    if( direction > 0 )
        m_field[ind]--;
    else
        m_field[ind]++;
}

//-----------------------------------------------------------
// dice functions

void BackgammonGame::setDice( Rolls rolls )
{
    m_dice1 = qMax( rolls.first, rolls.second );
    m_dice2 = qMin( rolls.first, rolls.second );
    
    m_dice1Used = false;
    m_dice2Used = false;
    
    emit( diceRolled( m_dice1, m_dice2 ) );
}

int BackgammonGame::getFirstDie()
{
    return m_dice1;
}

int BackgammonGame::getSecondDie()
{
    return m_dice2;
}

void BackgammonGame::setFirstDieUsed()
{
    m_dice1Used = true;
}

void BackgammonGame::setSecondDieUsed()
{
    m_dice2Used = true;
}

bool BackgammonGame::firstDieUsed()
{
    return m_dice1Used;
}
    
bool BackgammonGame::secondDieUsed()
{
    return m_dice2Used;
}

bool BackgammonGame::diceAreEqual()
{
    return m_dice1 == m_dice2;
}

void BackgammonGame::setDieUsed( int die )
{
    if( die == getFirstDie() )
        setFirstDieUsed();
    else if( die == getSecondDie() )
        setSecondDieUsed();
    else if( die < getFirstDie() )
        setFirstDieUsed();
    else if( die < getSecondDie() )
        setSecondDieUsed();
}

void BackgammonGame::resetDice()
{
    m_dice1Used = false;
    m_dice2Used = false;
    
    m_dice1 = -1;
    m_dice2 = -1;
}

//-----------------------------------------------------------
// current move functions

void BackgammonGame::setCurrentMoveSrc( int cell )
{
    m_currentMove.src = cell;
}

void BackgammonGame::setCurrentMoveDst( int cell )
{
    if( m_currentMove.src != cell )
        m_currentMove.dst = cell;
}

bool BackgammonGame::gotCurrentMoveSrc()
{
    return m_currentMove.src != -1; 
}

bool BackgammonGame::gotCurrentMoveDst()
{
    return m_currentMove.dst != -1;
}

void BackgammonGame::resetCurrentMove()
{
    m_currentMove.src = -1;
    m_currentMove.dst = -1;
}

void BackgammonGame::resetField()
{
    memcpy( m_field, m_initCheckerPos, FIELD_SIZE );
}

QList<int> BackgammonGame::getProposedMoves( int cell, const IPlayer& player )
{
    QList<int> ret;
    
    if( canUseFirstDie( player ) )
    {
        int index = getIndexFromSrc( cell, player.getDirection(), getFirstDie() );
        
        if( canMakeMove( player, Move( cell, index ) ) )
            ret.push_back(index);
    }
    
    if( canUseSecondDie( player ) )
    {
        int index = getIndexFromSrc( cell, player.getDirection(), getSecondDie() );
        
        if( canMakeMove( player, Move( cell, index ) ) )
            ret.push_back(index);
    }
    
    return ret;
}

void BackgammonGame::showProposedMoves( int src, const IPlayer& player )
{
    QList<int> propMoves = getProposedMoves( src, player );
    emit( proposedMoves( propMoves, player.color() ) );
}

void BackgammonGame::unshowProposedMoves()
{
    emit( hideProposedMoves() );
}
