/***************************************************************************
 *   Copyright (C) 2005 by Piotr Banasiak <piotrbanasiak@onet.pl>          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to                             *
 *   the Free Software Foundation, 51 Franklin Street, Fifth Floor,        *
 *   Boston, MA 02110-1301, USA.                                           *
 ***************************************************************************/
#include "game.h"
#include "config.h"
#include "board.h"
#include "aengine.h"
#include "abstractenginet.h"
#include <QDataStream>
#include <QFile>

Game::Game( AEngine *engine ):QObject()
{
   //engine is NOT ownet by game.
   pAbstractEngine=engine;
   newGame();
   pAbstractEngine->fillBook( Book  );
}
//----------------------------------------------------------

Game::~Game()
{
}
//----------------------------------------------------------
/*!
    \fn Game::getCurrentBoard() const
 */
Board* Game::getCurrentBoard() const
{
    Board *pBoard=gameHistory[ currentMove ];
    return pBoard;
}

//----------------------------------------------------------
/*!
    \fn Game::getCurrentBoard() const
    returns pointer to the possible moves that 
    can be made from current position
 */
const QList<LegalMove>* Game::getPossibleMoves()
{
    Q_ASSERT( pAbstractEngine );

    return &possibleMoves;
}
//----------------------------------------------------------
// be sure possible moves are always generated for current position
void Game::setEngine( AEngine *engine )
{
    Q_ASSERT( engine );
    pAbstractEngine=engine;
    
    Board *pBoard=getCurrentBoard();
    
    pAbstractEngine->genPossibleMoves( possibleMoves, pBoard );
}
//----------------------------------------------------------
// return false if game is over (after making move)
bool Game::makeMove( LegalMove &LM )
{
   Q_ASSERT( pAbstractEngine );
   //Q_ASSERT( moveNumber>=0);
   //Q_ASSERT( moveNumber<possibleMoves.count() );
   
   //egalMove LM=possibleMoves[moveNumber];
   Board *pB=new Board ( getCurrentBoard() );
   
   //remove all future moves if any
   while ( gameHistory.size()> currentMove+1 )
   {
      Board *pB=gameHistory.takeLast();
      delete pB;
   }
   gameHistory.append( pB );
   
   int oldMH=moveHistory.size();
   while ( moveHistory.size()> currentMove )
       moveHistory.takeLast();
   moveHistory.append( LM );
   emit historyChanged( oldMH, moveHistory.size() );
   
   pAbstractEngine->makeMove( pB, LM );
   const EngineInfo *pEI = pAbstractEngine->getEngineInfo();
   emit  engineInfoChanged( pEI );
   
   currentMove++;
   emit currentMoveChanged(currentMove);
   
   pAbstractEngine->genPossibleMoves( possibleMoves, pB );
   
   emit documentChanged();
   
   if(pB->movesToDraw<=0)
   {
     emit gameOver( draw );
     return false;
   }
   if( possibleMoves.isEmpty() )
   {
      qDebug( "game over ");
      if( currentMove %2 )
        emit gameOver( whiteWin );
      else
        emit gameOver( blackWin );
      return false;
   }
   return true;
}
//----------------------------------------------------------
bool Game::takeBack()
{
   Q_ASSERT( pAbstractEngine );
   if( !currentMove )return false;
   
   currentMove--;
   emit currentMoveChanged(currentMove);

//    //reverse players
//    int tmp=whitePlayer;
//    whitePlayer=blackPlayer;
//    blackPlayer=tmp;
   
   Board *pBoard=getCurrentBoard();
   pAbstractEngine->genPossibleMoves( possibleMoves, pBoard );

   emit documentChanged();
   return true;
}
//----------------------------------------------------------
bool Game::forward()
{
   Q_ASSERT( pAbstractEngine );
   if( currentMove==gameHistory.size()-1 )return false;
   
   currentMove++;
   emit currentMoveChanged(currentMove);
   
//    //reverse players
//    int tmp=whitePlayer;
//    whitePlayer=blackPlayer;
//    blackPlayer=tmp;
   
   Board *pBoard=getCurrentBoard();
   pAbstractEngine->genPossibleMoves( possibleMoves, pBoard );

   emit documentChanged();
   return true;
}
//----------------------------------------------------------
void Game::moveTo( int moveNr)
{
   Q_ASSERT( pAbstractEngine );
   if( moveNr<0 )return ;
   if( moveNr>=gameHistory.size() )return ;
   
   currentMove=moveNr+1;
   emit currentMoveChanged(currentMove);
   
   Board *pBoard=getCurrentBoard();
   pAbstractEngine->genPossibleMoves( possibleMoves, pBoard );

   emit documentChanged();
   return ;
}
//----------------------------------------------------------
void Game::newGame()
{
   Q_ASSERT( pAbstractEngine );

   gameHistory.clear();
   possibleMoves.clear();
   moveHistory.clear(); 
   emit historyChanged( 0, 0 );
   
   //board number 0 is starting position.
   currentMove=0;
   emit currentMoveChanged(currentMove);
   
   Board *pBoard=new Board( Conf().m_boardSize );
   gameHistory.append( pBoard );
   pAbstractEngine->genPossibleMoves( possibleMoves, pBoard );
   emit documentChanged();
}
//----------------------------------------------------------
// return true if game over
bool Game::getStatus( int &whiteM, int& whiteK, int &blackM, int& blackK, int& moveNr,int& toDraw)
{
  Board *pBoard=getCurrentBoard();
  pBoard->countPieces(whiteM,  whiteK, blackM,  blackK);
  toDraw=pBoard->movesToDraw;
  moveNr=currentMove;
  
  return possibleMoves.isEmpty();
}


const LegalMove *const Game::getLastMove() const
{
   if(currentMove<=0)return 0;
   
   const LegalMove *const pLM= &moveHistory[ currentMove-1];
   return pLM;
}

/*!
    \fn Game::save(QFile & file)
 */
bool Game::save(QFile & file)
{
    QDataStream out(&file);   // we will serialize the data into the file
    //int moves=moveHistory.count();
    out << moveHistory;
    return true;
}

QDataStream & operator<<( QDataStream & out, const LegalMove& LM )
{
  out << LM.path;
  out << LM.takenPieces;
  return out;
}

/*!
    \fn Game::load(QFile & file)
 */
bool Game::load(QFile & file)
{
    QDataStream in(&file);   // we will serialize the data from 

    in >> moveHistory;
    
    qDebug("moves :%d", moveHistory.count() );
   
   gameHistory.clear();
   possibleMoves.clear();
   
   currentMove=0;
   Board *pBoard=new Board( Conf().m_boardSize );
   gameHistory.append( pBoard );
   pAbstractEngine->genPossibleMoves( possibleMoves, pBoard );
   
   for( int j=0;j< moveHistory.count() ;j++ )
   {
       Board *pBoard=new Board ( gameHistory[ j ] );
       pAbstractEngine->makeMove( pBoard, moveHistory[j] );
       gameHistory.append( pBoard );
   }
   emit historyChanged( 0, moveHistory.size() );
   emit currentMoveChanged(currentMove);
   emit documentChanged();
   return true;
}

QDataStream & operator>>( QDataStream & in, LegalMove& LM )
{
   in >> LM.path;
   in >> LM.takenPieces;
   return in ;
}

bool Game::loadBook(QFile & file)
{
   QDataStream in(&file);
   QList<LegalMove> bookLine;
   
    qint32 size=8;
    QString rules;
    in >> size;
    in >> rules;
    
    in >> bookLine;
   
   Book.append( bookLine );
   qDebug( " Game::loadBook size:%d", Book.size() );
   
   QList<LegalMove> bl=Book[0];
   for( int j=0;j< bl.size();j++)
   {
      qDebug("move %d",j);
      bl[j].print();
   }
   return true;
}
//----------------------------------------
LegalMove Game::getBookMove(bool &found)
{
  found=false;
  
  QList<LegalMove> candidates;
  
  for( int j=0;j< Book.size();j++)
  {
     QList<LegalMove> bookLine=Book[j];
     //compare with history
     if( currentMove >= bookLine.size() )continue;//that line is too short
     
     bool tmpFound=true;
     for( int m=0;m< currentMove;m++)
     {
       LegalMove move=moveHistory[m];
       LegalMove moveB=bookLine[m];
       if( !(move == moveB)) 
       {
          tmpFound=false;
          break;
       }
     }
     //after break
     if( tmpFound )
     {
        LegalMove tmp=bookLine[currentMove];
        tmp.print();
        candidates.append( bookLine[currentMove] );
     }
  }
  
  //choose random candidate
  found=candidates.size();
  if(! found )
     return LegalMove();
  int r=rand()% candidates.size();
  return candidates[r] ;
}
//----------------------------------------
bool Game::isGameOver()
{
    Board *pBoard=getCurrentBoard();
    
    pAbstractEngine->genPossibleMoves( possibleMoves, pBoard );
    if( possibleMoves.isEmpty() )return true;
    
    //but gamve over is also if draw
    if(pBoard->movesToDraw<=0) return true;
    return false;
}

#include "game.moc"
