/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "Camera.h"

// FruitWars headers
#include "Globals.h"
#include "Math3D.h"

// System headers
#include <GL/glu.h>

using namespace Eigen;

static const float CAMSIZE = 1.0;
static const float CAMSTEPPART = 0.02;
static const float CAMSMOOTH = 4.0;


Camera::Camera()
    : m_up(0, 1, 0),
    m_view(0, 0, -1),
    m_right(1, 0, 0),
    m_pos(0, 0, 0),
    m_angle(0, 0, 0)
{
}

Camera::Camera(const Vector3f &pos)
    : m_up(0, 1, 0),
    m_view(0, 0, -1),
    m_right(1, 0, 0),
    m_pos(pos),
    m_angle(0, 0, 0)
{
}

Camera::~Camera()
{
}

void Camera::move(float dist)
{
    m_pos += m_view * dist;
}

void Camera::setPosition(const Vector3f &pos)
{
    m_pos = pos;
}

void Camera::setPosition(float x, float y, float z)
{
    m_pos = Vector3f(x, y, z);
}

Vector3f Camera::position() const
{
    return m_pos;
}

void Camera::rotateX(float angle)
{
    m_angle.x() = angle;
}

void Camera::rotateY(float angle)
{
    m_angle.y() = angle;
}

void Camera::rotateZ(float angle)
{
    m_angle.z() = angle;
}

void Camera::rotate(float xangle, float yangle, float zangle)
{
    m_angle = Vector3f(xangle, yangle, zangle);
}

void Camera::rotate(const Vector3f &angle)
{
    m_angle = angle;
}

void Camera::point(float x, float y, float z)
{
    point(Vector3f(x, y, z));
}

void Camera::point(const Vector3f &pos)
{
    float yangle = atan2(pos.x() - m_pos.x(), m_pos.z() - pos.z());
    Vector3f dif = pos - m_pos;
    float l = dif.norm();
    float xangle = asin( (m_pos.y() - pos.y()) / l);
    rotate(xangle, yangle, 0);
}

void Camera::strafe(float dist)
{
    m_pos += m_right * dist;
}

void Camera::update()
{
    m_up = Vector3f(0.0, 1.0, 0.0);
    m_view = Vector3f(0.0, 0.0, -1.0);
    m_right = Vector3f(1.0, 0.0, 0.0);

    applyRotations();
}

void Camera::render()
{
    gluLookAt(m_pos.x(), m_pos.y(), m_pos.z(),
        m_pos.x() + m_view.x(), m_pos.y() + m_view.y(), m_pos.z() + m_view.z(),
        m_up.x(), m_up.y(), m_up.z());
}

void Camera::applyRotations()
{
    Transform3f r(
        AngleAxisf(-m_angle.z(), Vector3f::UnitZ()) *
        AngleAxisf(-m_angle.y(), Vector3f::UnitY()) *
        AngleAxisf(-m_angle.x(), Vector3f::UnitX()));
    m_view = r * m_view;
    m_up = r * m_up;
    m_right = r * m_right;
}

ThirdPersonCamera::ThirdPersonCamera(const Vector3f &pos, float seaHeight, Terrain *ter)
    : Camera(pos),
    m_center(pos),
    m_dist(0),
    m_tPos(0, 0, 0),
    m_relPos(m_tPos),
    m_smooth(0.0),
    m_oldCenter(0, 0, 0),
    m_tAngle(m_angle),
    m_angSmooth(0),
    m_lastTime(time()),
    m_seaHeight(seaHeight),
    m_ter(ter)
{
}

ThirdPersonCamera::ThirdPersonCamera()
    : Camera()
{
}

ThirdPersonCamera::~ThirdPersonCamera()
{
}

void ThirdPersonCamera::setCenter(const Vector3f &pos)
{
    m_center = pos;
}

void ThirdPersonCamera::setDist(float dist)
{
    m_dist = dist;
}

void ThirdPersonCamera::setSmooth(float smooth)
{
    m_smooth = smooth;
}

void ThirdPersonCamera::setAngSmooth(float smooth)
{
    m_angSmooth = smooth;
}

void ThirdPersonCamera::update()
{
    m_up = Vector3f(0.0, 1.0, 0.0);
    m_view = Vector3f(0.0, 0.0, -1.0);
    m_right = Vector3f(1.0, 0.0, 0.0);

    m_angle = lerp(m_tAngle, m_angle, m_angSmooth);

    applyRotations();

    Vector3f tempPos = m_oldCenter - m_center;
    m_tPos = Vector3f(
            m_right.dot(tempPos) * CAMSMOOTH,
            m_up.dot(tempPos) * CAMSMOOTH,
            m_view.dot(tempPos) * CAMSMOOTH);
    m_relPos = lerp(m_tPos, m_relPos, m_smooth);
    m_pos = (m_view * (m_relPos.z() - m_dist))
        + (m_up * m_relPos.y())
        + (m_right * m_relPos.x())
        + m_center;
    float h = qMax(m_seaHeight, m_ter->getGroundHeight(m_pos));
    if (m_pos.y() < h + CAMSIZE) {
        m_pos.y() = h + CAMSIZE;
    }

    m_oldCenter = m_center;
}

void ThirdPersonCamera::moveCenter(float dist)
{
    m_center += m_view * dist;
}

void ThirdPersonCamera::strafeCenter(float dist)
{
    m_center += m_right * dist;
}

Eigen::Vector3f ThirdPersonCamera::targetAngle() const
{
    return m_tAngle;
}

void ThirdPersonCamera::setTargetAngle(const Eigen::Vector3f& tAngle)
{
    m_tAngle = tAngle;
}
