/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "Collision.h"

// FruitWars headers
#include "Globals.h"
#include "Instance.h"
#include "Math3D.h"

USING_PART_OF_NAMESPACE_EIGEN


static const float COLOFFSET = 0.01;
static const float COLOFFSET2 = 0.02;
static const float TERSLIDINGVAL = 0.80;
static const float BOUNCEFRICTION = 0.50;
static const float BANANABOUNCEFRICTION = 0.75;
static const int BANANAGROUNDSTEPS = 32;
static const int BANANAGROUNDITERS = 8;
static const int BANANAMAXCOLCOUNT = 32;

static void playerGroundCol();
static void playerGroundCol2();
static void bananaGroundColReact(float t);
static void bananaBasketColReact(float t, int basketId);
static void bananaPlayerColReact(float t, float relt, int playerId);
static float bananaPlayerColTime(float relt, int& playerId);
static float bananaBasketColTime(int& basketId);
static float bananaGroundColTime();
static void fruitGroundCol();
static void fruitFruitCol();
static void playerFruitCol();
static void fruitBasketCol();
static void playerPlayerCol();
static void playerBasketCol();
static void setOldPos();

void calcCollisions()
{
    playerGroundCol();

    int i = 0;
    float t = 1.0;
    float relt = 0.0;
    int basketId, playerId;
    while (i < BANANAMAXCOLCOUNT && t >= 0 && t <= 1) {
        float t1 = bananaBasketColTime(basketId);
        float t2 = bananaPlayerColTime(relt, playerId);
        float t3 = bananaGroundColTime();
        if (t1 < 0) t1 = 2;
        if (t2 < 0) t2 = 2;
        if (t3 < 0) t3 = 2;
        t = qMin(t1, qMin(t2, t3));
        if (t1 < t2 && t1 < t3 && t1 <= 1.0) {
            bananaBasketColReact(t1, basketId);
        }
        if (t2 < t1 && t2 < t3 && t2 <= 1.0) {
            bananaPlayerColReact(t2, relt, playerId);
        }
        if (t3 < t2 && t3 < t1 && t3 <= 1.0) {
            bananaGroundColReact(t3);
        }
        relt += t * (1.0 - relt);
        i++;
    }

    fruitFruitCol();
    fruitBasketCol();
    fruitGroundCol();
    playerFruitCol();
    playerPlayerCol();
    playerBasketCol();
    playerGroundCol2();
    setOldPos();
}

void playerGroundCol()
{
    for (unsigned int i = 0; i < nplayers; i++) {
        float h = myTer.getGroundHeight(myPlayers[i].getPos());;
        if (myPlayers[i].getPos().y() <= h) {
            Plane plane = myTer.getGroundPlane(myPlayers[i].getPos());
            if (plane.b >= TERSLIDINGVAL) {
                myPlayers[i].position(Vector3f(myPlayers[i].getPos().x(), h, myPlayers[i].getPos().z()));
                if (myPlayers[i].getIsJumping()) myPlayers[i].setLastDodgeTime(time());
                myPlayers[i].setIsJumping(false);
                myPlayers[i].setJumpSpeed(0.0);
                myPlayers[i].setDodgeSpeed(Vector3f(0, 0, 0));
            } else {
                float dist = -distToPlane(plane, myPlayers[i].getPos());
                myPlayers[i].position(myPlayers[i].getPos() + Vector3f(plane.a, plane.b, plane.c) * dist);
                if (myPlayers[i].getIsJumping()) myPlayers[i].setLastDodgeTime(time());
                myPlayers[i].setDodgeSpeed(Vector3f(0, 0, 0));
            }
        }
        BS bs = myPlayers[i].getColBS();
        if (bs.pos.y() + bs.radius < SEAHEIGHT) myPlayers[i].die();
    }
}

void playerGroundCol2()
{
    for (unsigned int i = 0; i < nplayers; i++) {
        float h = myTer.getGroundHeight(myPlayers[i].getPos());
        if (myPlayers[i].getPos().y() <= h) {
            Plane plane = myTer.getGroundPlane(myPlayers[i].getPos());
            myPlayers[i].position(Vector3f(myPlayers[i].getPos().x(), h, myPlayers[i].getPos().z()));
            if (myPlayers[i].getIsJumping()) myPlayers[i].setLastDodgeTime(time());
            if (plane.b >= TERSLIDINGVAL) {
                myPlayers[i].setIsJumping(false);
                myPlayers[i].setJumpSpeed(0.0);
            }
            myPlayers[i].setDodgeSpeed(Vector3f(0, 0, 0));
        }
    }
}

void fruitGroundCol()
{
    for (int i = 0; i < fruitSpawner.getFruitDataCount(); i++) {
        if (fruitSpawner.getFruitData(i).inGame) {
            Fruit* fruit = &fruitSpawner.getFruit(i);
            fruit->setVanishing(false);
            if (fruit->isGroundColOn()) {
                BS bs = fruit->colBS();
                float h = myTer.getGroundHeight(bs.pos);
                if (bs.pos.y() <= h + fruit->midHeight() + COLOFFSET) {
                    fruit->positionBS(Vector3f(bs.pos.x(), h + fruit->midHeight(), bs.pos.z()));
                    fruit->setFallSpeed(0.0);
                    fruit->setMoveSpeed(Vector3f(0.0, 0.0, 0.0));
                    if (h + fruit->midHeight() + bs.radius < SEAHEIGHT && fruit->carriedBy() < 0) {
                        fruit->setVanishing(true);
                        if (fruit->vanish() > 2.0 * bs.radius) {
                            fruitSpawner.respawnFruit(i);
                        }
                    }
                }
            }
        }
    }
}

void fruitFruitCol()
{
    for (int i = 0; i < fruitSpawner.getFruitDataCount(); i++) {
        if (fruitSpawner.getFruitData(i).inGame) {
            if (true) { // isOwner
                for (int j = i+1; j < fruitSpawner.getFruitDataCount(); j++) {
                    if (fruitSpawner.getFruitData(j).inGame) {
                        Fruit* fruita = &fruitSpawner.getFruit(i);
                        Fruit* fruitb = &fruitSpawner.getFruit(j);

                        BS bsa = fruita->colBS();
                        BS bsb = fruitb->colBS();

                        Vector3f dif = bsb.pos - bsa.pos;
                        float dist = dif.squaredNorm();
                        if (dist == 0) dist = 0.001;
                        if (dist < sqr(bsa.radius + bsb.radius)) {
                            dist = sqrt(dist);
                            float part = dist / (bsa.radius + bsb.radius);
                            dif.normalize();
                            Vector3f center = bsa.pos + dif * (bsa.radius * part);
                            fruita->positionBS(center - dif * (bsa.radius + COLOFFSET));
                            fruitb->positionBS(center + dif * (bsb.radius + COLOFFSET));
                        }
                    }
                }
            }
        }
    }
}

void playerFruitCol()
{
    for (int i = 0; i < fruitSpawner.getFruitDataCount(); i++) {
        if (true) { // isOwner
            if (fruitSpawner.getFruitData(i).inGame && fruitSpawner.getFruit(i).carriedBy() < 0) {
                bool col = false;
                for (unsigned int j = 0; j < nplayers; j++) {
                    if (myPlayers[j].getHolds() == -1 && !myPlayers[j].getDead()) {
                        Fruit* fruit = &fruitSpawner.getFruit(i);
                        BS bsf = fruit->colBS();
                        BS bsp = myPlayers[j].getColBS();
                        Vector3f dif = bsf.pos - bsp.pos;
                        float dist = dif.squaredNorm();
                        if (dist < sqr(bsf.radius + bsp.radius)) {
                            col = true;
                            if (fruit->carriedBy() == -1 && holdMode != 1) {
                                fruit->setCarriedBy(j);
                                myPlayers[j].setHolds(i);
                                myPlayers[j].pickUpFruit(i);
                                // Network
                            }
                            if (holdMode == 1) {
                                Vector3f speed = (myPlayers[j].getColBS().pos - myPlayers[j].getOldPos()) * (1.0 / myPlayers[j].getDt());
                                fruit->setFallSpeed(FRUITTHROWSPEED + throwInfluence * speed.y());
                                fruit->setMoveSpeed(myPlayers[j].getDirVec() * FRUITMOVESPEED + Vector3f(speed.x(), 0.0, speed.z()) * throwInfluence);
                            }
                        }
                    }
                }
                if (!col && fruitSpawner.getFruit(i).carriedBy() == -2) {
                    fruitSpawner.getFruit(i).setCarriedBy(-1);
                }
            }
        }
    }
}

void fruitBasketCol()
{
    for (int i = 0; i < fruitSpawner.getFruitDataCount(); i++) {
        if (fruitSpawner.getFruitData(i).inGame) {
            Fruit* fruit = &fruitSpawner.getFruit(i);
            fruit->setGroundColOn(true);
            fruit->disableWarpClip();
            for (int j = 0; j < nBaskets; j++) {
                BS bsa = fruit->colBS();
                BS bsb = myBaskets[j].getColBS();

                Vector3f dif = bsb.pos - bsa.pos;
                float dist2 = sqr(dif.x()) + sqr(dif.z());

                if (bsa.pos.y() < bsb.pos.y() && dist2 > bsb.radius * bsb.radius) {
                    float dist = dif.x() * dif.x() + dif.y() * dif.y() + dif.z() * dif.z();
                    if (dist == 0) dist = 0.01;
                    if (dist < (bsa.radius + bsb.radius) * (bsa.radius + bsb.radius)) {
                        dif.normalize();
                        fruit->positionBS(bsb.pos - dif * (bsa.radius + bsb.radius + COLOFFSET));
                        // and bounce
                        Vector3f tempSpeed = fruit->moveSpeed();
                        tempSpeed.y() = fruit->fallSpeed();
                        tempSpeed = reflect(tempSpeed, dif) * BOUNCEFRICTION;
                        fruit->setMoveSpeed(Vector3f(tempSpeed.x(), 0.0, tempSpeed.z()));
                        fruit->setFallSpeed(tempSpeed.y());
                    }
                } else {
                    dist2 = sqrt(dist2);

                    // Control the warp
                    if (dist2 < (bsa.radius + bsb.radius - COLOFFSET)) {
                        if (fruit->carriedBy() < 0) fruit->setBasketWarp(j);
                        fruit->setGroundColOn(false);
                        if (bsa.pos.y() < fruit->warp().getPos().y() - bsa.radius && dist2 <= bsb.radius + COLOFFSET && !fruit->isAttached()) {
                            if (true) { // isFruitOwner
                                fruitSpawner.respawnFruit(i);
                                myBaskets[j].addFruit(fruit);
                                // playSound
                                // send NWMID_FRUITSCORE
                            }
                        }
                    }

                    if (bsa.pos.y() > bsb.pos.y()) {
                        // Bounce of edge
                        float dist = sqr(dist2 - bsb.radius) + sqr(dif.y());
                        if (dist == 0) dist = 0.01;
                        if (dist < sqr(bsa.radius)) {
                            Vector3f bouncep = bsb.pos + Vector3f(dif.x(), 0.0, dif.z()) * (-bsb.radius / dist2);
                            dif = (bsa.pos - bouncep).normalized();
                            fruit->positionBS(bouncep + dif * (bsa.radius + COLOFFSET));
                            // and bounce
                            Vector3f tempSpeed = fruit->moveSpeed();
                            tempSpeed.y() = fruit->fallSpeed();
                            tempSpeed = reflect(tempSpeed, dif) * BOUNCEFRICTION;
                            fruit->setMoveSpeed(Vector3f(tempSpeed.x(), 0.0, tempSpeed.z()));
                            fruit->setFallSpeed(tempSpeed.y());
                        }
                    } else {
                        if (dist2 > bsb.radius - bsa.radius) {
                            dif = Vector3f(dif.x(), 0.0, dif.z()) * (-1.0 / dist2);
                            fruit->positionBS(Vector3f(bsb.pos.x(), bsa.pos.y(), bsb.pos.z()) + dif * (bsb.radius - (bsa.radius + COLOFFSET)));
                            // and bounce
                            Vector3f tempSpeed = fruit->moveSpeed();
                            tempSpeed.y() = fruit->fallSpeed();
                            tempSpeed = reflect(tempSpeed, dif) * BOUNCEFRICTION;
                            if (dif.dot(tempSpeed) > 0 || tempSpeed.y() > 0) {
                                tempSpeed = Vector3f(0, 0, 0);
                            }
                            fruit->setMoveSpeed(Vector3f(tempSpeed.x(), 0.0, tempSpeed.z()));
                            fruit->setFallSpeed(tempSpeed.y());
                        }
                    }
                }
            }
        }
    }
}

void playerPlayerCol()
{
    HumanPlayer* lplayer = &myPlayers[0];
    for (unsigned int j = 0; j < nplayers; j++) {
        if (j != 0) {
            BS bsa = lplayer->getColBS();
            BS bsb = myPlayers[j].getColBS();

            Vector3f dif = bsb.pos - bsa.pos;
            float dist = dif.squaredNorm();
            if (dist == 0) dist = 0.01;
            if (dist < sqr(bsa.radius + bsb.radius)) {
                dist = sqrt(dist);
                float part = dist / (bsa.radius + bsb.radius);
                dif.normalize();
                Vector3f center = bsa.pos + dif * (bsa.radius * part);
                lplayer->positionBS(center - dif * (bsa.radius + COLOFFSET));
                myPlayers[j].positionBS(center + dif * (bsb.radius + COLOFFSET));
            }
        }
    }
}

void playerBasketCol()
{
    for (unsigned int i = 0; i < nplayers; i++) {
        for (int j = 0; j < nBaskets; j++) {
            BS bsa = myPlayers[i].getColBS();
            BS bsb = myBaskets[j].getColBS();

            Vector3f dif = bsb.pos - bsa.pos;
            float dist = dif.squaredNorm();
            if (dist == 0) dist = 0.01;
            if (dist < sqr(bsa.radius + bsb.radius)) {
                dif.normalize();
                myPlayers[i].positionBS(bsb.pos - dif * (bsa.radius + bsb.radius + COLOFFSET));
            }
        }
    }
}

void setOldPos()
{
    for (unsigned int i = 0; i < nplayers; i++) {
        myPlayers[i].setOldPos(myPlayers[i].getColBS().pos);
        myPlayers[i].getBanana().setOldPos(myPlayers[i].getBanana().getColBS().pos);
    }
}

float bananaPlayerColTime(float relt, int& playerId)
{
    unsigned int i = 0; // local player
    Banana& banana = myPlayers[i].getBanana();
    float result = 2.0;
    if (banana.getState() > 0) {
        for (unsigned int j = 0; j < nplayers; j++) if (i != j) {
            Vector3f endp1 = banana.getColBS().pos;
            Vector3f beginp1 = banana.getOldPos();
            Vector3f endp2 = myPlayers[j].getColBS().pos;
            Vector3f beginp2 = lerp(myPlayers[j].getOldPos(), endp2, relt);
            Vector3f dif1 = endp1 - beginp1;
            Vector3f dif2 = endp2 - beginp2;
            Vector3f dif = dif1 - dif2;
            BS colbsb = myPlayers[i].getColBS();
            BS colbsa = banana.getColBS();

            float a = dif.squaredNorm();
            float b = 2.0 * (dif.x() * (beginp1.x() - beginp2.x()) + dif.y() * (beginp1.y() - beginp2.y()) + dif.z() * (beginp1.z() - beginp2.z()));
            float c = (beginp1 - beginp2).squaredNorm() - sqr(colbsb.radius + colbsa.radius);
            float d = sqr(b) - 4.0 * a * c;

            if (d >= 0 && a > 0) {
                d = sqrt(d);
                float t1 = (-b + d) / (2.0 * a);
                float t2 = (-b - d) / (2.0 * a);
                float t;
                if (t1 < 0) t = t2;
                else if (t2 < 0) t = t1;
                else t = qMin(t1, t2);
                if ((beginp1 - beginp2).dot(beginp1 - beginp2) < sqr(colbsb.radius + colbsa.radius)) {
                    t = 2.0;
                }
                if (t < result) {
                    result = t;
                    playerId = j;
                }
            }
        }
    }
    return result;
}

void bananaPlayerColReact(float t, float relt, int playerId)
{
    int i = 0; // local player
    Banana& banana = myPlayers[i].getBanana();
    int j = playerId;

    Vector3f endP1 = banana.getColBS().pos;
    Vector3f beginP1 = banana.getOldPos();
    Vector3f endP2 = myPlayers[j].getColBS().pos;
    Vector3f beginP2 = lerp(myPlayers[j].getOldPos(), endP2, relt);
    Vector3f dif1 = endP1 - beginP1;
    Vector3f dif2 = endP2 - beginP2;
    Vector3f dif = dif1 - dif2;
    BS colbsb = myPlayers[i].getColBS();
    BS colbsa = banana.getColBS();

    Vector3f p = lerp(beginP1, endP1, t);
    Vector3f p2 = lerp(beginP2, endP2, t);

    dif = (p - p2).normalized();
    p = p2 + dif * (colbsb.radius + colbsa.radius + COLOFFSET2);

    banana.setOldPos(p);

    float dt;
    if (banana.getThrowSpeed().x() != 0) {
        dt = (endP1.x() - beginP1.x()) / banana.getThrowSpeed().x();
    } else if (banana.getThrowSpeed().y() != 0) {
        dt = (endP1.y() - beginP1.y()) / banana.getThrowSpeed().y();
    } else if (banana.getThrowSpeed().z() != 0) {
        dt = (endP1.z() - beginP1.z()) / banana.getThrowSpeed().z();
    } else {
        dt = 0;
    }

    if (dif.dot(banana.getThrowSpeed()) <= 0) {
        banana.setThrowSpeed(reflect(banana.getThrowSpeed(), dif) * BANANABOUNCEFRICTION);
    }

    banana.positionBS(p);
    if (banana.getState() == 3) {
        banana.setState(2);
    }

    if (hitEffect != 2 && (friendlyFire || myPlayers[j].getTeamId() < 0 || myPlayers[j].getTeamId() != myPlayers[banana.getCarriedBy()].getTeamId())) {
        if (myPlayers[j].getStunTime() <= 0 && hitEffect != 2 && !myPlayers[j].getDead()) {
            myBaskets[myPlayers[i].getBaskId()].addScore(extraHitPoints);
            if (hitEffect == 1) {
                printf("%d killed %d\n", i, j);
                myPlayers[j].die();
            }
        }
        printf("player %d was hit by %d\n", j, i);
        // XXX send URHIT message
        myPlayers[j].setStunTime(playerStunTime);
    }
    // XXX playsound banana
}

float bananaBasketColTime(int& basketId)
{
    int i = 0; // local player
    Banana& banana = myPlayers[i].getBanana();
    float result = 2.0;
    if (banana.getState() > 0) {
        for (int j = 0; j < nBaskets; j++) {
            Vector3f endp = banana.getColBS().pos;
            Vector3f beginp = banana.getOldPos();
            Vector3f dif = endp - beginp;
            BS colbsb = myBaskets[j].getColBS();
            BS colbsa = banana.getColBS();

            float a = dif.squaredNorm();;
            float b = 2.0 * (dif.x() * (beginp.x() - colbsb.pos.x()) + dif.y() * (beginp.y() - colbsb.pos.y()) + dif.z() * (beginp.z() - colbsb.pos.z()));
            float c = (beginp - colbsb.pos).squaredNorm() - sqr(colbsb.radius + colbsa.radius);
            float d = sqr(b) - 4.0 * a * c;

            if (d >= 0 && a > 0) {
                d = sqrt(d);
                float t1 = (-b + d) / (2.0 * a);
                float t2 = (-b - d) / (2.0 * a);
                float t;
                if (t1 < 0) t = t2;
                else if (t2 < 0) t = t1;
                else t = qMin(t1, t2);
                if (t < result) {
                    result = t;
                    basketId = j;
                }
            }
        }
    }
    return result;
}

void bananaBasketColReact(float t, int basketId)
{
    int i = 0; // local player
    Banana& banana = myPlayers[i].getBanana();
    int j = basketId;

    Vector3f endp = banana.getColBS().pos;
    Vector3f beginp = banana.getOldPos();
    Vector3f dif = endp - beginp;
    BS colbsb = myBaskets[j].getColBS();
    BS colbsa = banana.getColBS();

    Vector3f p = lerp(beginp, endp, t);

    dif = (p - colbsb.pos).normalized();
    p = (colbsb.pos + dif * (colbsb.radius + colbsa.radius + COLOFFSET2));

    banana.setOldPos(p);

    float dt;
    if (banana.getThrowSpeed().x() != 0) dt = (endp.x() - beginp.x()) / banana.getThrowSpeed().x();
    else if (banana.getThrowSpeed().y() != 0) dt = (endp.y() - beginp.y()) / banana.getThrowSpeed().y();
    else if (banana.getThrowSpeed().z() != 0) dt = (endp.z() - beginp.z()) / banana.getThrowSpeed().z();
    else dt = 0;
    banana.setThrowSpeed(reflect(banana.getThrowSpeed(), dif) * BANANABOUNCEFRICTION);

    // XXX playsound banana

    p += banana.getThrowSpeed() * (dt * (1.0 - t));

    banana.positionBS(p);

    if (banana.getState() == 3) banana.setState(2);
}

float bananaGroundColTime()
{
    int i = 0; // local player
    Banana& banana = myPlayers[i].getBanana();
    float result = 2.0;

    if (banana.getState() > 0) {
        Vector3f endp = banana.getColBS().pos;
        Vector3f beginp = banana.getOldPos();

        int colStep = 0;
        float stepSize = 1.0 / BANANAGROUNDSTEPS;
        bool col = false;

        while (colStep <= BANANAGROUNDSTEPS && !col) {
            Vector3f p = lerp(beginp, endp, colStep * stepSize);
            float h = myTer.getGroundHeight(p);
            if (p.y() < h + banana.getColBS().radius) col = true;
            else colStep++;
        }

        if (col) {
            float stepa = (colStep - 1) * stepSize;
            float stepb = colStep * stepSize;
            for (int j = 0; j < BANANAGROUNDITERS; j++) {
                float stepc = (stepa + stepb) * 0.5;
                Vector3f p = lerp(beginp, endp, stepc);
                float h = myTer.getGroundHeight(p);
                if (p.y() < h + banana.getColBS().radius) {
                    stepb = stepc;
                } else {
                    stepa = stepc;
                }
            }
            result = qMax(stepb, 0.0f);
        }
    }
    return result;
}

void bananaGroundColReact(float t)
{
    int i = 0; // local player
    Banana& banana = myPlayers[i].getBanana();

    Vector3f endp = banana.getColBS().pos;
    Vector3f beginp = banana.getOldPos();

    Vector3f p = lerp(beginp, endp, t);
    p.y() = myTer.getGroundHeight(p) + banana.getColBS().radius + COLOFFSET2;
    banana.setOldPos(p);

    Plane bouncePlane = myTer.getGroundPlane(p);
    Vector3f ref = Vector3f(bouncePlane.a, bouncePlane.b, bouncePlane.c);
    float dt;
    if (banana.getThrowSpeed().x() != 0) dt = (endp.x() - beginp.x()) / banana.getThrowSpeed().x();
    else if (banana.getThrowSpeed().y() != 0) dt = (endp.y() - beginp.y()) / banana.getThrowSpeed().y();
    else if (banana.getThrowSpeed().z() != 0) dt = (endp.z() - beginp.z()) / banana.getThrowSpeed().z();
    else dt = 0;
    banana.setThrowSpeed(reflect(banana.getThrowSpeed(), ref) * BANANABOUNCEFRICTION);

    p += banana.getThrowSpeed() * (dt * (1.0 - t)) ;
    p.y() = qMax(p.y(), myTer.getGroundHeight(p) + banana.getColBS().radius + COLOFFSET2);

    // XXX playsound banana

    banana.positionBS(p);
    if (banana.getState() == 3) banana.setState(2);
}
