/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "Fruit.h"

// FruitWars headers
#include "Globals.h"
#include "Instance.h"
#include "Particles.h"
#include "Math3D.h"

// KDE headers
#include <KGlobal>
#include <KStandardDirs>

USING_PART_OF_NAMESPACE_EIGEN

static const int FRUITKINDS = 5;
static const float FRUITAPPLESCALE = 0.01;
static const float FRUITAPPLEBSRAD = 90.0;
static const float FRUITAPPLEBSOFFSET[] = {0.0, 0.0, 64.0};
static const float FRUITAPPLECOLRADIUS = 70.0;
static const float FRUITAPPLEHEIGHT = 60.0;
static const float FRUITORANGESCALE = 0.06;
static const float FRUITORANGEBSRAD = 13.0;
static const float FRUITORANGEBSOFFSET[] = {0.0, 0.0, 0.0};
static const float FRUITORANGECOLRADIUS = 10.0;
static const float FRUITORANGEHEIGHT = 9.0;
static const float FRUITLEMONSCALE = 0.02;
static const float FRUITLEMONBSRAD = 45.0;
static const float FRUITLEMONBSOFFSET[] = {0.0, 0.0, 10.0};
static const float FRUITLEMONCOLRADIUS = 38.0;
static const float FRUITLEMONHEIGHT = 24.0;
static const float FRUITANANASSCALE = 0.06;
static const float FRUITANANASBSRAD = 30.0;
static const float FRUITANANASBSOFFSET[] = {3.0, 15.0, 3.0};
static const float FRUITANANASCOLRADIUS = 12.0;
static const float FRUITANANASHEIGHT = 9.0;
static const float FRUITSTRAWSCALE = 0.01;
static const float FRUITSTRAWBSRAD = 70.0;
static const float FRUITSTRAWBSOFFSET[] = {-15.0, 0.0, 0.0};
static const float FRUITSTRAWCOLRADIUS = 50.0;
static const float FRUITSTRAWHEIGHT = 40.0;

static const float FRUITSPAWNHEIGHT = 70.0;
static const float FRUITMAXSPEED = 20.0;
static const float FRUITWARPSIZE = 4.0;
static const int FRUITBASKETCAP = 8;
static const float FRUITBASKETBSRAD = 35.0;
static const float FRUITBASKETBSOFFSET[] = {0.0, 24.0, 0.0};
static const float FRUITBASKETCOLRADIUS = 27.0;
static const float FRUITBASKETWARPDIF = 15.0;
static const float FRUITBASKETWARPSIZE = 1.8;
static const float FRUITVANISHSPEED = 2.0;
static const int STARSHOWCOUNT = 32;

static const char* FRUITFILES[] = {"data/apple.ogl", "data/orange.ogl", "data/lemon.ogl", "data/pineapple.ogl", "data/strawberry.ogl"};
static const char* FRUITNAMES[] = {"Apple", "Orange", "Lemon", "Pineapple", "Strawberry"};

Fruit::Fruit(Model3D* model, FruitType kind, int owner)
    : m_id(0)
    , m_kind(kind)
    , m_carriedBy(-1)
    , m_grav(GLOBALGRAV)
    , m_fallSpeed(0.0)
    , m_moveSpeed(0.0, 0.0, 0.0)
    , m_lastTime(time())
    , m_dt(0.0)
    , m_warp(Vector3f(0, 0, 0), FRUITWARPSIZE)
    , m_groundColOn(true)
    , m_vanishing(false)
    , m_vanish(0.0)
    , m_owner(owner)
{
    m_inst = Instance::load(model);

    switch (kind) {
        case ftApple: {
            m_inst->setBS(BS(Vector3f(FRUITAPPLEBSOFFSET[0], FRUITAPPLEBSOFFSET[1], FRUITAPPLEBSOFFSET[2]), FRUITAPPLEBSRAD));
            m_inst->scaleObj(Vector3f(FRUITAPPLESCALE, FRUITAPPLESCALE, FRUITAPPLESCALE));
            m_colBS = m_inst->getBS();
            m_colBS.radius = FRUITAPPLECOLRADIUS;
            m_midHeight = FRUITAPPLEHEIGHT;
            m_score = fruitAppleScore;
            break;
        }
        case ftOrange: {
            m_inst->setBS(BS(Vector3f(FRUITORANGEBSOFFSET[0], FRUITORANGEBSOFFSET[1], FRUITORANGEBSOFFSET[2]), FRUITORANGEBSRAD));
            m_inst->scaleObj(Vector3f(FRUITORANGESCALE, FRUITORANGESCALE, FRUITORANGESCALE));
            m_colBS = m_inst->getBS();
            m_colBS.radius = FRUITORANGECOLRADIUS;
            m_midHeight = FRUITORANGEHEIGHT;
            m_score = fruitOrangeScore;
            break;
        }
        case ftLemon: {
            m_inst->setBS(BS(Vector3f(FRUITLEMONBSOFFSET[0], FRUITLEMONBSOFFSET[1], FRUITLEMONBSOFFSET[2]), FRUITLEMONBSRAD));
            m_inst->scaleObj(Vector3f(FRUITLEMONSCALE, FRUITLEMONSCALE, FRUITLEMONSCALE));
            m_colBS = m_inst->getBS();
            m_colBS.radius = FRUITLEMONCOLRADIUS;
            m_midHeight = FRUITLEMONHEIGHT;
            m_score = fruitLemonScore;
            break;
        }
        case ftAnanas: {
            m_inst->setBS(BS(Vector3f(FRUITANANASBSOFFSET[0], FRUITANANASBSOFFSET[1], FRUITANANASBSOFFSET[2]), FRUITANANASBSRAD));
            m_inst->scaleObj(Vector3f(FRUITANANASSCALE, FRUITANANASSCALE, FRUITANANASSCALE));
            m_colBS = m_inst->getBS();
            m_colBS.radius = FRUITANANASCOLRADIUS;
            m_midHeight = FRUITANANASHEIGHT;
            m_score = fruitAnanasScore;
            break;
        }
        case ftStraw: {
            m_inst->setBS(BS(Vector3f(FRUITSTRAWBSOFFSET[0], FRUITSTRAWBSOFFSET[1], FRUITSTRAWBSOFFSET[2]), FRUITSTRAWBSRAD));
            m_inst->scaleObj(Vector3f(FRUITSTRAWSCALE, FRUITSTRAWSCALE, FRUITSTRAWSCALE));
            m_colBS = m_inst->getBS();
            m_colBS.radius = FRUITSTRAWCOLRADIUS;
            m_midHeight = FRUITSTRAWHEIGHT;
            m_score = fruitStrawScore;
            break;
        }
    }

    m_warp.setState(3);
    m_oldPos = m_colBS.pos;
}

Fruit::Fruit()
    : m_warp(Vector3f(0, 0, 0), FRUITWARPSIZE)
{
}

Fruit::~Fruit()
{
}

void Fruit::update()
{
    m_dt = (time() - m_lastTime) / 1000.0;
    m_lastTime = time();

    m_fallSpeed -= m_grav * m_dt;
    if (m_fallSpeed < -FRUITMAXSPEED) m_fallSpeed = -FRUITMAXSPEED;

    if (m_vanishing) {
        BS theBS = colBS();
        if (m_warp.getState() == 3) {
            m_warp.setState(0);
            m_warp.setClipSide(-1);
            m_warp.setSize(theBS.radius * 2.0);
            m_warp.setPos(theBS.pos + Vector3f(0.0, theBS.radius, 0.0));
        }
        if (m_warp.getState() == 1) {
            m_vanish = m_vanish + FRUITVANISHSPEED * m_dt;
            m_warp.setPos(theBS.pos + Vector3f(0.0, m_vanish - theBS.radius, 0.0));
            if (m_vanish < 2.0 * theBS.radius) {
                m_warp.setWaitTime(m_dt + 0.001);
            }
        }
    } else {
        m_vanish = 0.0;
    }

    if (m_warp.getState() < 3) m_warp.update(m_dt);

    //if owner is local player: XXX
    position(position() + Vector3f(m_moveSpeed.x() * m_dt, m_fallSpeed * m_dt, m_moveSpeed.z() * m_dt));
}

void Fruit::update2()
{
    m_dt = (time() - m_lastTime) / 1000.0;
    m_lastTime = time();

    if (m_warp.getState() < 3) m_warp.update(m_dt);
}

void Fruit::render()
{
    if (m_warp.getState() < 3 && m_warp.getClipSide() != 0) {
        glEnable(GL_CLIP_PLANE0);
        GLdouble eq[4] = {0.0, -1.0 * m_warp.getClipSide(), 0.0, m_warp.getPos().y() * m_warp.getClipSide()};
        glClipPlane(GL_CLIP_PLANE0, eq);
        m_inst->render();
        glDisable(GL_CLIP_PLANE0);
    } else {
        m_inst->render();
    }
}

void Fruit::startWarp()
{
    m_warp.setState(0);
    m_warp.setPos(colBS().pos - Vector3f(0.0, WARPDIF, 0.0));
    m_warp.setClipSide(1);
    m_warp.setScaleSpeed(WARPSCALESPEED);
    m_warp.setSize(FRUITWARPSIZE);
}

void Fruit::setBasketWarp(int bid)
{
    if (m_warp.getState() == 3) m_warp.setState(0);
    m_warp.setPos(myBaskets[bid].getColBS().pos - Vector3f(0.0, FRUITBASKETWARPDIF * FRUITBASKETSCALE, 0.0));
    m_warp.setWaitTime(WARPWAITTIME);
    m_warp.setClipSide(-1);
    m_warp.setSize(FRUITBASKETWARPSIZE);
    if (m_warp.getState() == 0) {
        float d = sqr(m_fallSpeed) - 4.0 * (position().y() - m_warp.getPos().y()) * 0.5 * m_grav;
        if (d >= 0) {
            d = sqrt(d);
            float t1 = (-m_fallSpeed + d) / m_grav;
            float t2 = (-m_fallSpeed - d) / m_grav;
            float t;
            if (t1 < 0) t = t2;
            else if (t2 < 0) t = t1;
            else t = qMin(t1, t2);
            if (t > 0) {
                m_warp.setScaleSpeed(qMax((float)(1.0 - m_warp.getScale()) / t, WARPSCALESPEED));
            }
        }
    } else {
        m_warp.setScaleSpeed(WARPSCALESPEED);
    }
}

void Fruit::disableWarpClip()
{
    if (m_warp.getClipSide() < 0 && !m_vanishing) m_warp.setClipSide(0);
}

void Fruit::warpRender()
{
    if (m_warp.getState() < 3 && m_warp.getPos().y() > SEAHEIGHT) m_warp.render();
}

void Fruit::warpRender2()
{
    if (m_warp.getState() < 3 && m_warp.getPos().y() <= SEAHEIGHT) m_warp.render();
}

void Fruit::scale(const Vector3f& scale)
{
    m_inst->scaleObj(scale);
}

void Fruit::rotateX(float ang)
{
    m_inst->rotateX(ang);
}

void Fruit::rotateY(float ang)
{
    m_inst->rotateY(ang);
}

void Fruit::rotateZ(float ang)
{
    m_inst->rotateZ(ang);
}

void Fruit::position(const Vector3f& pos)
{
    m_inst->position(pos);
}

void Fruit::positionBS(const Vector3f& pos)
{
    m_inst->position(pos - (m_inst->transform().linear() * m_colBS.pos)); //XXX
}

Vector3f Fruit::scale()
{
    return m_inst->getScale();
}

Vector3f Fruit::rotation()
{
    return m_inst->getRot();
}

Vector3f Fruit::position()
{
    return m_inst->getPos();
}

BS Fruit::colBS()
{
    return BS(m_inst->transform() * m_colBS.pos, m_colBS.radius * scale().x());
}

float Fruit::midHeight()
{
    return m_midHeight * scale().y();
}

bool Fruit::fruitMoved()
{
    return (colBS().pos - m_oldPos).norm() >= 0.01;
}

FruitBasket::FruitBasket(Model3D* model)
    : inst(Instance::load(model))
    , playId(0)
    , teamId(-1)
    , midHeight(FRUITBASKETHEIGHT)
    , score(0)
    , starShow(Vector3f(0, 0, 0), FRUITBASKETWARPSIZE - STARSHOWSIZE, STARSHOWCOUNT)
{
    for (int i = 0; i < FRUITKINDS; i++) contain.append(0);
    inst->setBS(BS(Vector3f(FRUITBASKETBSOFFSET[0], FRUITBASKETBSOFFSET[1], FRUITBASKETBSOFFSET[2]), FRUITBASKETBSRAD));
    inst->scaleObj(Vector3f(FRUITBASKETSCALE, FRUITBASKETSCALE, FRUITBASKETSCALE));
    colBS = inst->getBS();
    colBS.radius = FRUITBASKETCOLRADIUS;
}

FruitBasket::~FruitBasket()
{
}

void FruitBasket::addFruit(FruitType fid)
{
    contain[fid]++;
    bool diverse = true;
    for (int i = 0; i < FRUITKINDS; i++) {
        if (contain[i] == 0) diverse = false;
    }
    if (diverse) {
        score += diversityBonus;
        for (int i = 0; i < FRUITKINDS; i++) {
            contain[i]--;
        }
    }
    printf("Score: %d\n", score);
}

void FruitBasket::render()
{
    inst->render();
}

void FruitBasket::scale(const Vector3f& scale)
{
    inst->scaleObj(scale);
}

void FruitBasket::rotate(const Vector3f& rot)
{
    inst->rotate(rot);
}

void FruitBasket::rotateX(float ang)
{
    inst->rotateX(ang);
}

void FruitBasket::rotateY(float ang)
{
    inst->rotateY(ang);
}

void FruitBasket::rotateZ(float ang)
{
    inst->rotateZ(ang);
}

void FruitBasket::position(const Vector3f& pos)
{
    inst->position(pos);
    starShow.setSpawnPos(getColBS().pos - Vector3f(0.0, FRUITBASKETWARPDIF * FRUITBASKETSCALE, 0.0));
}

void FruitBasket::positionBS(const Vector3f& pos)
{
    inst->position(pos - (inst->transform().matrix().corner<3, 3>(Eigen::TopLeft) * colBS.pos));
}

int FruitBasket::getFruitID(int bid)
{
    return contain[bid];
}

Vector3f FruitBasket::getScale()
{
    return inst->getScale();
}

Vector3f FruitBasket::getRot()
{
    return inst->getRot();
}

Vector3f FruitBasket::getPos()
{
    return inst->getPos();
}

BS FruitBasket::getColBS()
{
    return BS(inst->transform() * colBS.pos, colBS.radius * getScale().x());
}

float FruitBasket::getMidHeight()
{
    return midHeight * getScale().y();
}

int FruitBasket::getContain(int id)
{
    return contain[id];
}

void FruitBasket::setContain(int id, int value)
{
    contain[id] = value;
}

FruitSpawner::FruitSpawner(int minCount, int maxCount, float minWait, float maxWait)
    : minCount(minCount)
    , maxCount(maxCount)
    , minWait(minWait)
    , maxWait(maxWait)
    , inGameCount(0)
    , lastTime(time())
    , dt(0.0)
{
    for (int i = 0; i < FRUITKINDS; i++) {
        FruitType kind = (FruitType) i;
        fruitModel.append(loadModel(KGlobal::dirs()->findResource("appdata", FRUITFILES[i])));
        for (int j = 0; j < maxCount; j++) {
            int id = i * maxCount + j;
            fruit.append(Fruit(fruitModel[i], kind, 0));
            fruit[id].setId(id);
        }
    }

    // Calculate the Prime
    thePrime = maxCount * FRUITKINDS - 1;
    bool isPrime = false;
    while (!isPrime) {
        isPrime = true;
        thePrime++;
        for (int i = 2; i < thePrime; i++) {
            if (thePrime % i == 0) isPrime = false;
        }
    }

    for (int i = 0; i < thePrime; i++) {
        fruitData.append(FruitData());
        fruitData[i].inGame = false;
        fruitData[i].respawning = false;
    }
}

FruitSpawner::FruitSpawner()
{
}

FruitSpawner::~FruitSpawner()
{
}

void FruitSpawner::render()
{
    for (int i = 0; i < maxCount * FRUITKINDS; i++) {
        if (fruitData[i].inGame) {
            fruit[i].render();
        }
    }
}

void FruitSpawner::warpRender()
{
    for (int i = 0; i < maxCount * FRUITKINDS; i++) {
        if (fruitData[i].inGame || fruitData[i].respawning) {
            fruit[i].warpRender();
        }
    }
}

void FruitSpawner::warpRender2()
{
    for (int i = 0; i < maxCount * FRUITKINDS; i++) {
        if (fruitData[i].inGame || fruitData[i].respawning) {
            fruit[i].warpRender2();
        }
    }
}

void FruitSpawner::update()
{
    dt = time() - lastTime;
    lastTime = time();

    // if Host
    if (inGameCount < minCount || (inGameCount < maxCount && frandom() < dt / (maxWait - ((float)(maxCount - inGameCount) / (maxCount - minCount + 1)) * (maxWait - minWait)))) {
        spawnFruit();
    }

    for (int i = 0; i < maxCount * FRUITKINDS; i++) {
        if (fruitData[i].inGame) {
            fruit[i].update();
        } else {
            fruit[i].update2();
            if (fruit[i].warp().getState() == 3 && fruitData[i].respawning) {
                fruitData[i].respawning = false;
                inGameCount--;
            }
        }
    }
}

void FruitSpawner::respawnFruit(int fid)
{
    fruitData[fid].inGame = false;
    fruitData[fid].respawning = true;
    fruit[fid].positionBS(Vector3f(-100.0, 100.0, -100.0));
    fruit[fid].setMoveSpeed(Vector3f(0, 0, 0));
    fruit[fid].setFallSpeed(0.0);
    fruit[fid].setVanish(0.0);
}

void FruitSpawner::spawnFruit()
{
    inGameCount++;

    int theId = random() % (maxCount * FRUITKINDS);
    int theStep = qMin(1L, random() % (maxCount * FRUITKINDS)); // XXX was eerst max???
    int check = 0;
    while ( ( (oneFruitType && ((int)theFruitType != theId % maxCount)) ||
        fruitData[theId].inGame || fruitData[theId].respawning || (theId >= maxCount * FRUITKINDS) )
        && check < 10000 ) {
        theId = (theId + theStep) % thePrime;
        check++;
    }

    if (check == 10000) {
        inGameCount--;
        return;
    }

    float h = SEAHEIGHT - 1;
    check = 0;
    Vector3f thePos;
    while (h < SEAHEIGHT && check < 10000) {
        float ang = frandom() * 2.0 * M_PI;
        float rad = frandom() * spawnRadius;
        thePos = Vector3f(sin(ang) * rad, FRUITSPAWNHEIGHT, cos(ang) * rad);
        h = myTer.getGroundHeight(Vector2f(thePos.x(), thePos.z()));
        check++;
    }

    fruitData[theId].inGame = true;
    fruit[theId].setRemoteId(theId);
    fruit[theId].position(thePos);
    fruit[theId].startWarp();
    fruit[theId].setVanishing(false);
//	fruit[theId].setOwner( getLocalPlayerId );

    // XXX send SPAWNFRUIT message
}

void FruitSpawner::spawnFruit(int /* type */, int owner, const Vector3f& pos, int id)
{
    inGameCount++;

    fruitData[id].inGame = true;
    fruit[id].position(pos);
    fruit[id].startWarp();
    fruit[id].setVanishing(false);
    fruit[id].setOwner(owner);
    fruit[id].setRemoteId(id);
}

Fruit& FruitSpawner::getFruit(int n)
{
    return fruit[n];
}

Model3D* FruitSpawner::getFruitModel(int n)
{
    return fruitModel[n];
}

FruitData FruitSpawner::getFruitData(int n)
{
    return fruitData[n];
}

void FruitSpawner::ownFruits(const HumanPlayer& owner)
{
    (void) owner;
    // XXX network stuff
}




