/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "Math3D.h"

#include <Eigen/LU>

#include <QtGlobal>

USING_PART_OF_NAMESPACE_EIGEN
using Eigen::Quaternionf;
using Eigen::Transform3f;
using Eigen::Translation3f;

Vector3f lerp(const Vector3f &a, const Vector3f &b, float u)
{
    return Vector3f(a.x() * (1.00 - u) + b.x() * u,
            a.y() * (1.00 - u) + b.y() * u,
            a.z() * (1.00 - u) + b.z() * u);
}

Vector3f reflect(const Vector3f &a, const Vector3f &b)
{
    return a - (b * 2 * a.dot(b));
}

void extractScaleRotation(const Transform3f& tr, Vector3f &scale, Quaternionf &rot)
{
    const float *data = tr.data();
    scale.x() = sqrt(sqr(data[0]) + sqr(data[1]) + sqr(data[2]));
    scale.y() = sqrt(sqr(data[4]) + sqr(data[5]) + sqr(data[6]));
    scale.z() = sqrt(sqr(data[8]) + sqr(data[9]) + sqr(data[10]));

    Transform3f tb = tr;
    float *b = tb.data();
    b[0] = data[0] / scale.x();
    b[1] = data[1] / scale.x();
    b[2] = data[2] / scale.x();
    b[4] = data[4] / scale.y();
    b[5] = data[5] / scale.y();
    b[6] = data[6] / scale.y();
    b[8] = data[8] / scale.z();
    b[9] = data[9] / scale.z();
    b[10] = data[10] / scale.z();

    float t = 1.00 + b[0] + b[5] + b[10];
    if (t > 0.0001) {
        float q = 2.0 * sqrt(t);
        rot.x() = (b[6] - b[9]) / q;
        rot.y() = (b[8] - b[2]) / q;
        rot.z() = (b[1] - b[4]) / q;
        rot.w() = 0.25 * q;
    } else {
        if (b[0] > b[5] && b[0] > b[10]) {
            t = sqrt(1.0 + b[0] - b[5] - b[10]) * 2.0;
            rot.x() = 0.25 * t;
            rot.y() = (b[4] + b[1]) / t;
            rot.z() = (b[2] + b[8]) / t;
            rot.w() = (b[6] - b[9]) / t;
        } else if (b[5] > b[10]) {
            t = sqrt(1.0 + b[5] - b[0] - b[10]) * 2.0;
            rot.y() = 0.25 * t;
            rot.x() = (b[4] + b[1]) / t;
            rot.w() = (b[8] - b[2]) / t;
            rot.z() = (b[6] + b[9]) / t;
        } else {
            t = sqrt(1.0 + b[10] - b[0] - b[5]) * 2.0;
            rot.z() = 0.25 * t;
            rot.w() = (b[1] - b[4]) / t;
            rot.x() = (b[2] + b[8]) / t;
            rot.y() = (b[6] + b[9]) / t;
        }
    }
}

void extractScaleRotation(const Transform3f& tr, Vector3f &scale, Vector3f &rot)
{
    const float* data = tr.data();

    scale.x() = sqrt(sqr(data[0]) + sqr(data[1]) + sqr(data[2]));
    scale.y() = sqrt(sqr(data[4]) + sqr(data[5]) + sqr(data[6]));
    scale.z() = sqrt(sqr(data[8]) + sqr(data[9]) + sqr(data[10]));

    Transform3f tb = tr;
    float *b = tb.data();
    b[0] = data[0] / scale.x();
    b[1] = data[1] / scale.x();
    b[2] = data[2] / scale.x();
    b[4] = data[4] / scale.y();
    b[5] = data[5] / scale.y();
    b[6] = data[6] / scale.y();
    b[8] = data[8] / scale.z();
    b[9] = data[9] / scale.z();
    b[10] = data[10] / scale.z();

    if (qAbs(b[8]) <= 0.001 && qAbs(b[10]) <= 0.001) {
        rot.y() = 0;
        rot.x() = M_PI / 2.0;
        rot.z() = atan2(-b[4], b[6]);
    } else {
        rot.y() = atan2(-b[8], b[10]);
        rot.z() = atan2(b[1], b[5]);
        rot.x() = asin(-b[9]);
    }
}

Quaternionf extractRotation(const Transform3f& t)
{
    Vector3f scale;
    Quaternionf rot;
    extractScaleRotation(t, scale, rot);
    return rot;
}

Transform3f viewInverse(const Transform3f& t)
{
    return Transform3f(t.linear().transpose()) * Translation3f(t.translation() * -1);
}

Matrix3f getNormMat(const Transform3f& t)
{
    Matrix3f rm = Matrix3f::Zero();
    const float* data = t.data();
    float* r = rm.data();
    if ((sqr(data[0]) + sqr(data[1]) + sqr(data[2])) <= 0.0000001)
    {
        r[0] = 1.00;
        return rm;
    }
    if ((sqr(data[4]) + sqr(data[5]) + sqr(data[6])) <= 0.0000001)
    {
        r[4] = 1.00;
        return rm;
    }
    if ((sqr(data[8]) + sqr(data[9]) + sqr(data[10])) <= 0.0000001)
    {
        r[8] = 1.00;
        return rm;
    }

    float invDet = 1.00 / t.linear().determinant();

    r[0] = (data[10] * data[5] - data[6] * data[9]) * invDet;
    r[1] = (data[6] * data[8] - data[10] * data[4]) * invDet;
    r[2] = (data[4] * data[9] - data[5] * data[8]) * invDet;
    r[3] = (data[2] * data[9] - data[1] * data[10]) * invDet;
    r[4] = (data[10] * data[0] - data[2] * data[8]) * invDet;
    r[5] = (data[1] * data[8] - data[0] * data[9]) * invDet;
    r[6] = (data[1] * data[6] - data[2] * data[5]) * invDet;
    r[7] = (data[2] * data[4] - data[6] * data[0]) * invDet;
    r[8] = (data[0] * data[5] - data[1] * data[4]) * invDet;
    return rm;
}
