/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
// glew.h must be before any other opengl related include files
#include <GL/glew.h>

#include "Model3D.h"

// FruitWars headers
#include "Globals.h"
#include "Math3D.h"
#include "Textures.h"

// KDE headers
#include <KGlobal>
#include <KStandardDirs>

USING_PART_OF_NAMESPACE_EIGEN
using Eigen::Quaternionf;
using Eigen::Scaling3f;
using Eigen::Transform3f;
using Eigen::Translation3f;

#ifdef DPRINTF
#undef DPRINTF
#endif
//#define DPRINTF(args...) fprintf(stderr, args)
#define DPRINTF(args...)
GLuint cartoonTex;

Key::Key()
{
}

Frame::Frame()
    : blendSum(0.0), priority(0)
{
}

Model3D* loadModel(const QString& fileName)
{
    QFile file(fileName);
    Header header;

    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) return 0;
    QTextStream stream(&file);
    stream >> header.identifier >> header.version >> header.ogltype;
    file.close();
    Model3D *model = NULL;
    if (header.identifier == "OGL" && header.version == "0100") {
        if (header.ogltype == "STATIC") {
            model = new StaticModel(header, fileName);
            model->load();
        } else if (header.ogltype == "FRAME") {
            model = new FrameModel(header, fileName);
            model->load();
        } else if (header.ogltype == "ANIM") {
            model = new AnimModel(header, fileName);
            model->load();
        } else if (header.ogltype == "BONE") {
            model = new BoneModel(header, fileName);
            model->load();
        }
    }
    return model;
}

Model3D::Model3D()
{
}

Model3D::~Model3D()
{
}

Model3D::Model3D(Header header, const QString& fileName)
    : oglFile(fileName),
    header(header)
{
    oglFile.open(QIODevice::ReadOnly | QIODevice::Text | QIODevice::Unbuffered);
    oglStream.setDevice(&oglFile);
}

void Model3D::load()
{
}

void Model3D::render()
{
}

void Model3D::setFrame(unsigned int, float)
{
}

void Model3D::setAnimSetBlend(unsigned int, float, float, int)
{
}

void Model3D::setAnimBlend(unsigned int, unsigned int, float)
{
}

void Model3D::calcAnim()
{
}

void Model3D::setTextures(const QVector<int>&)
{
}

Header Model3D::getHeader()
{
    return header;
}

void Model3D::skipInputLine()
{
    QString tmp = oglStream.readLine();
    DPRINTF("Skipped line '%s'\n", qPrintable(tmp));
}

int Model3D::readInt()
{
    int i;
    oglStream >> i;
    DPRINTF("Read int %d\n", i);
    return i;
}

float Model3D::readFloat()
{
    float f;
    oglStream >> f;
    DPRINTF("Read float %f\n", f);
    return f;
}

QString Model3D::readLine()
{
    QString s = oglStream.readLine();
    DPRINTF("Not read line '%s'\n", qPrintable(s));
    s = oglStream.readLine().trimmed();
    DPRINTF("Read line '%s'\n", qPrintable(s));
    return s;
}

StaticModel::StaticModel(Header header, const QString& fileName)
    : Model3D(header, fileName)
{
}

StaticModel::~StaticModel()
{
}

void StaticModel::load()
{
    int num;

    // Skip Header
    skipInputLine();

    // Read Materials
    num = readInt();
    materials.resize(num);
    for (int i = 0; i < num; i++) {
        readMatBlock(i);
    }

    // Read Meshes
    num = readInt();
    meshes.resize(num);
    for (int i = 0; i < num; i++) {
        readMeshBlock(i);
    }

    loadTextures();

    oglFile.close();
}

void StaticModel::setTextures(const QVector<int> &tex)
{
    for (int i = 0; i < tex.size(); i++) {
        materials[i].textureID = tex[i];
        materials[i].textured = (tex[i] != -1);
    }
}

void StaticModel::readMatBlock(int i)
{
    int p;
    materials[i].faceColor.r = readFloat();
    materials[i].faceColor.g = readFloat();
    materials[i].faceColor.b = readFloat();
    materials[i].faceColor.a = readFloat();
    materials[i].specular.r = readFloat();
    materials[i].specular.g = readFloat();
    materials[i].specular.b = readFloat();
    materials[i].specular.a = readFloat();
    materials[i].emissive.r = readFloat();
    materials[i].emissive.g = readFloat();
    materials[i].emissive.b = readFloat();
    materials[i].emissive.a = readFloat();
    materials[i].shininess = readFloat();
    p = readInt();
    materials[i].textured = (p == 1);
    if (materials[i].textured) {
        materials[i].texFileName = readLine();
    }
}

void StaticModel::readMeshBlock(int n)
{
    int fvf = readInt();
    meshes[n].hasNormals = fvf & 1;
    meshes[n].hasColors = fvf & 2;
    meshes[n].faceWraps = fvf & 4;
    meshes[n].hasTexCoords = fvf & 8;
    meshes[n].hasMaterial = fvf & 16;
    meshes[n].inFrame = fvf & 32;

    int num;
    // Read Vertices
    num = readInt();
    meshes[n].vertices.resize(num);
    for (int i = 0; i < num; i++) {
        meshes[n].vertices[i].pos.x() = readFloat();
        meshes[n].vertices[i].pos.y() = readFloat();
        meshes[n].vertices[i].pos.z() = readFloat();
    }

    // Read Faces
    num = readInt();
    meshes[n].faces.resize(num);
    for (int i = 0; i < num; i++) {
        meshes[n].faces[i].indices[0] = readInt();
        meshes[n].faces[i].indices[1] = readInt();
        meshes[n].faces[i].indices[2] = readInt();
    }

    // Read Normals
    if (meshes[n].hasNormals) {
        num = readInt();
        meshes[n].normals.resize(num);
        for (int i = 0; i < num; i++) {
            meshes[n].normals[i].x() = readFloat();
            meshes[n].normals[i].y() = readFloat();
            meshes[n].normals[i].z() = readFloat();
        }
        readInt();
        for (int i = 0; i < meshes[n].faces.size(); i++) {
            meshes[n].faces[i].normIndices[0] = readInt();
            meshes[n].faces[i].normIndices[1] = readInt();
            meshes[n].faces[i].normIndices[2] = readInt();
        }
        // Assign Vertex normals
        for (int i = 0; i < meshes[n].faces.size(); i++) {
            for (int j = 0; j < 3; j++) {
                int vid = meshes[n].faces[i].indices[j];
                int nid = meshes[n].faces[i].normIndices[j];
                meshes[n].vertices[vid].norm = meshes[n].normals[nid];
            }
        }
    } else {
        autoGenerateNormals(n);
    }

    // Read Colors
    if (meshes[n].hasColors) {
        readInt();
        for (int i = 0; i < meshes[n].vertices.size(); i++) {
            meshes[n].vertices[i].color.r = readFloat();
            meshes[n].vertices[i].color.g = readFloat();
            meshes[n].vertices[i].color.b = readFloat();
            meshes[n].vertices[i].color.a = readFloat();
        }
    }

    // Read Tex Coords
    if (meshes[n].hasTexCoords) {
        readInt();
        for (int i = 0; i < meshes[n].vertices.size(); i++) {
            meshes[n].vertices[i].texCoord[0] = readFloat();
            meshes[n].vertices[i].texCoord[1] = readFloat();
        }
    }

    // Read Mat ID
    if (meshes[n].hasMaterial) {
        meshes[n].matID = readInt();
    } else {
        meshes[n].matID = 0;
    }
}

void StaticModel::autoGenerateNormals(int n)
{
    meshes[n].normals.resize(meshes[n].vertices.size());
    for (int i = 0; i < meshes[n].normals.size(); i++) {
        meshes[n].normals[i] = Vector3f(0.0, 0.0, 0.0);
    }

    for (int i = 0; i < meshes[n].faces.size(); i++) {
        Vector3f norm = (meshes[n].vertices[meshes[n].faces[i].indices[1]].pos -
                        meshes[n].vertices[meshes[n].faces[i].indices[0]].pos).cross(
                    meshes[n].vertices[meshes[n].faces[i].indices[2]].pos -
                        meshes[n].vertices[meshes[n].faces[i].indices[0]].pos).normalized();
        meshes[n].normals[meshes[n].faces[i].indices[0]] += norm;
        meshes[n].normals[meshes[n].faces[i].indices[1]] += norm;
        meshes[n].normals[meshes[n].faces[i].indices[2]] += norm;
        meshes[n].faces[i].normIndices[0] = meshes[n].faces[i].indices[0];
        meshes[n].faces[i].normIndices[1] = meshes[n].faces[i].indices[1];
        meshes[n].faces[i].normIndices[2] = meshes[n].faces[i].indices[2];
    }

    for (int i = 0; i < meshes[n].normals.size(); i++) {
        meshes[n].normals[i].normalize();
    }

    // Assign Vertex normals
    for (int i = 0; i < meshes[n].faces.size(); i++) {
        for (int j = 0; j < 3; j++) {
            int vid = meshes[n].faces[i].indices[j];
            int nid = meshes[n].faces[i].normIndices[j];
            meshes[n].vertices[vid].norm = meshes[n].normals[nid];
        }
    }
}

void StaticModel::loadTextures()
{
    for (int i = 0; i < materials.size(); i++) {
        if (materials[i].textured) {
            GLuint tex;
            if (!loadTexture(KGlobal::dirs()->findResource("appdata", materials[i].texFileName), tex, glcontext)) {
                materials[i].textured = false;
                continue;
            }
            materials[i].textureID = tex;

            glBindTexture(GL_TEXTURE_2D, materials[i].textureID);
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        }
    }
}

void StaticModel::render()
{
    for (int i = 0; i < meshes.size(); i++) {
        renderMesh(i);
    }
}

void StaticModel::renderCartoony(const Vector3f &lightVec)
{
    for (int i = 0; i < meshes.size(); i++) {
        renderCartoonyMesh(i, lightVec);
    }
}

void StaticModel::renderMesh(int n)
{
    int mid = 0;
    if (meshes[n].hasMaterial) {
        mid = meshes[n].matID;
        if (materials[mid].textured) {
            glBindTexture(GL_TEXTURE_2D, materials[mid].textureID);
            glEnable(GL_TEXTURE_2D);
        }

        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, (GLfloat *) &materials[mid].faceColor);//XXX ik hoop dat dit werkt :)
        glMaterialfv(GL_FRONT, GL_SPECULAR, (GLfloat *) &materials[mid].specular);
        glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, (GLfloat *) &materials[mid].emissive);
        glMaterialf(GL_FRONT, GL_SHININESS, materials[mid].shininess);

        if (meshes[n].faceWraps) {
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        } else {
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        }
    }

    bool color = meshes[n].hasColors;
    bool norms = meshes[n].hasNormals;
    bool tex = meshes[n].hasTexCoords;

    if (color) {
        glEnable(GL_COLOR_MATERIAL);
    }

    glBegin(GL_TRIANGLES);
        for (int f = 0; f < meshes[n].faces.size(); f++) {
            VertexBuf vert;
            Vector3f norm;

            vert = meshes[n].vertices[meshes[n].faces[f].indices[0]];
            if (norms) {
                norm = meshes[n].normals[meshes[n].faces[f].normIndices[0]];
            } else {
                norm = meshes[n].normals[meshes[n].faces[f].indices[0]];
            }
            if (color) {
                glColor4f(vert.color.r, vert.color.g, vert.color.b, vert.color.a);
            }
            if (tex) {
                glTexCoord2f(vert.texCoord[0], vert.texCoord[1]);
            }
            glNormal3f(norm.x(), norm.y(), norm.z());
            glVertex3f(vert.pos.x(), vert.pos.y(), vert.pos.z());

            vert = meshes[n].vertices[meshes[n].faces[f].indices[1]];
            if (norms) {
                norm = meshes[n].normals[meshes[n].faces[f].normIndices[1]];
            } else {
                norm = meshes[n].normals[meshes[n].faces[f].indices[1]];
            }
            if (color) {
                glColor4f(vert.color.r, vert.color.g, vert.color.b, vert.color.a);
            }
            if (tex) {
                glTexCoord2f(vert.texCoord[0], vert.texCoord[1]);
            }
            glNormal3f(norm.x(), norm.y(), norm.z());
            glVertex3f(vert.pos.x(), vert.pos.y(), vert.pos.z());

            vert = meshes[n].vertices[meshes[n].faces[f].indices[2]];
            if (norms) {
                norm = meshes[n].normals[meshes[n].faces[f].normIndices[2]];
            } else {
                norm = meshes[n].normals[meshes[n].faces[f].indices[2]];
            }
            if (color) {
                glColor4f(vert.color.r, vert.color.g, vert.color.b, vert.color.a);
            }
            if (tex) {
                glTexCoord2f(vert.texCoord[0], vert.texCoord[1]);
            }
            glNormal3f(norm.x(), norm.y(), norm.z());
            glVertex3f(vert.pos.x(), vert.pos.y(), vert.pos.z());
        }
    glEnd();

    if (color) {
        glDisable(GL_COLOR_MATERIAL);
    }

    if (meshes[n].hasMaterial) {
        if (materials[mid].textured) {
            glDisable(GL_TEXTURE_2D);
        }
    }
}

void StaticModel::renderCartoonyMesh(int n, const Vector3f &lightVec)
{
    glNormal3f(lightVec.x(), lightVec.y(), lightVec.z());

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_1D, cartoonTex);
    glEnable(GL_TEXTURE_1D);

    int mid = 0;
    if (meshes[n].hasMaterial) {
        mid = meshes[n].matID;
        if (materials[mid].textured) {
            glActiveTexture(GL_TEXTURE1);
            glBindTexture(GL_TEXTURE_2D, materials[mid].textureID);
            glEnable(GL_TEXTURE_2D);
        }

        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, (GLfloat *) &materials[mid].faceColor);
        glMaterialfv(GL_FRONT, GL_SPECULAR, (GLfloat *) &materials[mid].specular);
        glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, (GLfloat *) &materials[mid].emissive);
        glMaterialf(GL_FRONT, GL_SHININESS, materials[mid].shininess);

        if (meshes[n].faceWraps) {
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        } else {
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        }
    }

    bool color = meshes[n].hasColors;
    bool norms = meshes[n].hasNormals;
    bool tex = meshes[n].hasTexCoords;
    glActiveTexture(GL_TEXTURE0);

    if (color) {
        glEnable(GL_COLOR_MATERIAL);
    }

    glBegin(GL_TRIANGLES);
        for (int f = 0; f < meshes[n].faces.size(); f++) {
            VertexBuf vert;
            Vector3f norm;

            vert = meshes[n].vertices[meshes[n].faces[f].indices[0]];
            if (norms) {
                norm = meshes[n].normals[meshes[n].faces[f].normIndices[0]];
            } else {
                norm = meshes[n].normals[meshes[n].faces[f].indices[0]];
            }
            if (color) {
                glColor4f(vert.color.r, vert.color.g, vert.color.b, vert.color.a);
            }
            glMultiTexCoord1f(GL_TEXTURE0, norm.dot(lightVec));
            if (tex) {
                glMultiTexCoord2f(GL_TEXTURE1, vert.texCoord[0], vert.texCoord[1]);
            }
            glVertex3f(vert.pos.x(), vert.pos.y(), vert.pos.z());

            vert = meshes[n].vertices[meshes[n].faces[f].indices[1]];
            if (norms) {
                norm = meshes[n].normals[meshes[n].faces[f].normIndices[1]];
            } else {
                norm = meshes[n].normals[meshes[n].faces[f].indices[1]];
            }
            if (color) {
                glColor4f(vert.color.r, vert.color.g, vert.color.b, vert.color.a);
            }
            glMultiTexCoord1f(GL_TEXTURE0, norm.dot(lightVec));
            if (tex) {
                glMultiTexCoord2f(GL_TEXTURE1, vert.texCoord[0], vert.texCoord[1]);
            }
            glVertex3f(vert.pos.x(), vert.pos.y(), vert.pos.z());

            vert = meshes[n].vertices[meshes[n].faces[f].indices[2]];
            if (norms) {
                norm = meshes[n].normals[meshes[n].faces[f].normIndices[2]];
            } else {
                norm = meshes[n].normals[meshes[n].faces[f].indices[2]];
            }
            if (color) {
                glColor4f(vert.color.r, vert.color.g, vert.color.b, vert.color.a);
            }
            glMultiTexCoord1f(GL_TEXTURE0, norm.dot(lightVec));
            if (tex) {
                glMultiTexCoord2f(GL_TEXTURE1, vert.texCoord[0], vert.texCoord[1]);
            }
            glVertex3f(vert.pos.x(), vert.pos.y(), vert.pos.z());
        }
    glEnd();

    if (color) {
        glDisable(GL_COLOR_MATERIAL);
    }

    if (meshes[n].hasMaterial) {
        if (materials[mid].textured) {
            glActiveTexture(GL_TEXTURE1);
            glDisable(GL_TEXTURE_2D);
        }
    }

    glActiveTexture(GL_TEXTURE0);
    glDisable(GL_TEXTURE_1D);

    // Render lines
    glEnable(GL_COLOR_MATERIAL);
    glColor3f(0.0, 0.0, 0.0);
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

    glLineWidth(4.0);
    glPointSize(4.0);

    glBegin(GL_TRIANGLES);
        for (int f = 0; f < meshes[n].faces.size(); f++) {
            VertexBuf vert;
            vert = meshes[n].vertices[meshes[n].faces[f].indices[0]];
            glVertex3f(vert.pos.x(), vert.pos.y(), vert.pos.z());
            vert = meshes[n].vertices[meshes[n].faces[f].indices[2]];
            glVertex3f(vert.pos.x(), vert.pos.y(), vert.pos.z());
            vert = meshes[n].vertices[meshes[n].faces[f].indices[1]];
            glVertex3f(vert.pos.x(), vert.pos.y(), vert.pos.z());
        }
    glEnd();

    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

    glColor3f(1.0, 1.0, 1.0);
    glDisable(GL_COLOR_MATERIAL);
}

void StaticModel::setFrame(unsigned int, float)
{
}

void StaticModel::setAnimSetBlend(unsigned int, float, float, int)
{
}

void StaticModel::setAnimBlend(unsigned int, int unsigned, float)
{
}

void StaticModel::calcAnim()
{
}

FrameModel::FrameModel(Header header, const QString& fileName)
    : StaticModel(header, fileName)
{
}

FrameModel::~FrameModel()
{
}

void FrameModel::load()
{
    int num;

    // Skip Header
    skipInputLine();

    // Read Materials
    num = readInt();
    materials.resize(num);
    for (int i = 0; i < num; i++) {
        readMatBlock(i);
    }

    // Read Meshes
    num = readInt();
    meshes.resize(num);
    for (int i = 0; i < num; i++) {
        readMeshBlock(i);
    }

    // Read Frames
    num = readInt();
    frames.resize(num);
    for (int i = 0; i < num; i++) {
        readFrameBlock(i);
    }
    num = readInt();
    rootFrames.resize(num);
    for (int i = 0; i < num; i++) {
        rootFrames[i] = readInt();
    }

    loadTextures();

    oglFile.close();
}

void FrameModel::setTextures(const QVector<int> &tex)
{
    for (int i = 0; i < tex.size(); i++) {
        materials[i].textureID = tex[i];
        materials[i].textured = (tex[i] != -1);
    }
}

void FrameModel::readFrameBlock(int n)
{
    frames[n].frameName = readLine();

    int num;
    // Read sub frames
    num = readInt();
    frames[n].subFrames.resize(num);
    for (int i = 0; i < num; i++) {
        frames[n].subFrames[i] = readInt();
    }

    // Read sub meshes
    num = readInt();
    frames[n].subMeshes.resize(num);
    for (int i = 0; i < num; i++) {
        frames[n].subMeshes[i] = readInt();
    }

    // Read transform matrix
    for (int i = 0; i < 16; i++) {
        frames[n].transform.data()[i] = readFloat();
    }

    frames[n].animPos = frames[n].transform.translation();
    extractScaleRotation(frames[n].transform, frames[n].animScale, frames[n].animRot);
}

void FrameModel::render()
{
    for (int i = 0; i < meshes.size(); i++) {
        if (!meshes[i].inFrame) {
            renderMesh(i);
        }
    }

    for (int i = 0; i < rootFrames.size(); i++) {
        renderFrame(rootFrames[i]);
    }
}

void FrameModel::renderFrame(int n) {
    glPushMatrix();
    glMultMatrixf(frames[n].transform.data());
    for (int i = 0; i < frames[n].subMeshes.size(); i++) {
        renderMesh(frames[n].subMeshes[i]);
    }
    for (int i = 0; i < frames[n].subFrames.size(); i++) {
        renderFrame(frames[n].subFrames[i]);
    }
    glPopMatrix();
}

void FrameModel::setFrame(unsigned int, float)
{
}

void FrameModel::setAnimSetBlend(unsigned int, float, float, int)
{
}

void FrameModel::setAnimBlend(unsigned int, int unsigned, float)
{
}

void FrameModel::calcAnim()
{
}

AnimModel::AnimModel(Header header, const QString& fileName)
    : FrameModel(header, fileName),
    addToOne(true)
{
}

AnimModel::~AnimModel()
{
}

void AnimModel::load()
{
    int num;

    // Skip Header
    skipInputLine();

    // Read Materials
    num = readInt();
    materials.resize(num);
    for (int i = 0; i < num; i++) {
        readMatBlock(i);
    }

    // Read Meshes
    num = readInt();
    meshes.resize(num);
    for (int i = 0; i < num; i++) {
        readMeshBlock(i);
    }

    // Read Frames
    num = readInt();
    frames.resize(num);
    for (int i = 0; i < num; i++) {
        readFrameBlock(i);
    }
    num = readInt();
    rootFrames.resize(num);
    for (int i = 0; i < num; i++) {
        rootFrames[i] = readInt();
    }

    // Read Anim sets
    num = readInt();
    animSets.resize(num);
    for (int i = 0; i < num; i++) {
        readAnimSetBlock(i);
    }

    loadTextures();

    oglFile.close();
}

void AnimModel::setTextures(const QVector<int> &tex)
{
    for (int i = 0; i < tex.size(); i++) {
        materials[i].textureID = tex[i];
        materials[i].textured = (tex[i] != -1);
    }
}

void AnimModel::readAnimSetBlock(int n)
{
    animSets[n].animSetName = readLine();
    animSets[n].blend = 0.00;
    animSets[n].priority = -1;
    animSets[n].time = 0.00;

    int num;
    // Read Anims
    num = readInt();
    animSets[n].anims.resize(num);
    for (int i = 0; i < num; i++) {
        readAnimBlock(n, i);
    }
}

void AnimModel::readAnimBlock(int n, int p)
{
    animSets[n].anims[p].frame = readInt();
    animSets[n].anims[p].blend = 1.00;

    int num;
    // Read KeySets
    num = readInt();
    animSets[n].anims[p].keySets.resize(num);
    for (int i = 0; i < num; i++) {
        readKeySetBlock(n, p, i);
    }
}

void AnimModel::readKeySetBlock(int n, int p, int q)
{
    KeySet* keySet = &animSets[n].anims[p].keySets[q];
    int num;

    num = readInt();
    keySet->keys.resize(num);
    keySet->animType = readInt();

    // Read the keys
    switch (keySet->animType) {
        case 0:
            for (int i = 0; i < keySet->keys.size(); i++) {
                Key* key = &keySet->keys[i];
                key->time = readInt();
                key->rot.w() = readFloat();
                key->rot.x() = readFloat();
                key->rot.y() = readFloat();
                key->rot.z() = readFloat();
                key->pos = Vector3f(0.0, 0.0, 0.0);
                key->scale = Vector3f(1.0, 1.0, 1.0);
                key->trans = key->rot;
            }
            break;
        case 1:
            for (int i = 0; i < keySet->keys.size(); i++) {
                Key* key = &keySet->keys[i];
                key->time = readInt();
                key->scale.x() = readFloat();
                key->scale.y() = readFloat();
                key->scale.z() = readFloat();
                key->pos = Vector3f(0.0, 0.0, 0.0);
                key->rot = Quaternionf(1.0, 0.0, 0.0, 0.0);
                key->trans = Scaling3f(key->scale);
            }
            break;
        case 2:
            for (int i = 0; i < keySet->keys.size(); i++) {
                Key* key = &keySet->keys[i];
                key->time = readInt();
                key->pos.x() = readFloat();
                key->pos.y() = readFloat();
                key->pos.z() = readFloat();
                key->scale = Vector3f(0.0, 0.0, 0.0);
                key->rot = Quaternionf(1.0, 0.0, 0.0, 0.0);
                key->trans = Translation3f(key->pos);
            }
            break;
        default:
            for (int i = 0; i < keySet->keys.size(); i++) {
                Key* key = &keySet->keys[i];
                key->time = readInt();
                for (int j = 0; j < 16; j++) {
                    key->trans.data()[j] = readFloat();
                }
                key->pos = key->trans.translation();
                extractScaleRotation(key->trans, key->scale, key->rot);
            }
            break;
    }

}

void AnimModel::render()
{
    for (int i = 0; i < meshes.size(); i++) {
        if (!meshes[i].inFrame) {
            renderMesh(i);
        }
    }

    for (int i = 0; i < rootFrames.size(); i++) {
        renderFrame(rootFrames[i]);
    }
}

void AnimModel::setFrame(unsigned int set, float time)
{
    if (set >= (unsigned int)animSets.size()) {
        return;
    }

    for (int i = 0; i < animSets[set].anims.size(); i++) {
        Anim* a = &animSets[set].anims[i];
        int f = a->frame;
        frames[f].transform.setIdentity();
        frames[f].animPos = Vector3f(0.0, 0.0, 0.0);
        frames[f].animScale = Vector3f(1.0, 1.0, 1.0);
        frames[f].animRot = Quaternionf(1.0, 0.0, 0.0, 0.0);
        for (int j = 0; j < a->keySets.size(); j++) {
            KeySet* ks = &a->keySets[j];
            int dTime;
            if (ks->keys.size() > 1) {
                dTime = ks->keys[ks->keys.size() - 1].time - ks->keys[ks->keys.size() - 2].time;
            } else {
                dTime = 0;
            }
            if (time >= ks->keys[0].time && time < ks->keys[ks->keys.size() - 1].time + dTime) {
                int x = 0;
                int y = ks->keys.size();
                if (y == 0) {
                    break;
                }
                while (x+1 != y) {
                    int h = (x + y) / 2;
                    if (time >= ks->keys[h].time) {
                        x = h;
                    } else {
                        y = h;
                    }
                }

                float part;
                Key* yKey;
                if (y == ks->keys.size()) {
                    part = (time - ks->keys[x].time) / dTime;
                    yKey = &ks->keys[0];
                } else {
                    part = (time - ks->keys[x].time) / (ks->keys[y].time - ks->keys[x].time);
                    yKey = &ks->keys[y];
                }

                switch (ks->animType) {
                    case 0:
                        frames[f].animRot = ks->keys[x].rot.slerp(part, yKey->rot);
                        break;
                    case 1:
                        frames[f].animScale = lerp(ks->keys[x].scale, yKey->scale, part);
                        break;
                    case 2:
                        frames[f].animPos = lerp(ks->keys[x].pos, yKey->pos, part);
                        break;
                    default:
                        frames[f].animRot = ks->keys[x].rot.slerp(part, yKey->rot);
                        frames[f].animScale = lerp(ks->keys[x].scale, yKey->scale, part);
                        frames[f].animPos = lerp(ks->keys[x].pos, yKey->pos, part);
                        break;
                }
            }
        }
    }

    for (int f = 0; f < frames.size(); f++) {
        frames[f].transform.fromPositionOrientationScale(frames[f].animPos, frames[f].animRot, frames[f].animScale);
    }
}

void AnimModel::setAnimSetBlend(unsigned int set, float time, float blend, int priority)
{
    if (set >= (unsigned int)animSets.size()) {
        return;
    }
    animSets[set].time = time;
    animSets[set].blend = blend;
    animSets[set].priority = priority;
}

void AnimModel::setAnimBlend(unsigned int set, unsigned int anim, float blend)
{
    if (set >= (unsigned int)animSets.size()) {
        return;
    }
    if (anim >= (unsigned int)animSets[set].anims.size()) {
        return;
    }
    animSets[set].anims[anim].blend = blend;
}

void AnimModel::calcAnim()
{
    // PreCalcs
    for (int f = 0; f < frames.size(); f++) {
        frames[f].blendSum = 0.00;
        frames[f].priority = 0;
    }

    for (int i = 0; i < animSets.size(); i++) {
        for (int j = 0; j < animSets[i].anims.size(); j++) {
            int fid = animSets[i].anims[j].frame;
            if (frames[fid].priority <= animSets[i].priority) {
                if (frames[fid].priority < animSets[i].priority) {
                    frames[fid].priority = animSets[i].priority;
                    frames[fid].blendSum = 0.00;
                }
                frames[fid].blendSum += animSets[i].blend * animSets[i].anims[j].blend;
            }
        }
    }

    for (int f = 0; f < frames.size(); f++) {
        if (frames[f].blendSum != 0.00) {
            frames[f].blendMult = 1.00 / frames[f].blendSum;
        } else {
            frames[f].blendMult = 0.00;
        }
        if (!addToOne && frames[f].blendMult != 0.00) {
            frames[f].blendMult = 1.00;
        }
        if (frames[f].blendMult != 0.00) {
            frames[f].transform.setIdentity();
            frames[f].animPos = Vector3f(0.0, 0.0, 0.0);
            frames[f].animScale = Vector3f(0.0, 0.0, 0.0);
            frames[f].animRot = Quaternionf(0.0, 0.0, 0.0, 0.0);
        }
    }

    // Calc animation
    for (int set = 0; set < animSets.size(); set++) {
        float time = animSets[set].time;
        for (int i = 0; i < animSets[set].anims.size(); i++) {
            Anim* a = &animSets[set].anims[i];
            int f = a->frame;
            if (frames[f].priority == animSets[set].priority) {
                for (int j = 0; j < a->keySets.size(); j++) {
                    KeySet* ks = &a->keySets[j];
                    float dTime;
                    if (ks->keys.size() > 1) {
                        dTime = ks->keys[ks->keys.size() - 1].time - ks->keys[ks->keys.size() - 2].time;
                    } else {
                        dTime = 0.0;
                    }
                    if (time >= ks->keys[0].time && time < ks->keys[ks->keys.size() - 1].time + dTime) {
                        int x = 0;
                        int y= ks->keys.size();
                        if (y == 0) {
                            break;
                        }
                        while (x+1 != y) {
                            int h = (x + y) / 2;
                            if (time >= ks->keys[h].time) {
                                x = h;
                            } else {
                                y = h;
                            }
                        }

                        float part;
                        Key* yKey;
                        if (y == ks->keys.size()) {
                            part = (time - ks->keys[x].time) / dTime;
                            yKey = &ks->keys[0];
                        } else {
                            part = (time - ks->keys[x].time) / (ks->keys[y].time - ks->keys[x].time);
                            yKey = &ks->keys[y];
                        }

                        switch (ks->animType) {
                            case 0:
                                frames[f].animRot.coeffs() += ks->keys[x].rot.slerp(part, yKey->rot).coeffs() *
                                    (frames[f].blendMult * animSets[set].blend * a->blend);
                                break;
                            case 1:
                                frames[f].animScale += lerp(ks->keys[x].scale, yKey->scale, part) *
                                    (frames[f].blendMult * animSets[set].blend * a->blend);
                                break;
                            case 2:
                                frames[f].animPos += lerp(ks->keys[x].pos, yKey->pos, part) *
                                    (frames[f].blendMult * animSets[set].blend * a->blend);
                                break;
                            default:
                                frames[f].animRot.coeffs() += ks->keys[x].rot.slerp(part, yKey->rot).coeffs() *
                                    (frames[f].blendMult * animSets[set].blend * a->blend);
                                frames[f].animScale += lerp(ks->keys[x].scale, yKey->scale, part) *
                                    (frames[f].blendMult * animSets[set].blend * a->blend);
                                frames[f].animPos += lerp(ks->keys[x].pos, yKey->pos, part) *
                                    (frames[f].blendMult * animSets[set].blend * a->blend);
                                break;
                        }
                    }
                }
            }
        }
    }

    // Calc Frame Transforms
    for (int f = 0; f < frames.size(); f++) {
        frames[f].animRot.normalize();
        frames[f].transform.fromPositionOrientationScale(frames[f].animPos, frames[f].animRot, frames[f].animScale);
    }

    // Reset Blend Values
    for (int i = 0; i < animSets.size(); i++) {
        animSets[i].blend = 0.00;
        animSets[i].priority = -1;
        for (int j = 0; j < animSets[i].anims.size(); j++) {
            animSets[i].anims[j].blend = 1.00;
        }
    }
}

BoneModel::BoneModel(Header header, const QString& fileName)
    : AnimModel(header, fileName)
{
}

BoneModel::~BoneModel()
{
}

void BoneModel::load()
{
    int num;

    // Skip Header
    skipInputLine();

    // Read Materials
    num = readInt();
    materials.resize(num);
    for (int i = 0; i < num; i++) {
        readMatBlock(i);
    }

    // Read Meshes
    num = readInt();
    meshes.resize(num);
    for (int i = 0; i < num; i++) {
        readMeshBlock(i);
    }

    // Read Frames
    num = readInt();
    frames.resize(num);
    for (int i = 0; i < num; i++) {
        readFrameBlock(i);
    }
    num = readInt();
    rootFrames.resize(num);
    for (int i = 0; i < num; i++) {
        rootFrames[i] = readInt();
    }

    // Read Anim sets
    num = readInt();
    animSets.resize(num);
    for (int i = 0; i < num; i++) {
        readAnimSetBlock(i);
    }

    loadTextures();

    oglFile.close();
}

void BoneModel::setTextures(const QVector<int> &tex)
{
    for (int i = 0; i < tex.size(); i++) {
        materials[i].textureID = tex[i];
        materials[i].textured = (tex[i] != -1);
    }
}

void BoneModel::readMeshBlock(int n)
{
    int fvf = readInt();
    meshes[n].hasNormals = fvf & 1;
    meshes[n].hasColors = fvf & 2;
    meshes[n].faceWraps = fvf & 4;
    meshes[n].hasTexCoords = fvf & 8;
    meshes[n].hasMaterial = fvf & 16;
    meshes[n].inFrame = fvf & 32;

    int num;
    // Read Vertices
    num = readInt();
    meshes[n].vertices.resize(num);
    for (int i = 0; i < num; i++) {
        meshes[n].vertices[i].pos.x() = readFloat();
        meshes[n].vertices[i].pos.y() = readFloat();
        meshes[n].vertices[i].pos.z() = readFloat();

        int nBoneWeights = readInt();
        meshes[n].vertices[i].boneWeights.resize(nBoneWeights);
        for (int j = 0; j < nBoneWeights; j++) {
            meshes[n].vertices[i].boneWeights[j].boneID = readInt();
            meshes[n].vertices[i].boneWeights[j].weight = readFloat();
        }
    }

    // Read Faces
    num = readInt();
    meshes[n].faces.resize(num);
    for (int i = 0; i < num; i++) {
        meshes[n].faces[i].indices[0] = readInt();
        meshes[n].faces[i].indices[1] = readInt();
        meshes[n].faces[i].indices[2] = readInt();
    }

    // Read Normals
    if (meshes[n].hasNormals) {
        DPRINTF("Reading normals\n");
        num = readInt();
        meshes[n].normals.resize(num);
        for (int i = 0; i < num; i++) {
            meshes[n].normals[i].x() = readFloat();
            meshes[n].normals[i].y() = readFloat();
            meshes[n].normals[i].z() = readFloat();
        }
        readInt();
        for (int i = 0; i < meshes[n].faces.size(); i++) {
            meshes[n].faces[i].normIndices[0] = readInt();
            meshes[n].faces[i].normIndices[1] = readInt();
            meshes[n].faces[i].normIndices[2] = readInt();
        }
        // Assign Vertex normals
        for (int i = 0; i < meshes[n].faces.size(); i++) {
            for (int j = 0; j < 3; j++) {
                int vid = meshes[n].faces[i].indices[j];
                int nid = meshes[n].faces[i].normIndices[j];
                meshes[n].vertices[vid].norm = meshes[n].normals[nid];
            }
        }
    } else {
        autoGenerateNormals(n);
    }

    // Read Colors
    if (meshes[n].hasColors) {
        readInt();
        for (int i = 0; i < meshes[n].vertices.size(); i++) {
            meshes[n].vertices[i].color.r = readFloat();
            meshes[n].vertices[i].color.g = readFloat();
            meshes[n].vertices[i].color.b = readFloat();
            meshes[n].vertices[i].color.a = readFloat();
        }
    } else {
        for (int i = 0; i < meshes[n].vertices.size(); i++) {
            meshes[n].vertices[i].color.r = 0;
            meshes[n].vertices[i].color.g = 0;
            meshes[n].vertices[i].color.b = 0;
            meshes[n].vertices[i].color.a = 0;
        }
    }

    // Read Tex Coords
    if (meshes[n].hasTexCoords) {
        readInt();
        for (int i = 0; i < meshes[n].vertices.size(); i++) {
            meshes[n].vertices[i].texCoord[0] = readFloat();
            meshes[n].vertices[i].texCoord[1] = readFloat();
        }
    }

    // Read Mat ID
    if (meshes[n].hasMaterial) {
        meshes[n].matID = readInt();
    }

    // Read bones
    meshes[n].nMaxBones = readInt();
    meshes[n].nMaxWeightsPerVertex = readInt();
    meshes[n].nMaxWeightsPerFace = readInt();
    num = readInt();
    meshes[n].bones.resize(num);
    for (int i = 0; i < num; i++) {
        meshes[n].bones[i].frameID = readInt();;
        for (int j = 0; j < 16; j++) {
            meshes[n].bones[i].offset.data()[j] = readFloat();
        }
    }

    // Setup arrays
    meshes[n].indices = new GLuint[meshes[n].faces.size() * 3];
    for (int i = 0, index = 0; i < meshes[n].faces.size(); i++) {
        for (int j = 0; j < 3; j++) {
            meshes[n].indices[index++] = meshes[n].faces[i].indices[j];
        }
    }
    meshes[n].vertexBuf = new VertexBuf[meshes[n].vertices.size()];
}

void BoneModel::render()
{
    glGetFloatv(GL_MODELVIEW_MATRIX, curViewMat2.data());
    curViewMat = curViewMat2;

    curNormMat2 = getNormMat(curViewMat2);
    curNormMat = curNormMat2;

    glEnable(GL_NORMALIZE);
    glDisable(GL_RESCALE_NORMAL);

    for (int i = 0; i < meshes.size(); i++) {
        if (!meshes[i].inFrame) {
            renderMesh(i);
        }
    }

    for (int i = 0; i < rootFrames.size(); i++) {
        renderFrame(rootFrames[i]);
    }

    glDisable(GL_NORMALIZE);
    glEnable(GL_RESCALE_NORMAL);
}

void BoneModel::renderCartoony(const Vector3f &lightVec)
{
    glGetFloatv(GL_MODELVIEW_MATRIX, curViewMat2.data());
    curViewMat = curViewMat2;

    curNormMat2 = getNormMat(curViewMat2);
    curNormMat = curNormMat2;

    glEnable(GL_NORMALIZE);
    glDisable(GL_RESCALE_NORMAL);

    for (int i = 0; i < meshes.size(); i++) {
        if (!meshes[i].inFrame) {
            renderCartoonyMesh(i, lightVec);
        }
    }

    for (int i = 0; i < rootFrames.size(); i++) {
        renderCartoonyFrame(rootFrames[i], lightVec);
    }

    glDisable(GL_NORMALIZE);
    glEnable(GL_RESCALE_NORMAL);
}

void BoneModel::renderFrame(int n)
{
    glPushMatrix();
    glLoadMatrixf(frames[n].transform.data());
    curViewMat = curViewMat2;
    curNormMat = getNormMat(curViewMat);
    for (int i = 0; i < frames[n].subMeshes.size(); i++) {
        renderMesh(frames[n].subMeshes[i]);
    }
    curViewMat = curViewMat2;
    curNormMat = curNormMat2;
    for (int i = 0; i < frames[n].subFrames.size(); i++) {
        renderFrame(frames[n].subFrames[i]);
    }
    glPopMatrix();
}

void BoneModel::renderCartoonyFrame(int n, const Vector3f &lightVec)
{
    glPushMatrix();
    glLoadMatrixf(frames[n].transform.data());
    curViewMat = curViewMat2;
    curNormMat = getNormMat(curViewMat);
    for (int i = 0; i < frames[n].subMeshes.size(); i++) {
        renderCartoonyMesh(frames[n].subMeshes[i], lightVec);
    }
    curViewMat = curViewMat2;
    curNormMat = curNormMat2;
    for (int i = 0; i < frames[n].subFrames.size(); i++) {
        renderCartoonyFrame(frames[n].subFrames[i], lightVec);
    }
    glPopMatrix();

}

void BoneModel::setupBones(int n)
{
    for (int i = 0; i < meshes[n].bones.size(); i++) {
        int fid = meshes[n].bones[i].frameID;
        meshes[n].bones[i].trans = /*fastMul(fastMul(*/Transform3f(curViewMat * frames[fid].transform) * meshes[n].bones[i].offset;
        meshes[n].bones[i].norm = getNormMat(meshes[n].bones[i].trans);
    }
}

Vector3f BoneModel::getVertPos(int mid, int vid)
{
    const VertexBone& vert = meshes[mid].vertices[vid];
    if (vert.boneWeights.size() == 0) {
        return curViewMat * vert.pos;
    } else {
        Matrix4f finalMat = Matrix4f::Zero();
        for (int w = 0; w < vert.boneWeights.size(); w++) {
            float weight = vert.boneWeights[w].weight;
            int bid = vert.boneWeights[w].boneID;
            finalMat += /*fastMad(*/meshes[mid].bones[bid].trans.matrix() * weight/*, finalMat)*/;
        }
        return Transform3f(finalMat) * vert.pos;
    }
}

void BoneModel::renderMesh(int n)
{
    setupBones(n);

    int mid = 0;
    if (meshes[n].hasMaterial) {
        mid = meshes[n].matID;
        if (materials[mid].textured) {
            glBindTexture(GL_TEXTURE_2D, materials[mid].textureID);
            glEnable(GL_TEXTURE_2D);
        }

        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, (GLfloat *) &materials[mid].faceColor);
        glMaterialfv(GL_FRONT, GL_SPECULAR, (GLfloat *) &materials[mid].specular);
        glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, (GLfloat *) &materials[mid].emissive);
        glMaterialf(GL_FRONT, GL_SHININESS, materials[mid].shininess);

        if (meshes[n].faceWraps) {
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        } else {
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        }
    }

    bool color = meshes[n].hasColors;
    bool tex = meshes[n].hasTexCoords;

    if (color) {
        glEnable(GL_COLOR_MATERIAL);
    }

    // Setup vertex array
    for (int v = 0; v < meshes[n].vertices.size(); v++) {
        const VertexBone& vert = meshes[n].vertices[v];
        Vector3f norm = vert.norm;
        Vector3f vertPos, vertNorm;
        if (vert.boneWeights.size() == 0) {
            vertPos = curViewMat * vert.pos;
            vertNorm = curNormMat * norm;
        } else {
            Matrix4f finalMat = Matrix4f::Zero();
            Matrix3f normMat;
            for (int w = 0; w < vert.boneWeights.size(); w++) {
                float weight = vert.boneWeights[w].weight;
                int bid = vert.boneWeights[w].boneID;
                finalMat += /*fastMad(*/meshes[n].bones[bid].trans.matrix() * weight/*, finalMat)*/;
                normMat += meshes[n].bones[bid].norm * weight;
            }
            vertPos = Transform3f(finalMat) * vert.pos;
            vertNorm = normMat * norm;
        }
        meshes[n].vertexBuf[v].pos = vertPos;
        meshes[n].vertexBuf[v].norm = vertNorm;
        if (color) {
            meshes[n].vertexBuf[v].color = vert.color;
        }
        if (tex) {
            meshes[n].vertexBuf[v].texCoord = vert.texCoord;
        }
    }

    // Set Vertex Buffer
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY);
    if (color) {
        glEnableClientState(GL_COLOR_ARRAY);
    }
    if (tex) {
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    }
    glVertexPointer(3, GL_FLOAT, sizeof(VertexBuf), &meshes[n].vertexBuf[0].pos.x());
    glNormalPointer(GL_FLOAT, sizeof(VertexBuf), &meshes[n].vertexBuf[0].norm);
    if (color) {
        glColorPointer(4, GL_FLOAT, sizeof(VertexBuf), &meshes[n].vertexBuf[0].color);
    }
    if (tex) {
        glTexCoordPointer(2, GL_FLOAT, sizeof(VertexBuf), &meshes[n].vertexBuf[0].texCoord[0]);
    }

    glPushMatrix();
    glLoadIdentity();

    // Draw em
    glDrawElements(GL_TRIANGLES, meshes[n].faces.size() * 3, GL_UNSIGNED_INT, meshes[n].indices);

    glPopMatrix();

    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);
    if (color) {
        glDisableClientState(GL_COLOR_ARRAY);
    }
    if (tex) {
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }

    if (color) {
        glDisable(GL_COLOR_MATERIAL);
    }

    if (meshes[n].hasMaterial) {
        if (materials[mid].textured) {
            glDisable(GL_TEXTURE_2D);
        }
    }
}

void BoneModel::renderCartoonyMesh(int n, const Vector3f &lightVec)
{
    setupBones(n);

    glNormal3f(lightVec.x(), lightVec.y(), lightVec.z());

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_1D, cartoonTex);
    glEnable(GL_TEXTURE_1D);
    glDisable(GL_TEXTURE_2D);

    int mid = 0;
    if (meshes[n].hasMaterial) {
        mid = meshes[n].matID;
        if (materials[mid].textured) {
            glActiveTexture(GL_TEXTURE1);
            glBindTexture(GL_TEXTURE_2D, materials[mid].textureID);
            glEnable(GL_TEXTURE_2D);
        }

        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, (GLfloat *) &materials[mid].faceColor);
        glMaterialfv(GL_FRONT, GL_SPECULAR, (GLfloat *) &materials[mid].specular);
        glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, (GLfloat *) &materials[mid].emissive);
        glMaterialf(GL_FRONT, GL_SHININESS, materials[mid].shininess);

        if (meshes[n].faceWraps) {
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        } else {
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        }
    }

    glActiveTexture(GL_TEXTURE0);

    bool color = meshes[n].hasColors;
    bool tex = meshes[n].hasTexCoords;

    if (color) {
        glEnable(GL_COLOR_MATERIAL);
    }

    // Setup vertex array
    for (int v = 0; v < meshes[n].vertices.size(); v++) {
        const VertexBone &vert = meshes[n].vertices[v];
        Vector3f norm = vert.norm;
        Vector3f vertPos, vertNorm;
        if (vert.boneWeights.size() == 0) {
            vertPos = curViewMat * vert.pos;
            vertNorm = curNormMat * norm;
        } else {
            Matrix4f finalMat = Matrix4f::Zero();
            Matrix3f normMat;
            for (int w = 0; w < vert.boneWeights.size(); w++) {
                float weight = vert.boneWeights[w].weight;
                int bid = vert.boneWeights[w].boneID;
                finalMat += /*fastMad(*/meshes[n].bones[bid].trans.matrix() * weight;
                normMat += meshes[n].bones[bid].norm * weight;
            }
            vertPos = Transform3f(finalMat) * vert.pos;
            vertNorm = normMat * norm;
        }
        meshes[n].vertexBuf[v].pos = vertPos;
        meshes[n].vertexBuf[v].norm = vertNorm;
        if (color) {
            meshes[n].vertexBuf[v].color = vert.color;
        }
        if (tex) {
            meshes[n].vertexBuf[v].texCoord = vert.texCoord;
        }
    }

    // Set Vertex Buffer
    glEnableClientState(GL_VERTEX_ARRAY);
    if (color) {
        glEnableClientState(GL_COLOR_ARRAY);
    }
    glVertexPointer(3, GL_FLOAT, sizeof(VertexBuf), meshes[n].vertexBuf);
    if (color) {
        glColorPointer(4, GL_FLOAT, sizeof(VertexBuf), &meshes[n].vertexBuf[0].color);
    }
    glClientActiveTexture(GL_TEXTURE0);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glTexCoordPointer(1, GL_FLOAT, sizeof(VertexBuf), &meshes[n].vertexBuf[0].tex2);
    glClientActiveTexture(GL_TEXTURE1);
    if (tex) {
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        glTexCoordPointer(2, GL_FLOAT, sizeof(VertexBuf), &meshes[n].vertexBuf[0].texCoord);
    }

    glActiveTexture(GL_TEXTURE0);

    glPushMatrix();
    glLoadIdentity();

    // Draw em
    glDrawElements(GL_TRIANGLES, meshes[n].faces.size() * 3, GL_UNSIGNED_INT, meshes[n].indices);

    glPopMatrix();

    if (color) {
        glDisableClientState(GL_COLOR_ARRAY);
    }
    glClientActiveTexture(GL_TEXTURE1);
    if (tex) {
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }
    glClientActiveTexture(GL_TEXTURE0);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);

    if (color) {
        glDisable(GL_COLOR_MATERIAL);
    }

    if (meshes[n].hasMaterial) {
        if (materials[mid].textured) {
            glActiveTexture(GL_TEXTURE_2D);
            glDisable(GL_TEXTURE_2D);
        }
    }

    glActiveTexture(GL_TEXTURE0);
    glDisable(GL_TEXTURE_1D);

    glEnable(GL_COLOR_MATERIAL);
    glColor3f(0.0, 0.0, 0.0);

    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

    glLineWidth(4.0);
    glPointSize(4.0);

    glCullFace(GL_FRONT);

    glPushMatrix();
    glLoadIdentity();

    // Draw em
    glDrawElements(GL_TRIANGLES, meshes[n].faces.size() * 3, GL_UNSIGNED_INT, meshes[n].indices);

    glPopMatrix();

    glCullFace(GL_BACK);

    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

    glColor3f(1.0, 1.0, 1.0);
    glDisable(GL_COLOR_MATERIAL);

    glDisableClientState(GL_VERTEX_ARRAY);
}

void BoneModel::setFrame(unsigned int set, float time)
{
    if (set >= (unsigned int)animSets.size()) {
        return;
    }

    for (int i = 0; i < animSets[set].anims.size(); i++) {
        Anim* a = &animSets[set].anims[i];
        int f = a->frame;
        frames[f].transform.setIdentity();
        frames[f].animPos = Vector3f(0.0, 0.0, 0.0);
        frames[f].animScale = Vector3f(1.0, 1.0, 1.0);
        frames[f].animRot = Quaternionf(1.0, 0.0, 0.0, 0.0);
        for (int j = 0; j < a->keySets.size(); j++) {
            KeySet* ks = &a->keySets[j];
            int dTime;
            if (ks->keys.size() > 1) {
                dTime = ks->keys[ks->keys.size() - 1].time - ks->keys[ks->keys.size() - 2].time;
            } else {
                dTime = 0;
            }
            if (time >= ks->keys[0].time && time < ks->keys[ks->keys.size() - 1].time + dTime) {
                int x = 0;
                int y = ks->keys.size();
                if (y == 0) {
                    break;
                }
                while (x+1 != y) {
                    int h = (x + y) / 2;
                    if (time >= ks->keys[h].time) {
                        x = h;
                    } else {
                        y = h;
                    }
                }

                float part;
                Key* yKey;
                if (y == ks->keys.size()) {
                    part = (float)(time - ks->keys[x].time) / dTime;
                    yKey = &ks->keys[0];
                } else {
                    part = (float)(time - ks->keys[x].time) / (ks->keys[y].time - ks->keys[x].time);
                    yKey = &ks->keys[y];
                }

                switch (ks->animType) {
                    case 0:
                        frames[f].animRot = ks->keys[x].rot.slerp(part, yKey->rot);
                        break;
                    case 1:
                        frames[f].animScale = lerp(ks->keys[x].scale, yKey->scale, part);
                        break;
                    case 2:
                        frames[f].animPos = lerp(ks->keys[x].pos, yKey->pos, part);
                        break;
                    default:
                        frames[f].animRot = ks->keys[x].rot.slerp(part, yKey->rot);
                        frames[f].animScale = lerp(ks->keys[x].scale, yKey->scale, part);
                        frames[f].animPos = lerp(ks->keys[x].pos, yKey->pos, part);
                        break;
                }
            }
        }
    }

    for (int f = 0; f < frames.size(); f++) {
        frames[f].transform.fromPositionOrientationScale(frames[f].animPos, frames[f].animRot, frames[f].animScale);
    }

    // Calc Combined matrices
    glPushMatrix();
    glLoadIdentity();
    for (int f = 0; f < rootFrames.size(); f++) {
        calcCombinedMatrix(rootFrames[f]);
    }
    glPopMatrix();
}

void BoneModel::calcCombinedMatrix(int n)
{
    glPushMatrix();
    glMultMatrixf(frames[n].transform.data());
    glGetFloatv(GL_MODELVIEW_MATRIX, frames[n].transform.data());
    for (int i = 0; i < frames[n].subFrames.size(); i++) {
        calcCombinedMatrix(frames[n].subFrames[i]);
    }
    glPopMatrix();
}


void BoneModel::setAnimSetBlend(unsigned int set, float time, float blend, int priority)
{
    if (set >= (unsigned int)animSets.size()) {
        printf("set is too high\n");
        return;
    }
    animSets[set].time = time;
    animSets[set].blend = blend;
    animSets[set].priority = priority;
}

void BoneModel::setAnimBlend(unsigned int set, unsigned int anim, float blend)
{
    if (set >= (unsigned int)animSets.size()) {
        return;
    }
    if (anim >= (unsigned int)animSets[set].anims.size()) {
        return;
    }
    animSets[set].anims[anim].blend = blend;
}

void BoneModel::calcAnim()
{
    // PreCalcs
    for (int f = 0; f < frames.size(); f++) {
        frames[f].blendSum = 0.00;
        frames[f].priority = 0;
    }

    for (int i = 0; i < animSets.size(); i++) {
        for (int j = 0; j < animSets[i].anims.size(); j++) {
            int fid = animSets[i].anims[j].frame;
            if (frames[fid].priority <= animSets[i].priority) {
                if (frames[fid].priority < animSets[i].priority) {
                    frames[fid].priority = animSets[i].priority;
                    frames[fid].blendSum = 0.00;
                }
                frames[fid].blendSum += animSets[i].blend * animSets[i].anims[j].blend;
            }
        }
    }

    for (int f = 0; f < frames.size(); f++) {
        if (frames[f].blendSum != 0.00) {
            frames[f].blendMult = 1.00 / frames[f].blendSum;
        } else {
            frames[f].blendMult = 0.00;
        }
        if (!addToOne && frames[f].blendMult != 0.00) {
            frames[f].blendMult = 1.00;
        }
        if (frames[f].blendMult != 0.00) {
            frames[f].transform.setIdentity();
            frames[f].animPos = Vector3f(0.0, 0.0, 0.0);
            frames[f].animScale = Vector3f(0.0, 0.0, 0.0);
            frames[f].animRot = Quaternionf(0.0, 0.0, 0.0, 0.0);
        }
    }

    // Calc animation
    for (int set = 0; set < animSets.size(); set++) {
        float time = animSets[set].time;
        for (int i = 0; i < animSets[set].anims.size(); i++) {
            Anim* a = &animSets[set].anims[i];
            int f = a->frame;
            if (frames[f].priority == animSets[set].priority) {
                for (int j = 0; j < a->keySets.size(); j++) {
                    KeySet* ks = &a->keySets[j];
                    int dTime;
                    if (ks->keys.size() > 1) {
                        dTime = ks->keys[ks->keys.size() - 1].time - ks->keys[ks->keys.size() - 2].time;
                    } else {
                        dTime = 0;
                    }
                    if (time >= ks->keys[0].time && time < ks->keys[ks->keys.size() - 1].time + dTime) {
                        int x = 0;
                        int y= ks->keys.size();
                        if (y == 0) {
                            break;
                        }
                        while (x+1 != y) {
                            int h = (x + y) / 2;
                            if (time >= ks->keys[h].time) {
                                x = h;
                            } else {
                                y = h;
                            }
                        }

                        float part;
                        Key* yKey;
                        if (y == ks->keys.size()) {
                            part = (float)(time - ks->keys[x].time) / dTime;
                            yKey = &ks->keys[0];
                        } else {
                            part = (float)(time - ks->keys[x].time) / (ks->keys[y].time - ks->keys[x].time);
                            yKey = &ks->keys[y];
                        }

                        switch (ks->animType) {
                            case 0:
                                frames[f].animRot.coeffs() += ks->keys[x].rot.slerp(part, yKey->rot).coeffs() *
                                    (frames[f].blendMult * animSets[set].blend * a->blend);
                                break;
                            case 1:
                                frames[f].animScale += lerp(ks->keys[x].scale, yKey->scale, part) *
                                    (frames[f].blendMult * animSets[set].blend * a->blend);
                                break;
                            case 2:
                                frames[f].animPos += lerp(ks->keys[x].pos, yKey->pos, part) *
                                    (frames[f].blendMult * animSets[set].blend * a->blend);
                                break;
                            default:
                                frames[f].animRot.coeffs() += ks->keys[x].rot.slerp(part, yKey->rot).coeffs() *
                                    (frames[f].blendMult * animSets[set].blend * a->blend);
                                frames[f].animScale += lerp(ks->keys[x].scale, yKey->scale, part) *
                                    (frames[f].blendMult * animSets[set].blend * a->blend);
                                frames[f].animPos += lerp(ks->keys[x].pos, yKey->pos, part) *
                                    (frames[f].blendMult * animSets[set].blend * a->blend);
                                break;
                        }
                    }
                }
            }
        }
    }

    // Calc Frame Transforms
    for (int f = 0; f < frames.size(); f++) {
        frames[f].animRot.normalize();
        frames[f].transform.fromPositionOrientationScale(frames[f].animPos, frames[f].animRot, frames[f].animScale);
    }

    // Calc Combined matrices
    glPushMatrix();
    glLoadIdentity();
    for (int f = 0; f < rootFrames.size(); f++) {
        calcCombinedMatrix(rootFrames[f]);
    }
    glPopMatrix();

    // Reset Blend Values
    for (int i = 0; i < animSets.size(); i++) {
        animSets[i].blend = 0.00;
        animSets[i].priority = -1;
        for (int j = 0; j < animSets[i].anims.size(); j++) {
            animSets[i].anims[j].blend = 1.00;
        }
    }
}

