/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef _MODEL3D_H
#define _MODEL3D_H

#include <GL/glu.h>

#include <QString>
#include <QVector>
#include <QFile>
#include <QTextStream>

#include <Eigen/Core>
#include <Eigen/Geometry>

#include "Array.h"

extern GLuint cartoonTex;

typedef struct Color {
    Color() {}
    Color(float _r, float _g, float _b, float _a) { r = _r; g = _g; b = _b; a = _a;}
    float r, g, b, a;
} /*__attribute__ ((packed))*/ Color;

class BoneWeight {
    public:
        int boneID;
        float weight;
};

class VertexBuf {
    public:
        Eigen::Vector3f pos, norm;
        Eigen::Vector2f texCoord, tex2;
        Color color;
};

class VertexBone {
    public:
        Eigen::Vector3f pos, norm;
        Eigen::Vector2f texCoord;
        Color color;
        Array<BoneWeight> boneWeights;
};

class Face {
    public:
        int indices[3], normIndices[3];
};

class Material {
    public:
        Color faceColor, specular, emissive;
        float shininess;
        bool textured;
        QString texFileName;
        int textureID;
};

class Bone {
    public:
        int frameID;
        Eigen::Transform3f offset;
        Eigen::Transform3f trans;
        Eigen::Matrix3f norm;

        EIGEN_MAKE_ALIGNED_OPERATOR_NEW
};

class Mesh {
    public:
        int matID;
        Array<VertexBuf> vertices;
        Array<Face> faces;
        Array<Eigen::Vector3f> normals;
        bool hasNormals, hasColors, faceWraps;
        bool hasTexCoords, hasMaterial, inFrame;
};

class MeshBone {
    public:
        int matID;
        Array<VertexBone> vertices;
        Array<Face> faces;
        Array<Eigen::Vector3f> normals;
        GLuint* indices;
        VertexBuf* vertexBuf;
        bool hasNormals, hasColors, faceWraps;
        bool hasTexCoords, hasMaterial, inFrame;
        int nMaxBones, nMaxWeightsPerVertex, nMaxWeightsPerFace;
        Array<Bone> bones;
};

class Frame {
    public:
        QString frameName;
        Array<int> subFrames, subMeshes;
        float blendMult, blendSum;
        int priority;
        Eigen::Vector3f animPos, animScale;
        Eigen::Quaternionf animRot;
        Eigen::Transform3f transform;

        Frame();

        EIGEN_MAKE_ALIGNED_OPERATOR_NEW
};

class Key {
    public:
        int time;
        Eigen::Vector3f pos, scale;
        Eigen::Quaternionf rot;
        Eigen::Transform3f trans;

        Key();

        EIGEN_MAKE_ALIGNED_OPERATOR_NEW
};

class KeySet {
    public:
        int animType;
        Array<Key> keys;
};

class Anim {
    public:
        float blend;
        int frame;
        Array<KeySet> keySets;
};

class AnimSet {
    public:
        QString animSetName;
        float time, blend;
        int priority;
        Array<Anim> anims;
};

class Header {
    public:
        QString identifier;
        QString version;
        QString ogltype;
};

class Model3D {
    private:
    protected:
        QFile oglFile;
        QTextStream oglStream;
        Header header;
        void skipInputLine();
        int readInt();
        float readFloat();
        QString readLine();
    public:
        Model3D();
        Model3D(Header header, const QString& fileName);
        virtual ~Model3D();
        virtual void load();
        virtual void render();
        virtual void setFrame(unsigned int set, float time);
        virtual void setAnimSetBlend(unsigned int set, float time, float blend, int priority = 0);
        virtual void setAnimBlend(unsigned int set, unsigned int anim, float blend);
        virtual void calcAnim();
        virtual void setTextures(const QVector<int> &tex);
        Header getHeader();
};

class StaticModel : public Model3D {
    protected:
        void readMatBlock(int n);
        void readMeshBlock(int n);
        void autoGenerateNormals(int n);
        void loadTextures();
        void renderMesh(int n);
        void renderCartoonyMesh(int n, const Eigen::Vector3f &lightVec);
    public:
        Array<Mesh> meshes;
        Array<Material> materials;

        StaticModel(Header header, const QString& fileName);
        ~StaticModel();
        void load();
        void render();
        void renderCartoony(const Eigen::Vector3f &lightVec);
        void setFrame(unsigned int set, float time);
        void setAnimSetBlend(unsigned int set, float time, float blend, int priority = 0);
        void setAnimBlend(unsigned int set, unsigned int anim, float blend);
        void calcAnim();
        void setTextures(const QVector<int> &tex);
};

class FrameModel : public StaticModel {
    protected:
        void readFrameBlock(int n);
        void renderFrame(int n);
    public:
        Array<Frame> frames;
        Array<int> rootFrames;

        FrameModel(Header header, const QString& fileName);
        ~FrameModel();
        void load();
        void render();
        void setFrame(unsigned int set, float time);
        void setAnimSetBlend(unsigned int set, float time, float blend, int priority = 0);
        void setAnimBlend(unsigned int set, unsigned int anim, float blend);
        void calcAnim();
        void setTextures(const QVector<int> &tex);
};

class AnimModel : public FrameModel {
    protected:
        void readAnimSetBlock(int n);
        void readAnimBlock(int n, int p);
        void readKeySetBlock(int n, int p, int q);
    public:
        Array<AnimSet> animSets;
        bool addToOne;

        AnimModel(Header header, const QString& fileName);
        ~AnimModel();
        void load();
        void render();
        void setFrame(unsigned int set, float time);
        void setAnimSetBlend(unsigned int set, float time, float blend, int priority = 0);
        void setAnimBlend(unsigned int set, unsigned int anim, float blend);
        void calcAnim();
        void setTextures(const QVector<int> &tex);
};

class BoneModel : public AnimModel {
    protected:
        Eigen::Transform3f curViewMat2;
        Eigen::Matrix3f curNormMat, curNormMat2;

        void readMeshBlock(int n);
        void calcCombinedMatrix(int n);
        void renderMesh(int n);
        void renderFrame(int n);
        void renderCartoonyMesh(int n, const Eigen::Vector3f &lightVec);
        void renderCartoonyFrame(int n, const Eigen::Vector3f &lightVec);
    public:
        Array<MeshBone> meshes;
        Eigen::Transform3f curViewMat;

        BoneModel(Header header, const QString& fileName);
        ~BoneModel();
        void load();
        void render();
        void renderCartoony(const Eigen::Vector3f &lightVec);
        void setFrame(unsigned int set, float time);
        void setAnimSetBlend(unsigned int set, float time, float blend, int priority = 0);
        void setAnimBlend(unsigned int set, unsigned int anim, float blend);
        void calcAnim();
        void setTextures(const QVector<int> &tex);
        void setupBones(int n);
        Eigen::Vector3f getVertPos(int mID, int vID);

        void dump();

        EIGEN_MAKE_ALIGNED_OPERATOR_NEW
};

Model3D* loadModel(const QString& fileName);

#endif // _MODEL3D_H
