/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "Particles.h"

// FruitWars headers
#include "Camera.h"
#include "GameMath.h"
#include "Globals.h"
#include "Instance.h"
#include "Math3D.h"
#include "Textures.h"
#include "Weapons.h"

// KDE headers
#include <KGlobal>
#include <KStandardDirs>

// System headers
#include <GL/gl.h>
#include <GL/glu.h>

USING_PART_OF_NAMESPACE_EIGEN
using Eigen::Rotation2Df;
using Eigen::Transform3f;
using Eigen::Translation3f;

GLuint warpTex;
GLuint shadowTex;
GLuint aimTex;
GLuint starTex;
GLuint trailTex;
GLuint starShowTex[3];
GLuint indicTex;

BillBoard::BillBoard(const Vector3f& pos, float size, int tex)
    : pos(pos)
    , spin(0.0)
    , size(size)
    , scale(1.0, 1.0)
    , tex(tex)
    , blended(false)
    , blendSFac(0)
    , blendDFac(0)
{
}

BillBoard::BillBoard()
{
}

void BillBoard::render(const Camera& cam) const
{
    BS tempBS(pos, scale.maxCoeff() * size * 1.50);
    if (bsInFrustum(instFrust, tempBS)) {
        if (blended) {
            glEnable(GL_BLEND);
            glBlendFunc(blendSFac, blendDFac);
        }

        glBindTexture(GL_TEXTURE_2D, tex);
        glEnable(GL_TEXTURE_2D);

        Vector3f view = (cam.position() - pos).normalized();
        Vector3f up = (Vector3f(0.0, 1.0, 0.0) - (view * view.y())).normalized();
        Vector3f right = view.cross(up);
        Vector2f up2(0, 1);
        Vector2f right2(1, 0);
        Matrix2f rotMat = Rotation2Df(spin).toRotationMatrix();
        up2 = (rotMat * up2) * (scale[1] * size);
        right2 = (rotMat * right2) * (scale[0] * size);

        Matrix4f theMat;
        theMat << right.x(), up.x(), view.x(), 0,
                right.y(), up.y(), view.y(), 0,
                right.z(), up.z(), view.z(), 0,
                0      , 0   , 0     , 1;

        Vector3f lightVec(lightDir[0], lightDir[1], lightDir[2]);
        lightVec = getNormMat(Transform3f(theMat)).transpose() * lightVec;
        glNormal3f(lightVec.x(), lightVec.y(), lightVec.z());

        glPushMatrix();
        glTranslatef(pos.x(), pos.y(), pos.z());
        glMultMatrixf(theMat.data());

        // Draw the billboard
        glBegin(GL_QUADS);
            glTexCoord2f(0.0, 0.0);
            glVertex3f(up2[0] - right2[0], up2[1] - right2[1], 0.0);
            glTexCoord2f(1.0, 0.0);
            glVertex3f(up2[0] + right2[0], up2[1] + right2[1], 0.0);
            glTexCoord2f(1.0, 1.0);
            glVertex3f(-up2[0] + right2[0], -up2[1] + right2[1], 0.0);
            glTexCoord2f(0.0, 1.0);
            glVertex3f(-up2[0] - right2[0], -up2[1] - right2[1], 0.0);
        glEnd();

        glPopMatrix();

        glDisable(GL_TEXTURE_2D);

        if (blended) {
            glDisable(GL_BLEND);
        }
    }
}

Warp::Warp(const Vector3f& pos, float size)
    : pos(pos)
    , size(size)
    , scale(0.0)
    , state(0)
    , scaleSpeed(WARPSCALESPEED)
    , waitTime(WARPWAITTIME)
    , clipSide(0)
{
}

Warp::~Warp()
{
}

void Warp::render()
{
    BS theBS(pos, size * scale);

    if (state < 3 && bsInFrustum(instFrust, theBS)) {
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

        glBindTexture(GL_TEXTURE_2D, warpTex);
        glEnable(GL_TEXTURE_2D);

        glBegin(GL_QUADS);
            glNormal3f(-0.49, 0.81, -0.32);
            glTexCoord2f(0.0, 0.0);
            glVertex3f(pos.x() - size * scale, pos.y(), pos.z() + size * scale);
            glTexCoord2f(0.0, 1.0);
            glVertex3f(pos.x() - size * scale, pos.y(), pos.z() - size * scale);
            glTexCoord2f(1.0, 1.0);
            glVertex3f(pos.x() + size * scale, pos.y(), pos.z() - size * scale);
            glTexCoord2f(1.0, 0.0);
            glVertex3f(pos.x() + size * scale, pos.y(), pos.z() + size * scale);

            glNormal3f(-0.49, 0.81, -0.32);
            glTexCoord2f(0.0, 0.0);
            glVertex3f(pos.x() - size * scale, pos.y(), pos.z() + size * scale);
            glTexCoord2f(1.0, 0.0);
            glVertex3f(pos.x() + size * scale, pos.y(), pos.z() + size * scale);
            glTexCoord2f(1.0, 1.0);
            glVertex3f(pos.x() + size * scale, pos.y(), pos.z() - size * scale);
            glTexCoord2f(0.0, 1.0);
            glVertex3f(pos.x() - size * scale, pos.y(), pos.z() - size * scale);
        glEnd();

        glDisable(GL_TEXTURE_2D);

        glDisable(GL_BLEND);
    }
}

void Warp::update(float dt)
{
    switch (state) {
        case 0: {
            scale += dt * scaleSpeed;
            if (scale >= 1.0) {
                scale = 1.0;
                state = 1;
            }
            break;
        }
        case 1: {
            waitTime -= dt;
            if (waitTime <= 0) {
                waitTime = WARPWAITTIME;
                state = 2;
            }
            break;
        }
        default: {
            scale = scale - dt * scaleSpeed;
            if (scale <= 0.0) {
                scale = 0.0;
                state = 3;
            }
            break;
        }
    }
}

Aim::Aim(unsigned int id)
    : bill(Vector3f(), AIMSIZE, aimTex)
    , depth(0)
    , tDepth(0)
    , terDepth(0)
    , smooth(AIMSMOOTH)
    , viewMat(Matrix4f::Identity())
    , projMat(Matrix4f::Identity())
    , playId(id)
{
    bill.setBlended(true);
    bill.setBlendSFac(GL_SRC_ALPHA);
    bill.setBlendDFac(GL_ONE_MINUS_SRC_ALPHA);
}

Aim::Aim()
{
}

void Aim::render()
{
    tDepth = qMin(tDepth, terDepth);
    depth += smooth * (tDepth - depth);
    bill.setPos(myPlayers[playId].getInst()->getKeyPos(1) + myPlayers[playId].getThrowVec() * depth);
    bill.setSpin(depth * 0.1);
    glNormal3f(lightDir[0], lightDir[1], lightDir[2]);
    bill.render(myCam);
}

void Aim::calcAim()
{
    Vector3f view = myPlayers[playId].getThrowVec();
    Vector3f up(
            sin(myPlayers[playId].getRot().y() - PLAYERROTOFFSETY) * cos((myCam.targetAngle().x() + M_PI/2) + PLAYERLOOKANGLE),
            sin((myCam.targetAngle().x() + M_PI/2) + PLAYERLOOKANGLE),
            -cos(myPlayers[playId].getRot().y() - PLAYERROTOFFSETY) * cos((myCam.targetAngle().x() + M_PI/2) + PLAYERLOOKANGLE));
    Vector3f right = view.cross(up);
    Vector3f playPos = myPlayers[playId].getInst()->getKeyPos(1);

    Transform3f posMat(Translation3f(-playPos.x(), -playPos.y(), -playPos.z()));
    Matrix4f m;
    m << right.x(), right.y(), right.z(), 0.0,
            up.x()   , up.y()   , up.z()   , 0.0,
            view.x() , view.y() , view.z() , 0.0,
            0.0    , 0.0    , 0.0    , 1.0;
    viewMat = m;
    viewMat = viewMat * posMat;

    // Calc Proj Mat
    Transform3f combMat = viewMat;

    // Project positions
    Vector3f projPos[8];
    for (unsigned int i = 0; i < nplayers; i++) {
        if (i != playId) {
            projPos[i] = combMat * myPlayers[i].getColBS().pos;
        }
        projPos[i].x() = (projPos[i].x() / projPos[i].z()) * 128.0;
        projPos[i].y() = (projPos[i].y() / projPos[i].z()) * 128.0;
    }

    Vector3f rayStart = myPlayers[playId].getInst()->getKeyPos(1);
    Vector3f rayDir = myPlayers[playId].getThrowVec();
    float dist = 0.0;
    while (dist <= WEAPONBANANARANGE && myTer.getGroundHeight(Vector2f(rayStart.x(), rayStart.z())) + WEAPONBANANASCALE * WEAPONBANANACOLRADIUS < rayStart.y()) {
        dist += TERRAINAIMSTEPSIZE;
        rayStart += rayDir * TERRAINAIMSTEPSIZE;
    }
    terDepth = dist;

    // Evaluate Kevil Curve
    float c = KEVILCURVEC;
    float k = KEVILCURVEK;
    float d = KEVILCURVED;
    float mw = KEVILCURVEMW;
    float mr = WEAPONBANANARANGE;

    float totalWeight = 0.0;
    for (unsigned int i = 0; i < nplayers; i++) {
        if (i != playId && projPos[i].x() >= -128.0 && projPos[i].x() <= 128.0 &&
                projPos[i].z() >= 0.0    && projPos[i].z() <= terDepth &&
                projPos[i].y() >= -128.0 && projPos[i].y() <= 128.0) {
            float weight = c / (projPos[i].x() * projPos[i].x() + projPos[i].y() * projPos[i].y() + d);
            if (weight > mw) {
                weight = mw;
            }
            totalWeight += weight;
        }
    }
    float theDepth = mr;
    totalWeight = totalWeight + k / (totalWeight * totalWeight);
    for (unsigned int i = 0; i < nplayers; i++) {
        if (i != playId && projPos[i].x() >= -128.0 && projPos[i].x() <= 128.0 &&
                projPos[i].z() >= 0.0    && projPos[i].z() <= terDepth &&
                projPos[i].y() >= -128.0 && projPos[i].y() <= 128.0) {
            float weight = c / (projPos[i].x() * projPos[i].x() + projPos[i].y() * projPos[i].y() + d);
            if (weight > mw) {
                weight = mw;
            }
            theDepth += (weight / totalWeight) * (projPos[i].z() - mr);
        }
    }
    tDepth = theDepth;
}

Stars::Stars(int id, float size)
    : playId(id)
    , size(size)
    , spin(0.0)
    , alpha(1.0)
    , trans()
{
    for (int i = 0; i < 4; i++) {
        bill[i] = BillBoard(Vector3f(0, 0, 0), STARSIZE, starTex);
        bill[i].setBlended(true);
        bill[i].setBlendSFac(GL_SRC_ALPHA);
        bill[i].setBlendDFac(GL_ONE_MINUS_SRC_ALPHA);
    }
}

Stars::~Stars()
{
}

void Stars::update(float dt)
{
    spin += dt * STARSPINSPEED;
}

void Stars::render()
{
    glEnable(GL_COLOR_MATERIAL);
    glColor4f(1.0, 1.0, 1.0, alpha);
    for (int i = 0; i < 4; i++) {
        Vector3f relPos(sin(0.5 * i * M_PI + spin) * size + STARRELPOS[0], STARRELPOS[1], cos(0.5 * i * M_PI + spin) * size + STARRELPOS[2]);
        bill[i].setPos(trans * relPos);
        bill[i].render(myCam);
    }
    glColor4f(1.0, 1.0, 1.0, 1.0);
    glDisable(GL_COLOR_MATERIAL);
}

Trail::Trail(int count)
    : trailBegin(0)
    , trails(count)
{
}

Trail::~Trail()
{
}

void Trail::update(const Vector3f &pos1, const Vector3f& pos2)
{
    trails[trailBegin].segA = pos1;
    trails[trailBegin].segB = pos2;
    trailBegin = (trailBegin + 1) % trails.size();
}

void Trail::render()
{
    glNormal3f(lightDir[0], lightDir[1], lightDir[2]);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glBindTexture(GL_TEXTURE_2D, trailTex);
    glEnable(GL_TEXTURE_2D);

    glBegin(GL_TRIANGLE_STRIP);
        for (unsigned int i = 0; i < trails.size(); i++) {
            int k = (trailBegin + i) % trails.size();
            glTexCoord2f(0.0, 1.0 - (float)i / (trails.size() - 1));
            glVertex3f(trails[k].segA.x(), trails[k].segA.y(), trails[k].segA.z());
            glTexCoord2f(1.0, 1.0 - (float)i / (trails.size() - 1));
            glVertex3f(trails[k].segB.x(), trails[k].segB.y(), trails[k].segB.z());
        }
    glEnd();

    glBegin(GL_TRIANGLE_STRIP);
        for (unsigned int i = 0; i < trails.size(); i++) {
            int k = (trailBegin + i) % trails.size();
            glTexCoord2f(1.0, 1.0 - (float)i / (trails.size() - 1));
            glVertex3f(trails[k].segB.x(), trails[k].segB.y(), trails[k].segB.z());
            glTexCoord2f(0.0, 1.0 - (float)i / (trails.size() - 1));
            glVertex3f(trails[k].segA.x(), trails[k].segA.y(), trails[k].segA.z());
        }
    glEnd();

    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
}

StarShow::StarShow(const Vector3f& pos, float rad, int count)
    : parts(count)
    , maxRadius(rad)
    , spawnPos(pos)
    , dt(0.0)
    , lastTime(time())
    , active(false)
{
    for (int i = 0; i < count; i++) {
        parts[i].bill = BillBoard(Vector3f(0, 0, 0), STARSHOWSIZE, starShowTex[i % 3]);
        parts[i].bill.setBlended(true);
        parts[i].bill.setBlendSFac(GL_SRC_ALPHA);
        parts[i].bill.setBlendDFac(GL_ONE);
    }
}

StarShow::~StarShow()
{
}

void StarShow::update()
{
    dt = (time() - lastTime) / 1000.0;
    lastTime = time();

    if (active) {
        active = false;
        for (unsigned int i = 0; i < parts.size(); i++) {
            parts[i].angle += dt * parts[i].angSpeed;
            parts[i].bill.setSpin(parts[i].bill.getSpin() + dt * 4.0 * M_PI);
            parts[i].hSpeed += dt * parts[i].acc;
            parts[i].height += dt * parts[i].hSpeed;
            parts[i].alpha -= dt * parts[i].alphaSpeed;
            if (parts[i].alpha <= 0.00) {
                parts[i].alpha = 0.0;
            } else {
                active = true;
            }
        }
    }
}

void StarShow::render()
{
    if (active) {
        glNormal3f(lightDir[0], lightDir[1], lightDir[2]);
        glEnable(GL_COLOR_MATERIAL);
        for (unsigned int i = 0; i < parts.size(); i++) {
            if (parts[i].alpha > 0) {
                glColor4f(1.0, 1.0, 1.0, parts[i].alpha);
                parts[i].bill.setPos(spawnPos + Vector3f(sin(parts[i].angle) * parts[i].radius, parts[i].height, cos(parts[i].angle) * parts[i].radius));
                parts[i].bill.render(myCam);
            }
        }
        glColor4f(1.0, 1.0, 1.0, 1.0);
        glDisable(GL_COLOR_MATERIAL);
    }
}

void StarShow::activate()
{
    for (unsigned int i = 0; i < parts.size(); i++) {
        parts[i].alpha = 1.0;
        parts[i].alphaSpeed = 0.4 + frandom() * 0.6;
        parts[i].angle = frandom() * 2.0 * M_PI;
        parts[i].angSpeed = M_PI + frandom() * 2.0 * M_PI;
        parts[i].radius = maxRadius * 0.5 + frandom() * maxRadius * 0.5;
        parts[i].height = 0.0;
        parts[i].hSpeed = frandom() * 4.0;
        parts[i].acc = frandom() * 5.0;
    }
    active = true;
}

Indicator::Indicator(float size, int pid, int bid)
    : spin(0.0)
    , bill(Vector3f(0, 0, 0), size, indicTex)
    , playId(pid)
    , baskId(bid)
    , show(true)
{
    bill.setBlended(true);
    bill.setBlendSFac(GL_SRC_ALPHA);
    bill.setBlendDFac(GL_ONE_MINUS_SRC_ALPHA);
    bill.setScale(Vector2f(0.8, 1.0));
}

Indicator::Indicator()
{
}

Indicator::~Indicator()
{
}

void Indicator::render()
{
    if (show) {
        spin = atan2(myBaskets[baskId].getPos().z() - myPlayers[playId].getPos().z(),
                myBaskets[baskId].getPos().x() - myPlayers[playId].getPos().x()) - (myPlayers[playId].getRot().y() + 3.0 * M_PI / 2.0);
        bill.setPos(myPlayers[playId].getColBS().pos + Vector3f(0.0, 2.0, 0.0));
        bill.setSpin(spin);
        bill.render(myCam);
    }
}




void createWarpTex()
{
    unsigned char data[65536];
    for (int i = 0; i < 128; i++) {
        for (int j = 0; j < 128; j++) {
            int k = (i * 128 + j) * 4;
            float dist = (i - 63.5) * (i - 63.5) + (j - 63.5) * (j - 63.5);
            float dist2 = sqrt(dist);
            data[k] = qMax(0, (int)round((1.0 - 0.01 * ((dist2 - 50.0) * (dist2 - 50.0))) * 255));
            data[k] = 0;
            data[k+1] = qMin(qMax(0, 100 + (int)round((dist / 4096.0) * 155)), 255);
            data[k+2] = qMax(0, (int)round((1.0 - 0.04 * ((dist2 - 56.0) * (dist2 - 56.0))) * 255));
            data[k+3] = qMin(qMax(0, (int)round(((63.0 - dist2) / 13.0) * 255)), 200);
        }
    }

    glGenTextures(1, &warpTex);
    glBindTexture(GL_TEXTURE_2D, warpTex);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); // only first two can be used
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // all of the above can be used
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, 128, 128, GL_RGBA, GL_UNSIGNED_BYTE, data);
}

void createTrailTex()
{
    unsigned char data[65536];
    for (int i = 0; i < 128; i++) {
        for (int j = 0; j < 128; j++) {
            int k = (i * 128 + j) * 4;
            data[k] = 255;
            data[k+1] = 255 - (int)round(((127.00 - i) / 127.00) * 255);
            data[k+2] = 0;
            data[k+3] = (int)round(sqrt(((127.00 - i) / 127.0) * sqrt((fabs(j - 63.5) / 63.5))) * 255);
        }
    }

    glGenTextures(1, &trailTex);
    glBindTexture(GL_TEXTURE_2D, trailTex);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); // only first two can be used
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // all of the above can be used
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, 128, 128, GL_RGBA, GL_UNSIGNED_BYTE, data);
}

void createCartoonTex()
{
    unsigned char data[1535];
    for (int i = 0; i < 512; i++) {
        int k = i * 3;
        data[k] = data[k+1] = data[k+2] = (int)round((i / 128) * 65 + 60);
    }

    glGenTextures(1, &cartoonTex);
    glBindTexture(GL_TEXTURE_1D, cartoonTex);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); // only first two can be used
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // all of the above can be used
    glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    gluBuild1DMipmaps(GL_TEXTURE_1D, GL_RGB, 512, GL_RGB, GL_UNSIGNED_BYTE, data);
}

void createShadowTex()
{
    unsigned char data[65536];
    for (int i = 0; i < 128; i++) {
        for (int j = 0; j < 128; j++) {
            int k = (i * 128 + j) * 4;
            float dist = (i - 63.5) * (i - 63.5) + (j - 63.5) * (j - 63.5);
            float dist2 = sqrt(dist);
            data[k] = 0;
            data[k+1] = 0;
            data[k+2] = 0;
            if (dist2 >= 64.0) {
                data[k+3] = 0;
            } else {
                data[k+3] = (int)round((cos(dist2 * M_PI / 128.0)) * 150);
            }
        }
    }

    glGenTextures(1, &shadowTex);
    glBindTexture(GL_TEXTURE_2D, shadowTex);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); // only first two can be used
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // all of the above can be used
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, 128, 128, GL_RGBA, GL_UNSIGNED_BYTE, data);
}

void createCartoonShadowTex()
{
    unsigned char data[65536];
    for (int i = 0; i < 128; i++) {
        for (int j = 0; j < 128; j++) {
            int k = (i * 128 + j) * 4;
            float dist = (i - 63.5) * (i - 63.5) + (j - 63.5) * (j - 63.5);
            float dist2 = sqrt(dist);
            data[k] = 0;
            data[k+1] = 0;
            data[k+2] = 0;
            if (dist2 >= 63.5) {
                data[k+3] = 0;
            } else {
                data[k+3] = 100;
            }
        }
    }

    glGenTextures(1, &shadowTex);
    glBindTexture(GL_TEXTURE_2D, shadowTex);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); // only first two can be used
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // all of the above can be used
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, 128, 128, GL_RGBA, GL_UNSIGNED_BYTE, data);
}

void createAimTex()
{
    loadTexture(KGlobal::dirs()->findResource("appdata", "data/crosshair2.tga"), aimTex, glcontext);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
}

void createIndicTex()
{
    loadTexture(KGlobal::dirs()->findResource("appdata", "data/arrow.tga"), indicTex, glcontext);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
}

void createStarTex()
{
    loadTexture(KGlobal::dirs()->findResource("appdata", "data/dizzy.tga"), starTex, glcontext);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
}

void createStarShowTex()
{
    loadTexture(KGlobal::dirs()->findResource("appdata", "data/redstar.tga"), starShowTex[0], glcontext);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    loadTexture(KGlobal::dirs()->findResource("appdata", "data/greenstar.tga"), starShowTex[1], glcontext);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    loadTexture(KGlobal::dirs()->findResource("appdata", "data/bluestar.tga"), starShowTex[2], glcontext);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
}




