/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef _TERRAIN_H
#define _TERRAIN_H

#include "GameMath.h"
#include "Model3D.h"
#include <vector>

enum TexMethod {
    SingleTex, ColorTex, DetailTex, MixTex, ColorMix
};

enum CubeSide {
    SKY_TOP = 0,
    SKY_FRONT = 1,
    SKY_BACK = 2,
    SKY_LEFT = 3,
    SKY_RIGHT = 4
};

class VertBufTer {
public:
    Eigen::Vector3f pos;
    Eigen::Vector3f norm;
    Color color;
    Eigen::Vector2f texCoord1, texCoord2, texCoord3;
};

class Patch {
public:
    int bx, bz, ex, ez;
    AABB bb;
    int nStrips;
    std::vector<std::vector<int> > indices;

    Patch() {}
};

class TerrainFinish;

class Terrain {
private:
    friend class TerrainFinish;
    int xSegs, zSegs;
    float xSegSize, zSegSize;
    float minHeight, maxHeight;
    int patchSize;
    std::vector<float> height;
    std::vector<Eigen::Vector3f> normal;
    std::vector<Color> color;
    Eigen::Vector3f pos;
    TexMethod texMeth;
    GLuint textures[3];
    Eigen::Vector2f texScale[3];
    VertBufTer* vertBuf;
    bool useVertBuf;
    int nPatches;
    std::vector<Patch> patch;
    void renderPatch(int n);
public:
    Terrain(int xSegs, int zSegs, float sizeX, float sizeZ, float minHeight, float maxHeight);
    Terrain();
    ~Terrain();
    void useVertexBuffer(bool u);
    void loadHeightMap(const QString& fileName);
    void loadColorMap(const QString& fileName);
    void setHeight(int x, int z, float y);
    void setNormal(int x, int z, const Eigen::Vector3f& norm);
    void setColor(int x, int z, const Color& col);
    void position(const Eigen::Vector3f& pos);
    void calcNormals(int bx, int bz, int ex, int ez);
    void setTexMethod(TexMethod mTex);
    void setPatchSize(int size);
    void setMinHeight(float h);
    void setMaxHeight(float h);
    void setTexture(int tex, GLuint texId);
    void setTexScale(int tex, const Eigen::Vector2f& scale);
    void loadTex(int tex, const QString& texFile);
    void calcSeaTex(int tex, Terrain* terrain, int size, float maxDif);
    void updateVertBuf();
    void render();
    float getGroundHeight(const Eigen::Vector2f& pos) const;
    float getGroundHeight(const Eigen::Vector3f& pos) const { return getGroundHeight(Eigen::Vector2f(pos.x(), pos.z())); }
    Plane getGroundPlane(const Eigen::Vector2f& pos) const;
    Plane getGroundPlane(const Eigen::Vector3f& pos) const { return getGroundPlane(Eigen::Vector2f(pos.x(), pos.z())); }
    float getHeight(int x, int z) const;
    Eigen::Vector3f getNormal(int x, int z) const;
    GLuint getTex(int t) const;
    Eigen::Vector2f getTexScale(int t) const;
    Color getColor(int x, int z) const;
    inline int getXSegs() const { return xSegs; }
    inline int getZSegs() const { return zSegs; }
    inline float getXSegSize() const { return xSegSize; }
    inline float getZSegSize() const { return zSegSize; }
    inline Eigen::Vector3f& getPos() { return pos; }
    inline TexMethod& getTexMeth() { return texMeth; }
    inline int getPatchSize() const { return patchSize; }
    inline float getMinHeight() const { return minHeight; }
    inline float getMaxHeight() const { return maxHeight; }
    inline bool getUseVertBuf() const { return useVertBuf; }
};

class SkyCube {
private:
    GLuint textures[5];
    Eigen::Vector3f pos;
    float size;
public:
    SkyCube(float size);
    SkyCube();
    void setTexture(int tex, GLuint texId);
    void setSize(float size);
    void loadTex(int tex, const QString& texName);
    void position(const Eigen::Vector3f& pos);
    void render();
    GLuint getTexture(int tex);
    inline Eigen::Vector3f& getPos() { return pos; }
    inline float getSize() const { return size; }
};

class TerrainFinish {
public:
    TerrainFinish(Terrain* terrain, float height, int splitCount, int tileCount, int xDistCount, int zDistCount, Color color1, Color color2);
    TerrainFinish();
    ~TerrainFinish();
    void render();
    void setEdgeColor(const QString& texFile);
    void calcEdgeSea(Terrain* terrain, float maxDif);
    void setOuterColor(int level, Color color);
    void setCornerColor(int level, Color color);
private:
    float height;
    Color color1;
    Color color2;
    int xDistCount, zDistCount;
    Terrain* terrain;
    int splitCount;
    float xSegSize, zSegSize;
    int tileCount;
    int nVertices;
    VertBufTer* vertBuf;
    int nPatches;
    std::vector<Patch> patch;
    void renderPatch(int n);
    int getVertId(int x, int z);
};

#endif // _TERRAIN_H

