/*
    gamefumainwindow.cpp - Gamefu Application Main Window

    Copyright (c) 2005-2006 by Michaël Larouche       <larouche@kde.org>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/

#include "gamefumainwindow.h"
#include "gamefuview.h"

// Qt includes
#include <qapplication.h>
#include <qfile.h>
#include <qlabel.h>
#include <qstringlist.h>
#include <qtimer.h>
#include <qvaluelist.h>

// KDE includes
#include <kaction.h>
#include <kdebug.h>
#include <kfiledialog.h>
#include <kiconloader.h>
#include <kio/netaccess.h>
#include <klocale.h>
#include <kmainwindow.h>
#include <kstdaction.h>
#include <kurl.h>
#include <kxmlguifactory.h>

// For configuration dialog.
#include <kcmultidialog.h>
#include <ksettings/dialog.h>
#include <kwin.h>

// Gamefu include
#include <collectionaccess.h>
#include <collectionbuildjob.h>
#include <gamelauncher.h>
#include <gamefupreferences.h>
#include <notifymanager.h>
#include <systemmanager.h>
#include <threadweaver.h>
#include <gamefuplugin.h>
#include <gamefupluginmanager.h>

#include "gamefuapplication.h"
#include "firsttimewizard.h"

class GamefuMainWindow::Private
{
public:
	Private()
		: m_view(0),launcher(0)
	{}

	GamefuView *m_view;
	Gamefu::GameLauncher *launcher;
};

// TODO: Use a singleShot init() method.
// TODO: Make sure MainWindow is used properly, make sure it does nothing when loading plugins.
GamefuMainWindow::GamefuMainWindow()
    : KMainWindow(0, "GamefuMainWindow"), d(new Private())
{
	// Init Collection acces.
	Gamefu::CollectionAccess::self()->init();

	// Set the XML GUI file.
	setXMLFile("gamefuui.rc");
	
	// Create the main view.
	d->m_view = new GamefuView(this, "m_view");
	
	setCentralWidget(d->m_view);

	// Create the Game Launcher.
	d->launcher = new Gamefu::GameLauncher(this);
	
	// Setup the actions.
	setupActions();
	// Setup the Qt signals/slots connection
	setupConnections();

	// Restore the size from settings.
	readSettings();

	// Check if it's the first launch of Gamefu. Build the initial collection.
	checkFirstTime();
}

GamefuMainWindow::~GamefuMainWindow()
{
	kdDebug() << endl;

	delete d;
}

void GamefuMainWindow::setupActions()
{
	KStdAction::open(this, SLOT(fileOpen()), actionCollection());
	KStdAction::quit(kapp, SLOT(quit()), actionCollection());

	setStandardToolBarMenuEnabled(true);

	KStdAction::preferences(this, SLOT(showPreferences()), actionCollection());

	new KAction(i18n("First time Wizard"), SmallIcon("wizard"), 0, this, SLOT(showFirstTimeWizard()), actionCollection(), "collectionWizard");
	new KAction(i18n("Update Collection"), SmallIcon("reload"), 0, this, SLOT(buildCollection()), actionCollection(), "collectionUpdate");

	createGUI();
}

void GamefuMainWindow::setupConnections()
{
	connect(Gamefu::CollectionAccess::self(), SIGNAL(databaseError(const QString& )), this, SLOT(displayError(const QString&)));

	connect(d->m_view, SIGNAL(romExecuted(const QString& )), this, SLOT(executeRom(const QString& )));
	
	connect(Gamefu::PluginManager::self(), SIGNAL(pluginLoaded(Gamefu::Plugin*)), this, SLOT(slotPluginLoaded(Gamefu::Plugin*)));
	connect(Gamefu::PluginManager::self(), SIGNAL(pluginUnloaded(Gamefu::Plugin*)), this, SLOT(slotPluginUnloaded(Gamefu::Plugin*)));
}

bool GamefuMainWindow::queryExit()
{
	kdDebug() << endl;

	saveSettings();
	d->m_view->saveSettings();

	//Gamefu::PluginManager::self()->shutdown();

	// Always delete this main window.
	deleteLater();

	return true;
}

void GamefuMainWindow::checkFirstTime()
{
	if( GamefuPreferences::self()->gamefuFirstTime() )
	{
		showFirstTimeWizard();
	}
	else
	{
		show();
		fillRomList();
	}
}

void GamefuMainWindow::showFirstTimeWizard()
{
	FirstTimeWizard *wizard = new FirstTimeWizard(this, "firsttime_wizard");
	wizard->show();
	connect(wizard, SIGNAL(accepted()), this, SLOT(wizardAccepted()));
}

void GamefuMainWindow::wizardAccepted()
{
	kdDebug() << endl;

	show();

	if( GamefuPreferences::self()->gamefuFirstTime() )
	{
		GamefuPreferences::self()->setGamefuFirstTime(false);
		GamefuPreferences::self()->writeConfig();
	}

	// Don't start the collection building now, wait a little to show the main window.
	QTimer::singleShot(1000, this, SLOT(buildCollection()));
}

void GamefuMainWindow::fileOpen()
{
	// Get all available mime types from System manager
	QStringList romMimeType = Gamefu::SystemManager::self()->getRegistredMimeTypes();
	romMimeType.append( "application/x-zip" ); // Add ZIP file.
	KURL url = KFileDialog::getOpenURL(QString::null, romMimeType.join(" "), this, i18n("Open Location"));
	
	if(url.isEmpty())
		return;

	if(url.isLocalFile())
	{
		executeRom( url.path() );
	}
	else
	{
		Gamefu::NotifyManager::self()->longMessage( i18n("It doesn't support loading ROM from remote file at the moment."), Gamefu::NotifyManager::Sorry );
	}
}

void GamefuMainWindow::showPreferences()
{
	KSettings::Dialog *settings = new KSettings::Dialog( KSettings::Dialog::Static, this );
	settings->show();
}

void GamefuMainWindow::displayError(const QString &errorMessage)
{
	Gamefu::NotifyManager::self()->longMessage( errorMessage, Gamefu::NotifyManager::Error );
}

void GamefuMainWindow::executeRom(const QString &romUrl)
{
	if( !d->launcher->launchGame(romUrl) )
	{
		// TODO: Display ROM name from database instead.
		Gamefu::NotifyManager::self()->longMessage( 
					i18n("An error occured when trying to launch this rom:<br>%1"
							"<br><br>Please make sure that Gamefu's kfile_plugin and mimetypes are correctly registred or make sure this is a valid ROM file. Running \"kbuildsycoca\" will help.")
						.arg( KURL::fromPathOrURL(romUrl).fileName() ),
						Gamefu::NotifyManager::Error );
	}
}

void GamefuMainWindow::fillRomList()
{
	d->m_view->fillRoms();
}

void GamefuMainWindow::buildCollection()
{
	// Fill KURL list with selected Directories from Preferences.
	KURL::List urlList;
	QStringList selectedDirectories = GamefuPreferences::self()->collectionDirectories();
	QStringList::const_iterator it;
	QStringList::const_iterator itEnd = selectedDirectories.end();

	for( it = selectedDirectories.begin(); it != itEnd; ++it )
	{
		urlList.append( KURL::fromPathOrURL( *it ) );
	}

	Gamefu::CollectionBuildJob *job = new Gamefu::CollectionBuildJob(this, urlList, GamefuPreferences::self()->collectionRecursive());
	connect(job, SIGNAL(collectionBuildFinished()), this, SLOT(fillRomList()));
	connect(job, SIGNAL(collectionBuildFinished()), d->m_view, SLOT(fillFilters()));

	ThreadWeaver::instance()->onlyOneJob(job);
}

void GamefuMainWindow::readSettings()
{
	if( !GamefuPreferences::self()->mainWindowSize().isEmpty() )
		resize( GamefuPreferences::self()->mainWindowSize() );
}

void GamefuMainWindow::saveSettings()
{
	// Save size preferences
	GamefuPreferences::self()->setMainWindowSize( size() );
	GamefuPreferences::self()->writeConfig();
}

void GamefuMainWindow::slotPluginLoaded(Gamefu::Plugin *plugin)
{
	guiFactory()->addClient(plugin);
}

void GamefuMainWindow::slotPluginUnloaded(Gamefu::Plugin *plugin)
{
	guiFactory()->removeClient(plugin);
}

#include "gamefumainwindow.moc"
