/*
    gamefuview.cpp - Gamefu Main View

    Copyright (c) 2005      by Michaël Larouche       <larouche@kde.org>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/

#include "gamefuview.h"

// Qt includes
#include <qlayout.h>
#include <qsplitter.h>

// KDE includes
#include <kdebug.h>
#include <klistview.h>
#include <klistviewsearchline.h>
#include <klocale.h>
#include <kio/job.h>
#include <kmessagebox.h>

// Gamefu includes
#include "collectionaccess.h"
#include "collectionviewer.h"
#include "filterlist.h"
#include "gamefupreferences.h"
#include "metadataeditdialog.h"
#include "notifymanager.h"
#include "styledromdisplay.h"
#include "threadweaver.h"

#include "gamefustatusbar.h"
#include "statusbarnotify.h"
#include "listcollectionjob.h"

class GamefuView::Private
{
public:
	Private()
	 : display(0), filterList(0), collectionViewer(0), mainSplitter(0), verticalSplitter(0), statusBar(0)
	{}

	StyledRomDisplay *display;
	FilterList *filterList;
	CollectionViewer *collectionViewer;
	QSplitter *mainSplitter;
	QSplitter *verticalSplitter;
	GamefuStatusBar *statusBar;
};

GamefuView::GamefuView(QWidget *parent, const char *name)
 : QWidget(parent, name), d(new Private)
{
	QGridLayout *mainLayout = new QGridLayout(this, 2, 1);
	d->mainSplitter = new QSplitter(QSplitter::Horizontal, this);

	// Create the ROM browser
	unsigned int opts = (DISPLAY_ROMNAME | DISPLAY_SYSTEM | DISPLAY_TITLESCREEN |
		DISPLAY_DEVELOPER | DISPLAY_YEAR | DISPLAY_COUNTRY | DISPLAY_GENRE | DISPLAY_SCREENSHOTS);
	d->display = new StyledRomDisplay(opts, d->mainSplitter, "rom_html_display");

	d->verticalSplitter = new QSplitter(QSplitter::Vertical, d->mainSplitter);

	// Create the filter lists.
	d->filterList = new FilterList(d->verticalSplitter);
	connect(d->filterList, SIGNAL(refreshRoms(const QString&)), this, SLOT(fillRoms(const QString&)));

	d->collectionViewer = new CollectionViewer(d->verticalSplitter);
	connect(d->collectionViewer, SIGNAL(itemExecuted(const QString &)), this, SIGNAL(romExecuted( const QString& )));
	connect(d->collectionViewer, SIGNAL(itemSelected(const QString &)), d->display, SLOT(setRom(const QString& )));
	connect(this, SIGNAL(romsDeletedOrTrashed()), this, SLOT(updateCollectionTotal()));
	// Dispatched calls.
	connect(d->collectionViewer, SIGNAL(itemEditMetadata(const KURL::List&)), this, SLOT(editRomsInformation(const KURL::List&)));
	connect(d->collectionViewer, SIGNAL(itemDeleteRoms(const KURL::List&)), this, SLOT(deleteRoms(const KURL::List&)));
	connect(d->collectionViewer, SIGNAL(itemTrashRoms(const KURL::List&)), this, SLOT(trashRoms(const KURL::List&)));

	mainLayout->addWidget(d->mainSplitter, 0, 0);

	d->statusBar = new GamefuStatusBar(this);
	mainLayout->addWidget(d->statusBar, 1, 0);

	// Set notify backend.
	Gamefu::NotifyManager::self()->setNotifyBackend( new StatusBarNotify(d->statusBar, this, "StatusBarNotify") );
	// Fill the filters.
	fillFilters();

	// Restore the size from settings.
	readSettings();
}


GamefuView::~GamefuView()
{
	kdDebug() << endl;
	delete d;
}

void GamefuView::clearRomList()
{
	d->collectionViewer->clear();
}

void GamefuView::fillFilters()
{
	d->filterList->fillFilterLists();
}

void GamefuView::fillRoms(const QString &query)
{
	ListCollectionJob *listJob = new ListCollectionJob(this, d->collectionViewer, query);
	connect(listJob, SIGNAL(listCollectionFinished()), this, SLOT(updateCollectionTotal()));
	
	ThreadWeaver::instance()->onlyOneJob(listJob);
}

void GamefuView::updateSelectedRomInformation(const QString &romUrl)
{
	d->collectionViewer->updateSelectedRomInformation(romUrl);
	fillFilters();
}

void GamefuView::updateSelectedRomInformation(const KURL::List &romList)
{
	Q_UNUSED(romList);
	d->collectionViewer->updateSelectedRoms();
	fillFilters();
}

void GamefuView::readSettings()
{
	// Restore the size preferences
	if( !GamefuPreferences::self()->viewSplitterSizes().empty() )
		d->verticalSplitter->setSizes( GamefuPreferences::self()->viewSplitterSizes() );

	if( !GamefuPreferences::self()->mainWindowSplitterSizes().empty() )
		d->mainSplitter->setSizes( GamefuPreferences::self()->mainWindowSplitterSizes() );
}

void GamefuView::saveSettings()
{
	// Save size preferences
	GamefuPreferences::self()->setViewSplitterSizes( d->verticalSplitter->sizes() );
	GamefuPreferences::self()->setMainWindowSplitterSizes( d->mainSplitter->sizes() );
	GamefuPreferences::self()->writeConfig();
}

void GamefuView::editRomsInformation(const KURL::List &romList)
{
	kdDebug() << endl;

	MetadataEditDialog *editDlg;
	if( romList.count() == 1 )
	{
		KURL url = romList.first();
		QString romUrl = url.path();
		editDlg = new MetadataEditDialog(romUrl, romUrl, this);
	}
	else
	{
		editDlg = new MetadataEditDialog(romList, i18n("Information for one file", "Information for %n files", romList.count()), this);
	}

	// connect rom modification (Qt disconnects when editDlg is destroyed)
	connect(editDlg, SIGNAL(metadataChanged(const QString&)), this, SLOT(updateSelectedRomInformation(const QString&)));
	connect(editDlg, SIGNAL(metadataChanged(const KURL::List&)), this, SLOT(updateSelectedRomInformation(const KURL::List&)));
	// Update sidebar on modified item.
	connect(editDlg, SIGNAL(metadataChanged(const QString&)), d->display, SLOT(setRom(const QString& )));

	editDlg->show();
}

void GamefuView::deleteRoms(const KURL::List &romList)
{
	if( KMessageBox::questionYesNo(this, i18n("Are you sure you want to delete this file ?", "Are you sure you want to delete %n files.", romList.count()), i18n("Gamefu")) == KMessageBox::Yes )
	{
		KIO::Job *job = KIO::del(romList, false, true); // For first debug, show progress.
		connect(job, SIGNAL(result(KIO::Job*)), this, SLOT(romJobDone(KIO::Job*)));
	}
}

void GamefuView::trashRoms(const KURL::List &romList)
{
	if( KMessageBox::questionYesNo(this, i18n("Are you sure you want to trash this file ?", "Are you sure you want to trash %n files.", romList.count()), i18n("Gamefu")) == KMessageBox::Yes )
	{
		KIO::Job *job = KIO::trash(romList, true); // For first debug, show progress.
		connect(job, SIGNAL(result(KIO::Job*)), this, SLOT(romJobDone(KIO::Job*)));
	}
}

void GamefuView::romJobDone(KIO::Job *job)
{
	if( job->error() )
	{
		Gamefu::NotifyManager::self()->longMessage(i18n("An error occured while trying to delete or trash ROM file(s).\n%1").arg(job->errorString()), Gamefu::NotifyManager::Error);
		return;
	}

	KURL::List romList;
	// Determine which type is the resulting job
	// They don't have the same method to retrieve source URLs.
	KIO::DeleteJob *deleteJob = dynamic_cast<KIO::DeleteJob*>(job);
	KIO::CopyJob *copyJob = dynamic_cast<KIO::CopyJob*>(job);

	if(deleteJob)
	{
		kdDebug() << "Delete job" << endl;
		romList = deleteJob->urls();
	}
	else if(copyJob)
	{
		kdDebug() << "Trash job" << endl;
		romList = copyJob->srcURLs();
	}
	else
	{
		return;
	}

	if( !romList.isEmpty() )
	{
		KURL::List::ConstIterator it, itEnd = romList.constEnd();
		for(it = romList.constBegin(); it != itEnd; ++it)
		{
			KURL temp = *it;
			kdDebug() << "Removing from database: " << temp.path() << endl;
			Gamefu::CollectionAccess::self()->removeRomFromDatabase( temp.path() );
		}
		// Remove the items(there are selected) from the list view.
		d->collectionViewer->removeSelectedItems();

		Gamefu::NotifyManager::self()->shortMessage( i18n("Selected ROMs were removed succesfully.") );
		emit romsDeletedOrTrashed();
	}
}

int GamefuView::collectionSize() const
{
	return d->collectionViewer->collectionSize();
}

void GamefuView::updateCollectionTotal()
{
	d->statusBar->updateCollectionTotal( collectionSize() );
}

#include "gamefuview.moc"
