/*
    kfile_nds.cpp - NDS(Nintendo DS) ROM image KFile Plugin

    Copyright (c) 2006      by Thomas Köckerbauer     <k0055217@students.uni-linz.ac.at>

    Ported to Qt and adapted into Gamefu by:
    Copyright (c) 2006      by Michaël Larouche       <larouche@kde.org>

     *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/
#include "kfile_nds.h"

// KDE includes
#include <kgenericfactory.h>
#include <kglobal.h>
#include <klocale.h>

#include "ndsfile.h"

#include <iostream>

typedef KGenericFactory<ndsPlugin> ndsFactory;

K_EXPORT_COMPONENT_FACTORY(kfile_nds, ndsFactory( "kfile_nds" ))

ndsPlugin::ndsPlugin(QObject *parent, const char *name, const QStringList &args)
 : KFilePlugin(parent, name, args)
{
	KFileMimeTypeInfo* info = addMimeTypeInfo( "application/x-nds" );
	
	KFileMimeTypeInfo::GroupInfo* groupGeneral = 0L;
	groupGeneral = addGroupInfo(info, "romGeneralInfo", i18n("General"));
	
	KFileMimeTypeInfo::ItemInfo* item;

	item = addItemInfo(groupGeneral, "internalName", i18n("Internal Name"), QVariant::String);
	setHint(item, KFileMimeTypeInfo::Name);
	addItemInfo(groupGeneral, "country", i18n("Country"), QVariant::String);
	item = addItemInfo(groupGeneral, "editor", i18n("Developer"), QVariant::String);
	setHint(item, KFileMimeTypeInfo::Author);
	item = addItemInfo(groupGeneral, "romSize", i18n("Size"), QVariant::Int);
	setSuffix(item, i18n("MBit"));
	
	KFileMimeTypeInfo::GroupInfo* groupNds = 0L;
	groupNds = addGroupInfo(info, "ndsInfo", i18n("NDS Info"));
	addItemInfo(groupNds, "ndsGameCode", i18n("Game code"), QVariant::String);
	addItemInfo(groupNds, "ndsTitle", i18n("Title"), QVariant::String);
	item = addItemInfo(groupNds, "ndsDescription", i18n("Description"), QVariant::String);
	setHint(item, KFileMimeTypeInfo::Description);
	addItemInfo(groupNds, "ndsLoader", i18n("Loader"), QVariant::String);
	addItemInfo(groupNds, "ndsDownloadPlay", i18n("DS Download Play"), QVariant::String);
}

int ndsPlugin::getCurrentLanguage()
{
	int languageCode = 0;

	// Read the current language configured.
	QString language = KGlobal::locale()->language();

	if( language.startsWith("de") )
		languageCode = NDSFile::LANG_German;
	else if( language.startsWith("en") )
		languageCode = NDSFile::LANG_English;
	else if( language.startsWith("jp") )
		languageCode = NDSFile::LANG_Japanese;
	else if( language.startsWith("fr") )
		languageCode = NDSFile::LANG_French;
	else if( language.startsWith("it") ) 
		languageCode = NDSFile::LANG_Italian;
	else if (language.startsWith("es")) 
		languageCode = NDSFile::LANG_Spanish;

	return languageCode;
}

QString ndsPlugin::getCountryNameFromCode(const QString &countryCode)
{
	QString country;
	if( countryCode.contains("JPN") )
	{
		country = i18n("Japan");
	}
	else if( countryCode.contains("USA") )
	{
		country = i18n("America");
	}
	else if( countryCode.contains("EUR") )
	{
		country = i18n("Europe");
	}
	else if( countryCode.contains("ITA") )
	{
		country = i18n("Italy");
	}
	else if( countryCode.contains("SPA") )
	{
		country = i18n("Spain");
	}
	else if( countryCode.contains("HOL") )
	{
		country = i18n("Holland");
	}
	else if( countryCode.contains("KOR") )
	{
		country = i18n("Korea");
	}

	return country;
}

bool ndsPlugin::readInfo( KFileMetaInfo& info, uint /*what*/)
{
	if( info.path().isEmpty() )
		return false;

	NDSFile ndsFile(info.path());

	if( !ndsFile.isValid() )
		return false;

	ndsFile.readInfo();

	QString internalName, title, gameCode, description, loader, countryCode, maker, country;
	int romSize = 0;
	bool dsDownloadPlay = false;

	internalName = ndsFile.internalName();
	gameCode = ndsFile.gameCode();
	maker = ndsFile.maker();
	countryCode = ndsFile.countryCode();
	romSize = ndsFile.romSize();
	gameCode = QString::fromUtf8("%1 (NTR-%2-%3)").arg(gameCode).arg(gameCode).arg(countryCode);
	country= getCountryNameFromCode(countryCode);

	if( !ndsFile.isGstDump() )
	{
		romSize = ndsFile.romSize();
		
		NDSFile::LongTitleLang foundLanguage;
		foundLanguage = (NDSFile::LongTitleLang)getCurrentLanguage();

		title = ndsFile.title( foundLanguage );
		if( title.isEmpty() )
		{
			foundLanguage = NDSFile::LANG_English;
			title = ndsFile.title( foundLanguage );
	
			if( title.isEmpty() )
			{
				foundLanguage = NDSFile::LANG_Japanese;
				title = ndsFile.title( NDSFile::LANG_Japanese );
			}
		}
		description = ndsFile.description( foundLanguage );

		dsDownloadPlay = ndsFile.isDsDownloadPlay();
	}

	if( ndsFile.hasEmbeddedLoader() )
		loader = i18n("Embedded loader");
	else if( ndsFile.isGstDump() )
		loader = i18n("GST Dump");
	else if( ndsFile.hasNDSmallLoader() ) 
		loader = i18n("NDSLoader/NDSmall loader (rename to .nds.gba)");
	else 
		loader = i18n("no known loader present");

	KFileMetaInfoGroup groupGeneral = appendGroup(info, "romGeneralInfo");
	if( title != '.' )
		appendItem(groupGeneral, "internalName", internalName);
	appendItem(groupGeneral, "country", country);
	appendItem(groupGeneral, "editor", maker);
	appendItem(groupGeneral, "romSize", romSize);

	KFileMetaInfoGroup groupNds = appendGroup(info, "ndsInfo");
	appendItem(groupNds, "ndsGameCode", gameCode);
	appendItem(groupNds, "ndsTitle", title);
	appendItem(groupNds, "ndsDescription", description);
	appendItem(groupNds, "ndsLoader", loader);
	appendItem(groupNds, "ndsDownloadPlay", dsDownloadPlay ? i18n("Yes") : i18n("No") );

	return true;
}

#include "kfile_nds.moc"
