/*
    collectionbuildjob.cpp - Build database

    Copyright (c) 2006      by Michaël Larouche       <larouche@kde.org>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/
#include "collectionbuildjob.h"

// Gamefu includes
#include "collectionaccess.h"
#include "rommetadatafactory.h"
#include "rommetadata.h"
#include "rommetainformation.h"
#include "systemmanager.h"

// Qt includes
#include <qvaluelist.h>
#include <qvaluestack.h>

// KDE includes
#include <kdebug.h>
#include <kdirlister.h>
#include <klocale.h>

namespace Gamefu 
{

class CollectionBuildJob::Private
{
public:
	Private()
	 : recursive(false), scanFinished(false), dirLister(0)
	{
	}
	~Private()
	{
		delete dirLister;
	}
	bool recursive;

	volatile bool scanFinished;

	KDirLister *dirLister;
	QValueStack<KURL> directoriesStack;
	QValueList<KURL> directoryList;
};

CollectionBuildJob::CollectionBuildJob(QObject *parent, const KURL::List &directoryList, bool recursive)
	: ThreadWeaver::DependentJob(parent, "CollectionBuildJob"), d(new Private)
{
	d->recursive = recursive;
	setDirectoryList(directoryList);

	d->dirLister = new KDirLister();
	QStringList mimeTypeList(Gamefu::SystemManager::self()->getRegistredMimeTypes());

	// Add the mimetype for directories (for recursing into sub-directories)
	mimeTypeList.append( QString::fromUtf8("inode/directory") );
	// Add mimetype for ZIP files.
	mimeTypeList.append( QString::fromUtf8("application/x-zip") );

	d->dirLister->setMimeFilter(mimeTypeList);
	connect(d->dirLister, SIGNAL(completed(const KURL &)), this, SLOT(slotDirectoryFinished()));
	connect(d->dirLister, SIGNAL(newItems(const KFileItemList &)), this, SLOT(slotNewFiles(const KFileItemList& )));
	connect(d->dirLister, SIGNAL(completed()), this, SLOT(slotListingComplete()));
	
	setDescription( i18n("Building Collection") );
}

CollectionBuildJob::~CollectionBuildJob()
{
	delete d;
}

void CollectionBuildJob::setDirectoryList(const KURL::List &directoryList)
{
	// Erase the old list.
	d->directoriesStack.clear();
	d->directoryList.clear();

	// Add the directories to the URL stack.
	KURL::List::const_iterator it;
	KURL::List::const_iterator end = directoryList.end();
	for(it = directoryList.begin(); it != end; ++it)
	{
		d->directoriesStack.push(*it);
		d->directoryList.append(*it);
	}
}

void CollectionBuildJob::slotNewFiles(const KFileItemList &list)
{
	KFileItem *item;
	QPtrListIterator<KFileItem> it( list );

	while ( (item = it.current()) != 0 ) 
	{
		// Stop loop when the job is aborted.
		if( isAborted() )
		{
			kdDebug() << "Job was aborted." << endl;
			break;
		}

		// Do not try to extract metainformation from Directory.
		if(item->isDir())
		{
			// If we recurse into sub-directories, add to the stack.
			if(d->recursive)
			{
				kdDebug() << "Adding a new directory to the directories stack." << endl;
				d->directoriesStack.push(item->url());
				d->directoryList.append(item->url());
			}
		}

		++it;
	}
}

void CollectionBuildJob::slotInsertItems(const KFileItemList &list)
{
	KFileItem *item;
	QPtrListIterator<KFileItem> it( list );

	while ( (item = it.current()) != 0 )
	{
		// Stop loop when the job is aborted.
		if( isAborted() )
		{
			kdDebug() << "Job was aborted." << endl;
			break;
		}

		// Do not try to extract metainformation from Directory.
		if( !item->isDir() )
		{
			kdDebug() << "Listing " << item->url().prettyURL() << endl;

			// Do not extract the meta-information if the rom is already in database.
			if( !CollectionAccess::self()->isRomUrlInDatabase( item->url().path() ) )
			{
				RomMetaData *metaDataExtractor = RomMetaDataFactory::self()->createRomMetaData(item->url().path(), item->mimetype());
				RomMetaInformation romMetaInfo = metaDataExtractor->extractMetaInformation(item);

				// Satefy check because a zip file can contain no valid ROM image.
				if( romMetaInfo.isValid() )
				{
					kdDebug() << "Insert new \"" << item->url().fileName() << "\" in database." << endl;
					CollectionAccess::self()->insertNewRom(romMetaInfo);
				}
			}
		}

		incrementProgress();
		++it;
	}
}

void CollectionBuildJob::slotListingComplete()
{
	d->scanFinished = true;
}

void CollectionBuildJob::slotDirectoryFinished()
{
	kdDebug() << endl;
	// Start another scanning if the directories stack is not empty
	if( !d->directoriesStack.isEmpty() )
	{
		kdDebug() << "Starting another directory." << endl;
		d->dirLister->openURL(d->directoriesStack.pop(), true);
	}
}

bool CollectionBuildJob::doJob()
{
	kdDebug() << endl;
	
	if( !d->directoriesStack.isEmpty() )
	{
		kdDebug() << "Start listing of directories" << endl;
		if( !d->dirLister->openURL(d->directoriesStack.pop(), true) )
		{
			// Abort the job
			abort();
		}
	}

	while( !isAborted() )
	{
		if(d->scanFinished)
		{
			kdDebug() << "BEGIN ROM TRANSACTION..." << endl;
			CollectionAccess::self()->beginRomTransaction();

			QValueList<KURL>::const_iterator it, itEnd = d->directoryList.constEnd();
			// Get total of items listed to begin show the progress bar.
			int totalItems = 0;
			for(it = d->directoryList.constBegin(); it != itEnd; ++it)
			{
				KFileItemList fileList = d->dirLister->itemsForDir(*it);
				totalItems += fileList.count();
			}
			setProgressTotalSteps(totalItems);
			for(it = d->directoryList.constBegin(); it != itEnd; ++it)
			{
				KFileItemList fileList = d->dirLister->itemsForDir(*it);
				slotInsertItems(fileList);
			}

			Gamefu::CollectionAccess::CommitMode commitMode;
			if( isAborted() )
			{
				kdDebug() << "Job was aborted. Drop temp table in collection" << endl;
				commitMode = Gamefu::CollectionAccess::CommitNone;
			}
			else
			{
				kdDebug() << "Commit new roms in collection." << endl;
				commitMode = Gamefu::CollectionAccess::CommitAll;
			}

			CollectionAccess::self()->endRomTransaction(commitMode);

			break;
		}
	}

	return !isAborted();
}

void CollectionBuildJob::completeJob()
{
	emit collectionBuildFinished();
}

}

#include "collectionbuildjob.moc"
