/*
    collectionbuildjob.h - Build database

    Copyright (c) 2006      by Michaël Larouche       <larouche@kde.org>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/
#ifndef GAMEFUCOLLECTIONBUILDJOB_H
#define GAMEFUCOLLECTIONBUILDJOB_H

#include <kdemacros.h>
#include <kurl.h>
#include "rommetainformation.h" // Forward declaration of container typedef is a pain.
#include <kfileitem.h> // idem as above.

#include "threadweaver.h"

namespace Gamefu 
{

/**
 * @brief Threaded Collection building using ThreadWeaver::Job
 * Collection Builder is the service that list a list of directories defined in a configuration file. 
 * After retriving the metadata, it adds the info to the database.
 *
 * During listing of all the files, if a file is a zip file, the collection builder uncompress the zip file to see 
 * if the zip file contains a ROM image. Using the temp file (unzipped), the Collection Builder will be able to retrive 
 * metadata from the file.
 *
 * For each file, Collection Builder should use a class obtained from RomMetaDataFactory. 
 * If no specific class is required, the RomMetaDataFactory will return the DefaultRomMetaData 
 * that read the metadata from KFileMetaInfo. 
 * This will be useful for exemple for the MAME system to use MAME database to retrive meta information.
 *
 * When it has retrived the meta information, it push the meta information to database using Collection Access service.
 *
 * @author Michaël Larouche <larouche@kde.org>
*/
class CollectionBuildJob : public ThreadWeaver::DependentJob
{
	Q_OBJECT
public:
	/**
	 * @brief Create a new collection building job.
	 * @param parent The job watch for the parent, if the parent gets deleted, the job abort.
	 * @param directoryList List of directories to scan.
	 * @param recursive Recurse through sub-directories.
	 */
	CollectionBuildJob(QObject *parent, const KURL::List &directoryList, bool recursive = true);
	/**
	 * d-tor
	 */
	~CollectionBuildJob();

signals:
	/**
	 * Collection build is finished.
	 */
	void collectionBuildFinished();

private slots:
	/**
	 * Called when KDirLister found new files.
	 * Now it's just do the recursive magic.
	 */
	void slotNewFiles(const KFileItemList& list);
	/**
	 * @brief Insert items into database.
	 * Insert the file list into the database. It expect filtered items.
	 * Called when the listing of KDirLister is finished.
	 */
	void slotInsertItems(const KFileItemList &list);
	/**
	 * Called when KDirLister has finished all its job.
	 */
	void slotListingComplete();
	/**
	 * Called when KDirLister has finished to list a directory.
	 */
	void slotDirectoryFinished();

private:
	/**
	 * @internal
	 * @brief Set directory list internally.
	 * Helper method that setup the directory queue(directories to list)
	 * and the directories list (directories listed).
	 * @param directoryList directory queue
	 */
	void setDirectoryList(const KURL::List &directoryList);

	/**
	 * @internal
	 * This is here the real proceeding is done.
	 * Do the job and return the status.
	 * @return true if the job completed succesfully.
	 */
	virtual bool doJob();
	/**
	 * @internal
	 * Emit collectionBuildFinished().
	 */
	virtual void completeJob();

	class Private;
	Private *d;
};

}

#endif
