/*
    systemmanager.cpp - Singleton class to manager loaded System plugins.

    Copyright (c) 2005      by Michaël Larouche       <larouche@kde.org>

    Portions from Kopete
    Copyright (c) 2001-2005 Kopete Developers <kopete-devel@kde.org>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/
#include "systemmanager.h"

// Qt include

// KDE includes
#include <kdebug.h>
#include <kstaticdeleter.h>

// Gamefu include
#include "gamefuplugin.h"
#include "gamefupluginmanager.h"
#include "system.h"

namespace Gamefu
{

class SystemManager::Private
{
public:
	/**
	 * It's all the registred MIME type loaded by the system plugins.
	 */
	QStringList registredMimeTypes;
	/**
	 * All loaded systems
	 */
	SystemList loadedSystems;
};

static KStaticDeleter<SystemManager> systemManagerDeleter;

SystemManager *SystemManager::s_self = 0L;
SystemManager *SystemManager::self()
{
	if(!s_self)
	{
		systemManagerDeleter.setObject(s_self, new SystemManager);
	}

	return s_self;
}

SystemManager::SystemManager(QObject *parent, const char *name)
 : QObject(parent, name), d(new Private())
{
	// Convert loaded plugins to system if needed.
	connect( Gamefu::PluginManager::self(), SIGNAL(pluginLoaded(Gamefu::Plugin *)), this, SLOT(slotPluginLoaded(Gamefu::Plugin *)));
}

SystemManager::~SystemManager()
{
	kdDebug() << endl;
	delete d;
}

QStringList SystemManager::getRegistredMimeTypes() const
{
	return d->registredMimeTypes;
}

QString SystemManager::guessSystemNameFromMimeType(const QString &mimeType) const
{
	QString systemDisplayName;
	SystemList::ConstIterator it;
	SystemList::ConstIterator itEnd = d->loadedSystems.constEnd();

	for( it = d->loadedSystems.constBegin(); it != itEnd; ++it )
	{
		System *system = static_cast<System*>(*it);
		QStringList mimeTypeList = system->getRegistredSystemMimeTypes();
		for(QStringList::Iterator stringIt = mimeTypeList.begin(); stringIt != mimeTypeList.end(); ++stringIt)
		{
			// Current system use that mime type.
			if(*stringIt == mimeType)
			{
				systemDisplayName = system->systemName();
				break;
			}
		}
	}

	return systemDisplayName;
}

SystemList SystemManager::loadedSystems() const
{
	return d->loadedSystems;
}

void SystemManager::slotPluginLoaded(Gamefu::Plugin *plugin)
{
	System *system = dynamic_cast<System*>(plugin);
	// This is a system plugin)
	if(system)
	{
		kdDebug() << "Casting plugin to system " << system->systemName() << endl;
		// Adding system mimetype to registred mimetypes.
		d->registredMimeTypes += system->getRegistredSystemMimeTypes();
		d->loadedSystems.append( system );
	}
}

}
#include "systemmanager.moc"
