/*
    Copyright 2009  Stanislas Marquis <smarquis@chaosorigin.com>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "json.h"

int json_parse( char* s, char** ptr )
{
    assert( s );

    while ( *s && isspace( *s ) )
        ++s;

    *ptr = s;
    if ( !*s )
        return JSON_END;

    switch ( *s )
    {
        case '{': return JSON_LBRACE;
        case '}': return JSON_RBRACE;
        case '"': return JSON_STRING;
        case ':': return JSON_COLON;
        case ',': return JSON_COMMA;
        case '[': return JSON_LBRACKET;
        case ']': return JSON_RBRACKET;
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case '+':
        case '-':
        case '.':
            return JSON_NUMBER;
        default:
            if ( !strncmp( s, "true", 4 ) )
                return JSON_TRUE;
            if ( !strncmp( s, "false", 5 ) )
                return JSON_FALSE;
            if ( !strncmp( s, "null", 4 ) )
                return JSON_NULL;
            return JSON_END;
    }
}

Bool json_string( char* s, char** start, char** end, char** next )
{
    assert( *s == '"' );

    char* p;
    ++s;
    if ( !*s )
        return False;
    *start = s;

    process:

    for ( ; *s; ++s )
    {
        switch( *s )
        {
            case '"':
                *end = s - 1;
                *next = s + 1;
                return True;
            case '\\':
                goto checkbs;
            default:
                continue;
        }
    }

    *end = s;
    return False;

    checkbs:

    p = s + 1;
    switch ( *p )
    {
        case '\\':
            s += 2;
            goto process;
        case '"':
        case '/':
        case 'b':
        case 'f':
        case 'n':
        case 'r':
        case 't':
            ++s;
            goto process;
        case 'u':
            goto process;/* todo */
        default:
            *next = *end = s;
            return False;
    }
}

Bool json_number( char* s, char** start, char** end, char** next )
{
    assert( *s );

    *start = s;
    const size_t sz = strspn( s, "0123456789+-." );
    if ( !sz )
        return False;
    *end = *start + sz - 1;
    *next = *start + sz;
    return True;
}

int json_number_int( char* start, char* end )
{
    assert( end - start < 127 );

    char buf[128];
    strncpy( buf, start, end - start + 1 );
    buf[ end - start + 1 ] = '\0';

    errno = 0;
    int i = strtol( buf, NULL, 10 );

    if ( errno )
    {
        Debug(
            printf( "-- json_number_int: cant convert string:\n"
                    "%s\n", buf );
        )
        return 0;
    }
    return i;
}

double json_number_float( char* start, char* end )
{
    assert( end - start < 127 );

    char buf[128];
    strncpy( buf, start, end - start + 1 );
    buf[ end - start + 1 ] = '\0';

    errno = 0;
    double d = strtod( buf, NULL );

    if ( errno )
    {
        Debug(
            printf( "-- json_number_float: cant convert string:\n"
                    "%s\n", buf );
        )
        return 0;
    }
    return d;
}

#ifndef NDEBUG

void json_unittest()
{
    char js[] = "{ \"objname\": \"objval\","
                "  \"test_int\" : 3,"
                "  \"test_float\" : 2.5"
                "}\n";

    Bool b;
    int ret, i;
    double d;
    char *ptr = js;
    char* st = NULL;
    char* nd = NULL;

    ret = json_parse( ptr, &ptr );
    assert( ret == JSON_LBRACE );

    ret = json_parse( ptr + 1, &ptr );
    assert( ret == JSON_STRING );

    b = json_string( ptr, &st, &nd, &ptr );
    assert( b );
    assert( strncmp( st, "objname", nd - st ) == 0 );

    ret = json_parse( ptr, &ptr );
    assert( ret == JSON_COLON );

    ret = json_parse( ptr + 1, &ptr );
    assert( ret == JSON_STRING );

    b = json_string( ptr, &st, &nd, &ptr );
    assert( b );
    assert( strncmp( st, "objval", nd - st ) == 0 );

    ret = json_parse( ptr, &ptr );
    assert( ret == JSON_COMMA );

    ret = json_parse( ptr + 1, &ptr );
    assert( ret == JSON_STRING );

    b = json_string( ptr, &st, &nd, &ptr );
    assert( b );
    assert( strncmp( st, "test_int", nd - st ) == 0 );

    ret = json_parse( ptr, &ptr );
    assert( ret == JSON_COLON );

    ret = json_parse( ptr + 1, &ptr );
    assert( ret == JSON_NUMBER );

    b = json_number( ptr, &st, &nd, &ptr );
    assert( b );

    i = json_number_int( st, nd );
    assert( i == 3 );

    ret = json_parse( ptr, &ptr );
    assert( ret == JSON_COMMA );

    ret = json_parse( ptr + 1, &ptr );
    assert( ret == JSON_STRING );

    b = json_string( ptr, &st, &nd, &ptr );
    assert( b );
    assert( strncmp( st, "test_float", nd - st ) == 0 );

    ret = json_parse( ptr, &ptr );
    assert( ret == JSON_COLON );

    ret = json_parse( ptr + 1, &ptr );
    assert( ret == JSON_NUMBER );

    b = json_number( ptr, &st, &nd, &ptr );
    assert( b );

    d = json_number_float( st, nd );
    assert( d == 2.5 );

    ret = json_parse( ptr, &ptr );
    assert( ret == JSON_RBRACE );

    ret = json_parse( ptr + 1, &ptr );
    assert( ret == JSON_END );
}

#endif /* !NDEBUG */
