/***************************************************************************
 *   Copyright 2009 Zeng Huan <zh.issac@gmail.com>
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "blender-texture.h"

#include <cmath>
#include <QColor>

Kolf::BlenderTexture::BlenderTexture(const QImage& image)
	: QImage(image.convertToFormat(QImage::Format_ARGB32))
{
	m_range[0] = 0;
	m_range[1] = 128;
	m_range[2] = 255;
}

int* Kolf::BlenderTexture::range()
{
	return m_range;
}

void Kolf::BlenderTexture::setRange(int min, int opt, int max)
{
	m_range[0] = min;
	m_range[1] = opt;
	m_range[2] = max;
}

QColor Kolf::BlenderTexture::pixel(int x, int y, int z, Kolf::BlenderCurve curveShape) const
{
	//retrieve color
	const uint* dataPtr = reinterpret_cast<const uint*>(bits());
	x %= width();
	y %= height();
	QColor color = QColor::fromRgb(dataPtr[x + y * width()]);
	//change alpha depending on Z coordinate
	int min = m_range[0], opt = m_range[1], max = m_range[2];
	if (z == opt)
		color.setAlphaF(1.0);
	else if (z >= min && z < opt)
		color.setAlphaF(curve(opt - z, opt - min, curveShape));
	else if (z > opt && z <= max)
		color.setAlphaF(curve(z - opt, max - opt, curveShape));
	else
		color.setAlphaF(0.0);
	return color;
}

qreal Kolf::BlenderTexture::curve(int delta, int range, Kolf::BlenderCurve curveShape) const
{
	//The return value should range in 0.0 ~ 1.0
	if(range == 0)
		return 0.0;
	qreal input = delta / (qreal)range;
	qreal x1 = 0.5, y1 = 0.8;
	qreal x2 = 0.8, y2 = 0.3;
	switch (curveShape)
	{
		case LinearCurve:
			return 1.0 - input;
		case LogarithmicCurve:
			return qMax(1.0 - std::log(input + 1.0), 0.0);
		case PotentialCurve:
			return qMax(std::pow(1.0 - input, 1.0/3.0), 0.0);
		case PolylineCurve:
			/*
			 * 1  = k * 0 + b
			 * y1 = k * x1 + b
			 * => k = (y1 -1) / x1, b = 1
			 *
			 */
			if(input <= x1)
				return (y1 - 1.0) / x1 * input + 1.0;
			/*
			 * y1 = k * x1 + b
			 * y2 = k * x2 + b
			 *
			 */
			if(input > x1 && input < x2)
				return ((y1 -y2) * input + (x1 - x2) * y1 - (y1 - y2) * x1) / (x1 - x2);
			if(input >= x2)
				return y2;
		default: //This should never be reached.
			return 1.0;
	}
}
