/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "geometry-wall.h"

#include <cmath>
#include <ode/ode.h>

const qreal Kolf::WallGeometry::Thickness = 2.0;

Kolf::WallGeometry::WallGeometry(dSpaceID spaceID, Kolf::Object* object)
	: Kolf::Geometry(dCreateGeomTransform(spaceID), object)
	, m_innerID(dCreateBox(0, 0, 0, 0)) //null-size box (and no space)
{
	dGeomTransformSetGeom(id(), m_innerID);
	dGeomTransformSetInfo(id(), 1); //see the ODE user guide for the meaning of "1"
}

Kolf::WallGeometry::~WallGeometry()
{
	dGeomDestroy(m_innerID);
}

Kolf::VectorF Kolf::WallGeometry::extent() const
{
	return m_extent;
}

void Kolf::WallGeometry::setExtent(const Kolf::VectorF& extent)
{
	if (m_extent != extent)
	{
		m_extent = extent;
		Kolf::Geometry::update();
		//set new size of inner geometry
		QVectorF extent2D = extent.to2D();
		const qreal extent2DLength = sqrt(extent2D * extent2D);
		dGeomBoxSetLengths(m_innerID, 2 * extent2DLength, Thickness, extent.z());
		//set new rotation for outer geometry
		const qreal angle = atan2(extent.y(), extent.x());
		dMatrix3 matrix;
		dRFromAxisAndAngle(matrix, 0, 0, 1, angle);
		dGeomSetRotation(m_innerID, matrix);
	}
}

void Kolf::WallGeometry::createOutlines()
{
	const QVectorF extent2D(m_extent.x(), m_extent.y());
	const qreal extentAngle = atan2(m_extent.y(), m_extent.x());
	const qreal paddingAngle = extentAngle + acos(-1.0f) / 2; //acos(-1) == pi
	const qreal padding = Kolf::Geometry::ActivationOutlinePadding;
	QVectorF padding2D(padding * cos(paddingAngle), padding * sin(paddingAngle));
	//create outline (essentially a rectangle that is aligned with the wall)
	QPainterPath interactionOutline;
	interactionOutline.moveTo(extent2D + padding2D);
	interactionOutline.lineTo(extent2D - padding2D);
	interactionOutline.lineTo(-extent2D - padding2D);
	interactionOutline.lineTo(-extent2D + padding2D);
	interactionOutline.closeSubpath();
	QPainterPath activationOutline;
	activationOutline.moveTo(extent2D + padding2D);
	activationOutline.arcTo(
		QRectF(extent2D.x() - padding, extent2D.y() - padding, 2 * padding, 2 * padding),
		-paddingAngle * 180.0 / acos(-1.0f), 180.0
	);
	activationOutline.lineTo(-extent2D - padding2D);
	activationOutline.arcTo(
		QRectF(-extent2D.x() - padding, -extent2D.y() - padding, 2 * padding, 2 * padding),
		180.0 - paddingAngle * 180.0 / acos(-1.0f), 180.0
	);
	activationOutline.closeSubpath();
	//use this outline for both requested properties
	setActivationOutline(activationOutline);
	setInteractionOutline(interactionOutline);
}
