/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_GEOMETRY_H
#define KOLF_GEOMETRY_H

#include "physicsengine.h"
#include <QPainterPath>
struct dMass;

namespace Kolf
{
	class Object;

	/**
	 * \class Geometry
	 * \since 2.0
	 *
	 * A geometry represents a part of the shape of a Kolf::Object. Simple objects can be represented by a single geometry (e.g. boxes and spheres). while the shape of a complex object may have to be constructed from multiple geometries. (Note though that Kolf::Objects can currently have only one geometry object.)
	 *
	 * Every geometry has a reference point. Refer to the documentation for the subclasses for a description of their reference points. The reference point of the geometry has to match the reference point of the object it is used with because the implementation of Kolf::Object always keeps them in sync.
	 */
	class Geometry
	{
		public:
			virtual ~Geometry();
			///Sets the mass for this geometry.
			void setMass(qreal mass);

			///Returns the object that this geometry belongs to.
			Kolf::Object* object() const;

			///Returns the two-dimensional activation outline, i.e. the area of this geometry (plus some defined padding). The editor overlay is supposed to activate the editor interface of the object associated with this geometry, if a mouse click occurs inside the activation outline.
			QPainterPath activationOutline() const;
			///Returns the two-dimensional interaction outline, i.e. the exact outline of this geometry. The editor overlay is supposed to move the object associated with this geometry, if the mouse is clicked inside the interaction outline and then dragged.
			///\note The interaction outline is allowed to be greater than the object's exact shape for the sake of usability (e.g. a diagonal wall cannot be precisely hit with a mouse cursor).
			QPainterPath interactionOutline() const;
		protected:
			///Creates a new Kolf::Geometry on the given \a object. The required \a id can be created through one of ODE's dCreate... methods.
			Geometry(dGeomID id, Kolf::Object* object);
			///Returns the mass structure assigned to this geometry.
			dMass mass() const;
			///Sets the mass structure for this geometry.
			void setMass(const dMass& mass);
			///Returns the internal ID of this geometry (for usage with the ODE interface).
			dGeomID id() const;
			friend class PhysicsEngine; //The PhysicsEngine needs to know the dGeomID.

			///Call this method when the parameters of this geometry change (usually in setter methods of the subclass).
			void update();
			///Reimplement this method to create the outlines of this geometry and pass them to setActivationOutline() and setInteractionOutline(). You will not have to call this function in subclass implementations, it's invoked by Kolf::Geometry::update.
			virtual void createOutlines() = 0;
			///Call this method from createInteractionAreas(). See activationOutline() for a description of this property.
			void setActivationOutline(const QPainterPath& outline);
			///Call this method from createInteractionAreas(). See interactionOutline() for a description of this property.
			void setInteractionOutline(const QPainterPath& outline);
			///Use this padding as distance between the exact InteractionOutline and the fuzzy ActivationOutline.
			static const int ActivationOutlinePadding;
		private:
			dGeomID m_id;
			Kolf::Object* m_object;
			QPainterPath m_activationOutline, m_interactionOutline;
	};
}

#endif // KOLF_GEOMETRY_H
