/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_MATH_CUBOIDF_H
#define KOLF_MATH_CUBOIDF_H

#include "math-pointf.h"
#include "math-sizef.h"
#include <algorithm>
#include <QDebug>
#include <QMetaType>

namespace Kolf
{
	/**
	 * \class CuboidF
	 * \since 2.0
	 * This class is the analogon of QRectF for the threedimensional space. Its interface is an adoption of a subset of the QRectF interface:
	 * \li The bounds are referred to as "lower X", "upper X", "lower Y", and so on.
	 * \li The bound values can be retrieved through functions like lowerX(), upperX(), etc.
	 * \li To change bound values, use setLowerX() etc. (if the other bounds should remain unchanged) or moveLowerX() etc. (if the size may not change).
	 * \li You can also pass Kolf::PointF or Kolf::SizeF structures through the setLowerBounds(), setUpperBounds() or setSize() methods.
	 */
	class CuboidF
	{
		private:
			qreal m_x1, m_y1, m_z1, m_x2, m_y2, m_z2;
		public:
			inline CuboidF()
				: m_x1(0), m_y1(0), m_z1(0)
				, m_x2(0), m_y2(0), m_z2(0)
			{
			}
			inline CuboidF(const Kolf::PointF& lowerBounds, const Kolf::SizeF& size)
				: m_x1(lowerBounds.x()), m_y1(lowerBounds.y()), m_z1(lowerBounds.z())
				, m_x2(lowerBounds.x() + size.xSpread()), m_y2(lowerBounds.y() + size.ySpread()), m_z2(lowerBounds.z() + size.zSpread())
			{
			}
			inline CuboidF(const Kolf::PointF& lowerBounds, const Kolf::PointF& upperBounds)
				: m_x1(lowerBounds.x()), m_y1(lowerBounds.y()), m_z1(lowerBounds.z())
				, m_x2(upperBounds.x()), m_y2(upperBounds.y()), m_z2(upperBounds.z())
			{
			}
			inline CuboidF(qreal x, qreal y, qreal z, qreal xSpread, qreal ySpread, qreal zSpread)
				: m_x1(x), m_y1(y), m_z1(z)
				, m_x2(x + xSpread), m_y2(y + ySpread), m_z2(z + zSpread)
			{
			}
			inline CuboidF(const Kolf::CuboidF& cuboid)
				: m_x1(cuboid.m_x1), m_y1(cuboid.m_y1), m_z1(cuboid.m_z1)
				, m_x2(cuboid.m_x2), m_y2(cuboid.m_y2), m_z2(cuboid.m_z2)
			{
			}

			///Ensures that the size of this rectangle is valid (i.e. the lower bounds are lower than the upper ones).
			void normalize()
			{
				if (m_x1 > m_x2)
					std::swap(m_x1, m_x2);
				if (m_y1 > m_y2)
					std::swap(m_y1, m_y2);
				if (m_z1 > m_z2)
					std::swap(m_z1, m_z2);
			}

			inline qreal lowerX() const
			{
				return m_x1;
			}
			inline qreal upperX() const
			{
				return m_x2;
			}
			inline qreal lowerY() const
			{
				return m_y1;
			}
			inline qreal upperY() const
			{
				return m_y2;
			}
			inline qreal lowerZ() const
			{
				return m_z1;
			}
			inline qreal upperZ() const
			{
				return m_z2;
			}
			inline Kolf::PointF lowerBounds() const
			{
				return Kolf::PointF(m_x1, m_y1, m_z1);
			}
			inline Kolf::PointF upperBounds() const
			{
				return Kolf::PointF(m_x2, m_y2, m_z2);
			}

			inline void setLowerX(qreal x)
			{
				m_x1 = x;
			}
			inline void setUpperX(qreal x)
			{
				m_x2 = x;
			}
			inline void setLowerY(qreal y)
			{
				m_y1 = y;
			}
			inline void setUpperY(qreal y)
			{
				m_y2 = y;
			}
			inline void setLowerZ(qreal z)
			{
				m_z1 = z;
			}
			inline void setUpperZ(qreal z)
			{
				m_z2 = z;
			}
			inline void setLowerBounds(const Kolf::PointF& point)
			{
				m_x1 = point.x();
				m_y1 = point.y();
				m_z1 = point.z();
			}
			inline void setUpperBounds(const Kolf::PointF& point)
			{
				m_x2 = point.x();
				m_y2 = point.y();
				m_z2 = point.z();
			}

			inline void moveLowerX(qreal x1)
			{
				m_x2 += x1 - m_x1;
				m_x1 = x1;
			}
			inline void moveUpperX(qreal x2)
			{
				m_x1 += x2 - m_x2;
				m_x2 = x2;
			}
			inline void moveLowerY(qreal y1)
			{
				m_y2 += y1 - m_y1;
				m_y1 = y1;
			}
			inline void moveUpperY(qreal y2)
			{
				m_y1 += y2 - m_y2;
				m_y2 = y2;
			}
			inline void moveLowerZ(qreal z1)
			{
				m_z2 += z1 - m_z1;
				m_z1 = z1;
			}
			inline void moveUpperZ(qreal z2)
			{
				m_z1 += z2 - m_z2;
				m_z2 = z2;
			}

			inline qreal xSpread() const
			{
				return m_x2 - m_x1;
			}
			inline qreal ySpread() const
			{
				return m_y2 - m_y1;
			}
			inline qreal zSpread() const
			{
				return m_z2 - m_z1;
			}
			inline Kolf::SizeF size() const
			{
				return Kolf::SizeF(m_x2 - m_x1, m_y2 - m_y1, m_z2 - m_z1);
			}

			inline void setXSpread(qreal xSpread)
			{
				m_x2 = m_x1 + xSpread;
			}
			inline void setYSpread(qreal ySpread)
			{
				m_y2 = m_y1 + ySpread;
			}
			inline void setZSpread(qreal zSpread)
			{
				m_z2 = m_z1 + zSpread;
			}
			inline void setSize(const Kolf::SizeF& size)
			{
				m_x2 = m_x1 + size.xSpread();
				m_y2 = m_y1 + size.ySpread();
				m_z2 = m_z1 + size.zSpread();
			}

			inline bool operator==(const Kolf::CuboidF& cuboid) const
			{
				return m_x1 == cuboid.m_x1 && m_x2 == cuboid.m_x2
					&& m_y1 == cuboid.m_y1 && m_y2 == cuboid.m_y2
					&& m_z1 == cuboid.m_z1 && m_z2 == cuboid.m_z2;
			}
			inline bool operator!=(const Kolf::CuboidF& cuboid) const
			{
				return m_x1 != cuboid.m_x1 || m_x2 != cuboid.m_x2
					|| m_y1 != cuboid.m_y1 || m_y2 != cuboid.m_y2
					|| m_z1 != cuboid.m_z1 || m_z2 != cuboid.m_z2;
			}
	};

}

inline QDataStream& operator<<(QDataStream& stream, const Kolf::CuboidF& cuboid)
{
	return stream
		<< double(cuboid.lowerX()) << double(cuboid.upperX())
		<< double(cuboid.lowerY()) << double(cuboid.upperY())
		<< double(cuboid.lowerZ()) << double(cuboid.upperZ());
}
inline QDataStream& operator>>(QDataStream& stream, Kolf::CuboidF& cuboid)
{
	double temp;
	stream >> temp; cuboid.setLowerX(temp);
	stream >> temp; cuboid.setUpperX(temp);
	stream >> temp; cuboid.setLowerY(temp);
	stream >> temp; cuboid.setUpperY(temp);
	stream >> temp; cuboid.setLowerZ(temp);
	stream >> temp; cuboid.setUpperZ(temp);
	return stream;
}
inline QDebug operator<<(QDebug debug, const Kolf::CuboidF& cuboid)
{
	debug.nospace() << "Kolf::CuboidF(" << cuboid.lowerX() << ", " << cuboid.lowerY() << ", " << cuboid.lowerZ() << "  " << cuboid.xSpread() << " x " << cuboid.ySpread() << " x " << cuboid.zSpread() << ")";
	return debug.space();
}

Q_DECLARE_METATYPE(Kolf::CuboidF)

#endif // KOLF_MATH_CUBOIDF_H
