/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_MATH_POINTF_H
#define KOLF_MATH_POINTF_H

#include <QDataStream>
#include <QDebug>
#include <QMetaType>
#include <QPointF>

namespace Kolf
{
	/**
	 * \class PointF
	 * \since 2.0
	 * This is a three-dimensional equivalent of QPointF. The interface is the same as for QPointF, with the following exceptions:
	 * \li Z dimensions have been added everywhere.
	 * \li The toPoint() method is not available.
	 */
	class PointF
	{
		private:
			qreal m_x, m_y, m_z;
		public:
			inline PointF()
				: m_x(0), m_y(0), m_z(0)
			{
			}
			explicit inline PointF(const QPointF& point)
				: m_x(point.x()), m_y(point.y()), m_z(0)
			{
			}
			inline PointF(const Kolf::PointF& point)
				: m_x(point.m_x), m_y(point.m_y), m_z(point.m_z)
			{
			}
			inline PointF(qreal x, qreal y, qreal z)
				: m_x(x), m_y(y), m_z(z)
			{
			}

			inline bool isNull() const
			{
				return m_x == 0 && m_y == 0 && m_z == 0;
			}

			inline qreal& rx()
			{
				return m_x;
			}
			inline qreal& ry()
			{
				return m_y;
			}
			inline qreal& rz()
			{
				return m_z;
			}
			inline qreal x() const
			{
				return m_x;
			}
			inline qreal y() const
			{
				return m_y;
			}
			inline qreal z() const
			{
				return m_z;
			}

			inline void setX(qreal x)
			{
				m_x = x;
			}
			inline void setY(qreal y)
			{
				m_y = y;
			}
			inline void setZ(qreal z)
			{
				m_z = z;
			}

			inline QPointF to2D() const
			{
				return QPointF(m_x, m_y);
			}

			inline Kolf::PointF& operator=(const Kolf::PointF& point)
			{
				m_x = point.m_x;
				m_y = point.m_y;
				m_z = point.m_z;
				return *this;
			}
			inline Kolf::PointF& operator+=(const Kolf::PointF& point)
			{
				m_x += point.m_x;
				m_y += point.m_y;
				m_z += point.m_z;
				return *this;
			}
			inline Kolf::PointF& operator-=(const Kolf::PointF& point)
			{
				m_x -= point.m_x;
				m_y -= point.m_y;
				m_z -= point.m_z;
				return *this;
			}
			inline Kolf::PointF& operator*=(qreal factor)
			{
				m_x *= factor;
				m_y *= factor;
				m_z *= factor;
				return *this;
			}
			inline Kolf::PointF& operator/=(qreal divisor)
			{
				m_x /= divisor;
				m_y /= divisor;
				m_z /= divisor;
				return *this;
			}

			inline const Kolf::PointF operator+(const Kolf::PointF& point) const
			{
				return Kolf::PointF(m_x + point.m_x, m_y + point.m_y, m_z + point.m_z);
			}
			inline const Kolf::PointF operator-(const Kolf::PointF& point) const
			{
				return Kolf::PointF(m_x - point.m_x, m_y - point.m_y, m_z - point.m_z);
			}
			inline const Kolf::PointF operator*(qreal factor) const
			{
				return Kolf::PointF(m_x * factor, m_y * factor, m_z * factor);
			}
			inline const Kolf::PointF operator/(qreal divisor) const
			{
				return Kolf::PointF(m_x / divisor, m_y / divisor, m_z / divisor);
			}
			inline qreal operator*(const Kolf::PointF& point) const
			{
				return m_x * point.m_x + m_y * point.m_y + m_z * point.m_z;
			}

			inline bool operator==(const Kolf::PointF& point) const
			{
				return m_x == point.m_x && m_y == point.m_y && m_z == point.m_z;
			}
			inline bool operator!=(const Kolf::PointF& point) const
			{
				return m_x != point.m_x || m_y != point.m_y || m_z != point.m_z;
			}
	};

	typedef Kolf::PointF VectorF;
}

typedef QPointF QVectorF;

inline qreal operator*(const QPointF& point1, const QPointF& point2)
{
	return point1.x() * point2.x() + point1.y() * point2.y();
}

inline const Kolf::PointF operator-(const Kolf::PointF& point) //negation operator
{
	return Kolf::PointF(-point.x(), -point.y(), -point.z());
}
inline const Kolf::PointF operator*(qreal factor, const Kolf::PointF& point)
{
	return Kolf::PointF(point.x() * factor, point.y() * factor, point.z() * factor);
}

inline QDataStream& operator<<(QDataStream& stream, const Kolf::PointF& point)
{
	stream << double(point.x()) << double(point.y()) << double(point.z());
	return stream;
}
inline QDataStream& operator>>(QDataStream& stream, Kolf::PointF& point)
{
	stream >> point.rx();
	stream >> point.ry();
	stream >> point.rz();
	return stream;
}
inline QDebug operator<<(QDebug debug, const Kolf::PointF& point)
{
	debug.nospace() << "Kolf::PointF(" << point.x() << ", " << point.y() << ", " << point.z() << ")";
	return debug.space();
}

Q_DECLARE_METATYPE(Kolf::PointF)

#endif // KOLF_MATH_POINTF_H
