/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "object-blackhole.h"
#include "ball.h"
#include "geometry-sphere.h"
#include "object-hole.h"
#include "objectregistry.h"
#include "overlay-blackhole.h"

#include <cmath>
#include <QGraphicsEllipseItem>
#include <QGraphicsLineItem>
#include <QPen>
#include <QTimer>

REGISTER_KOLF_OBJECT(Kolf::BlackHoleObject, ki18n("Black hole"), KIcon())

const qreal Kolf::BlackHoleObject::ExitHalfLength = 20;

Kolf::BlackHoleObject::BlackHoleObject(dWorldID worldID, dSpaceID spaceID)
	: Kolf::Object(worldID, spaceID)
	, m_entryItem(new QGraphicsEllipseItem(QRectF(-Kolf::HoleObject::HoleRadius, -Kolf::HoleObject::HoleRadius, 2 * Kolf::HoleObject::HoleRadius, 2 * Kolf::HoleObject::HoleRadius), this))
	, m_exitItem(new QGraphicsLineItem(QLineF(QPointF(-ExitHalfLength, 0), QPointF(ExitHalfLength, 0)), this))
	, m_exitDelay(0)
{
	QPen exitPen(Qt::black);
	exitPen.setWidth(5);
	m_exitItem->setPen(exitPen);
	//create geometry
	Kolf::SphereGeometry* geometry = new Kolf::SphereGeometry(spaceID, this);
	geometry->setRadius(Kolf::HoleObject::HoleRadius);
}

Kolf::BlackHoleObject::~BlackHoleObject()
{
	delete m_entryItem;
	delete m_exitItem;
}

int Kolf::BlackHoleObject::exitDelay() const
{
	return m_exitDelay;
}

Kolf::PointF Kolf::BlackHoleObject::exitOffset() const
{
	return m_exitOffset;
}

Kolf::VectorF Kolf::BlackHoleObject::exitVelocity() const
{
	return m_exitVelocity;
}

void Kolf::BlackHoleObject::setExitDelay(int exitDelay)
{
	if (m_exitDelay != exitDelay)
	{
		m_exitDelay = qMax(1, exitDelay);
		propagateUpdate();
	}
}

void Kolf::BlackHoleObject::setExitOffset(const Kolf::VectorF& exitOffset)
{
	if (m_exitOffset != exitOffset)
	{
		m_exitOffset = exitOffset;
		m_exitItem->setPos(exitOffset.x(), exitOffset.y());
		propagateUpdate();
	}
}

void Kolf::BlackHoleObject::setExitVelocity(const Kolf::VectorF& exitVelocity)
{
	if (m_exitVelocity != exitVelocity)
	{
		m_exitVelocity = exitVelocity;
		if (m_exitVelocity.isNull())
			//logic check: disallow zero exit velocity
			m_exitVelocity.rx() = 1;
		//normalize exitVelocity vector
		const QPointF exitVelocity2D(exitVelocity.x(), exitVelocity.y());
		const QPointF nExitVelocity2D = exitVelocity2D / sqrt(exitVelocity2D * exitVelocity2D);
		//exit item line vector is perpendicular to the exit direction vector (in X-Y space)
		const QPointF lineVector(-nExitVelocity2D.y() * ExitHalfLength, nExitVelocity2D.x() * ExitHalfLength);
		m_exitItem->setLine(QLineF(-lineVector, lineVector));
		propagateUpdate();
	}
}

QList<Kolf::Ball*> ballsWithActiveTraders; //needed by the helper class below
///\class BlackHoleActionTrader
///An instance of this class is created when a ball enters a black hole. It makes the ball pseudo-invisible by placing it off-scene, and starts a timer to get it back to the black hole's exit.
///\see Kolf::BlackHoleObject::collideWithBall
class BlackHoleActionTrader : public QObject
{
	private:
		Kolf::BlackHoleObject* m_object;
		Kolf::Ball* m_ball;
		int m_timerID;
	public:
		BlackHoleActionTrader(Kolf::BlackHoleObject* object, Kolf::Ball* ball)
			: m_object(object)
			, m_ball(ball)
			, m_timerID(startTimer(object->exitDelay()))
		{
			m_ball->setPosition(Kolf::PointF(-1000, -1000, -1000)); //that should be out of the visible area -> ball seems invisible for now
			m_ball->setVelocity(Kolf::VectorF(-1000, -1000, -1000)); //ball does still have to move! (or it would think that its move has finished)
			ballsWithActiveTraders << ball;
		}
	protected:
		virtual void timerEvent(QTimerEvent* event)
		{
			Q_UNUSED(event)
			m_ball->setPosition(m_object->exitOffset() + m_object->position());
			m_ball->setVelocity(m_object->exitVelocity());
			//delete object
			killTimer(m_timerID);
			ballsWithActiveTraders.removeAll(m_ball);
			deleteLater();
		}
};

Kolf::Object::SimulationFlags Kolf::BlackHoleObject::simulationFlags() const
{
	return Kolf::Object::CustomInteraction;
}

void Kolf::BlackHoleObject::doCustomInteraction(Kolf::Object* object)
{
	Kolf::Ball* ball = qobject_cast<Kolf::Ball*>(object);
	if (ball && !ballsWithActiveTraders.contains(ball))
		new BlackHoleActionTrader(this, ball);
}

Kolf::Overlay* Kolf::BlackHoleObject::createOverlay()
{
	return new Kolf::BlackHoleOverlay(this);
}

#include "object-blackhole.moc"
