/***************************************************************************
 *   Copyright 2008-2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "object-rectblock.h"
#include "geometry-cuboid.h"
#include "objectregistry.h"
#include "overlay-rect.h"

#include <QBrush>
#include <QGraphicsRectItem>
#include <QPen>

REGISTER_KOLF_OBJECT(Kolf::RectBlockObject, ki18n("Rectangular block"), KIcon())

Kolf::RectBlockObject::RectBlockObject(dWorldID worldID, dSpaceID spaceID)
	: Kolf::Object(worldID, spaceID)
	, m_geometry(new Kolf::CuboidGeometry(spaceID, this))
	, m_item(new QGraphicsRectItem(QRectF(), this))
	, m_vertexArray(0)
	, m_normalArray(0)
{
	m_item->setPen(QPen(color()));
	m_item->setBrush(color());
}

Kolf::RectBlockObject::~RectBlockObject()
{
	delete m_item;
	delete m_vertexArray;
	delete m_normalArray;
}

Kolf::SizeF Kolf::RectBlockObject::size() const
{
	return m_geometry->size();
}

void Kolf::RectBlockObject::setSize(const Kolf::SizeF& size)
{
	if (m_geometry->size() != size)
	{
		m_geometry->setSize(size);
		const QRectF displayRect(-size.xSpread() / 2, -size.ySpread() / 2, size.xSpread(), size.ySpread());
		m_item->setRect(displayRect);
		propagateUpdate();
	}
}

void Kolf::RectBlockObject::objectEvent(int type, const QVariant& data)
{
	if (type == Kolf::Object::ColorChangeEvent)
	{
		const QColor color = data.value<QColor>();
		m_item->setPen(QPen(color));
		m_item->setBrush(color);
	}
}

Kolf::Overlay* Kolf::RectBlockObject::createOverlay()
{
	return new Kolf::RectOverlay(this);
}

void Kolf::RectBlockObject::createSpatialGeometry()
{
	float x = position().x();
	float y = position().y();
	float z = position().z();
	float w = size().xSpread();
	float h = size().ySpread();
	float d = size().zSpread();;
	
	m_vertexArray = new float[72];
	numVerts = 72;
	m_vertexArray[0] = (float)(x - w/2);
	m_vertexArray[1] = (float)(y + h/2);
	m_vertexArray[2] = (float)(z + d);
	m_vertexArray[3] = (float)(x + w/2);
	m_vertexArray[4] = (float)(y + h/2);
	m_vertexArray[5] = (float)(z + d);
	m_vertexArray[6] = (float)(x + w/2);
	m_vertexArray[7] = (float)(y + h/2);
	m_vertexArray[8] = (float)(z);
	m_vertexArray[9] = (float)(x - w/2);
	m_vertexArray[10] = (float)(y + h/2);
	m_vertexArray[11] = (float)(z);

	m_vertexArray[12] = (float)(x + w/2);
	m_vertexArray[13] = (float)(y - h/2);
	m_vertexArray[14] = (float)(z);
	m_vertexArray[15] = (float)(x - w/2);
	m_vertexArray[16] = (float)(y - h/2);
	m_vertexArray[17] = (float)(z);
	m_vertexArray[18] = (float)(x - w/2);
	m_vertexArray[19] = (float)(y + h/2);
	m_vertexArray[20] = (float)(z);
	m_vertexArray[21] = (float)(x + w/2);
	m_vertexArray[22] = (float)(y + h/2);
	m_vertexArray[23] = (float)(z);

	m_vertexArray[24] = (float)(x + w/2);
	m_vertexArray[25] = (float)(y + h/2);
	m_vertexArray[26] = (float)(z + d);
	m_vertexArray[27] = (float)(x + w/2);
	m_vertexArray[28] = (float)(y + h/2);
	m_vertexArray[29] = (float)(z);
	m_vertexArray[30] = (float)(x + w/2);
	m_vertexArray[31] = (float)(y - h/2);
	m_vertexArray[32] = (float)(z);
	m_vertexArray[33] = (float)(x + w/2);
	m_vertexArray[34] = (float)(y - h/2);
	m_vertexArray[35] = (float)(z + d);

	m_vertexArray[36] = (float)(x + w/2);
	m_vertexArray[37] = (float)(y + h/2);
	m_vertexArray[38] = (float)(z + d);
	m_vertexArray[39] = (float)(x + w/2);
	m_vertexArray[40] = (float)(y - h/2);
	m_vertexArray[41] = (float)(z + d);
	m_vertexArray[42] = (float)(x - w/2);
	m_vertexArray[43] = (float)(y - h/2);
	m_vertexArray[44] = (float)(z + d);
	m_vertexArray[45] = (float)(x - w/2);
	m_vertexArray[46] = (float)(y + h/2);
	m_vertexArray[47] = (float)(z + d);

	m_vertexArray[48] = (float)(x - w/2);
	m_vertexArray[49] = (float)(y + h/2);
	m_vertexArray[50] = (float)(z);
	m_vertexArray[51] = (float)(x - w/2);
	m_vertexArray[52] = (float)(y + h/2);
	m_vertexArray[53] = (float)(z + d);
	m_vertexArray[54] = (float)(x - w/2);
	m_vertexArray[55] = (float)(y - h/2);
	m_vertexArray[56] = (float)(z + d);
	m_vertexArray[57] = (float)(x - w/2);
	m_vertexArray[58] = (float)(y - h/2);
	m_vertexArray[59] = (float)(z);

	m_vertexArray[60] = (float)(x + w/2);
	m_vertexArray[61] = (float)(y - h/2);
	m_vertexArray[62] = (float)(z);
	m_vertexArray[63] = (float)(x - w/2);
	m_vertexArray[64] = (float)(y - h/2);
	m_vertexArray[65] = (float)(z);
	m_vertexArray[66] = (float)(x - w/2);
	m_vertexArray[67] = (float)(y - h/2);
	m_vertexArray[68] = (float)(z + d);
	m_vertexArray[69] = (float)(x + w/2);
	m_vertexArray[70] = (float)(y - h/2);
	m_vertexArray[71] = (float)(z + d);

	m_normalArray = new float[72];
	m_normalArray[0] = 0.0;
	m_normalArray[1] = 1.0;
	m_normalArray[2] = 0.0;
	m_normalArray[3] = 0.0;
	m_normalArray[4] = 1.0;
	m_normalArray[5] = 0.0;
	m_normalArray[6] = 0.0;
	m_normalArray[7] = 1.0;
	m_normalArray[8] = 0.0;
	m_normalArray[9] = 0.0;
	m_normalArray[10] = 1.0;
	m_normalArray[11] = 0.0;

	m_normalArray[12] = 0.0;
	m_normalArray[13] = 0.0;
	m_normalArray[14] = -1.0;
	m_normalArray[15] = 0.0;
	m_normalArray[16] = 0.0;
	m_normalArray[17] = -1.0;
	m_normalArray[18] = 0.0;
	m_normalArray[19] = 0.0;
	m_normalArray[20] = -1.0;
	m_normalArray[21] = 0.0;
	m_normalArray[22] = 0.0;
	m_normalArray[23] = -1.0;
	
	m_normalArray[24] = -1.0;
	m_normalArray[25] = 0.0;
	m_normalArray[26] = 0.0;
	m_normalArray[27] = -1.0;
	m_normalArray[28] = 0.0;
	m_normalArray[29] = 0.0;
	m_normalArray[30] = -1.0;
	m_normalArray[31] = 0.0;
	m_normalArray[32] = 0.0;
	m_normalArray[33] = -1.0;
	m_normalArray[34] = 0.0;
	m_normalArray[35] = 0.0;
	
	m_normalArray[36] = 0.0;
	m_normalArray[37] = 0.0;
	m_normalArray[38] = 1.0;
	m_normalArray[39] = 0.0;
	m_normalArray[40] = 0.0;
	m_normalArray[41] = 1.0;
	m_normalArray[42] = 0.0;
	m_normalArray[43] = 0.0;
	m_normalArray[44] = 1.0;
	m_normalArray[45] = 0.0;
	m_normalArray[46] = 0.0;
	m_normalArray[47] = 1.0;
	
	m_normalArray[48] = 1.0;
	m_normalArray[49] = 0.0;
	m_normalArray[50] = 0.0;
	m_normalArray[51] = 1.0;
	m_normalArray[52] = 0.0;
	m_normalArray[53] = 0.0;
	m_normalArray[54] = 1.0;
	m_normalArray[55] = 0.0;
	m_normalArray[56] = 0.0;
	m_normalArray[57] = 1.0;
	m_normalArray[58] = 0.0;
	m_normalArray[59] = 0.0;
	
	m_normalArray[60] = 0.0;
	m_normalArray[61] = -1.0;
	m_normalArray[62] = 0.0;
	m_normalArray[63] = 0.0;
	m_normalArray[64] = -1.0;
	m_normalArray[65] = 0.0;
	m_normalArray[66] = 0.0;
	m_normalArray[67] = -1.0;
	m_normalArray[68] = 0.0;
	m_normalArray[69] = 0.0;
	m_normalArray[70] = -1.0;
	m_normalArray[71] = 0.0;
	rotateVertexArray(numVerts, m_vertexArray);
	rotateNormalArray(numVerts, m_normalArray);
	//FIXME: return the proper size of the array.
	//Note: You could just return a QVector<float>, which can be converted to float* with the data() method.
}

float* Kolf::RectBlockObject::getNormalGeometry()
{
	return m_normalArray;
}

float* Kolf::RectBlockObject::getVertexGeometry()
{
	return m_vertexArray;
}

#include "object-rectblock.moc"
