/***************************************************************************
 *   Copyright 2008-2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "object.h"
#include "overlay.h"

#include <cmath>
#include <ode/ode.h>

Kolf::Object::Object(dWorldID worldID, dSpaceID spaceID)
	: QObject()
	, m_bodyID(dBodyCreate(worldID))
	, m_worldID(worldID)
	, m_spaceID(spaceID)
	, m_geometry(0)
	, m_overlay(0)
	, m_2DRotationAngle(0)
	, m_blockUpdates(false)
	, m_color(Qt::gray)
{
	Q_UNUSED(spaceID) //This parameter is required just for consistency with all subclass constructors (which may need the space ID for constructing geometries).
	//objects will most definitely not have mouse interaction
	setAcceptedMouseButtons(0);
	//Note: Kolf::Scene::addObject() is not called here because it needs a fully constructed Kolf::Object (including all subclass constructors!).
}

Kolf::Object::~Object()
{
	delete m_geometry;
	qDeleteAll(m_subObjects);
	dBodyDestroy(m_bodyID);
}

const Kolf::Geometry* Kolf::Object::geometry() const
{
	return m_geometry;
}

Kolf::Overlay* Kolf::Object::overlay(bool createIfNecessary)
{
	//the overlay is created once it is requested
	if (!m_overlay && createIfNecessary)
	{
		m_overlay = createOverlay();
		Q_ASSERT(m_overlay != 0);
		m_overlay->setZValue(100); //should be above object representation
		m_overlay->update(); //initialize the overlay's parameters
	}
	return m_overlay;
}

void Kolf::Object::blockUpdates()
{
	m_blockUpdates = true;
}

void Kolf::Object::unblockUpdates(bool propagateChangesAsUpdate)
{
	m_blockUpdates = false;
	if (propagateChangesAsUpdate)
		propagateUpdate();
}

void Kolf::Object::propagateUpdate()
{
	//Note: Overlays still need to be updated even in blocking mode.
	if (m_overlay)
		m_overlay->update();
	if (!m_blockUpdates)
		emit updated();
}

bool Kolf::Object::isEditable() const
{
	return true;
}

bool Kolf::Object::isRotateable() const
{
	return true;
}

QColor Kolf::Object::color() const
{
	return m_color;
}

Kolf::PointF Kolf::Object::position() const
{
	const dReal* coords = dBodyGetPosition(m_bodyID);
	return Kolf::PointF(coords[0], coords[1], coords[2]);
}

int Kolf::Object::rotation() const
{
	//warning: this is a 4x3 matrix, not a 3x3 matrix -> matrix[i, j] = matrix[4 * i + j]
	const dReal* matrix = dBodyGetRotation(m_bodyID);
	//apply rotation to a unit vector in X direction
	const qreal rotatedX = matrix[0];
	const qreal rotatedY = matrix[4];
	return atan2(rotatedY, rotatedX) / acos(-1.0f) * 180.0;
}

Kolf::VectorF Kolf::Object::velocity() const
{
	const dReal* odeVelocity = dBodyGetLinearVel(m_bodyID);
	return Kolf::VectorF(odeVelocity[0], odeVelocity[1], odeVelocity[2]);
}

void Kolf::Object::setColor(const QColor& color)
{
	m_color = color;
	objectEvent(ColorChangeEvent, qVariantFromValue(color));
	propagateUpdate();
}

void Kolf::Object::setPosition(const Kolf::PointF& position)
{
	dBodySetPosition(m_bodyID, position.x(), position.y(), position.z());
	if (scene())
	{
		setPos(QPointF(position.x(), position.y()));
		setZValue(position.z());
	}
	objectEvent(PositionChangeEvent, qVariantFromValue(position));
	if (!m_blockUpdates)
		emit updated();
}

void Kolf::Object::setRotation(int angle)
{
	//apply to ODE
	dQuaternion rot;
	dQFromAxisAndAngle(rot, 0.0, 0.0, 1.0, angle / 180.0 * acos(-1.0f));
	dBodySetQuaternion(m_bodyID, rot);
	//apply to graphics item
	int angleStep = angle - m_2DRotationAngle;
	if (angleStep == 0)
		return; //esp. do not emit updated() signal in this case!
	rotate(angleStep);
	m_2DRotationAngle = angle;
	//propagate
	objectEvent(RotationChangeEvent, angle);
	if (!m_blockUpdates)
		emit updated();
}

void Kolf::Object::setVelocity(const Kolf::VectorF& velocity)
{
	dBodySetLinearVel(m_bodyID, velocity.x(), velocity.y(), velocity.z());
	objectEvent(VelocityChangeEvent, qVariantFromValue(velocity));
}

void Kolf::Object::prepareKinematicState()
{
	bool isMovable = simulationFlags() & Kolf::Object::KinematicSimulation;
	dBodySetGravityMode(m_bodyID, (int) isMovable);
}

void Kolf::Object::updateKinematicState()
{
	if (simulationFlags() & Kolf::Object::KinematicSimulation)
	{
		setPosition(position());
		//setRotation(rotation()); //note that this removes any non-2D rotations
		setVelocity(velocity());
	}
}

QList<Kolf::Object*> Kolf::Object::subObjects() const
{
	return m_subObjects;
}

void Kolf::Object::addSubObject(Kolf::Object* object)
{
	if (m_subObjects.contains(object))
		return;
	m_subObjects << object;
	emit subObjectAdded(object);
}

void Kolf::Object::removeSubObject(Kolf::Object* object)
{
	if (!m_subObjects.contains(object))
		return;
	m_subObjects.removeAll(object);
	emit subObjectAboutToBeRemoved(object);
}

//default implementation for virtual Kolf::Object methods

Kolf::Object::SimulationFlags Kolf::Object::simulationFlags() const
{
	return Kolf::Object::DynamicSimulation;
}

void Kolf::Object::doCustomInteraction(Kolf::Object* object)
{
	Q_UNUSED(object)
}

void Kolf::Object::objectEvent(int type, const QVariant& data)
{
	Q_UNUSED(type)
	Q_UNUSED(data)
}

void Kolf::Object::createSpatialGeometry()
{
}

float* Kolf::Object::getNormalGeometry()
{
	return 0;
}

float* Kolf::Object::getVertexGeometry()
{
	return 0;
}

void Kolf::Object::rotateNormalArray(int numItems, float* array)
{
	if(rotation() != 0)
	{
		float pi = acos(-1.0f);
		float sinAngle = sin(((360-rotation())*pi) / 180);
		float cosAngle = cos(((360-rotation())*pi) / 180);

		for(int i = 0; i < numItems; i += 3)
		{
			float oldX = array[i];
			float oldY = array[i + 1];

			array[i] = ((oldX * cosAngle) + (oldY * sinAngle));
			array[i + 1] = ((oldY * cosAngle) - (oldX * sinAngle));
		}
	}
}

void Kolf::Object::rotateVertexArray(int numItems, float* array)
{
	if(rotation() != 0)
	{
		float pi = acos(-1.0f);
		float sinAngle = sin(((360-rotation())*pi) / 180);
		float cosAngle = cos(((360-rotation())*pi) / 180);

		for(int i = 0; i < numItems; i += 3)
		{
			float xDistance = position().x() - array[i];
			float yDistance = position().y() - array[i + 1];

			array[i] = ((xDistance * cosAngle) + (yDistance * sinAngle)) + position().x();
			array[i + 1] = ((yDistance * cosAngle) - (xDistance * sinAngle)) + position().y();
		}
	}
}

//default implementation for QGraphicsItem

QRectF Kolf::Object::boundingRect() const
{
	return QRectF();
}

void Kolf::Object::paint(QPainter*, const QStyleOptionGraphicsItem*, QWidget*)
{
}

#include "object.moc"
