/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_OBJECTREGISTRY_H
#define KOLF_OBJECTREGISTRY_H

#include <QMap>
#include <KIcon>
#include <KLocalizedString>
#include <ode/common.h>

namespace Kolf
{
	class Object;
	class Scene;
	typedef Kolf::Object*(*ObjectCreator)(dWorldID worldID, dSpaceID spaceID);

	/**
	 * \class ObjectRegistry
	 * \since 2.0
	 *
	 * Provides data and instance management for Kolf::Object subclasses.
	 * \sa REGISTER_KOLF_OBJECT
	 */
	class ObjectRegistry
	{
		public:
			///Returns a list of all Kolf::Object subclasses that are known to this registry.
			static QList<QByteArray> knownClasses();
			///Returns the translated name of the given Kolf::Object subclass (or an empty QString if this class is not registered). This name can be used in user-visible places.
			static QString translatedName(const QByteArray& className);
			///Returns the icon of the given Kolf::Object subclass (or an empty QIcon if this class is not registered). This icon can be used in user-visible places.
			static QIcon icon(const QByteArray& className);

			///Creates an instance of the Kolf::Object subclass with the given \a className. If such a class does not exist (or is not binded), 0 is returned.
			static Kolf::Object* createObject(const QByteArray& className, Kolf::Scene* scene);
			///\overload which can be used if the target scene is not known, but an object on the target scene is available
			static Kolf::Object* createObject(const QByteArray& className, Kolf::Object* otherObjectOnScene);
		private:
			struct ClassData
			{
				KLocalizedString m_name;
				QIcon m_icon;
				Kolf::ObjectCreator m_creator;
			};

			friend class ObjectRegistrar;
			QMap<QByteArray, Kolf::ObjectRegistry::ClassData> m_classData;
	};

	///\internal
	///This class needs to be publicly accessible for the REGISTER_KOLF_OBJECT macro.
	class ObjectRegistrar
	{
		public:
			ObjectRegistrar(const char* className, const KLocalizedString& name, const QIcon& icon, Kolf::ObjectCreator creator);
	};

	///\internal
	///This method needs to be publicly accessible for the REGISTER_KOLF_OBJECT macro.
	template<typename ObjectClass> Kolf::Object* __createObject(dWorldID worldID, dSpaceID spaceID)
	{
		return new ObjectClass(worldID, spaceID);
	}
}

/**
 * \def REGISTER_KOLF_OBJECT
 * \relates Kolf::Object
 * \relates Kolf::ObjectRegistry
 *
 * Registers a Kolf::Object to make it visible to the game engine. For any Kolf::Object subclass, you will have to use this macro in the implementation file (.cpp extension) in the following pattern:
 * \code
 * REGISTER_KOLF_OBJECT(Kolf::FooObject, ki18n("Foo"), KIcon("foo-bar-baz"))
 * \endcode
 */
#define REGISTER_KOLF_OBJECT(ObjectClass, Name, Icon) \
	namespace { \
		const char __kolf_current_class[] = #ObjectClass; \
		Kolf::ObjectRegistrar __kolf_registrar(__kolf_current_class, Name, Icon, &Kolf::__createObject<ObjectClass>); \
	}

#endif // KOLF_OBJECTREGISTRY_H
