/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "overlay-blackhole.h"
#include "handle-point.h"
#include "object.h"

#include <QGraphicsLineItem>
#include <QPen>

Kolf::BlackHoleOverlay::BlackHoleOverlay(Kolf::Object* object)
	: Kolf::Overlay(object)
	, m_exitLine(new QGraphicsLineItem(this))
	, m_exitVelocityLine(new QGraphicsLineItem(this))
	, m_exitHandle(new Kolf::PointHandle(Kolf::PointHandle::CircleShape, this))
	, m_exitVelocityHandle(new Kolf::PointHandle(Kolf::PointHandle::TriangleShape, this))
{
	m_exitLine->setPen(Qt::DashLine);
	m_exitVelocityLine->setPen(Qt::DashLine);
	addHandle(m_exitLine);
	addHandle(m_exitVelocityLine);
	addHandle(m_exitHandle);
	addHandle(m_exitVelocityHandle);
	connect(m_exitHandle, SIGNAL(moveRequest(const QPointF&)), this, SLOT(moveHandle(const QPointF&)));
	connect(m_exitVelocityHandle, SIGNAL(moveRequest(const QPointF&)), this, SLOT(moveHandle(const QPointF&)));
}

void Kolf::BlackHoleOverlay::update()
{
	Kolf::Overlay::update();
	const Kolf::PointF exitOffset = object()->property("ExitOffset").value<Kolf::PointF>();
	const Kolf::VectorF exitVelocity = object()->property("ExitVelocity").value<Kolf::VectorF>();
	const QPointF exitOffset2D(exitOffset.x(), exitOffset.y());
	const QPointF exitVelocity2D(exitVelocity.x(), exitVelocity.y());
	m_exitLine->setLine(QLineF(QPointF(), exitOffset2D));
	m_exitVelocityLine->setLine(QLineF(exitOffset2D, exitOffset2D + exitVelocity2D));
	m_exitHandle->setPos(exitOffset2D);
	//for the exit velocity handle, we need to construct a QTransform because we also need a rotation
	QTransform t;
	t.translate(exitOffset2D.x() + exitVelocity2D.x(), exitOffset2D.y() + exitVelocity2D.y());
	qreal angleRadians = atan2(exitVelocity2D.y(), exitVelocity2D.x());
	t.rotate(angleRadians / acos(-1.0f) * 180);
	m_exitVelocityHandle->setTransform(t);
}

void Kolf::BlackHoleOverlay::moveHandle(const QPointF& handleScenePos)
{
	const Kolf::PointHandle* handle = qobject_cast<Kolf::PointHandle*>(sender());
	if (!handle)
		return;
	const QVectorF handlePos2D = mapFromScene(handleScenePos);
	if (handle == m_exitHandle)
	{
		Kolf::PointF exitOffset = object()->property("ExitOffset").value<Kolf::PointF>();
		exitOffset.rx() = handlePos2D.x();
		exitOffset.ry() = handlePos2D.y();
		object()->setProperty("ExitOffset", qVariantFromValue(exitOffset));
	}
	else //handle == m_exitVelocityHandle
	{
		const Kolf::PointF exitOffset = object()->property("ExitOffset").value<Kolf::PointF>();
		Kolf::VectorF exitVelocity = object()->property("ExitVelocity").value<Kolf::VectorF>();
		exitVelocity.rx() = handlePos2D.x() - exitOffset.x();
		exitVelocity.ry() = handlePos2D.y() - exitOffset.y();
		object()->setProperty("ExitVelocity", qVariantFromValue(exitVelocity));
	}
}

#include "overlay-blackhole.moc"
