/***************************************************************************
 *   Copyright 2008-2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "overlay.h"
#include "handle-point.h"
#include "object.h"
#include "overlayareaitem.h"

#include <QBrush>
#include <QCursor>
#include <QGraphicsScene>
#include <QGraphicsView>
#include <QPen>

const qreal Kolf::Overlay::MinimumObjectDimension = 20;
const qreal Kolf::Overlay::RotatorHandleDistance = 100;

Kolf::Overlay::Overlay(Kolf::Object* object)
	: QGraphicsItem(object)
	, m_object(object)
	, m_state(Kolf::Overlay::Passive)
	, m_activatorItem(new Kolf::OverlayAreaItem(Kolf::OverlayAreaItem::Clickable | Kolf::OverlayAreaItem::Hoverable, this))
	, m_interactorItem(new Kolf::OverlayAreaItem(Kolf::OverlayAreaItem::Clickable | Kolf::OverlayAreaItem::Draggable, this))
	, m_rotatorHandle(0)
	, m_rotatorHelper(0)
{
	//overlays have to be shown explicitly
	hide();
	//overlays themselves do not have mouse interaction
	setAcceptedMouseButtons(0);
	//initialize area items
	m_activatorItem->setZValue(1);
	connect(m_activatorItem, SIGNAL(hoverEntered()), this, SLOT(activatorEntered()));
	connect(m_activatorItem, SIGNAL(hoverLeft()), this, SLOT(activatorLeft()));
	connect(m_activatorItem, SIGNAL(clicked(int)), this, SLOT(activatorClicked(int)));
	m_interactorItem->setCursor(Qt::SizeAllCursor); //TODO: OpenHandCursor and ClosedHandCursor
	m_interactorItem->setZValue(2);
	connect(m_interactorItem, SIGNAL(clicked(int)), this, SLOT(activatorClicked(int))); //note that interactor item is over activator in hovered mode
	connect(m_interactorItem, SIGNAL(dragged(const QPointF&)), this, SLOT(interactorDragged(const QPointF&)));
	//create rotator if necessary - note that the position must only be set once, as the coordinate system of the object rotates the rotator handle as well
	if (object->isRotateable())
	{
		QPointF rotatorHandlePos(Kolf::Overlay::RotatorHandleDistance, 0);
		m_rotatorHandle = new Kolf::PointHandle(Kolf::PointHandle::CircleShape, this);
		m_rotatorHandle->setPos(rotatorHandlePos);
		connect(m_rotatorHandle, SIGNAL(moveRequest(const QPointF&)), this, SLOT(rotatorDragged(const QPointF&)));
		m_rotatorHelper = new QGraphicsLineItem(QLineF(QPointF(), rotatorHandlePos), this);
		m_rotatorHelper->setPen(Qt::DashLine);
		//show rotator handle together with the other handles
		addHandle(m_rotatorHandle);
		addHandle(m_rotatorHelper);
	}
	//apply passive state - we need to change to some other state to prevent the setState method from returning early
	m_state = Active;
	setState(Passive);
}

Kolf::Object* Kolf::Overlay::object() const
{
	return m_object;
}

void Kolf::Overlay::update()
{
	//update geometry outlines
	const Kolf::Geometry* geometry = m_object->geometry();
	if (geometry)
	{
		const QPainterPath activationOutline = geometry->activationOutline();
		m_activatorItem->setPath(activationOutline);
		m_interactorItem->setPath(geometry->interactionOutline());
	}
}

void Kolf::Overlay::addHandle(QGraphicsItem* handle)
{
	m_handles << handle;
	handle->setParentItem(this);
	handle->setVisible(m_state == Kolf::Overlay::Active);
	handle->setZValue(3); //should always be above the OverlayAreaItems
}

void Kolf::Overlay::addHandle(Kolf::PointHandle* handle)
{
	addHandle((QGraphicsItem*) handle);
	connect(handle, SIGNAL(moveStarted()), this, SLOT(handleInteractionStarted()));
	connect(handle, SIGNAL(moveEnded()), this, SLOT(handleInteractionEnded()));
}

Kolf::Overlay::State Kolf::Overlay::state() const
{
	return m_state;
}

void Kolf::Overlay::setState(Kolf::Overlay::State state)
{
	if (m_state == state)
		return;
	m_state = state;
	//get new inner properties
	QColor interactorColor(Qt::green);
	bool interactorActive = state == Hovered || state == Active;
	bool handlesActive = state == Active;
	switch (state)
	{
		case Kolf::Overlay::Hovered:
			interactorColor.setAlphaF(0.3);
			break;
		case Kolf::Overlay::Active:
			interactorColor.setAlphaF(0.6);
			break;
		default: //to avoid compiler warning because of missing branches
			break;
	}
	//apply new inner properties
	//TODO: create nice fade animations with Kolf::OverlayAreaItem::setActive
	m_interactorItem->setVisible(interactorActive);
	m_interactorItem->setBrush(interactorColor);
	foreach (QGraphicsItem* handle, m_handles)
		handle->setVisible(handlesActive);
	//propagate changes
	emit stateChanged();
}

void Kolf::Overlay::handleInteractionStarted()
{
	m_object->blockUpdates();
}

void Kolf::Overlay::handleInteractionEnded()
{
	m_object->unblockUpdates(true); //true == propagate changes in object's properties as update
}

void Kolf::Overlay::activatorEntered()
{
	if (m_state == Kolf::Overlay::Passive)
		setState(Kolf::Overlay::Hovered);
}

void Kolf::Overlay::activatorLeft()
{
	if (m_state == Kolf::Overlay::Hovered)
		setState(Kolf::Overlay::Passive);
}

void Kolf::Overlay::activatorClicked(int button)
{
	Q_UNUSED(button)
	setState(Kolf::Overlay::Active);
}

void Kolf::Overlay::interactorDragged(const QVectorF& distance)
{
	if (m_state == Kolf::Overlay::Active)
	{
		Kolf::VectorF distance3D(distance.x(), distance.y(), 0);
		m_object->setPosition(m_object->position() + distance3D);
	}
}

void Kolf::Overlay::rotatorDragged(const QPointF& targetScenePos)
{
	const QPointF targetPos = mapFromScene(targetScenePos);
	//note that handle is always at (RotatorHandleDistance, 0) because it is rotated together with the object - measured angles are difference angles
	qreal rotationRadians = atan2(targetPos.y(), targetPos.x());
	int rotationDegrees = rotationRadians / acos(-1.0f) * 180;
	m_object->setRotation(m_object->rotation() + rotationDegrees);
}

//default implementation for QGraphicsItem

QRectF Kolf::Overlay::boundingRect() const
{
	return QRectF();
}

void Kolf::Overlay::paint(QPainter*, const QStyleOptionGraphicsItem*, QWidget*)
{
}

#include "overlay.moc"
