/***************************************************************************
 *   Copyright 2008-2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "putter.h"
#include "ball.h"

#include <QCursor>
#include <QGraphicsLineItem>
#include <QGraphicsScene>
#include <QGraphicsSceneMouseEvent>
#include <QGraphicsView>
#include <QPen>

Kolf::Putter::Putter(QGraphicsScene* scene)
	: m_line(new QGraphicsLineItem(this))
	, m_ball(0)
	, m_direction(0)
{
	scene->addItem(this);
	setAcceptedMouseButtons(Qt::LeftButton | Qt::RightButton);
	setVisible(false);
	m_line->setPen(Qt::DashLine);
}

Kolf::Putter::~Putter()
{
	delete m_line;
}

Kolf::Ball* Kolf::Putter::ball() const
{
	return m_ball;
}

void Kolf::Putter::show(Kolf::Ball* ball)
{
	m_ball = ball;
	m_direction = 0;
	changeLine(findCursorScenePos());
	setVisible(true);
	grabMouse();
}

void Kolf::Putter::hide()
{
	if (m_ball) //mouse grabbing is only active if there is a ball, no point in deactivating it else
		ungrabMouse();
	m_ball = 0;
	setVisible(false);
}

void Kolf::Putter::mouseLeaveEvent()
{
	static const QPointF invalidPoint(-1.0, -1.0); //this point is outside the scene
	changeLine(invalidPoint);
}

void Kolf::Putter::mouseMoveEvent(QGraphicsSceneMouseEvent* event)
{
	changeLine(event->pos());
}

void Kolf::Putter::mousePressEvent(QGraphicsSceneMouseEvent* event)
{
	if (!m_line->isVisible())
		return;
	if (event->button() == Qt::LeftButton)
	{
		//read velocity from line vector
		const QLineF line = m_line->line();
		const QVectorF diffVector = line.p2() - line.p1();
		//set velocity
		static const qreal lengthToVelocityScalingFactor = 1;
		const QVectorF newVelocity = diffVector / lengthToVelocityScalingFactor;
		m_ball->setVelocity(Kolf::VectorF(newVelocity.x(), newVelocity.y(), 0));
		//wait for further instructions
		emit putted(m_ball);
	} else if (event->button() == Qt::RightButton)
	{
		//change direction of velocity vector
		m_direction = (m_direction + 1) % 4;
		changeLine(findCursorScenePos());
	}
}

void Kolf::Putter::changeLine(const QPointF& cursorPosition)
{
	if (!m_ball)
		return;
	if (cursorPosition.isNull()) //might indicate the same as the check below if something returned from findCursorScenePos() has been passed to this method
		return;
	//show line only if cursor is inside scene rect
	if (scene()->sceneRect().contains(cursorPosition))
		m_line->show();
	else
	{
		m_line->hide();
		return;
	}
	//calculate line vector from ball position and cursor position
	const Kolf::PointF ballPosition = m_ball->position();
	const QPointF ballPosition2D(ballPosition.x(), ballPosition.y());
	QVectorF velocityVector(cursorPosition - ballPosition2D);
	switch (m_direction) //rotate velocity vector according to requested rotation
	{
		case 1: //90° rotation
			velocityVector = QVectorF(-velocityVector.y(), velocityVector.x());
			break;
		case 2: //180° rotation = inversion
			velocityVector = -velocityVector;
			break;
		case 3: //270° rotation
			velocityVector = QVectorF(velocityVector.y(), -velocityVector.x());
			break;
	}
	m_line->setLine(QLineF(ballPosition2D, ballPosition2D + velocityVector));
}

QPointF Kolf::Putter::findCursorScenePos() const
{
	const QList<QGraphicsView*> views = scene()->views();
	foreach (QGraphicsView* view, views)
		if (view->underMouse())
			return view->mapToScene(view->mapFromGlobal(QCursor::pos()));
	return QPointF();
}

//default implementation for QGraphicsItem

QRectF Kolf::Putter::boundingRect() const
{
	return QRectF();
}

void Kolf::Putter::paint(QPainter*, const QStyleOptionGraphicsItem*, QWidget*)
{
}

#include "putter.moc"
