/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "utils-bytearraycodec.h"
#include "math-cuboidf.h"
#include "math-pointf.h"
#include "math-sizef.h"

#include <QBuffer>
#include <QImage>

namespace Kolf
{
	namespace ByteArrayCodecWorker
	{
		QByteArray encode(const QList<qreal>& values);
		QByteArray encode(const Kolf::PointF& value);
		QByteArray encode(const Kolf::SizeF& value);
		QByteArray encode(const Kolf::CuboidF& value);
		QByteArray encode(const QImage& value);

		QList<qreal> decode(const QByteArray& value);
		Kolf::PointF decode(const QByteArray& value, const Kolf::PointF& defaultValue);
		Kolf::SizeF decode(const QByteArray& value, const Kolf::SizeF& defaultValue);
		Kolf::CuboidF decode(const QByteArray& value, const Kolf::CuboidF& defaultValue);
		QImage decode(const QByteArray& value, const QImage& defaultValue);
	};
}

//BEGIN Kolf::ByteArrayCodec

const Kolf::ByteArrayCodec* Kolf::ByteArrayCodec::instance = new Kolf::ByteArrayCodec; //needs to be initialized when the application starts to register the QVariant data types

Kolf::ByteArrayCodec::ByteArrayCodec()
	: m_pointFTypeID(qRegisterMetaType<Kolf::PointF>())
	, m_sizeFTypeID(qRegisterMetaType<Kolf::SizeF>())
	, m_cuboidFTypeID(qRegisterMetaType<Kolf::CuboidF>())
{
}

bool Kolf::ByteArrayCodec::mathAwareCompare(const QVariant& value1, const QVariant& value2)
{
	//fallback to normal equality checking if appropriate
	int type1 = value1.type(), type2 = value2.type();
	if (type1 != QVariant::UserType || type2 != QVariant::UserType)
		return value1 == value2;
	//Qt data structures cannot be equal to Kolf data structures
	if (type1 != type2) //i.e. one of them != QVariant::UserType
		return false;
	//checking only between identical datatypes
	int utype1 = value1.userType(), utype2 = value2.userType();
	if (utype1 != utype2)
		return false;
	//compare values by type
	if (utype1 == instance->m_pointFTypeID)
		return value1.value<Kolf::PointF>() == value2.value<Kolf::PointF>();
	else if (utype1 == instance->m_sizeFTypeID)
		return value1.value<Kolf::SizeF>() == value2.value<Kolf::SizeF>();
	else if (utype1 == instance->m_cuboidFTypeID)
		return value1.value<Kolf::CuboidF>() == value2.value<Kolf::CuboidF>();
	else //type unknown to Kolf - fallback to normal equality checking
		return value1 == value2;
}

QByteArray Kolf::ByteArrayCodec::encode(const QVariant& value)
{
	if (value.type() == QVariant::Image)
		return Kolf::ByteArrayCodecWorker::encode(value.value<QImage>());
	else if (value.type() != QVariant::UserType)
		return QByteArray();
	const int userType = value.userType();
	if (userType == instance->m_pointFTypeID)
		return Kolf::ByteArrayCodecWorker::encode(value.value<Kolf::PointF>());
	else if (userType == instance->m_sizeFTypeID)
		return Kolf::ByteArrayCodecWorker::encode(value.value<Kolf::SizeF>());
	else if (userType == instance->m_cuboidFTypeID)
		return Kolf::ByteArrayCodecWorker::encode(value.value<Kolf::CuboidF>());
	else
		return QByteArray();
}

QVariant Kolf::ByteArrayCodec::decode(const QByteArray& value, const QVariant& defaultValue)
{
	switch (defaultValue.type())
	{
		case QVariant::Bool:
			return QString::fromLatin1(value).toLower() == QLatin1String("true");
		case QVariant::Image:
			return Kolf::ByteArrayCodecWorker::decode(value, defaultValue.value<QImage>());
		case QVariant::UserType:
			break; //use the following switch statement to find the return value
		default:
			QVariant convertedValue(value);
			convertedValue.convert(defaultValue.type());
			return convertedValue;
	}
	const int userType = defaultValue.userType();
	if (userType == instance->m_pointFTypeID)
		return qVariantFromValue(Kolf::ByteArrayCodecWorker::decode(value, defaultValue.value<Kolf::PointF>()));
	else if (userType == instance->m_sizeFTypeID)
		return qVariantFromValue(Kolf::ByteArrayCodecWorker::decode(value, defaultValue.value<Kolf::SizeF>()));
	else if (userType == instance->m_cuboidFTypeID)
		return qVariantFromValue(Kolf::ByteArrayCodecWorker::decode(value, defaultValue.value<Kolf::CuboidF>()));
	else
		return value;
}

//END Kolf::ByteArrayCodec

//BEGIN Kolf::ByteArrayCodecWorker

QByteArray Kolf::ByteArrayCodecWorker::encode(const QList<qreal>& values)
{
	QByteArray result;
	result.reserve(5 * values.count()); //reserve some space that should be sufficient in most cases
	bool first = true;
	foreach (qreal value, values)
	{
		//desired result: "23,42,13.37" -> place a comma before every number except the first one
		if (first)
			first = false;
		else
			result += ',';
		result += QByteArray::number((double) value);
	}
	return result;
}

QByteArray Kolf::ByteArrayCodecWorker::encode(const Kolf::PointF& value)
{
	return Kolf::ByteArrayCodecWorker::encode(QList<qreal>() << value.x() << value.y() << value.z());
}

QByteArray Kolf::ByteArrayCodecWorker::encode(const Kolf::SizeF& value)
{
	return Kolf::ByteArrayCodecWorker::encode(QList<qreal>() << value.xSpread() << value.ySpread() << value.zSpread());
}

QByteArray Kolf::ByteArrayCodecWorker::encode(const Kolf::CuboidF& value)
{
	return Kolf::ByteArrayCodecWorker::encode(QList<qreal>() << value.lowerX() << value.upperX() << value.lowerY() << value.upperY() << value.lowerZ() << value.upperZ());
}

QByteArray Kolf::ByteArrayCodecWorker::encode(const QImage& value)
{
	QBuffer buffer;
	value.save(&buffer, "PNG");
	return buffer.data();
}

QList<qreal> Kolf::ByteArrayCodecWorker::decode(const QByteArray& value)
{
	QList<QByteArray> parts = value.split(',');
	QList<qreal> result;
	foreach (const QByteArray& part, parts)
		result << part.toDouble();
	return result;
}

Kolf::PointF Kolf::ByteArrayCodecWorker::decode(const QByteArray& value, const Kolf::PointF& defaultValue)
{
	const QList<qreal> parts = decode(value);
	if (parts.count() == 3)
		return Kolf::PointF(parts[0], parts[1], parts[2]);
	else
		return defaultValue;
}

Kolf::SizeF Kolf::ByteArrayCodecWorker::decode(const QByteArray& value, const Kolf::SizeF& defaultValue)
{
	const QList<qreal> parts = decode(value);
	if (parts.count() == 3)
		return Kolf::SizeF(parts[0], parts[1], parts[2]);
	else
		return defaultValue;
}

Kolf::CuboidF Kolf::ByteArrayCodecWorker::decode(const QByteArray& value, const Kolf::CuboidF& defaultValue)
{
	const QList<qreal> parts = decode(value);
	if (parts.count() == 6)
		return Kolf::CuboidF(Kolf::PointF(parts[0], parts[1], parts[2]), Kolf::SizeF(parts[3], parts[4], parts[5]));
	else
		return defaultValue;
}

QImage Kolf::ByteArrayCodecWorker::decode(const QByteArray& value, const QImage& defaultValue)
{
	QImage image(defaultValue);
	image.loadFromData(QByteArray::fromBase64(value));
	return image;
}

//END Kolf::ByteArrayCodecWorker
