/***************************************************************************
 *   Copyright 2009 Zeng Huan <zh.issac@gmail.com>
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "utils-heightmap.h"

#include <cmath>

//BEGIN geometry helper functions

/*
 * Make a right-handed cross product, that means: The vector triplet (a, b, a × b) is right-handed.
 * Formula: a × b = (a1b2 - a2b1, a2b0 - a0b2, a0b1 - a1b0) with C-style indices
 * WARNING: The size of all array parameters is assumed to be 3 without further checking.
 */
template <typename T> void crossProduct(T* result, T* a, T* b)
{
	result[0] = a[1] * b[2] - a[2] * b[1];
	result[1] = a[2] * b[0] - a[0] * b[2];
	result[2] = a[0] * b[1] - a[1] * b[0];
}

template <typename T> void normalizeVector(T* vector, int size)
{
	//calculate length of vector
	T length = 0.0;
	for (int i = 0; i < size; ++i)
		length += vector[i] * vector[i];
	length = std::sqrt(length);
	//divide by length (but leave null vectors as-is)
	if (length != 0.0)
		for (int i = 0; i < size; ++i)
			vector[i] /= length;
}

//END geometry helper functions

Kolf::Heightmap::Heightmap()
{
}

Kolf::Heightmap::Heightmap(const QImage& image)
	: QImage(image)
{
}
Kolf::Heightmap::Heightmap(const QString& heightmapPath)
	: QImage(heightmapPath)
{
}

Kolf::Heightmap::~Heightmap()
{
}

bool Kolf::Heightmap::isDataValid() const
{
	//QImage::Format_Invalid == 0
	//QImage::Format_Mono == 1
	//QImage::Format_MonoLSB == 2
	return !isNull() && format() >= 3;
}

int Kolf::Heightmap::pixel(int x, int y) const
{
	return qMin(qGray(QImage::pixel(x % width(), y % height())) + 1, 255);
}

Kolf::HeightmapRenderer::HeightmapRenderer()
{
}

Kolf::HeightmapRenderer::HeightmapRenderer(const QImage& image)
	: Heightmap(image)
{
}
Kolf::HeightmapRenderer::HeightmapRenderer(const QString& heightmapPath)
	: Heightmap(heightmapPath)
{
}

Kolf::HeightmapRenderer::~HeightmapRenderer()
{
}

Utils::StaticVector<int> Kolf::HeightmapRenderer::indices() const
{
	return m_indices;
}

Utils::StaticVector<float> Kolf::HeightmapRenderer::textureCoordinates() const
{
	return m_textureCoordinates;
}

Utils::StaticVector<dReal> Kolf::HeightmapRenderer::vertices() const
{
	return m_vertices;
}

Utils::StaticVector<float> Kolf::HeightmapRenderer::normals() const
{
	return m_normals;
}

//BEGIN test for ode

Utils::StaticVector<Triangle_t> Kolf::HeightmapRenderer::odeIndices() const
{
	return m_odeIndices;
}

Utils::StaticVector<Vertex_t> Kolf::HeightmapRenderer::odeVertices() const
{
	return m_odeVertices;
}

//END test for ode

void Kolf::HeightmapRenderer::computeData(int width, int height, const Kolf::SizeF& terrainSize)
{
	if (isDataValid())
	{
		computeIndices(width, height);
		computeVertices(width, height, terrainSize);
		computeNormals(width, height);
	}
}

void Kolf::HeightmapRenderer::computeIndices(int width, int height)
{
	//The number of total triangles is  2 * width * height, so the number of the indices is 3 times of that.
	m_indices.resize(3 * 2 * width * height);
	int* ptr = m_indices.data();
	/*
	 *       .-------> x
	 *       | 1 3 5
	 *       | 2 4 6
	 *    y  v
	 *
	 *    like above, the index should be {1, 2, 3, 3, 2, 4, 3, 4, 5, 5, 4, 6}
	 */
	for( int y = 0; y < height; ++y)
		for(int x = 0; x < width; ++x)
		{
			*ptr++ =  y      * (width + 1) +  x;
			*ptr++ = (y + 1) * (width + 1) +  x;
			*ptr++ =  y      * (width + 1) + (x + 1);
			*ptr++ =  y      * (width + 1) + (x + 1);
			*ptr++ = (y + 1) * (width + 1) +  x;
			*ptr++ = (y + 1) * (width + 1) + (x + 1);
		}

	//BEGIN test for ode
	m_odeIndices.resize(2 * width * height);
	Triangle_t* oPtr = m_odeIndices.data();
	for( int y = 0; y < height; ++y)
		for(int x = 0; x < width; ++x)
		{
			(oPtr->indices)[0] =  y      * (width + 1) +  x;
			(oPtr->indices)[1] = (y + 1) * (width + 1) +  x;
			(oPtr->indices)[2] =  y      * (width + 1) + (x + 1);
			++oPtr;
			(oPtr->indices)[0] =  y      * (width + 1) + (x + 1);
			(oPtr->indices)[1] = (y + 1) * (width + 1) +  x;
			(oPtr->indices)[2] = (y + 1) * (width + 1) + (x + 1);
			++oPtr;
		}
	//END test for ode
}

void Kolf::HeightmapRenderer::computeVertices(int width, int height, const Kolf::SizeF& terrainSize)
{
	/**
	 *       z ^  ^ y (height)        This is our coordinate system.
	 *         | /
	 *         |/
	 *         .-----> x (width)
	 */
	float densityX = 1.0f / width;
	float densityY = 1.0f / height;
	float xScale = terrainSize.xSpread();
	float yScale = terrainSize.ySpread();
	float zScale = terrainSize.zSpread() / 256.0f;

	//We map [0, 1] to [0, width] and [0, height]. That is (width + 1) dots per row, and (height + 1) dots per column.
	//the number of vertexes, 3 components per vertex
	m_vertices.resize(3 * (width + 1) * (height + 1));
	dReal* vPtr = m_vertices.data();
	//the number of texture item, 2 components per item
	m_textureCoordinates.resize(2 * (width + 1) * (height + 1));
	float* tPtr = m_textureCoordinates.data();

	for(int y = 0; y <= height; ++y)
		for(int x = 0; x <= width; ++x)
		{
			*vPtr++ =  xScale * densityX * (float)x;
			*vPtr++ =  yScale * densityY * (float)y;
			*vPtr++ =  zScale * pixel(densityX * (float)x, densityY * (float)y);
			*tPtr++ = densityX * (float)x;
			//Our coordinate system is different from QGLWidget's(they are symmetirc about the x-axis), and we use the bindTexture() function which QGLWidget provided, then we have to do a inversion in the y-axis in order to get a right texture mapping.
			*tPtr++ = 1 - densityY * (float)y;
		}

	//BEGIN test for ode
	m_odeVertices.resize((width + 1) * (height + 1));
	Vertex_t* oPtr = m_odeVertices.data();

	for(int y = 0; y <= height; ++y)
		for(int x = 0; x <= width; ++x)
		{
			(oPtr->vertex)[0] = xScale * densityX * (float)x;
			(oPtr->vertex)[1] = yScale * densityY * (float)y;
			(oPtr->vertex)[2] = zScale * pixel(densityX * (float)x, densityY * (float)y);
			(oPtr->vertex)[3] = 1.0;
			++oPtr;
		}
	//END test for ode
}

void Kolf::HeightmapRenderer::computeNormals(int width, int height)
{
	m_normals.resize(3 * (width + 1) * (height + 1));
	QVector<float> tmp_normals(m_normals.size());
	/*           y ^ in
	 *            /
	 *  left     /      right
	 *    <-----+-----> x
	 *         /
	 *        /
	 *       v out
	 *  For each point in the vertices array, we compute 4 vectors for it,
	 *  left, right, in, out, then we make cross product of (in, left),
	 *  (left, out), (out, right), (right, in) separately, the average of
	 *  these four cross product is approximation of the normal of the point.
	 */
	for(int y = 0; y <= height; ++y)
		for(int x = 0; x <= width; ++x)
		{
			//in: (x, y - 1, z1) - (x, y, z2)
			float in[3] = {0, 0, 0};
			if (y > 0)
			{
				for(int i = 0; i < 3; ++i)
					in[i] = m_vertices[3 * ((y - 1) * (width + 1) + x) + i] - m_vertices[3 * (y * (width + 1) + x) + i];
			}
			//out: (x, y + 1, z1) - (x, y, z2)
			float out[3] = {0, 0, 0};
			if (y < height)
			{
				for(int i = 0; i < 3; ++i)
					out[i] = m_vertices[3 * ((y + 1) * (width + 1) + x) + i] - m_vertices[3 * (y * (width + 1) + x) + i];
			}
			//left: (x - 1, y, z1) - (x, y, z2)
			float left[3] = {0, 0, 0};
			if (x > 0)
			{
				for(int i = 0; i < 3; ++i)
					left[i] = m_vertices[3 * (y * (width + 1) + x - 1) + i] - m_vertices[3 * (y * (width + 1) + x) + i];
			}
			//right: (x + 1, y, z1) - (x, y, z2)
			float right[3] = {0, 0, 0};
			if (x < width)
			{
				for(int i = 0; i < 3; ++i)
					right[i] = m_vertices[3 * (y * (width + 1) + x + 1) + i] - m_vertices[3 * (y * (width + 1) + x) + i];
			}
			//compute sum of all cross products
			float normal[3] = {0, 0, 0};
			float sum[3] = {0, 0, 0};
			//cross product of (in, left)
			if (x > 0 && y > 0)
			{
				crossProduct(normal, in, left);
				normalizeVector(normal, 3);
				for(int i = 0; i < 3; ++i)
					sum[i] += normal[i];
			}
			//cross product of (left, out)
			if (x > 0 && y < height)
			{
				crossProduct(normal, left, out);
				normalizeVector(normal, 3);
				for(int i = 0; i < 3; ++i)
					sum[i] += normal[i];
			}
			//cross product of (out, right)
			if (x < width && y < height)
			{
				crossProduct(normal, out, right);
				normalizeVector(normal, 3);
				for(int i = 0; i < 3; ++i)
					sum[i] += normal[i];
			}
			//cross product of (right, in)
			if (x < width && y > 0)
			{
				crossProduct(normal, right, in);
				normalizeVector(normal, 3);
				for(int i = 0; i < 3; ++i)
					sum[i] += normal[i];
			}

			normalizeVector(sum, 3);
			for(int i = 0; i < 3; ++i)
				tmp_normals[3 * (y * (width + 1) + x) + i] = sum[i];
		}
	//Smooth out the normals
	const float smoothRatio = 0.5;
	for(int y = 0; y <= height; ++y)
		for(int x = 0; x <= width; ++x)
		{
			float sum[3] = {0, 0, 0};
			for(int i = 0; i < 3; ++i)
				sum[i] = tmp_normals[3 * (y * (width + 1) + x) + i];
			if (x > 0)
			{
				for(int i = 0; i < 3; ++i)
					sum[i] += tmp_normals[3 * (y * (width + 1) + x - 1) + i] * smoothRatio;
			}
			if (x < width)
			{
				for(int i = 0; i < 3; ++i)
					sum[i] += tmp_normals[3 * (y * (width + 1)+ x + 1) + i] * smoothRatio;
			}
			if (y > 0)
			{
				for(int i = 0; i < 3; ++i)
					sum[i] += tmp_normals[3 * ((y - 1) * (width + 1) + x) + i] * smoothRatio;
			}
			if (y < height)
			{
				for(int i = 0; i < 3; ++i)
					sum[i] += tmp_normals[3 * ((y + 1) * (width + 1) + x) + i] * smoothRatio;
			}

			normalizeVector(sum, 3);
			for(int i = 0; i < 3; ++i)
				m_normals[3 * (y * (width + 1) + x) + i] = sum[i];
		}
}
