/***************************************************************************
 *   Copyright 2009 Zeng Huan <zh.issac@gmail.com>
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_UTILS_HEIGHTMAP_H
#define KOLF_UTILS_HEIGHTMAP_H

#include "math-sizef.h"
#include "../utils/staticvector.h"

#include "ode/common.h"
#include <QImage>

struct Vertex_t { dVector3 vertex; };
struct Triangle_t { int indices[3]; };

namespace Kolf
{
	/**
	 * \class Heightmap
	 * \since 2.0
	 *
	 * This class represents a heightmap. Its pixel values are always returned in grayscale (8 bit per pixel).
	 */
	class Heightmap : public QImage
	{
		public:
			Heightmap();
			Heightmap(const QImage& image);
			Heightmap(const QString& heightmapPath);
			virtual ~Heightmap();
			///Checks if the source data is valid, we need an at least 8-bit per pixel image.
			bool isDataValid() const;

			///Returns the color (in grayscale) of the pixel at coordinates (x, y).
			int pixel(int x, int y) const;
	};

	/**
	 * \class HeightmapRenderer
	 * \since 2.0
	 *
	 * This class is used to prepare the data(vertice, normals, indices, texture coordinates) for rendering 3d terrain.
	 */
	class HeightmapRenderer : public Heightmap
	{
		public:
			HeightmapRenderer();
			HeightmapRenderer(const QImage& image);
			HeightmapRenderer(const QString& heightmapPath);
			~HeightmapRenderer();

			///Returns the indices of the terrain vertexes as a triangle mesh. The first three indices form the first triangle, the next three indices form the second triangle, and so on.
			Utils::StaticVector<int> indices() const;
			///Returns the coordinates of the terrain vertexes in the texture. In OpenGL, the texture has a coordinate system [0,1] x [0,1].
			///Format: [s0, t0, s1, t1, s2, t2, ...] where sn and tn are the coordinates of the n-th terrain vertex in the texture coordinate system. The order is the same as for the vertices array.
			Utils::StaticVector<float> textureCoordinates() const;
			///Returns the coordinates of the terrain vertexes.
			///Format: [x0, y0, z0, x1, y1, z1, ...]. The order is from top left to bottom right, scanning horizotally line per line.
			Utils::StaticVector<dReal> vertices() const;
			///Returns the coordaintes of the terrain normals.
			///\see vertices() for format of return value
			Utils::StaticVector<float> normals() const;

			//BEGIN test for ode
			Utils::StaticVector<Triangle_t> odeIndices() const;
			Utils::StaticVector<Vertex_t> odeVertices() const;
			//END test for ode

			///Computes the various indices, vertices and normals arrays in this object.
			void computeData(int width, int height, const Kolf::SizeF& terrainSize);
		private:
			void computeIndices(int width, int height);
			void computeVertices(int width, int height, const Kolf::SizeF& terrainSize);
			void computeNormals(int width, int height);

			Utils::StaticVector<int> m_indices;
			Utils::StaticVector<float> m_textureCoordinates;
			Utils::StaticVector<dReal> m_vertices;
			Utils::StaticVector<float> m_normals;
			//BEGIN test for ode
			Utils::StaticVector<Triangle_t> m_odeIndices;
			Utils::StaticVector<Vertex_t> m_odeVertices;
			//END test for ode
	};

}

#endif // KOLF_UTILS_HEIGHTMAP_H
