/***************************************************************************
 *   Copyright 2008-2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "course.h"
#include "../elements/scene.h"
#include "../elements/managedobjectgroup.h"

#include <QFileInfo>
#include <KConfig>
#include <KConfigGroup>
#include <KMessageBox>
#include <KStandardDirs>

namespace Kolf
{
	const QString courseFileNamePattern("kolf-ng/courses/%1.kolf");
	//group names for KConfigGroup
	const QString mainGroupKey("Kolf 2");
	const QString holeGroupKey("Hole %1");
	//key names for KConfigGroup
	const QString mainVersionKey("Version");
	const QString mainHoleCountKey("HoleCount");
	const QString mainTitleKey("Name");
	const QString holeParKey("Par");
}

QStringList Kolf::Course::availableCourses()
{
	QStringList paths = KStandardDirs().findAllResources("data", Kolf::courseFileNamePattern.arg("*"), KStandardDirs::NoDuplicates);
	QStringList courses;
	//extract course names
	foreach (const QString& path, paths)
		courses << QFileInfo(path).baseName();
	return courses;
}

QString Kolf::Course::courseTitle(const QString& courseName)
{
	const QString fileName = KStandardDirs::locate("data", Kolf::courseFileNamePattern.arg(courseName));
	if (fileName.isEmpty())
		return QString();
	KConfig config(fileName);
	KConfigGroup mainInfoGroup(&config, Kolf::mainGroupKey);
	return mainInfoGroup.readEntry(Kolf::mainTitleKey, QString());
}

Kolf::Course::Course(const QString& courseName)
	: m_courseName(courseName)
{
	//open course
	const QString fileName = KStandardDirs::locate("data", Kolf::courseFileNamePattern.arg(courseName));
	if (fileName.isEmpty())
		return;
	KConfig config(fileName);
	//main info: check version
	KConfigGroup mainInfoGroup(&config, Kolf::mainGroupKey);
	const int version = mainInfoGroup.readEntry(Kolf::mainVersionKey, 0);
	if (version != 1)
		return;
	//main info: get hole count and name
	const int holeCount = mainInfoGroup.readEntry(Kolf::mainHoleCountKey, 0);
	if (holeCount == 0)
		return;
	m_title = mainInfoGroup.readEntry(Kolf::mainTitleKey, QString());
	if (m_title.isEmpty())
		return;
	//read holes - Note: In the group captions, hole numbering starts at one.
	for (int i = 1; i <= holeCount; ++i)
	{
		KConfigGroup holeGroup(&config, Kolf::holeGroupKey.arg(i));
		m_parScores << holeGroup.readEntry(Kolf::holeParKey, 0);
		Kolf::Scene* scene = new Kolf::Scene;
		scene->objectGroup()->deserialize(&holeGroup);
		m_scenes << scene;
	}
}

Kolf::Course::~Course()
{
	qDeleteAll(m_scenes);
}

int Kolf::Course::holeCount() const
{
	return m_parScores.count();
}

QString Kolf::Course::title() const
{
	return m_title;
}

int Kolf::Course::parForHole(int holeIndex) const
{
	return m_parScores.value(holeIndex, 0);
}

Kolf::Scene* Kolf::Course::hole(int holeIndex) const
{
	return m_scenes.value(holeIndex);
}

void Kolf::Course::save() const
{
	KConfig config(KStandardDirs::locateLocal("data", Kolf::courseFileNamePattern.arg(m_courseName)));
	//write main info
	KConfigGroup mainInfoGroup(&config, Kolf::mainGroupKey);
	mainInfoGroup.writeEntry(Kolf::mainVersionKey, 1);
	const int holeCount = m_parScores.count();
	mainInfoGroup.writeEntry(Kolf::mainHoleCountKey, holeCount);
	mainInfoGroup.writeEntry(Kolf::mainTitleKey, m_title);
	//write holes - Note: In the group captions, hole numbering starts at one.
	for (int i = 0; i < holeCount; ++i)
	{
		KConfigGroup holeGroup(&config, Kolf::holeGroupKey.arg(i + 1));
		holeGroup.writeEntry(Kolf::holeParKey, m_parScores[i]);
		m_scenes[i]->objectGroup()->serialize(&holeGroup);
		m_scenes[i]->objectGroup()->setClean();
	}
}

#include "course.moc"
