/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *   Copyright 2009 Zeng Huan <zh.issac@gmail.com>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "game-editor.h"
#include "component-basicviews.h"
#include "course.h"
#include "player.h"
#include "../editor/blender-manager.h"
#include "../editor/blender-widget.h"
#include "../editor/hmcreator-drawingboard.h"
#include "../editor/hmcreator-toolbox.h"
#include "../elements/ball.h"
#include "../elements/overlay.h"
#include "../elements/managedobjectgroup.h"
#include "../elements/scene.h"
#include "../elements/utils-heightmap.h"

#include <QApplication>
#include <KAction>
#include <KActionCollection>
#include <KIcon>
#include <KLocalizedString>
#include <KMainWindow>
#include <KStandardAction>

//BEGIN Kolf::EditorGameUI

Kolf::EditorGameUI::EditorGameUI()
	: Kolf::GameUI(QLatin1String("kolf-editor"))
	, m_saveAction(KStandardAction::save(this, SIGNAL(actionSave()), actionCollection()))
	, m_undoAction(KStandardAction::undo(0, "", actionCollection()))
	, m_redoAction(KStandardAction::redo(0, "", actionCollection()))
{
	//action: go to previous hole
	const KIcon prevIcon(QApplication::isLeftToRight() ? "go-previous" : "go-next");
	KAction* prevAct = new KAction(prevIcon, i18n("Previous hole"), actionCollection());
	prevAct->setToolTip(i18n("Go to previous hole"));
	actionCollection()->addAction("kolf_level_previous", prevAct);
	connect(prevAct, SIGNAL(triggered()), this, SIGNAL(actionGoToPreviousHole()));
	//action: go to next hole
	const KIcon nextIcon(QApplication::isLeftToRight() ? "go-next" : "go-previous");
	KAction* nextAct = new KAction(nextIcon, i18n("Next hole"), actionCollection());
	nextAct->setToolTip(i18n("Go to next hole"));
	actionCollection()->addAction("kolf_level_next", nextAct);
	connect(nextAct, SIGNAL(triggered()), this, SIGNAL(actionGoToNextHole()));

	//page widgets: basic views
	addPageWidget(new Kolf::View2DComponent);

	//Blender's part
	Kolf::BlenderWidget* blenderWidget = new Kolf::BlenderWidget;
	addPageWidget(blenderWidget);

	//Heightmap creator's part
	Kolf::DrawingBoard* drawingBoard = new Kolf::DrawingBoard;
	addPageWidget(drawingBoard);
	addDockWidget(new Kolf::ToolBox(drawingBoard));
	connect(drawingBoard, SIGNAL(heightmapChanged(const Kolf::Heightmap&)),
		blenderWidget->manager(), SLOT(setHeightmap(const Kolf::Heightmap&)));

 	//Action: texture blender -- add textures
 	KAction* addTexAction = new KAction(KIcon("archive-insert"),i18n("Add texture"), actionCollection());
	addTexAction->setToolTip(i18n("Add a new texture"));
	actionCollection()->addAction("blender-addTexture", addTexAction);
	connect(addTexAction, SIGNAL(triggered()), blenderWidget->manager(), SLOT(addTextures()));

 	//Action: texture blender -- delete textures
	KAction* delTexAction = new KAction(KIcon("archive-remove"), i18n("Delete texture"), actionCollection());
	delTexAction->setToolTip(i18n("Delete a texture in the list"));
	actionCollection()->addAction("blender-delTexture", delTexAction);
	connect(delTexAction, SIGNAL(triggered()), blenderWidget->manager(), SLOT(delTextures()));

}

void Kolf::EditorGameUI::setCleanDispatcher(bool clean)
{
	setClean(clean);
	m_saveAction->setEnabled(!clean);
}

//END Kolf::EditorGameUI

//BEGIN Kolf::EditorGame

Kolf::EditorGame::EditorGame(Kolf::Course* course)
	: Kolf::Game(course, new Kolf::EditorGameUI)
{
	connect(gameUI(), SIGNAL(actionSave()), course, SLOT(save()));
	connect(gameUI(), SIGNAL(actionGoToPreviousHole()), this, SLOT(actionGoToPreviousHole()));
	connect(gameUI(), SIGNAL(actionGoToNextHole()), this, SLOT(actionGoToNextHole()));
}

Kolf::EditorGame::~EditorGame()
{
	//TODO: necessary?
	currentHoleAboutToBeChanged(); //deinitialize everything
}

bool Kolf::EditorGame::startGame()
{
	gameUI()->setWindowTitle(course()->title());
	setCurrentHole(0);
	return true;
}

void Kolf::EditorGame::currentHoleAboutToBeChanged()
{
	if (currentScene())
	{
		Kolf::ManagedObjectGroup* group = currentScene()->objectGroup();
		Kolf::EditorGameUI* gameUI = qobject_cast<Kolf::EditorGameUI*>(this->gameUI());
		disconnect(group, 0, this, 0);
		disconnect(group, 0, gameUI, 0);
		disconnect(group, 0, gameUI->m_redoAction, 0);
		disconnect(group, 0, gameUI->m_undoAction, 0);
		disconnect(gameUI->m_redoAction, 0, group, 0);
		disconnect(gameUI->m_undoAction, 0, group, 0);
	}
}

void Kolf::EditorGame::currentHoleChanged()
{
	const int currentHole = this->currentHole();
	const int holeCount = course()->holeCount();
	const bool isValidHole = currentHole < holeCount;
	//toggle action availability
	const bool isFirstHole = currentHole == 0;
	const bool isLastHole = currentHole == holeCount - 1;
	Kolf::EditorGameUI* gameUI = qobject_cast<Kolf::EditorGameUI*>(this->gameUI());
	gameUI->actionCollection()->action("kolf_level_previous")->setEnabled(isValidHole && !isFirstHole);
	gameUI->actionCollection()->action("kolf_level_next")->setEnabled(isValidHole && !isLastHole);
	//initiate more stuff
	if (isValidHole)
	{
		//activate overlays
		Kolf::Scene* scene = currentScene();
		foreach (Kolf::Object* object, scene->objects())
			if (object->isEditable())
			{
				Kolf::Overlay* overlay = object->overlay();
				overlay->show();
				connect(overlay, SIGNAL(stateChanged()), this, SLOT(overlayStateChanged()));
			}
		//hook up object group to game
		Kolf::ManagedObjectGroup* group = scene->objectGroup();
		connect(group, SIGNAL(objectCreated(Kolf::ManagedObject*)), this, SLOT(objectCreated(Kolf::ManagedObject*)));
		//hook up object group to interface
		connect(group, SIGNAL(cleanChanged(bool)), gameUI, SLOT(setCleanDispatcher(bool)));
		gameUI->setCleanDispatcher(group->isClean());
		connect(group, SIGNAL(canRedoChanged(bool)), gameUI->m_redoAction, SLOT(setEnabled(bool)));
		gameUI->m_redoAction->setEnabled(group->canRedo());
		connect(group, SIGNAL(canUndoChanged(bool)), gameUI->m_undoAction, SLOT(setEnabled(bool)));
		gameUI->m_undoAction->setEnabled(group->canUndo());
		connect(gameUI->m_redoAction, SIGNAL(triggered()), group, SLOT(redo()));
		connect(gameUI->m_undoAction, SIGNAL(triggered()), group, SLOT(undo()));
		//FIXME: pushState does not work for the ball - We somehow have to communicate the ball position change back to the environment object.
	}
}

void Kolf::EditorGame::overlayStateChanged()
{
	Kolf::Overlay* overlay = qobject_cast<Kolf::Overlay*>(sender());
	if (!overlay)
		return;
	//ensure that only one overlay is active
	if (overlay->state() >= Kolf::Overlay::Active)
		foreach (Kolf::Object* otherObject, currentScene()->objects())
			if (otherObject->isEditable())
			{
				Kolf::Overlay* otherOverlay = otherObject->overlay();
				if (otherOverlay != overlay && otherOverlay->state() >= Kolf::Overlay::Active)
					otherOverlay->setState(Kolf::Overlay::Passive);
			}
}

void Kolf::EditorGame::actionGoToPreviousHole()
{
	setCurrentHole(currentHole() - 1);
}

void Kolf::EditorGame::actionGoToNextHole()
{
	setCurrentHole(currentHole() + 1);
}

void Kolf::EditorGame::objectCreated(Kolf::ManagedObject* managedObject)
{
	managedObject->object()->overlay()->show();
}

//END Kolf::EditorGame

#include "game-editor.moc"
