/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_GAME_H
#define KOLF_GAME_H

#include <QObject>

namespace Kolf
{
	class Course;
	class GameUI;
	class Player;
	class Scene;

	/**
	 * \class Game
	 * \since 2.0
	 *
	 * The central game management class which should react on the movements of the ball and turn control to the players. Subclasses will have to extend the base implementation to do the actual actions needed to make it a game (i.e. player control management and such), and implement an own subclass of GameUI for each game type.
	 */
	class Game : public QObject
	{
		Q_OBJECT
		public:
			///\note This class takes ownership of the course and the game UI.
			///\note A pointer of Kolf::GameUI can only be generated by a public constructor of a Kolf::GameUI subclass, and those subclasses in most cases correspond to Kolf::Game subclasses, which you are expected to construct instead.
			Game(Kolf::Course* course, Kolf::GameUI* gameUI);
			virtual ~Game();
		public:
			Kolf::Course* course() const;
			Kolf::GameUI* gameUI() const;

			///Returns the par player which holds the par scores.
			Kolf::Player* parPlayer() const;
			///Returns the players (excluding the par player).
			QList<Kolf::Player*> players() const;

			///Starts the game (or fails if no players have been added to the game or if the course does not contain any holes).
			virtual bool startGame() = 0;
			///Returns the index of the current hole.
			int currentHole() const;
			///Returns the scene of the current hole.
			Kolf::Scene* currentScene() const;
		Q_SIGNALS:
			void currentHoleAboutToBeChanged(int newHoleIndex, int oldHoleIndex);
			void currentHoleChanged(int newHoleIndex, int oldHoleIndex);
			void sceneChanged(Kolf::Scene* scene);
			void playerAdded(Kolf::Player* player);
		protected:
			///Changes the current hole.
			void setCurrentHole(int newHole);
			///Adds the given \a player.
			void addPlayer(Kolf::Player* player);

			///Reimplement this method to react on changes of the current hole. The default implementation does nothing.
			virtual void currentHoleAboutToBeChanged();
			///Reimplement this method to react on changes of the current hole. The default implementation does nothing.
			virtual void currentHoleChanged();
		private:
			int m_currentHole; //counting starts at zero
			Kolf::Course* m_course;
			Kolf::GameUI* m_gameUI;
			QList<Kolf::Player*> m_players;
			Kolf::Player* m_parPlayer; //an imaginary dummy player which holds the par scores
	};
}

#endif // KOLF_GAME_H
