/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_GAMEUI_PAGEWIDGET_H
#define KOLF_GAMEUI_PAGEWIDGET_H

#include <QWidget>

namespace Kolf
{
	class Scene;

	/**
	 * \class GameUIPageWidget
	 * \since 2.0
	 *
	 * This is a wrapper class for widgets that appear in the central tab widget of the Kolf main window. It holds a reference to the current Kolf::Scene, and notifies the subclass implementation if the current scene instance changes.
	 */
	class GameUIPageWidget : public QWidget
	{
		Q_OBJECT
		public:
			///Construct a new GameUIPageWidget. The given \a label will appear on the tab bar. If the label contains an ampersand, the letter following the ampersand is used as a shortcut for the tab (as usual).
			GameUIPageWidget(const QString& label);

			///Returns a pointer to the current scene instance, or 0 if no scene instance is set currently.
			Kolf::Scene* currentScene() const;
			///The label that was specified for this page.
			QString label() const;
		public Q_SLOTS:
			///This slot is called by the game (or the game UI, respectively) when the scene changes.
			void setScene(Kolf::Scene* scene);
		Q_SIGNALS:
			///This signal is emitted when the page widget has become visible.
			void shown();
			///This signal is emitted after the page widget has been hidden.
			void hidden();
		protected:
			///If the whole content of this page consists only of one widget, use this convenience function to display it on the page. It will create a simple layout and put the given \a contentWidget in it. If you need more complex layouts, use QWidget::setLayout.
			///\note The usual ownership rules for widgets and layouts apply. Refer to the Qt documentation if in doubt.
			void setContent(QWidget* contentWidget);
			///Reimplement this function to react on a changing scene instance. (The default implementation does nothing.) You can safely do any cleanup on the old scene, it is guaranteed that the old scene was not yet deleted when this function is called.
			///\warning It is perfectly possible that one of both parameters is a null pointer.
			virtual void currentSceneChanged(Kolf::Scene* oldScene, Kolf::Scene* newScene);

			virtual void showEvent(QShowEvent* event);
			virtual void hideEvent(QHideEvent* event);
		private:
			Kolf::Scene* m_scene;
			QString m_label;
	};
}

#endif // KOLF_GAMEUI_PAGEWIDGET_H
