/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "gameui.h"
#include "gameui-dockwidget.h"
#include "gameui-pagewidget.h"

#include <QDockWidget>
#include <KTabWidget>
#include <KXMLGUIClient>
#include <KXMLGUIFactory>
#include <KXmlGuiWindow>

//BEGIN utilities

class KXMLGUIClient_WithXMLFileAccess : public KXMLGUIClient
{
	//This subclass is necessary because KXMLGUIClient does not expose the setXMLFile() method publicly.
	public:
		KXMLGUIClient_WithXMLFileAccess(const QString& interfaceName)
		{
			setXMLFile(QString::fromLatin1("%1ui.rc").arg(interfaceName));
		}
};

Kolf::GameUIParent::GameUIParent(KXmlGuiWindow* theWindow, KTabWidget* theTabWidget)
	: window(theWindow)
	, tabWidget(theTabWidget)
{
}

//END utilities

Kolf::GameUI::GameUI(const QString& interfaceName)
	: m_uiParent(Kolf::GameUIParent(0, 0))
	, m_uiClient(new KXMLGUIClient_WithXMLFileAccess(interfaceName))
	, m_clean(true)
{
}

Kolf::GameUI::~GameUI()
{
	delete m_uiClient;
	qDeleteAll(m_dockWidgets);
	qDeleteAll(m_pageWidgets);
}

KActionCollection* Kolf::GameUI::actionCollection()
{
	return m_uiClient->actionCollection();
}

void Kolf::GameUI::addDockWidget(Kolf::GameUIDockWidget* dockWidget)
{
	//check input
	if (!dockWidget || m_dockWidgets.contains(dockWidget))
		return;
	//insert dock widget
	m_dockWidgets << dockWidget;
	connect(this, SIGNAL(sceneChanged(Kolf::Scene*)), dockWidget, SLOT(setScene(Kolf::Scene*)));
	if (m_uiParent.window)
		m_uiParent.window->addDockWidget(dockWidget->preferredArea(), dockWidget);
}

void Kolf::GameUI::removeDockWidget(Kolf::GameUIDockWidget* dockWidget)
{
	//check input
	if (!m_dockWidgets.contains(dockWidget))
		return;
	//remove dock widget
	m_dockWidgets.removeAll(dockWidget);
	disconnect(this, 0, dockWidget, 0);
	if (m_uiParent.window)
		m_uiParent.window->removeDockWidget(dockWidget);
}

void Kolf::GameUI::addPageWidget(Kolf::GameUIPageWidget* pageWidget)
{
	//check input
	if (!pageWidget || m_pageWidgets.contains(pageWidget))
		return;
	//insert page widget
	m_pageWidgets << pageWidget;
	connect(this, SIGNAL(sceneChanged(Kolf::Scene*)), pageWidget, SLOT(setScene(Kolf::Scene*)));
	if (m_uiParent.tabWidget)
		m_uiParent.tabWidget->addTab(pageWidget, pageWidget->label());
}

void Kolf::GameUI::removePageWidget(Kolf::GameUIPageWidget* pageWidget)
{
	//check input
	if (!m_pageWidgets.contains(pageWidget))
		return;
	//remove page widgets
	m_pageWidgets.removeAll(pageWidget);
	disconnect(this, 0, pageWidget, 0);
	if (m_uiParent.tabWidget)
		m_uiParent.tabWidget->removeTab(m_uiParent.tabWidget->indexOf(pageWidget));
}

void Kolf::GameUI::setWindowTitle(const QString& title)
{
	if (m_windowTitle == title)
		return;
	m_windowTitle = title;
	if (m_uiParent.window)
		m_uiParent.window->setCaption(m_windowTitle, !m_clean);
}

void Kolf::GameUI::setClean(bool clean)
{
	if (m_clean == clean)
		return;
	m_clean = clean;
	if (m_uiParent.window)
		m_uiParent.window->setCaption(m_windowTitle, !m_clean);
}

Kolf::GameUIParent Kolf::GameUI::uiParent() const
{
	return m_uiParent;
}

void Kolf::GameUI::setUIParent(const Kolf::GameUIParent& uiParent)
{
	//Note: The members of uiParent are especially allowed to be null pointers.
	if (m_uiParent.window != uiParent.window)
	{
		//remove from old window
		if (m_uiParent.window)
		{
			//XMLGUI client
			m_uiParent.window->factory()->removeClient(m_uiClient);
			//dock widgets
			foreach (Kolf::GameUIDockWidget* dockWidget, m_dockWidgets)
				m_uiParent.window->removeDockWidget(dockWidget);
		}
		//add to new window
		if (uiParent.window)
		{
			//XMLGUI client
			uiParent.window->factory()->addClient(m_uiClient);
			//dock widgets
			foreach (Kolf::GameUIDockWidget* dockWidget, m_dockWidgets)
				uiParent.window->addDockWidget(dockWidget->preferredArea(), dockWidget);
			//set window title
			uiParent.window->setCaption(m_windowTitle, !m_clean);
		}
	}
	if (m_uiParent.tabWidget != uiParent.tabWidget)
	{
		//remove pages from old tab widget
		if (m_uiParent.tabWidget)
		{
			foreach (Kolf::GameUIPageWidget* pageWidget, m_pageWidgets)
				m_uiParent.tabWidget->removeTab(m_uiParent.tabWidget->indexOf(pageWidget));
		}
		//add pages to new tab widget
		if (uiParent.tabWidget)
		{
			foreach (Kolf::GameUIPageWidget* pageWidget, m_pageWidgets)
				uiParent.tabWidget->addTab(pageWidget, pageWidget->label());
		}
	}
	//done
	m_uiParent = uiParent;
}

void Kolf::GameUI::unsetUIParent()
{
	setUIParent(Kolf::GameUIParent(0, 0));
}

#include "gameui.moc"
