/***************************************************************************
 *   Copyright 2008-2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_PLAYER_H
#define KOLF_PLAYER_H

#include <QObject>
#include <QPointer>
#include <QString>
#include <QVector>

namespace Kolf
{
	class Ball;
	class Scene;
	class View;

	/**
	 * \class Player
	 * \since 2.0
	 *
	 * Representation for a player, which manages scores and passes control to this player. One Kolf::Player instance is not meant to be used with multiple Kolf::Game instances. This class is an abstract base class, the instantiable subclasses represent different types of players.
	 */
	class Player : public QObject
	{
		Q_OBJECT
		public:
			Player(const QString& name, unsigned int holeCount);
			virtual ~Player();

			///\note The ball is regularly deleted (e.g. when the scene changes). In this case, the next call to this getter will create a new Kolf::Ball.
			Kolf::Ball* ball() const;
			int hue() const;
			QString name() const;
			void setHue(int hue);
			void setName(const QString& name);

			int score(int hole) const;
			int currentScore() const;
			int totalScore() const;
			void setScore(int hole, int score);
			void setCurrentHole(int hole);
			void setCurrentScore(int score);
		public Q_SLOTS:
			///The subclass implementation of this slot should enable the putting interface (for human players) or do the putting (for computer players).
			virtual void giveControl() = 0;
			///The subclass implementation of this slot should disable the putting interface for human players if it's active.
			virtual void takeControl() = 0;
			///This slot will change the scene of this player.
			///\warning This will delete the ball on the old scene.
			void setScene(Kolf::Scene* scene);
		Q_SIGNALS:
			///This signal should be emitted by the subclass implementation when the player has putted. The subclass implementation should then also increase the score of this player.
			void ballPutted();
			///This signal is emitted when a score changes. (This signal does not have to be emitted by subclasses when you use setScore() and setCurrentScore().)
			void scoreChanged(Kolf::Player* player, int hole);
		private:
			mutable QPointer<Kolf::Ball> m_ball;
			Kolf::Scene* m_scene;

			int m_hue;
			QString m_name;
			QVector<int> m_scores;
			int m_totalScore; //This one is requested very often, it's useful to cache it.
			int m_currentHole;
	};
}

#endif // KOLF_PLAYER_H
