/***************************************************************************
 *   Copyright 2008, 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "object-ball.h"
#include "objectfactory.h"
#include "overlay-empty.h"
#include "shape-circle.h"
#include "utils-pointf.h"

#include <QPen>
#include <Box2D/Dynamics/b2Body.h>

const qreal Kolf::BallObject::BallRadius = 0.1;

REGISTER_KOLF_OBJECT(Kolf::BallObject, {
	classObject->screenName = ki18n("Ball (start position for all players)");
	classObject->editorMayCreateInstances = false;
	classObject->groupName = QLatin1String("01-basics");
})

Kolf::BallObject::BallObject(b2World* world)
	: Kolf::Object(world)
	, m_shape(new Kolf::CircleShape(BallRadius))
	, m_item(new QGraphicsEllipseItem(QRectF(-BallRadius, -BallRadius, 2 * BallRadius, 2 * BallRadius), this))
	, m_hue(-1)
	, m_inHole(false)
	, m_stoppedSignalSent(false)
{
	m_item->setPen(QPen(Qt::black));
	setHue(0); //initialize brush with something meaningful
	addShape(m_shape);
	//The following has to be consistent with setInHole(false).
	setSimulationType(Kolf::Object::DynamicSimulation);
	body()->SetLinearDamping(0.3);
}

int Kolf::BallObject::hue() const
{
	return m_hue;
}

void Kolf::BallObject::setHue(int hue)
{
	if (m_hue != hue)
	{
		m_hue = hue;
		m_item->setBrush(QColor::fromHsv(hue % 360, 255, 255));
	}
}

bool Kolf::BallObject::isInHole() const
{
	return m_inHole;
}

void Kolf::BallObject::setInHole(bool inHole)
{
	if (m_inHole == inHole)
		return;
	if ((m_inHole = inHole))
		setSimulationType(Kolf::Object::NoSimulation);
	else
		setSimulationType(Kolf::Object::DynamicSimulation);
}

Kolf::Overlay* Kolf::BallObject::createOverlay()
{
	//There's nothing you can do with a ball except for moving it around, so an empty overlay is enough.
	return new Kolf::EmptyOverlay(this);
}

void Kolf::BallObject::objectEvent(Kolf::Object::EventType type, const QVariant& data)
{
	if (type == Kolf::Object::VelocityChangeEvent)
	{
		QPointF velocity = data.value<QPointF>();
		//stop ball if velocity is too small
		const qreal velocitySquared = velocity * velocity;
		static const qreal threshold = Kolf::BallObject::BallRadius * Kolf::BallObject::BallRadius;
		if (velocitySquared > 0 && velocitySquared < threshold)
		{
			setVelocity(QPointF());
			m_stoppedSignalSent = false;
		}
		//emit stopped() signal as necessary
		velocity = this->velocity();
		if (velocity * velocity > 0)
			m_stoppedSignalSent = false;
		else if (!m_stoppedSignalSent)
		{
			emit stopped();
			m_stoppedSignalSent = true;
		}
	}
}

#include "object-ball.moc"
