/***************************************************************************
 *   Copyright 2008, 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "object-hole.h"
#include "object-ball.h"
#include "objectfactory.h"
#include "overlay-empty.h"
#include "scene-contactlistener.h"
#include "shape-circle.h"

#include <QPen>

const qreal Kolf::HoleObject::SizeFactor = 1.25;
const qreal Kolf::HoleObject::CaptureSizeFactor = 0.25;

REGISTER_KOLF_OBJECT(Kolf::HoleObject, {
	classObject->screenName = ki18n("Hole (target)");
	classObject->groupName = QLatin1String("01-basics");
})

Kolf::HoleObject::HoleObject(b2World* world)
	: Kolf::Object(world)
	, m_shape(new Kolf::CircleShape(Kolf::HoleObject::SizeFactor * Kolf::BallObject::BallRadius))
	, m_captureShape(new Kolf::CircleShape(Kolf::HoleObject::CaptureSizeFactor * Kolf::BallObject::BallRadius))
{
	//avoid to compute the radius too often
	const qreal r = m_shape->radius();
	m_item = new QGraphicsEllipseItem(QRectF(-r, -r, 2 * r, 2 * r), this);
	//normal initialization
	m_item->setPen(QPen(Qt::black));
	m_shape->setTraits(Kolf::Shape::VirtualShape);
	addShape(m_shape);
	m_captureShape->setTraits(Kolf::Shape::ParticipatesInCollisionDetection);
	addShape(m_captureShape);
}

Kolf::Overlay* Kolf::HoleObject::createOverlay()
{
	//There's nothing you can do with a hole except for moving it around, so an empty overlay is enough.
	return new Kolf::EmptyOverlay(this);
}

void holeBallInteraction(const Kolf::ContactData& data)
{
	//We do not use qobject_cast, but check the meta-objects, in order to return from this function as soon as possible if the interaction is not a hole-ball interaction. The comparison is a safe and complete type check because there are no subclasses below BallObject and HoleObject.
	const QMetaObject* mo1 = data.object1->metaObject();
	const QMetaObject* mo2 = data.object2->metaObject();
	Kolf::BallObject* ball; Kolf::HoleObject* hole;
	if (mo1 == &Kolf::BallObject::staticMetaObject && mo2 == &Kolf::HoleObject::staticMetaObject)
	{
		ball = reinterpret_cast<Kolf::BallObject*>(data.object1);
		hole = reinterpret_cast<Kolf::HoleObject*>(data.object2);
	}
	else if (mo1 == &Kolf::HoleObject::staticMetaObject && mo2 == &Kolf::BallObject::staticMetaObject)
	{
		hole = reinterpret_cast<Kolf::HoleObject*>(data.object1);
		ball = reinterpret_cast<Kolf::BallObject*>(data.object2);
	}
	else
		return;
	//This is a hole-ball interaction.
	ball->setInHole(true);
	ball->setPos(hole->pos());
	ball->setVelocity(QPointF());
}

REGISTER_KOLF_CONTACT_CALLBACK(holeBallInteraction)

#include "object-hole.moc"
