/***************************************************************************
 *   Copyright 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_OBJECTFACTORY_H
#define KOLF_OBJECTFACTORY_H

#include <QObject>
#include <KIcon>
#include <KLocalizedString>
class b2World;

namespace Kolf
{
	class Object;
	class Scene;
	typedef Kolf::Object*(*ObjectCreator)(b2World*);

	/**
	 * \class ObjectFactory
	 * \since 2.0
	 *
	 * Provides data and instance management for Kolf::Object subclasses. The factory encapsulates a Kolf::Scene instance, and allows to create arbitrary objects on this scene. It keeps a database of all instantiable Kolf::Object subclasses and provides methods for querying this database.
	 * \sa REGISTER_KOLF_OBJECT
	 */
	class ObjectFactory : public QObject
	{
		Q_OBJECT
		public:
			struct Class
			{
				Q_DISABLE_COPY(Class) //to speed things up because algorithms can become simpler internally
				public:
					Class();
					const QMetaObject* metaObject;

					KLocalizedString screenName;
					QIcon icon;
					bool editorMayCreateInstances;
					QString groupName;
			};

			///Creates a new object factory, that will create objects on the given \a scene.
			///You will usually want to use Kolf::Scene::objectFactory() instead of this constructor.
			ObjectFactory(Kolf::Scene* scene, b2World* world);

			///Returns a list of metadata for all known instantiable Kolf::Object subclasses.
			QList<const Kolf::ObjectFactory::Class*> knownObjectClasses() const;
			///Returns metadata for the Kolf::Object subclass with the given name, or 0 if no such class is known.
			const Kolf::ObjectFactory::Class* objectClass(const QByteArray& objectClassName) const;
			///Returns a new instance of the given Kolf::Object subclass, or 0 if invalid metadata has been given. The object is created on the scene which is associated to this factory. If a \a parentObject is given, the object will not be managed by Kolf::Scene.
			Kolf::Object* create(const Kolf::ObjectFactory::Class* objectClass, Kolf::Object* parentObject = 0);
			///Returns a new instance of the Kolf::Object subclass with the given name, or 0 if no such class is known. The object is created on the scene which is associated to this factory. If a \a parentObject is given, the object will not be managed by Kolf::Scene.
			Kolf::Object* create(const QByteArray& objectClassName, Kolf::Object* parentObject = 0);
		Q_SIGNALS:
			void objectCreated(Kolf::Object* object);
		private:
			Kolf::Object* create(Kolf::ObjectCreator creator, Kolf::Object* parentObject);

			Kolf::Scene* m_scene;
			b2World* m_world;
	};

	typedef void(*MetadataProvider)(Kolf::ObjectFactory::Class*);

	///\internal
	///This class needs to be publicly accessible for the REGISTER_KOLF_OBJECT macro.
	class ObjectRegistrar
	{
		public:
			ObjectRegistrar(const QMetaObject* object, Kolf::ObjectCreator creator, Kolf::MetadataProvider provider);
	};

	///\internal
	///This method needs to be publicly accessible for the REGISTER_KOLF_OBJECT macro.
	template<typename ObjectClass> Kolf::Object* __createObject(b2World* world)
	{
		return new ObjectClass(world);
	}
}

/**
 * \macro REGISTER_KOLF_OBJECT
 * \relates Kolf::Object
 * \relates Kolf::ObjectFactory
 *
 * Registers a Kolf::Object to make it visible to the game engine. For any Kolf::Object subclass, you will have to use this macro in the implementation file (.cpp extension) in the following pattern:
 * \code
 * REGISTER_KOLF_OBJECT(Kolf::FooObject, {
 *     classObject->screenName = ki18n("Foo");
 *     classObject->icon = KIcon("foo-bar-baz");
 *     classObject->groupName = QLatin1String("02-basic-obstacles");
 * })
 * \endcode
 * Just replace the values behind the assignment operators by metadata suitable for your subclass. For known group names, see the code of the other REGISTER_KOLF_OBJECT macro usages.
 * \note You have to include the Q_OBJECT macro in the definition of your Kolf::Object subclass to make this work.
 */
#define REGISTER_KOLF_OBJECT(ObjectClass, MetadataProviderCode) \
	static void __kolf_metadata_provider(Kolf::ObjectFactory::Class* classObject) MetadataProviderCode \
	static Kolf::ObjectRegistrar __kolf_registrar(&ObjectClass::staticMetaObject, &Kolf::__createObject<ObjectClass>, &__kolf_metadata_provider);

#endif // KOLF_OBJECTFACTORY_H
