/***************************************************************************
 *   Copyright 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_SCENE_CONTACTLISTENER_H
#define KOLF_SCENE_CONTACTLISTENER_H

#include <QList>
#include <Box2D/Dynamics/b2WorldCallbacks.h>

namespace Kolf
{
	class Object;
	class Shape;

	///Describes a collision contact between two shapes of two objects.
	///\see ContactListener
	struct ContactData
	{
		Kolf::Object* object1;
		Kolf::Object* object2;
		Kolf::Shape* shape1;
		Kolf::Shape* shape2;
	};

	///Records contacts between different shapes during physical simulation.
	///After each step of the physical simulation, the scene will call all contact callbacks that have been registered through the REGISTER_KOLF_CONTACT_CALLBACK macro for each of these contacts.
	class ContactListener : public b2ContactListener
	{
		public:
			///Clears the contact list.
			void clear();
			///Returns all contacts that have been made since the last clear() call.
			QList<ContactData> encounteredContacts() const;

			//implementation of b2ContactListener
			virtual void BeginContact(b2Contact* contact);
		private:
			QList<ContactData> m_encounteredContacts;
	};

	typedef void(*ContactCallback)(const Kolf::ContactData&);
	///Returns a list of all contact callbacks which have been registered through the REGISTER_KOLF_CONTACT_CALLBACK macro.
	QList<Kolf::ContactCallback> knownContactCallbacks();

	///\internal
	///This class needs to be publicly accessible for the REGISTER_KOLF_CONTACT_CALLBACK macro.
	class ContactCallbackRegistrar
	{
		public:
			ContactCallbackRegistrar(Kolf::ContactCallback callback);
	};
}

/**
 * \macro REGISTER_KOLF_CONTACT_CALLBACK
 * \relates Kolf::ContactCallback
 * \relates Kolf::ContactListener
 *
 * Registers a function of Kolf::ContactCallback type to make it visible to the Kolf::ContactListener. The function will then be called each time when two Kolf::Objects touch. Use this macro in the following way:
 * \code
 * void fooBar(const Kolf::ContactData& contactData) { ... }
 * REGISTER_KOLF_CONTACT_CALLBACK(fooBar)
 * \endcode
 */
#define REGISTER_KOLF_CONTACT_CALLBACK(FunctionName) \
	static Kolf::ContactCallbackRegistrar __kolf_contact_callback_registrar(&FunctionName);

#endif // KOLF_SCENE_CONTACTLISTENER_H
