/***************************************************************************
 *   Copyright 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "utils-undostack.h"

//BEGIN Utils::UndoStack

Utils::UndoStack::UndoStack(QObject* parent)
	: QObject(parent)
	, m_indexCount(0)
	, m_cleanIndex(-1)
	, m_currentIndex(-1)
{
}

bool Utils::UndoStack::canUndo() const
{
	return m_currentIndex > 0;
}

bool Utils::UndoStack::canRedo() const
{
	return m_currentIndex != -1 && m_currentIndex + 1 < m_indexCount;
}

bool Utils::UndoStack::isClean() const
{
	return m_cleanIndex != -1 && m_currentIndex == m_cleanIndex;
}

int Utils::UndoStack::count() const
{
	return m_indexCount;
}

int Utils::UndoStack::cleanIndex() const
{
	return m_cleanIndex;
}

int Utils::UndoStack::currentIndex() const
{
	return m_currentIndex;
}

void Utils::UndoStack::setIndex(int index)
{
	//normalization
	if (m_indexCount <= 0)
		return;
	if (index < 0)
		index = 0;
	if (index >= m_indexCount)
		index = m_indexCount - 1;
	if (m_currentIndex == index)
		return;
	//track state changes
	const bool oldCanRedo = canRedo(), oldCanUndo = canUndo(), oldClean = isClean();
	//apply state
	applyState(index);
	m_currentIndex = index;
	//report state changes
	emit indexChanged(m_currentIndex);
	const bool newCanRedo = canRedo(), newCanUndo = canUndo(), newClean = isClean();
	if (oldClean != newClean)
		emit cleanChanged(newClean);
	if (oldCanRedo != newCanRedo)
		emit canRedoChanged(newCanRedo);
	if (oldCanUndo != newCanUndo)
		emit canUndoChanged(newCanUndo);
}

void Utils::UndoStack::undo()
{
	setIndex(m_currentIndex - 1);
}

void Utils::UndoStack::redo()
{
	setIndex(m_currentIndex + 1);
}

void Utils::UndoStack::setClean()
{
	if (m_cleanIndex == m_currentIndex)
		return;
	m_cleanIndex = m_currentIndex;
	emit cleanChanged(true);
}

void Utils::UndoStack::trimStateList()
{
	if (m_currentIndex < 0)
		return;
	//track state changes (canUndo will never change in this step)
	const bool oldCanRedo = canRedo(), oldClean = isClean();
	//remove unreachable states from state history
	removeStatesAbove(m_currentIndex);
	m_indexCount = m_currentIndex + 1;
	//reset clean index if it points to one of the deleted states
	if (m_cleanIndex >= m_indexCount)
		m_cleanIndex = -1;
	//report state changes
	const bool newCanRedo = false, newClean = isClean();
	if (oldClean != newClean)
		emit cleanChanged(newClean);
	if (oldCanRedo != newCanRedo)
		emit canRedoChanged(newCanRedo);
}

void Utils::UndoStack::beginPushState()
{
	trimStateList();
}

void Utils::UndoStack::endPushState()
{
	//track state changes (canRedo will never change in this step)
	const bool oldCanUndo = canUndo(), oldClean = isClean();
	//perform update
	m_currentIndex = m_indexCount;
	++m_indexCount;
	//report state changes
	emit indexChanged(m_currentIndex);
	emit statePushed(m_currentIndex);
	const bool newCanUndo = true, newClean = isClean();
	if (oldClean != newClean)
		emit cleanChanged(newClean);
	if (oldCanUndo != newCanUndo)
		emit canUndoChanged(newCanUndo);
}

//END Utils::UndoStack
//BEGIN Utils::UndoGroup

Utils::UndoGroup::UndoGroup(QObject* parent)
	: Utils::UndoStack(parent)
{
}

Utils::UndoGroup::~UndoGroup()
{
	qDeleteAll(m_stacks);
}

QList<Utils::UndoStack*> Utils::UndoGroup::stacks() const
{
	return m_stacks;
}

void Utils::UndoGroup::addStack(Utils::UndoStack* stack)
{
	addStacks(QList<Utils::UndoStack*>() << stack);
}

void Utils::UndoGroup::addStacks(QList<Utils::UndoStack*> stacks)
{
	//append to substacks list
	if (stacks.isEmpty())
		return;
	const int oldStacksCount = m_stacks.count();
	m_stacks.append(stacks);
	const int newStacksCount = m_stacks.count();
	//update states to the new substacks count
	QList<Utils::UndoGroup::State>::iterator it1 = m_states.begin(), it2 = m_states.end();
	for (; it1 != it2; ++it1)
	{
		Utils::UndoGroup::State& state = *it1;
		state.resize(m_stacks.count()); //The new elements are set to 0, which is exactly the state index which we want.
	}
	//add states for the substates in the new substacks
	beginPushState(); //this is similar to a push of multiple states
	Utils::UndoGroup::State state = m_states.value(m_states.count() - 1);
	if (m_states.isEmpty())
		state = Utils::UndoGroup::State(m_stacks.count(), 0);
	for (int substackIndex = oldStacksCount; substackIndex < newStacksCount; ++substackIndex)
	{
		Utils::UndoStack* substack = m_stacks[substackIndex];
		connect(substack, SIGNAL(statePushed(int)), this, SLOT(pushSubstate(int)));
		substack->trimStateList(); //abuse to trim the state list to the current index
		const int substateIndexCount = substack->count();
		for (int substateIndex = 0; substateIndex < substateIndexCount; ++substateIndex)
		{
			if (state[substackIndex] == substateIndex && count() > 0)
				continue; //no change from the previous state
			state[substackIndex] = substateIndex;
			m_states << state;
			endPushState();
			emit substatePushed(substack, substateIndex);
			beginPushState(); //more states might be added
		}
	}
}

void Utils::UndoGroup::setClean()
{
	this->Utils::UndoStack::setClean();
	foreach (Utils::UndoStack* stack, m_stacks)
		stack->setClean();
}

void Utils::UndoGroup::applyState(int index)
{
	Utils::UndoGroup::State state = m_states[index];
	const int substackCount = m_stacks.count();
	for (int substackIndex = 0; substackIndex < substackCount; ++substackIndex)
		m_stacks[substackIndex]->setIndex(state[substackIndex]);
}

void Utils::UndoGroup::removeStatesAbove(int index)
{
	while (index < m_states.count() - 1)
		m_states.removeLast();
}

void Utils::UndoGroup::pushSubstate(int index)
{
	//copy last state
	Utils::UndoGroup::State state = m_states.value(m_states.count() - 1);
	if (m_states.isEmpty())
		state = Utils::UndoGroup::State(m_stacks.count(), 0);
	//modify state
	Utils::UndoStack* substack = qobject_cast<Utils::UndoStack*>(sender());
	const int substackIndex = m_stacks.indexOf(substack);
	if (substackIndex < 0)
		return;
	state[substackIndex] = index;
	beginPushState();
	m_states << state;
	endPushState();
	//done
	emit substatePushed(substack, index);
}

//END Utils::UndoGroup

#include "utils-undostack.moc"
