/***************************************************************************
 *   Copyright 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef UTILS_UNDOSTACK_H
#define UTILS_UNDOSTACK_H

#include <QObject>
#include <QVector>

namespace Utils
{
	/**
	 * \class UndoStack
	 *
	 * This class behaves similar to QUndoStack, but with an important difference: Its interface is based on states, not on commands, because the complexity of the manipulated objects is expected to be small.
	 */
	class UndoStack : public QObject
	{
		Q_OBJECT
		public:
			UndoStack(QObject* parent = 0);

			///Returns whether there is a previous state available for undo.
			bool canUndo() const;
			///Returns whether there is a following state available for redo.
			bool canRedo() const;
			///Returns whether the stack is in the clean state.
			bool isClean() const;

			///Returns the number of states on the stack.
			int count() const;
			///Returns the index of the clean state. This is the index at which setClean() was called.
			///A stack may not have a clean index. This happens if a document is saved, some commands are undone, then a new command is pushed. Since push() deletes all the undone commands before pushing the new command, the stack can't return to the clean state again. In this case, this function returns -1.
			int cleanIndex() const;
			///Returns the index of the current state, or -1 if no states have been pushed onto this stack yet.
			int currentIndex() const;

			///Delete all states after the current state (not including the current state!).
			void trimStateList();
		public Q_SLOTS:
			///Moves to the state with the given index.
			void setIndex(int index);
			///Moves to the previous state (if there is one).
			void undo();
			///Moves to the next state (if there is one).
			void redo();
			///Defines the current state to be the clean state.
			virtual void setClean();
		protected:
			///Reimplement this method to apply the state with the given \a index. (You are expected to save the states in your subclass implementation.)
			///\warning Do not call this method from your subclass implementation.
			virtual void applyState(int index) = 0;
			///Reimplement this method to remove all states above the state with the given \a index (but not that state!). This is the only method in which your subclass implementation is allowed to remove states from its state storage without messing up the data in the base class!
			///\warning Do not call this method from your subclass implementation.
			virtual void removeStatesAbove(int index) = 0;
			///Call this method from your pushState() functions before you insert the state into the data storage of your subclass. At the moment, this is only a synonym for trimStateList().
			///\warning This method will likely change the state count through removeStatesAbove().
			void beginPushState();
			///Call this method from your pushState() functions after you have inserted the state into the data storage of your subclass.
			void endPushState();
		Q_SIGNALS:
			///This signal is emitted whenever the value of canRedo() changes.
			void canRedoChanged(bool canRedo);
			///This signal is emitted whenever the value of canUndo() changes.
			void canUndoChanged(bool canUndo);
			///This signal is emitted whenever the stack enters or leaves the clean state.
			void cleanChanged(bool clean);
			///This signal is emitted whenever another state is entered. The state is given by the \a index.
			void indexChanged(int index);
			///This signal is emitted when a new state has been pushed. The parameter is the \a index of this new state.
			//NOTE: The availability of this signal is the main reason why we cannot use QUndoStack directly. This signal is absolutely necessary for the implementation of Utils::UndoGroup.
			void statePushed(int index);
		private:
			int m_indexCount, m_cleanIndex, m_currentIndex;
	};

	/**
	 * \class UndoGroup
	 *
	 * This UndoStack merges several UndoStacks into one. It especially keeps track of the chronological order in which states where inserted into the underlying stacks.
	 *
	 * \warning Stuff will break if you undo(), redo() or setIndex() on the stacks contained in this group after having inserted them into the group.
	 */
	class UndoGroup : public Utils::UndoStack
	{
		Q_OBJECT
		public:
			UndoGroup(QObject* parent = 0);
			virtual ~UndoGroup();

			///\overload
			void addStack(Utils::UndoStack* stack);
			///Add some \a stacks to this group. The group will take ownership of these stacks. If the stacks already have states, these are appended at the end of the group stack (preserving the order of the list given to this method).
			///\warning In the new \a stacks, any states after the current state are removed.
			void addStacks(QList<Utils::UndoStack*> stacks);
			///Returns a list of all stacks in this group.
			///\warning Stuff will break if you undo(), redo() or setIndex() on the stacks
			QList<Utils::UndoStack*> stacks() const;

			virtual void setClean();
		Q_SIGNALS:
			void substatePushed(Utils::UndoStack* stack, int index);
		protected:
			virtual void applyState(int index);
			virtual void removeStatesAbove(int index);
		private Q_SLOTS:
			void pushSubstate(int index);
		private:
			QList<Utils::UndoStack*> m_stacks;
			//NOTE: A state is described by the state indexes of the stacks. The element index in the QVector is the same as the element index in m_stacks.
			typedef QVector<int> State;
			QList<Utils::UndoGroup::State> m_states;
	};
}

#endif // UTILS_UNDOSTACK_H
