/***************************************************************************
 *   Copyright 2008, 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "course.h"
#include "../base/managedobjectgroup.h"
#include "../base/putter-simple.h"
#include "../base/scene.h"

#include <QFileInfo>
#include <KConfig>
#include <KConfigGroup>
#include <KStandardDirs>

namespace Kolf
{
	static const QString courseFileNamePattern("kolf2/courses/%1.kolf2");
	//group names for KConfigGroup
	static const QString mainGroupKey("Kolf 2");
	static const QString holeGroupKey("Hole %1");
	//key names for KConfigGroup
	static const QString mainVersionKey("Version");
	static const QString mainTitleKey("Name");
	static const QString holeParKey("Par");
}

//BEGIN Kolf::Hole

Kolf::Hole::Hole(int parScore)
	: m_parScore(parScore)
	, m_scene(new Kolf::Scene)
	, m_managedObjectGroup(new Kolf::ManagedObjectGroup(m_scene->objectFactory()))
	, m_putter(new Kolf::SimplePutter(m_scene))
{
}

Kolf::Hole::~Hole()
{
	delete m_putter;
	delete m_managedObjectGroup;
	delete m_scene;
}

int Kolf::Hole::parScore() const
{
	return m_parScore;
}

void Kolf::Hole::setParScore(int parScore)
{
	m_parScore = qMax(1, parScore);
}

Kolf::Scene* Kolf::Hole::scene() const
{
	return m_scene;
}

Kolf::ManagedObjectGroup* Kolf::Hole::managedObjectGroup() const
{
	return m_managedObjectGroup;
}

Kolf::Putter* Kolf::Hole::putter() const
{
	return m_putter;
}

//END Kolf::Hole
//BEGIN Kolf::Course

QStringList Kolf::Course::availableCourses()
{
	const QStringList paths = KStandardDirs().findAllResources("data", Kolf::courseFileNamePattern.arg("*"), KStandardDirs::NoDuplicates);
	QStringList courses;
	//extract course names
	foreach (const QString& path, paths)
		courses << QFileInfo(path).baseName();
	return courses;
}

QString Kolf::Course::courseTitle(const QString& courseName)
{
	const QString fileName = KStandardDirs::locate("data", Kolf::courseFileNamePattern.arg(courseName));
	if (fileName.isEmpty())
		return QString();
	KConfig config(fileName);
	KConfigGroup mainInfoGroup(&config, Kolf::mainGroupKey);
	return mainInfoGroup.readEntry(Kolf::mainTitleKey, QString());
}

Kolf::Course::Course(const QString& courseName)
	: m_courseName(courseName)
{
	//open course
	const QString fileName = KStandardDirs::locate("data", Kolf::courseFileNamePattern.arg(courseName));
	if (fileName.isEmpty())
		return;
	KConfig config(fileName);
	//main info: check version
	KConfigGroup mainInfoGroup(&config, Kolf::mainGroupKey);
	const int version = mainInfoGroup.readEntry(Kolf::mainVersionKey, 0);
	if (version != 1)
		return;
	//determine hole count - NOTE: In the group captions, hole numbering starts at one.
	int holeCount = 0;
	while (config.hasGroup(Kolf::holeGroupKey.arg(holeCount + 1)))
		++holeCount;
	//main info: get name
	m_title = mainInfoGroup.readEntry(Kolf::mainTitleKey, QString());
	if (m_title.isEmpty())
		return;
	//read holes - NOTE: In the group captions, hole numbering starts at one.
	for (int i = 1; i <= holeCount; ++i)
	{
		KConfigGroup holeGroup(&config, Kolf::holeGroupKey.arg(i));
		const int parScore = holeGroup.readEntry(Kolf::holeParKey, 0);
		Kolf::Hole* hole = new Kolf::Hole(parScore);
		hole->managedObjectGroup()->deserialize(&holeGroup);
		m_holes << hole;
	}
}

Kolf::Course::~Course()
{
	QList<Kolf::Hole*>::const_iterator it1 = m_holes.begin(), it2 = m_holes.end();
	for (; it1 != it2; ++it1)
		delete *it1;
	//NOTE: qDeleteAll does not work here, because ~Hole is a protected friend.
}

QString Kolf::Course::title() const
{
	return m_title;
}

QList<Kolf::Hole*> Kolf::Course::holes() const
{
	return m_holes;
}

void Kolf::Course::save() const
{
	KConfig config(KStandardDirs::locateLocal("data", Kolf::courseFileNamePattern.arg(m_courseName)));
	//write main info
	KConfigGroup mainInfoGroup(&config, Kolf::mainGroupKey);
	mainInfoGroup.writeEntry(Kolf::mainVersionKey, 1);
	mainInfoGroup.writeEntry(Kolf::mainTitleKey, m_title);
	//write holes - NOTE: In the group captions, hole numbering starts at one.
	for (int i = 0; i < m_holes.count(); ++i)
	{
		Kolf::Hole* hole = m_holes.at(i);
		KConfigGroup holeGroup(&config, Kolf::holeGroupKey.arg(i + 1));
		holeGroup.writeEntry(Kolf::holeParKey, hole->parScore());
		hole->managedObjectGroup()->serialize(&holeGroup);
		hole->managedObjectGroup()->setClean();
	}
}

//END Kolf::Course

#include "course.moc"
