/***************************************************************************
 *   Copyright 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "game-editor.h"
#include "course.h"
#include "scoremodel.h"
#include "../base/managedobjectgroup.h"
#include "../base/objectfactory.h"
#include "../base/scene.h"
#include "../interface/interface.h"

#include <QDockWidget>
#include <QToolButton>
#include <QVBoxLayout>
#include <KLocalizedString>
#include <KSeparator>

//BEGIN Kolf::EditorToolbox

Kolf::EditorToolbox::EditorToolbox(Kolf::EditorGame* game)
	: m_game(game)
	, m_buttonLayout(new QVBoxLayout)
{
	QVBoxLayout* mainLayout = new QVBoxLayout;
	mainLayout->addLayout(m_buttonLayout);
	mainLayout->addStretch();
	setLayout(mainLayout);
}

static bool textLessThan(QToolButton* tb1, QToolButton* tb2)
{
	return tb1->text() < tb2->text();
}

void Kolf::EditorToolbox::bootHole(Kolf::Hole* hole)
{
	QList<const Kolf::ObjectFactory::Class*> classes = hole->scene()->objectFactory()->knownObjectClasses();
	QMap<QString, QList<QToolButton*> > buttonGroups;
	foreach (const Kolf::ObjectFactory::Class* objectClass, classes)
	{
		if (!objectClass->editorMayCreateInstances)
			continue;
		//create button
		QToolButton* button = new QToolButton;
		button->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);
		button->setAutoRaise(true);
		button->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed);
		button->setIcon(objectClass->icon);
		button->setText(objectClass->screenName.toString());
		//store button (it is inserted into the layout later)
		buttonGroups[objectClass->groupName] << button;
		m_buttons << button;
	}
	//organise buttons into layout (NOTE: QMap iteration gives values sorted by key)
	QMap<QString, QList<QToolButton*> >::iterator it1 = buttonGroups.begin(), it2 = buttonGroups.end();
	bool firstGroup = true;
	for (; it1 != it2; ++it1)
	{
		//add separator before group
		if (firstGroup)
			firstGroup = false;
		else
			m_buttonLayout->addWidget(new KSeparator(Qt::Horizontal));
		//add buttons of group (after sorting)
		QList<QToolButton*>& buttonGroup = *it1;
		qSort(buttonGroup.begin(), buttonGroup.end(), &textLessThan);
		foreach (QToolButton* button, buttonGroup)
			m_buttonLayout->addWidget(button);
	}
}

static void flushLayout(QLayout* layout)
{
	QLayoutItem *child;
	while ((child = layout->takeAt(0)))
		delete child;
}

void Kolf::EditorToolbox::shutdownHole(Kolf::Hole* hole)
{
	Q_UNUSED(hole)
	flushLayout(m_buttonLayout);
}

//END Kolf::EditorToolbox
//BEGIN Kolf::EditorGame

Kolf::EditorGame::EditorGame(Kolf::Course* course, Kolf::Interface* interface)
	: Kolf::Game(course, interface)
	, m_toolboxDock(new QDockWidget(i18n("Editor toolbox")))
	, m_toolbox(new Kolf::EditorToolbox(this))
{
	foreach (Kolf::Hole* hole, course->holes())
		connect(hole->managedObjectGroup(), SIGNAL(cleanChanged(bool)), this, SLOT(cleanChanged()));
	connect(interface, SIGNAL(saveRequest()), this, SLOT(saveRequest()));
	scoreModel()->setEditable(true);
	//configure toolbox
	m_toolboxDock->setWidget(m_toolbox);
	m_toolboxDock->setObjectName("EditorToolboxDock");
	m_toolboxDock->setFeatures(QDockWidget::DockWidgetMovable | QDockWidget::DockWidgetFloatable);
	m_toolboxDock->setAllowedAreas(Qt::LeftDockWidgetArea | Qt::RightDockWidgetArea);
	interface->addDockWidget(Qt::LeftDockWidgetArea, m_toolboxDock);
}

Kolf::EditorGame::~EditorGame()
{
	delete m_toolboxDock;
}

void Kolf::EditorGame::bootHole(Kolf::Hole* hole)
{
	hole->scene()->setOverlaysActive(true);
	Kolf::ManagedObjectGroup* mog = hole->managedObjectGroup();
	interface()->setRedoEnabled(mog->canRedo());
	interface()->setUndoEnabled(mog->canUndo());
	connect(mog, SIGNAL(canRedoChanged(bool)), interface(), SLOT(setRedoEnabled(bool)));
	connect(mog, SIGNAL(canUndoChanged(bool)), interface(), SLOT(setUndoEnabled(bool)));
	connect(interface(), SIGNAL(redoRequest()), mog, SLOT(redo()));
	connect(interface(), SIGNAL(undoRequest()), mog, SLOT(undo()));
	m_toolbox->bootHole(hole);
}

void Kolf::EditorGame::shutdownHole(Kolf::Hole* hole)
{
	hole->scene()->setOverlaysActive(false);
	Kolf::ManagedObjectGroup* mog = hole->managedObjectGroup();
	disconnect(mog, 0, interface(), 0);
	disconnect(interface(), 0, mog, 0);
	m_toolbox->shutdownHole(hole);
}

void Kolf::EditorGame::cleanChanged()
{
	bool isClean = true;
	foreach (Kolf::Hole* hole, course()->holes())
		if (!hole->managedObjectGroup()->isClean())
		{
			isClean = false;
			break;
		}
	interface()->setClean(isClean);
}

void Kolf::EditorGame::saveRequest()
{
	course()->save();
	interface()->setClean(true);
}

//END Kolf::EditorGame

#include "game-editor.moc"
