/***************************************************************************
 *   Copyright 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "game.h"
#include "course.h"
#include "player.h"
#include "scoremodel.h"
#include "../base/scene.h"
#include "../interface/interface.h"

Kolf::Game::Game(Kolf::Course* course, Kolf::Interface* interface)
	: m_currentHoleIndex(-1)
	, m_course(course)
	, m_interface(interface)
	, m_scoreModel(new Kolf::ScoreModel(this))
{
	m_interface->setHoleCount(m_course->holes().count());
	m_interface->setCurrentHole(m_currentHoleIndex);
	m_interface->setWindowTitle(m_course->title());
	connect(m_interface, SIGNAL(holeSelected(int)), this, SLOT(setCurrentHoleIndex(int)));
	//the following commands provide sane defaults
	m_interface->setClean(true);
	m_interface->setUndoEnabled(false);
	m_interface->setRedoEnabled(false);
}

Kolf::Game::~Game()
{
	setCurrentHoleIndex(-1); //deinitialize everything
	delete m_course;
	qDeleteAll(m_players);
}

int Kolf::Game::currentHoleIndex() const
{
	return m_currentHoleIndex;
}

Kolf::Course* Kolf::Game::course() const
{
	return m_course;
}

Kolf::Interface* Kolf::Game::interface() const
{
	return m_interface;
}

QList<Kolf::Player*> Kolf::Game::players() const
{
	return m_players;
}

Kolf::ScoreModel* Kolf::Game::scoreModel() const
{
	return m_scoreModel;
}

void Kolf::Game::addPlayer(Kolf::Player* player)
{
	m_players << player;
	m_scoreModel->announcePlayerAdded();
}

Kolf::Hole* Kolf::Game::currentHole() const
{
	return m_course->holes().value(m_currentHoleIndex);
}

void Kolf::Game::setCurrentHoleIndex(int holeIndex)
{
	if (m_currentHoleIndex == holeIndex)
		return;
	Kolf::Hole* oldHole = m_course->holes().value(m_currentHoleIndex);
	Kolf::Hole* newHole = m_course->holes().value(holeIndex);
	//announce changes
	if (oldHole)
		shutdownHole(oldHole);
	//find subobjects for new hole
	Kolf::Scene* newScene = 0;
	Kolf::Putter* newPutter = 0;
	if (newHole)
	{
		newScene = newHole->scene();
		newPutter = newHole->putter();
	}
	//change hole
	m_currentHoleIndex = holeIndex;
	m_interface->setCurrentHole(m_currentHoleIndex);
	m_interface->setScene(newScene);
	foreach (Kolf::Player* player, m_players)
		player->setScene(newScene, newPutter);
	//announce changes
	if (newHole)
		bootHole(newHole);
	if (holeIndex > m_course->holes().count())
		emit finished();
	m_scoreModel->announceCurrentHoleChanged();
}

#include "game.moc"
