/***************************************************************************
 *   Copyright 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

//TODO: FIXME: check KUIT/context everywhere when done
#include "scoremodel.h"
#include "course.h"
#include "game.h"
#include "player.h"

#include <QApplication>
#include <KColorScheme>
#include <KIcon>
#include <KLocalizedString>

Kolf::ScoreModel::ScoreModel(Kolf::Game* game)
	: m_game(game)
	, m_editable(false)
	, m_currentHoleIndex(-1)
	, m_totalScores(1)
{
	Kolf::Course* course = game->course();
	m_totalScores[0] = 0;
	foreach (Kolf::Hole* hole, course->holes())
	{
		QVector<int> holeScores(1);
		holeScores[0] = hole->parScore();
		m_totalScores[0] += hole->parScore();
		m_scores << holeScores;
	}
}

void Kolf::ScoreModel::setEditable(bool editable)
{
	m_editable = editable;
}

void Kolf::ScoreModel::announcePlayerAdded()
{
	//add scores for new player
	const int playerCount = m_game->players().count();
	QList<QVector<int> >::iterator it1 = m_scores.begin(), it2 = m_scores.end();
	for (; it1 != it2; ++it1)
	{
		QVector<int>& holeScores = *it1;
		holeScores.resize(playerCount + 1);
		holeScores[playerCount] = 0;
	}
	m_totalScores.resize(playerCount + 1);
	m_totalScores[playerCount] = 0;
	//announce change
	beginInsertColumns(QModelIndex(), playerCount, playerCount);
	endInsertColumns();
}

void Kolf::ScoreModel::announceCurrentHoleChanged()
{
	const int oldHoleIndex = m_currentHoleIndex;
	const int newHoleIndex = m_currentHoleIndex = m_game->currentHoleIndex();
	const int holeCount = m_game->course()->holes().count();
	const int playerCount = m_game->players().count();
	//report header data changes
	if (oldHoleIndex >= 0 && oldHoleIndex < holeCount)
		emit headerDataChanged(Qt::Vertical, oldHoleIndex, oldHoleIndex);
	if (newHoleIndex >= 0 && newHoleIndex < holeCount)
		emit headerDataChanged(Qt::Vertical, newHoleIndex, newHoleIndex);
	//report data change for the data row of the old hole (background brush has changed)
	QModelIndex startIndex = index(oldHoleIndex, 0);
	QModelIndex endIndex = index(oldHoleIndex, playerCount + 1);
	emit dataChanged(startIndex, endIndex);
	//report data change for the data row of the new hole (background brush has changed)
	startIndex = index(newHoleIndex, 0);
	endIndex = index(newHoleIndex, playerCount + 1);
	emit dataChanged(startIndex, endIndex);
}

int Kolf::ScoreModel::score(Kolf::Hole* hole, Kolf::Player* player) const
{
	const int holeIndex = m_game->course()->holes().indexOf(hole);
	if (holeIndex == -1)
		return 0;
	int playerIndex = 0;
	if (player)
	{
		playerIndex = m_game->players().indexOf(player) + 1; //"+1" for the par scores
		if (playerIndex == 0)
			return 0;
	}
	return m_scores[holeIndex][playerIndex];
}

void Kolf::ScoreModel::setScore(Kolf::Hole* hole, Kolf::Player* player, int score)
{
	const int holeIndex = m_game->course()->holes().indexOf(hole);
	if (holeIndex == -1)
		return;
	int playerIndex = 0;
	if (player)
	{
		playerIndex = m_game->players().indexOf(player) + 1; //"+1" for the par scores
		if (playerIndex == 0)
			return;
	}
	int oldScore = m_scores[holeIndex][playerIndex];
	m_scores[holeIndex][playerIndex] = score;
	m_totalScores[playerIndex] += score - oldScore;
	//report data change for the single score
	QModelIndex dataIndex = index(holeIndex, playerIndex);
	emit dataChanged(dataIndex, dataIndex);
	//report data change for the total score
	dataIndex = index(m_game->course()->holes().count() + 1, playerIndex + 1);
	emit dataChanged(dataIndex, dataIndex);
}

//BEGIN QAbstractTableModel reimplementation

int Kolf::ScoreModel::rowCount(const QModelIndex& parent) const
{
	if (parent.isValid())
		return 0;
	return m_game->course()->holes().count() + 1; //"+1" for the total score
}

int Kolf::ScoreModel::columnCount(const QModelIndex& parent) const
{
	if (parent.isValid())
		return 0;
	return m_game->players().count() + 1; //"+1" for the par scores
}

Qt::ItemFlags Kolf::ScoreModel::flags(const QModelIndex& index) const
{
	Qt::ItemFlags flags = Qt::ItemIsEnabled;
	//are par scores editable?
	if (m_editable)
		if (index.column() == 0 && index.row() < m_game->course()->holes().count())
			flags |= Qt::ItemIsSelectable | Qt::ItemIsEditable;
	return flags;
}

QVariant Kolf::ScoreModel::data(const QModelIndex& index, int role) const
{
	if (index.parent().isValid())
		return QVariant();
	const int holeIndex = index.row();
	const int playerIndex = index.column();
	//retrieve data
	switch (role)
	{
		case Qt::DisplayRole: case Qt::EditRole:
			//show score
			if (holeIndex == m_game->course()->holes().count())
				return m_totalScores[playerIndex];
			else
			{
				if (playerIndex == 0 || holeIndex >= m_currentHoleIndex)
					return m_scores[holeIndex][playerIndex];
				else
				{
					//player has finished this hole -> show difference to par together with score
					return i18nc("@item:intable \"score (difference)\" on scorecard", "%1 (%2)",
						m_scores[holeIndex][playerIndex],
						m_scores[holeIndex][playerIndex] - m_scores[holeIndex][0]
					);
				}
			}
		case Qt::BackgroundRole: {
			KColorScheme::BackgroundRole bgRole = KColorScheme::NormalBackground;
			//mark current hole
			if (holeIndex == m_currentHoleIndex)
				bgRole = KColorScheme::NeutralBackground;
			//create brush
			return KStatefulBrush(KColorScheme::View, bgRole).brush(QPalette::Active);
		}
		default:
			return QVariant();
	}
}

QVariant Kolf::ScoreModel::headerData(int section, Qt::Orientation orientation, int role) const
{
	switch (role)
	{
		case Qt::DisplayRole:
			if (orientation == Qt::Vertical)
			{
				if (section == m_game->course()->holes().count())
					return i18nc("@title:column in scorecard (total score)", "Total");
				else
					return i18nc("@title:column in scorecard", "Hole %1", section + 1); //"+1" because user-visible hole numbers start at 1
			}
			else //orientation == Qt::Horizontal
			{
				if (section == 0)
					return i18nc("@title:column in scorecard", "Par score");
				else
					return m_game->players()[section - 1]->name();
			}
		case Qt::DecorationRole:
			//mark current hole
			if (orientation == Qt::Vertical && section == m_game->currentHoleIndex())
				return KIcon(QApplication::isLeftToRight() ? "arrow-right" : "arrow-left");
			else
				return QVariant();
		case Qt::BackgroundRole:
			//show player color
			if (orientation == Qt::Horizontal && section != 0) //section == 0 is par
				return QColor::fromHsv(m_game->players()[section - 1]->hue(), 255, 255);
		default:
			return QVariant();
	}
}

bool Kolf::ScoreModel::setData(const QModelIndex& index, const QVariant& value, int role)
{
	if (role != Qt::EditRole)
		return false;
	//allow to edit parscores only
	if (index.column() != 0)
		return false;
	//find hole to edit
	const int holeIndex = index.row();
	Kolf::Hole* hole = m_game->course()->holes().value(holeIndex);
	if (!hole)
		return false;
	//edit par score and total par score
	const int oldParScore = m_scores[holeIndex][0];
	const int newParScore = m_scores[holeIndex][0] = value.toInt();
	hole->setParScore(newParScore);
	m_totalScores[0] += newParScore - oldParScore;
	//announce changes (also to the total par score)
	emit dataChanged(index, index);
	const QModelIndex totalIndex = this->index(m_game->course()->holes().count(), 0);
	emit dataChanged(totalIndex, totalIndex);
	return true;
}

//END QAbstractTableModel reimplementation
