/***************************************************************************
 *   Copyright 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "interface.h"
#include "view.h"

#include <KAction>
#include <KActionCollection>
#include <KLocalizedString>
#include <KSelectAction>
#include <KStandardGameAction>
#include <KXmlGuiWindow>

Kolf::Interface::Interface(KXmlGuiWindow* window)
	: QObject(window)
	, m_window(window)
	, m_view(new Kolf::View)
	, m_currentHole(-1)
	, m_holeCount(0)
	, m_clean(true)
{
	window->setCentralWidget(m_view);
	KActionCollection* parent = window->actionCollection();
	KAction* act;
	//Kolf -> New
	m_actions[Kolf::NewAction] = act = KStandardGameAction::gameNew(window, 0, 0); //FIXME: connect to newgame slot in window
	parent->addAction("kolf_main_new", act);
	//Kolf -> Save
	m_actions[Kolf::SaveAction] = act = KStandardGameAction::save(this, SIGNAL(saveRequest()), 0);
	parent->addAction("kolf_main_save", act);
	//Kolf -> Quit
	m_actions[Kolf::QuitAction] = act = KStandardGameAction::quit(window, SLOT(close()), 0);
	parent->addAction("kolf_main_quit", act);
	//Hole -> Undo
	m_actions[Kolf::UndoAction] = act = KStandardGameAction::undo(this, SIGNAL(undoRequest()), 0);
	parent->addAction("kolf_hole_undo", act);
	//Hole -> Redo
	m_actions[Kolf::RedoAction] = act = KStandardGameAction::redo(this, SIGNAL(redoRequest()), 0);
	parent->addAction("kolf_hole_redo", act);
	//Course -> Previous hole
	m_actions[Kolf::PreviousHoleAction] = act = KStandardAction::back(this, SLOT(actionTriggeredMapper()), 0);
	act->setText(i18nc("@action:inmenu", "&Previous hole"));
	parent->addAction("kolf_course_previoushole", act);
	//Course -> Next hole
	m_actions[Kolf::NextHoleAction] = act = KStandardAction::forward(this, SLOT(actionTriggeredMapper()), 0);
	act->setText(i18nc("@action:inmenu", "&Next hole"));
	parent->addAction("kolf_course_nexthole", act);
	//Course -> Choose hole
	m_actions[Kolf::ChooseHoleAction] = act = new KSelectAction(i18nc("@action:inmenu", "Switch to hole"), this);
	parent->addAction("kolf_course_choosehole", act);
	connect(act, SIGNAL(triggered(int)), this, SLOT(actionTriggeredMapper(int)));
	//further setup
	setupHoleActions();
}

void Kolf::Interface::addDockWidget(Qt::DockWidgetArea area, QDockWidget* widget)
{
	m_window->addDockWidget(area, widget);
	//TODO: properly load dock widget position
}

void Kolf::Interface::setActionEnabled(Kolf::Action action, bool enabled)
{
	if (action >= Kolf::PreviousHoleAction && action <= Kolf::ChooseHoleAction)
		return;
	m_actions[action]->setEnabled(enabled);
}

void Kolf::Interface::setCurrentHole(int hole)
{
	if (m_currentHole == hole)
		return;
	m_currentHole = hole;
	setupHoleActions();
}

void Kolf::Interface::setHoleCount(int holes)
{
	if (holes < 0)
		holes = -1; //normalization to make algorithms easier
	if (m_holeCount == holes)
		return;
	m_holeCount = holes;
	setupHoleActions();
}

void Kolf::Interface::setScene(QGraphicsScene* scene)
{
	m_view->setScene(scene);
}

void Kolf::Interface::setWindowTitle(const QString& windowTitle)
{
	m_windowTitle = windowTitle;
	m_window->setCaption(m_windowTitle, !m_clean);
}

void Kolf::Interface::setClean(bool clean)
{
	m_clean = clean;
	m_actions[Kolf::SaveAction]->setEnabled(!m_clean);
	m_window->setCaption(m_windowTitle, !m_clean);
}

void Kolf::Interface::setRedoEnabled(bool redoEnabled)
{
	m_actions[Kolf::RedoAction]->setEnabled(redoEnabled);
}

void Kolf::Interface::setUndoEnabled(bool undoEnabled)
{
	m_actions[Kolf::UndoAction]->setEnabled(undoEnabled);
}

void Kolf::Interface::actionTriggeredMapper()
{
	QObject* sender = this->sender();
	if (sender == m_actions[Kolf::PreviousHoleAction])
		emit holeSelected(m_currentHole - 1);
	else if (sender == m_actions[Kolf::NextHoleAction])
		emit holeSelected(m_currentHole + 1);
}

void Kolf::Interface::actionTriggeredMapper(int index)
{
	emit holeSelected(index);
}

void Kolf::Interface::setupHoleActions()
{
	//toggle state of "Previous/Next hole" actions
	if (m_holeCount <= 0 || m_currentHole < 0 || m_currentHole >= m_holeCount)
	{
		m_actions[Kolf::PreviousHoleAction]->setEnabled(false);
		m_actions[Kolf::NextHoleAction]->setEnabled(false);
	}
	else
	{
		m_actions[Kolf::PreviousHoleAction]->setEnabled(m_currentHole != 0);
		m_actions[Kolf::NextHoleAction]->setEnabled(m_currentHole != m_holeCount - 1);
	}
	//populate "Choose hole" menu
	const KLocalizedString chEntryCaption = ki18nc("@item:inmenu", "Hole %1");
	KSelectAction* const chAct = qobject_cast<KSelectAction*>(m_actions[Kolf::ChooseHoleAction]);
	for (int i = chAct->actions().count() - 1; i >= m_holeCount; ++i)
		chAct->removeAction(chAct->actions().value(i)); //remove unneeded entries
	for (int i = chAct->actions().count(); i < m_holeCount; ++i)
		chAct->addAction(chEntryCaption.subs(i + 1).toString()); //+1 because user-visible numbering starts at 1
	//set check state in menu
	QList<QAction*> chActs = chAct->actions();
	for (int i = 0; i < m_holeCount; ++i)
		chActs[i]->setChecked(i == m_currentHole);
}

#include "interface.moc"
