/***************************************************************************
 *   Copyright 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "utils-transposeproxymodel.h"

Utils::TransposeProxyModel::TransposeProxyModel()
	: m_sourceModel(0)
{
}

void Utils::TransposeProxyModel::setSourceModel(QAbstractTableModel* sourceModel)
{
	if (sourceModel == m_sourceModel)
		return;
	if (m_sourceModel)
		disconnect(m_sourceModel, 0, this, 0);
	m_sourceModel = sourceModel;
	reset();
	//connect things properly
	if (sourceModel)
	{
		connect(sourceModel, SIGNAL(columnsAboutToBeInserted(const QModelIndex&, int, int)), this, SLOT(slotColumnsAboutToBeInserted(const QModelIndex&, int, int)));
		connect(sourceModel, SIGNAL(columnsAboutToBeRemoved(const QModelIndex&, int, int)), this, SLOT(slotColumnsAboutToBeRemoved(const QModelIndex&, int, int)));
		connect(sourceModel, SIGNAL(columnsInserted(const QModelIndex&, int, int)), this, SLOT(slotColumnsInserted(const QModelIndex&, int, int)));
		connect(sourceModel, SIGNAL(columnsRemoved(const QModelIndex&, int, int)), this, SLOT(slotColumnsRemoved(const QModelIndex&, int, int)));
		connect(sourceModel, SIGNAL(dataChanged(const QModelIndex&, const QModelIndex&)), this, SLOT(slotDataChanged(const QModelIndex&, const QModelIndex&)));
		connect(sourceModel, SIGNAL(headerDataChanged(Qt::Orientation, int, int)), this, SLOT(slotHeaderDataChanged(Qt::Orientation, int, int)));
		connect(sourceModel, SIGNAL(layoutAboutToBeChanged()), this, SIGNAL(layoutAboutToBeChanged()));
		connect(sourceModel, SIGNAL(layoutChanged()), this, SIGNAL(layoutChanged()));
		connect(sourceModel, SIGNAL(modelAboutToBeReset()), this, SIGNAL(modelAboutToBeReset()));
		connect(sourceModel, SIGNAL(modelReset()), this, SIGNAL(modelReset()));
		connect(sourceModel, SIGNAL(rowsAboutToBeInserted(const QModelIndex&, int, int)), this, SLOT(slotRowsAboutToBeInserted(const QModelIndex&, int, int)));
		connect(sourceModel, SIGNAL(rowsAboutToBeRemoved(const QModelIndex&, int, int)), this, SLOT(slotRowsAboutToBeRemoved(const QModelIndex&, int, int)));
		connect(sourceModel, SIGNAL(rowsInserted(const QModelIndex&, int, int)), this, SLOT(slotRowsInserted(const QModelIndex&, int, int)));
		connect(sourceModel, SIGNAL(rowsRemoved(const QModelIndex&, int, int)), this, SLOT(slotRowsRemoved(const QModelIndex&, int, int)));
	}
}

QAbstractTableModel* Utils::TransposeProxyModel::sourceModel() const
{
	return m_sourceModel;
}

QModelIndex Utils::TransposeProxyModel::mapFromSource(const QModelIndex& sourceIndex) const
{
	if (sourceIndex.model() != m_sourceModel)
		return QModelIndex();
	else
		return index(sourceIndex.column(), sourceIndex.row());
}

QModelIndex Utils::TransposeProxyModel::mapToSource(const QModelIndex& proxyIndex) const
{
	if (proxyIndex.model() != this)
		return QModelIndex();
	else if (!m_sourceModel)
		return QModelIndex();
	else
		return m_sourceModel->index(proxyIndex.column(), proxyIndex.row());
}

//BEGIN QAbstractTableModel reimplementations

int Utils::TransposeProxyModel::columnCount(const QModelIndex& parent) const
{
	if (!m_sourceModel)
		return 0;
	else
		return m_sourceModel->rowCount(mapToSource(parent));
}

int Utils::TransposeProxyModel::rowCount(const QModelIndex& parent) const
{
	if (!m_sourceModel)
		return 0;
	else
		return m_sourceModel->columnCount(mapToSource(parent));
}

Qt::ItemFlags Utils::TransposeProxyModel::flags(const QModelIndex& index) const
{
	if (!m_sourceModel)
		return Qt::NoItemFlags;
	else
		return m_sourceModel->flags(mapToSource(index));
}

QVariant Utils::TransposeProxyModel::data(const QModelIndex& index, int role) const
{
	if (!m_sourceModel)
		return QVariant();
	else
		return m_sourceModel->data(mapToSource(index), role);
}

QVariant Utils::TransposeProxyModel::headerData(int section, Qt::Orientation orientation, int role) const
{
	const Qt::Orientation transposedOrientation = (orientation == Qt::Horizontal) ? Qt::Vertical : Qt::Horizontal;
	if (!m_sourceModel)
		return QVariant();
	else
		return m_sourceModel->headerData(section, transposedOrientation, role);
}

bool Utils::TransposeProxyModel::setData(const QModelIndex& index, const QVariant& value, int role)
{
	if (!m_sourceModel)
		return false;
	else
		return m_sourceModel->setData(mapToSource(index), value, role);
}

bool Utils::TransposeProxyModel::setHeaderData(int section, Qt::Orientation orientation, const QVariant& value, int role)
{
	const Qt::Orientation transposedOrientation = (orientation == Qt::Horizontal) ? Qt::Vertical : Qt::Horizontal;
	if (!m_sourceModel)
		return false;
	else
		return m_sourceModel->setHeaderData(section, transposedOrientation, value, role);
}

//END QAbstractTableModel reimplementations

//BEGIN QAbstractTableModel signal forwarding

void Utils::TransposeProxyModel::slotColumnsAboutToBeInserted(const QModelIndex& parent, int start, int end)
{
	beginInsertRows(mapFromSource(parent), start, end);
}

void Utils::TransposeProxyModel::slotColumnsAboutToBeRemoved(const QModelIndex& parent, int start, int end)
{
	beginRemoveRows(mapFromSource(parent), start, end);
}

void Utils::TransposeProxyModel::slotColumnsInserted(const QModelIndex& parent, int start, int end)
{
	Q_UNUSED(parent)
	Q_UNUSED(start)
	Q_UNUSED(end)
	endInsertRows();
}

void Utils::TransposeProxyModel::slotColumnsRemoved(const QModelIndex& parent, int start, int end)
{
	Q_UNUSED(parent)
	Q_UNUSED(start)
	Q_UNUSED(end)
	endRemoveRows();
}

void Utils::TransposeProxyModel::slotDataChanged(const QModelIndex& topLeft, const QModelIndex& bottomRight)
{
	emit dataChanged(mapFromSource(topLeft), mapFromSource(bottomRight));
}

void Utils::TransposeProxyModel::slotHeaderDataChanged(Qt::Orientation orientation, int first, int last)
{
	const Qt::Orientation transposedOrientation = (orientation == Qt::Horizontal) ? Qt::Vertical : Qt::Horizontal;
	emit headerDataChanged(transposedOrientation, first, last);
}

void Utils::TransposeProxyModel::slotRowsAboutToBeInserted(const QModelIndex& parent, int start, int end)
{
	beginInsertColumns(mapFromSource(parent), start, end);
}

void Utils::TransposeProxyModel::slotRowsAboutToBeRemoved(const QModelIndex& parent, int start, int end)
{
	beginRemoveColumns(mapFromSource(parent), start, end);
}

void Utils::TransposeProxyModel::slotRowsInserted(const QModelIndex& parent, int start, int end)
{
	Q_UNUSED(parent)
	Q_UNUSED(start)
	Q_UNUSED(end)
	endInsertColumns();
}

void Utils::TransposeProxyModel::slotRowsRemoved(const QModelIndex& parent, int start, int end)
{
	Q_UNUSED(parent)
	Q_UNUSED(start)
	Q_UNUSED(end)
	endRemoveColumns();
}

//END QAbstractTableModel signal forwarding

#include "utils-transposeproxymodel.moc"
