/****************************************************************************
 * Copyright (C) 2007-2008 by Michal Sciubidlo <michal.sciubidlo@gmail.com> *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 2 of the License, or      *
 *   (at your option) any later version.                                    *
 ****************************************************************************/

#include <iostream>
#include <cstdlib>
#include <string>
#include <getopt.h>

#include <QString>
#include <QStringList>
#include <QDomDocument>
#include <QDomNode>
#include <QFile>
#include <QTextStream>
#include <QTextCodec>

#include <kstandarddirs.h>

using namespace std;

struct DawgNode
{
	char letter;
	bool isEndOfWord, written, visiting;
	DawgNode *children, *next, *prev, *redirect, *redirNext;
	int childNum, childDepth, offset;
};

const int WORD_END = 64;
const int NO_NEXT = 128;
const int MAX_WORD_SIZE = 15;

using namespace std;

void addWord(const QString& word, const int offset, DawgNode*& root, const QMap<QChar, char> &map)
{
	static int nodeCount = 0;
	struct DawgNode *prevNode, *currNode;
	prevNode = 0;

	//If this is a new section of the tree, create a new node
	if (root == 0)
	{
		root = new DawgNode();
		root->letter = map.value(word[offset-1]);
		root->isEndOfWord = false;
		root->children = 0;
		root->next = 0;
		root->prev = 0;
		root->offset = -1;
		root->written = false;
		root->redirect = 0;
		root->visiting = false;
		currNode = root;

		nodeCount++;
	}
	else
	{
		currNode = root;

		// Search down the current level of the tree and see if the current
		// letter in the word is already in the tree at this section.
		while (currNode != 0)
		{
			if (currNode->letter == map.value(word[offset-1]))
				break;

			//If this letter isn't in this part of the tree, add a new node for it.
			if (  map.value(word[offset-1]) < currNode->letter )
			{
				struct DawgNode * newNode = new DawgNode();
				newNode->letter = map.value(word[offset-1]);
				newNode->isEndOfWord = false;
				newNode->children = 0;
				newNode->next = currNode;
				newNode->offset = -1;
				newNode->written = false;
				newNode->redirect = 0;
				newNode->visiting = false;
	
				newNode->prev = prevNode;
	
				if ( prevNode == 0 )
					root = newNode;
				else
					prevNode->next = newNode;
				break;
			}
	
			prevNode = currNode;
			currNode = currNode->next;
		}

		// If we got through the search, the new node belongs at the end	
		if ( currNode == 0 )
		{
			currNode = new DawgNode();
			currNode->letter = map.value(word[offset-1]);
			currNode->isEndOfWord = false;
			currNode->children = 0;
			currNode->next = 0;
			currNode->prev = prevNode;
			currNode->offset = -1;
			currNode->written = false;
			currNode->redirect = 0;
			currNode->visiting = false;
	
			prevNode->next = currNode;
	
			nodeCount++;
		}
	}

	if ( (offset) == word.size() )
		currNode->isEndOfWord = true;
	else
		addWord(word, offset+1, currNode->children, map);
}

/*updateCounts computes the number of children and maximum depths for each tree node.*/

void updateCounts(DawgNode * const startNode, DawgNode ** const redirNodes, const int letterCount)
{
	static int visitedCount = 0;
	DawgNode *currNode;

	currNode = startNode;

	while ( currNode != 0 )
	{
		currNode->childNum = 0;
		currNode->childDepth = 0;
		visitedCount++;

		DawgNode * currChild = currNode->children;
		updateCounts(currChild, redirNodes, letterCount);
		while ( currChild != 0 )
		{
			currNode->childNum = currNode->childNum + 1 + currChild->childNum;

			if ( currChild->childDepth > currNode->childDepth )
				currNode->childDepth = currChild->childDepth;

			currChild = currChild->next;
		}
		currNode->childDepth += 1;

		currNode->redirNext = redirNodes[currNode->childDepth*letterCount+currNode->letter];
		redirNodes[currNode->childDepth*letterCount+currNode->letter] = currNode;

		currNode = currNode->next;
	}
}

/* NodesAreEquivalend returns true if two nodes are equivalent in depth,
  number of children, and node values. */

bool isSameNode(const DawgNode * const first, const DawgNode * const second)
{
	if ( first == second )
		return true;

	if ( (first == 0) || (second == 0) )
		return false;

	if ( (first->letter != second->letter) || (first->isEndOfWord != second->isEndOfWord) ||
	(first->childNum != second->childNum) || (first->childDepth != second->childDepth) )
		return false;

	if ( !isSameNode(first->children, second->children) )
		return false;

	if ( !isSameNode(first->next, second->next) )
		return false;

	return true;
}

/* combineNodes loops through all the nodes and searches for identical ones.
  When it finds an equivalent node, it sets the redirect pointer in the
  other node. The redirect keep the program from trying to change all
  references to the redirected node. When the nodes are written to a file,
  a node with a redirect is not written. */

void combineNodes(DawgNode ** const redirNodes, const int letterCount)
{
	DawgNode *currNode, *compareNode;

	for (int i = 0; i <= MAX_WORD_SIZE; i++)
	for (int j = 0; j < letterCount; j++)
	{
		currNode = redirNodes[i*letterCount+j];

		while ( currNode != 0 )
		{
			if ( currNode->redirect != 0 )
			{
				currNode = currNode->redirNext;
				continue;
			}

			compareNode = currNode->redirNext;

			while ( compareNode != 0 )
			{
				if ( compareNode->redirect != 0 )
				{
					compareNode = compareNode->redirNext;
					continue;
				}

				if ( isSameNode(currNode, compareNode) )
				compareNode->redirect = currNode;

				compareNode = compareNode->redirNext;
			}

			currNode = currNode->redirNext;
		}
	}
}

/* getOffsets figures out where in the tree file each node will
  be written. The computations need to be done before nodes can be
  written because some nodes may contain a forward reference. */

void getOffsets(DawgNode * const node, const int depth)
{
	static int visitedCount = 0;
	static int offset = 1;
	DawgNode *currNode, *startNode;

	if ( node == 0 )
		return;

	if ( node->offset >= 0 )
		return;

	visitedCount++;
	startNode = node;

	while ( startNode->prev != 0 )
		startNode = startNode->prev;

	if ( depth > 0 )
	{
		currNode = startNode;

		while ( currNode != 0 )
		{
			currNode->offset = offset;
			offset++;
			currNode = currNode->next;
		}
	}

	if ( depth == 0 )
	{
		currNode = startNode;

		while ( currNode != 0 )
		{
			currNode->offset = offset;
			offset++;
			currNode = currNode->next;
		}
	}

	currNode = startNode;

	while ( currNode != 0 )
	{
		if ( (currNode->children != 0) && (currNode->children->redirect != 0) )
			getOffsets(currNode->children->redirect, depth+1);
		else
			getOffsets(currNode->children, depth+1);

		currNode = currNode->next;
	}
}

void writeNodeEntry(const DawgNode * const node, QFile &file)
{
	int temp;
	char buf[4];

	if ( node->children != 0 )
	{
		if ( node->children->redirect != 0 )
			temp = node->children->redirect->offset << 8;
		else
			temp = node->children->offset << 8;
	}
	else
		temp = 0;

	temp += node->letter;

	if ( node->isEndOfWord )
		temp = temp + WORD_END;

	if ( node->next == 0 )
		temp = temp + NO_NEXT;

	buf[0] = temp >> 24;
	buf[1] = temp >> 16;
	buf[2] = temp >> 8;
	buf[3] = temp ;
	file.write(buf, 4);
}

// writeNode traverses the tree and writeNodeEntry each unique node

void writeNode(DawgNode * const node, const int depth, QFile &file)
{
	static int visitedCount = 0;
	DawgNode *currNode, *startNode;

	if (( node == 0) || ( node->written ))
		return;

	visitedCount++;
	startNode = node;

	while ( startNode->prev != 0 )
		startNode = startNode->prev;
	
	if ( depth > 0 )
	{
		currNode = startNode;
	
		while ( currNode != 0 )
		{
			writeNodeEntry(currNode, file);
			currNode->written = true;
			currNode = currNode->next;
		}
	}

	if ( depth == 0 )
	{
		currNode = startNode;
	
		while ( currNode != 0 )
		{
			writeNodeEntry(currNode, file);
			currNode->written = true;
			currNode = currNode->next;
		}
	}

	currNode = startNode;
	
	while ( currNode != 0 )
	{
		if ( (currNode->children != 0) && (currNode->children->redirect != 0) )
			writeNode(currNode->children->redirect, depth+1, file);
		else
			writeNode(currNode->children, depth+1, file);

		currNode = currNode->next;
	}
}

void readLetters(const QString &languageFileName, QMap<QChar, char> &map, int &letterCount)
{
	QFile letters(languageFileName);
	if (!letters.open(QIODevice::ReadOnly))
	{
		cout << "Can't read language file" << endl;
		exit(EXIT_SUCCESS);
	}
	QDomDocument doc("tileset");
	if (!doc.setContent(&letters))
	{
		cout << "language file is corrupted" << endl;
		exit(EXIT_SUCCESS);
	}
	letters.close();

	bool found, atEnd;
	QDomNode n = doc.documentElement().firstChild();
	found = false;
	atEnd = false;
	// find the tiles child
	while (!found && !atEnd)
	{
		if (n.isNull()) atEnd = true;
		else if (!n.toElement().isNull() && n.toElement().tagName() == "tiles") found = true;
		else n = n.nextSibling();
	}
	if (found)
		n = n.firstChild();
	else
	{
		cout << "tere is no ""tiles"" section in tiles file" << endl;
		exit(EXIT_SUCCESS);
	}

	letterCount = 0;

	while(!n.isNull())
	{
		QDomElement e = n.toElement();
		if(!e.isNull() && e.tagName() == "tile")
		{
			QChar s;
			int q = 0;
			uint score = 0;
			bool hadletter = false, hadQuantity = false, hadScore = false;
			
			QDomNode n2 = e.firstChild();
			while(!n2.isNull())
			{
				QDomElement e2 = n2.toElement();
				if (e2.tagName() == "letter")
				{
					if (e2.text().isEmpty()) //removes blanks
					{
						s = e2.text().at(0).toLower(); //TODO: for now letteer = one char ("ca" won't work)
						hadletter = true;
					}
				}
				else if (e2.tagName() == "quantity")
				{
					q = e2.text().toInt();
					hadQuantity = true;
				}
				else if (e2.tagName() == "score")
				{
					score = e2.text().toInt();
					hadScore = true;
				}
				
				n2 = n2.nextSibling();
			}
			if (hadletter && hadQuantity && hadScore)
			{
				map.insert(s, (char)letterCount);
				letterCount++;
			}
		}
		n = n.nextSibling();
	}
}

void readWords(const QString &wordsFileName, const QMap<QChar, char> &map, DawgNode *& mainRoot, const QString& codec)
{
	QFile file(wordsFileName);
	if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
	{
		cout << "can't open words list file" << endl;
		exit(EXIT_SUCCESS);
	}

	int wordCount = 0;
	int toLong = 0;
	int badChar = 0;

	QTextStream in(&file);
	QString line;

	if (codec.isEmpty() == false)
		in.setCodec(codec.toLatin1().data());

	while (!in.atEnd())
	{
		line = in.readLine();

		if (line.size() > MAX_WORD_SIZE)
		{
			toLong++;
			continue;
		}

		for (int i = 0; i < line.size(); i++)
		{
			if (map.contains(line[i]) == false)
			{
				badChar++;
				continue;
			}
		}

		addWord(line, 1, mainRoot, map);
		wordCount++;
	}
	file.close();
	cerr << "Words count: " << wordCount << endl;
	if (toLong)
		cerr << "Words ignored (length > 15): " << toLong << endl;
	if (badChar)
		cerr << "Words with chars not present in dictionary: " << badChar << endl;
}

void saveDawg(QString &outputFileName, DawgNode * mainRoot)
{
	QFile outFile(outputFileName);
	if (!outFile.open(QIODevice::Truncate | QIODevice::WriteOnly))
	{
		cout << "can't open output file" << endl;
		exit(EXIT_SUCCESS);
	}

	unsigned char b[4];
	b[3] = WORD_END | NO_NEXT;
	b[2] = 1;
	b[1] = b[0] = 0;
	outFile.write((const char*)b, 4);

	getOffsets(mainRoot, 0);
	writeNode(mainRoot, 0, outFile);
	outFile.close();
}

void readCommandLine(int argc, char *argv[], QString& wordsFileName, QString& languageFileName, QString& outputFileName, QString& codec)
{
	static struct option long_options[] = {
		{"wordslist", 1, 0, 'w'},
		{"xml-language", 1, 0, 'x'},
		{"language", 1, 0, 'l'},
		{"outfile", 1, 0, 'o'},
		{"show-languages", 0, 0, 's'},
		{"show-codecs", 0, 0, 'd'},
		{"codec", 1, 0, 'c'},
		{"help", 0, 0, 'h'},
		{0, 0, 0, 0}
	};

	QString language;
	bool xmlFileName = false;
	KStandardDirs kdirs;

	while (1)
	{
		int option_index = 0;

		int c = getopt_long(argc, argv, "w:x:l:o:c:dhs", long_options, &option_index);
		if (c == -1)
			break;

		QStringList::const_iterator iter;
		QStringList list;
		QList<QByteArray> codecList;

		switch (c) 
		{
		case 'w':
			wordsFileName = optarg;
		break;

		case 'x':
			xmlFileName = true;
			languageFileName = optarg;
		break;

		case 'l':
			if (xmlFileName) 
				break;
			language = optarg;
		break;

		case 'o':
			outputFileName = optarg;
		break;

		case 's':
			list = kdirs.findAllResources("data", "kombination/tiles/*");
			cout << "Available languages:" << endl;
			
			for (iter = list.constBegin(); iter != list.constEnd(); ++iter)
				cout << (*iter).toLocal8Bit().constData() << endl;
			exit(EXIT_SUCCESS);
		break;

		case 'c':
			codec = optarg;
		break;

		case 'd':
			codecList = QTextCodec::availableCodecs();
			for (int i = 0; i < codecList.size(); ++i)
			{
				cout << codecList.at(i).data() << endl;
         		}
			exit(EXIT_SUCCESS);
		break;

		case 'h':
		default:
			cout << "usage: " << argv[0] << " [options]" << endl;
			cout << "Option               Usage" << endl;
			cout << "-s --show-languages  Shows installed languages." << endl;
			cout << "-w --wordslist       Path to words list." << endl;
			cout << "-l --language        Language code. -x overwrites it." << endl;
			cout << "-x --xmllanguag      Path to xml file with language description." << endl;
			cout << "-o --outfile         File to save generated dictionary." << endl;
			cout << "-c --codec           worlds list encoding." << endl;
			cout << "-d --show-codecs     gives list of available codes." << endl;
			cout << "-h --help            Shows this help." << endl;
			exit(EXIT_SUCCESS);
		break;
               }
	}

	if (xmlFileName == false)
	{
		static QString baseDictPath("kombination/tiles/%1.xml");
		languageFileName = kdirs.findResource("data", baseDictPath.arg(language));
	}
}

int main(int argc, char *argv[])
{
	QString wordsFileName;
	QString languageFileName;
	QString outputFileName;
	QString codec;
	QMap<QChar, char> map;
	DawgNode * mainRoot = 0;
	DawgNode ** redirNodes;
	int letterCount;

	readCommandLine(argc, argv, wordsFileName, languageFileName, outputFileName, codec);
	readLetters(languageFileName, map, letterCount);

	redirNodes = new DawgNode*[letterCount*(MAX_WORD_SIZE+1)];
	for (int i = 0; i <= MAX_WORD_SIZE; i++)
	for (int j = 0; j < letterCount; j++)
	{
		redirNodes[letterCount*i+j] = 0;
	}

	cerr << "Reading words..." << endl;
	readWords(wordsFileName, map, mainRoot, codec);
	cerr << "Updating words data..." << endl;
	updateCounts(mainRoot, redirNodes, letterCount);
	cerr << "Creating dictionary (can tak few minutes)..." << endl;
	combineNodes(redirNodes, letterCount);
	saveDawg(outputFileName, mainRoot);
	cerr << "Done" << endl;
	return 0;
}
