/***************************************************************************
 *   Copyright (C) 2007 by Michal Sciubidlo <michal.sciubidlo@gmail.com>   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "dictionary.h"
#include "tile.h"
#include "tilefactory.h"

#include <QFile>
#include <QDir>
#include <QByteArray>
#include <sonnet/speller.h>
#include <kstandarddirs.h>

#include <kdebug.h>

static QString baseDictPath("kombination/dictionaries/%1");

dictionary::dictionary(const QString &language) : m_language(language), isUsingAspell(false), kSpellDict(0)
{
	//TODO: KHotNewStuff2 if there is no file
	QString path = KStandardDirs::locate("data", baseDictPath.arg(m_language));
	if (path.isEmpty())
	{
		isUsingAspell = true;
		kSpellDict = new Sonnet::Speller(language);

		kDebug() << "using OLD dictionary " <<language << endl;
		return;
	}
	kDebug() << "using NEW dictionary "<<language << endl;

	QFile file(path);
	if (file.open(QIODevice::ReadOnly))
	{
		letters = new letter[ file.size()/4 ];
		uchar buf[1024];
		int mainIndex = 0;
		
		while (!file.atEnd()) {
        		qint64 lineLength = file.read((char*)buf, sizeof(buf));
        		for(int i = 0; i < lineLength; i += 4, mainIndex++)
        		{
				letters[mainIndex].nextAddress = (buf[i]<<16) + (buf[i+1]<<8) + buf[i+2];
				letters[mainIndex].letter = buf[i+3] & 63; //6 bits
				letters[mainIndex].isWordEnd = buf[i+3] & 64;
				letters[mainIndex].isSequenceEnd = buf[i+3] & 128;
        		}
        	}		
		file.close();

		//creating mapping from unicode characters to codes used in dictionary
		//code in dictionary is easy: first letter in language xml = 0, second = 1 etc
		//TODO: fileFactory is also created with new game in server.cpp...
		tileFactory tf;
		tf.readTileSet(m_language);
		//m_rh = tf.getReplacementsHandler();
		//m_exclusiveTileList = tf.exclusiveTileList();

		QList< QPair<QString, int> > tiles = tf.tileStringList();
		int moveBack = 0;
		for(int i=0; i<tiles.size(); i++)
		{
			//skip blank
			if (tiles.at(i).second == 0)
			{
				moveBack++;
				continue;
			}

			coddingMap.insert(tiles.at(i).first, i-moveBack);
			decoddingMap.insert(i-moveBack, tiles.at(i).first);
		}
	}
}

dictionary::~dictionary()
{
	if (kSpellDict)
		delete kSpellDict;
	if (letters)
		delete [] letters;
}

bool dictionary::isCorrect(const QString &word)
{
	if (isUsingAspell)
		return kSpellDict->isCorrect(word);
	
	//kDebug() << "checking using new dictionary" << endl;

	int address = 1;
	qint8 com_letter = letters[address].letter;
	int i = 0;
	//kDebug() << "starting letter code: " << com_letter << endl;
	//kDebug() << "next address: " << letters[address].nextAddress << endl;
	//kDebug() << "word size = " << word.size() << endl;
	while(i < word.size())
	{
		//kDebug() << "i = " << i << endl;
		//kDebug() << "letter: " << word[i] << " code: " << coddingMap.value(word[i]) << endl;
		if ((i != 0) && (address == 0)) return false; //stop dictionary overflow
		if (coddingMap.value(word[i]) == com_letter)
		{
			//kDebug() << "found match" << endl;
			i++;
			if (i < word.size())
			{
				address = letters[address].nextAddress;
				//kDebug() << "next address: " << address << endl;
				com_letter = letters[address].letter;
				//kDebug() << "next letter: " << com_letter << endl;
			}
			else
				return letters[address].isWordEnd;
		}
		else
		{
			//kDebug() << "didn't found match" << address << endl;
			//kDebug() << "but next address: " << letters[address].nextAddress << endl;
			if (letters[address].isSequenceEnd) 
			{
				return false;
			}
			address++;
			//kDebug() << "new address: " <<  address << endl;
			com_letter = letters[address].letter;
			//kDebug() << "new letter: " <<  com_letter << endl;
		}
	}

	return letters[address].isWordEnd;
}

QString& dictionary::language()
{
	return m_language;
}

QStringList dictionary::languages()
{
	QString path = KStandardDirs::locate("data", baseDictPath.arg(""));
	return QDir(path).entryList(QDir::Files|QDir::NoSymLinks|QDir::NoDotAndDotDot|QDir::Readable);
}

QString dictionary::decodeLetter(qint8 code)
{
	return decoddingMap.value(code);
}

qint8 dictionary::letterCode(const QString &letter)
{
	return coddingMap.value(letter);
}

letter* dictionary::getInnerTable()
{
	return letters;
}
