/***************************************************************************
 *   Copyright (C) 2005-2007 by Albert Astals Cid <aacid@kde.org>          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "game.h"
#include "player.h"
#include "aiplayer.h"
#include "protocol.h"
#include "tilefactory.h"
#include "dictionary.h"
#include <kdebug.h>
#include <math.h>

#include <QTimer>

template<class x>
void addPlayer(const QList<x> &list, const spectator *s, bool isPlayer)
{
	foreach(x it, list)
	{
		aiPlayer *pp = dynamic_cast<aiPlayer *>(it);
		if (pp)
		{
			if (pp != s) pp->addPlayer(s->nick(), isPlayer);
		}
		else
		{
			if (it != s) it->addPlayer(s->nick(), isPlayer);
		}
	}
}

template<class x>
void sendStatus(const QList<x> &list, const player *p)
{
	foreach(x it, list)
	{
		aiPlayer *pp = dynamic_cast<aiPlayer *>(it);
		if (pp)
		{
			if (pp != p) pp->sendStatus(p->nick(), p->ready() ? STATUS_PLAYER_READY : STATUS_PLAYER_NOT_READY);
		}
		else
		{
			if (it != p) it->sendStatus(p->nick(), p->ready() ? STATUS_PLAYER_READY : STATUS_PLAYER_NOT_READY);
		}
	}
}

template<class x>
void nextTurn(const QList<x> &list, const QString &nick)
{
	foreach(x it, list)
	{
		aiPlayer *pp = dynamic_cast<aiPlayer *>(it);
		if (pp)
			pp->nextTurn(nick);
		else
			it->nextTurn(nick);
	}
}

template<class x>
void sendTiles(const QList<x> &list, const player *p, const tilePositionList &tiles, bool wordsOk, int score)
{
	foreach(x it, list)
	{
		aiPlayer *pp = dynamic_cast<aiPlayer *>(it);
		if (pp)
		{
			/*if (pp != p) */pp->sendTiles(tiles);
			if (wordsOk) pp->correctWord(score);
			else pp->unknownWords();
		}
		else
		{
			if (it != p) it->sendTiles(tiles);
			if (wordsOk) it->correctWord(score);
			else it->unknownWords();
		}
	}
}

template<class x>
void gameEnded(const QList<x> &list, const QString &nickFinisher, const QString &nickWinner)
{
	foreach (x it, list)
	{
		aiPlayer *pp = dynamic_cast<aiPlayer *>(it);
		if (pp)
			pp->gameEnded(nickFinisher, nickWinner);
		else
			it->gameEnded(nickFinisher, nickWinner);
	}
}

template<class x>
void passedTurn(const QList<x> &list, const QString &nick)
{
	foreach(x it, list)
	{
		aiPlayer *pp = dynamic_cast<aiPlayer *>(it);
		if (pp)
			pp->passedTurn(nick);
		else
			it->passedTurn(nick);
	}
}

template<class x>
void exchangedTiles(const QList<x> &list, const QString &nick, int nTiles)
{
	foreach(x it, list)
	{
		aiPlayer *pp = dynamic_cast<aiPlayer *>(it);
		if (pp)
			pp->exchangedTiles(nick, nTiles);
		else
			it->exchangedTiles(nick, nTiles);
	}
}

game::game(int id, const QString &name, int maxPlayers, const QString &comment, dictionary *dict, bool allowSpectators) : m_id(id), m_name(name), m_maxPlayers(maxPlayers), m_comment(comment), m_dict(dict), m_allowSpectators(allowSpectators), m_started(false)
{
	//m_dict->setLanguage(languageCode);
	tileFactory tf;
	tf.readTileSet(m_dict->language());	
	m_board.setTileStringList(tf.tileStringList());
	m_board.setDictionary(dict);

	m_rh = tf.getReplacementsHandler();
	m_exclusiveTileList = tf.exclusiveTileList();
}

game::~game()
{
	// TODO
}

void game::addAi(const QString &name, const QString &languageCode)
{
	if (numberOfPlayers() >= maxNumberOfPlayers())
		return;
	
	tileFactory tf;
	tf.readTileSet(m_dict->language());
	board * aiBoard = new board();
	aiBoard->setTileStringList(tf.tileStringList());
	aiBoard->setDictionary(m_dict);
	tileBox *aiTileBox = new tileBox();
	aiTileBox->newGame(tf.tiles());

	aiPlayer *p = new aiPlayer(name, aiBoard, aiTileBox);
	connect(p, SIGNAL(readyChanged(player *)), this, SLOT(checkCanStartGame(player *)));
	connect(p, SIGNAL(gotTiles(player *, const tilePositionList &)), this, SLOT(gotTiles(player *, const tilePositionList &)));
	connect(p, SIGNAL(votingChanged()), this, SLOT(checkVotingResult()));
	connect(p, SIGNAL(passTurn(player*)), this, SLOT(passTurn(player*)));
	connect(p, SIGNAL(exchangeTiles(player*, const QList<tile*> &)), this, SLOT(exchangeTiles(player*, const QList<tile*> &)));
	m_players.append(p);
	p->setCurrentGame(this);

	//notify the player all the players already in
	QList< QPair<QString, int> > players;
	foreach(const player *pAux, m_players)
	{
		if (p != pAux)
		{
			players << QPair<QString, int>(pAux->nick(), pAux->ready() ? STATUS_PLAYER_READY : STATUS_PLAYER_NOT_READY);
		}
	}
	foreach(const spectator *sAux, m_spectators)
	{
		players << QPair<QString, int>(sAux->nick(), STATUS_SPECTATOR);
	}
	p->acceptedInGame(players, m_dict->language());

	// notify all the players we have a new player
	::addPlayer(m_players, (spectator *)p, true);
	::addPlayer(m_spectators, (spectator *)p, true);
	p->tellIfReady();
}

spectator *game::addPlayer(spectator *s, bool joinAsSpectator)
{
	if (joinAsSpectator)
	{
		if (allowsSpectators())
		{
			m_spectators.append(s);
			addPlayerHelper(s);
			return s;
		}
		else
		{
			// TODO CHEATER!
			return 0;
		}
	}
	else
	{
		if (numberOfPlayers() < maxNumberOfPlayers())
		{
			player *p = new player(s);
			connect(p, SIGNAL(readyChanged(player *)), this, SLOT(checkCanStartGame(player *)));
			connect(p, SIGNAL(gotTiles(player *, const tilePositionList &)), this, SLOT(gotTiles(player *, const tilePositionList &)));
			connect(p, SIGNAL(votingChanged()), this, SLOT(checkVotingResult()));
			connect(p, SIGNAL(passTurn(player*)), this, SLOT(passTurn(player*)));
			connect(p, SIGNAL(exchangeTiles(player*, const QList<tile*> &)), this, SLOT(exchangeTiles(player*, const QList<tile*> &)));
			m_players.prepend(p);
			addPlayerHelper(p);
			return p;
		}
		else
		{
			s->tooManyPlayers();
			return 0;
		}
	}
}

void game::addPlayerHelper(spectator *s)
{
	s->setCurrentGame(this);
	player *p = dynamic_cast<player *>(s);

	//notify the player all the players already in
	QList< QPair<QString, int> > players;
	foreach(const player *pAux, m_players)
	{
		if (p != pAux)
		{
			players << QPair<QString, int>(pAux->nick(), pAux->ready() ? STATUS_PLAYER_READY : STATUS_PLAYER_NOT_READY);
		}
	}
	foreach(const spectator *sAux, m_spectators)
	{
		if (s != sAux)
		{
			players << QPair<QString, int>(sAux->nick(), STATUS_SPECTATOR);
		}
	}
	s->acceptedInGame(players, m_dict->language());

	// notify all the players we have a new player
	::addPlayer(m_players, s, p);
	::addPlayer(m_spectators, s, p);
	if (m_started || !p)
	{
		s->startGame();
		s->nextTurn(m_players[m_currentPlayer]->nick());
	}
}

int game::id() const
{
	return m_id;
}

QString game::name() const
{
	return m_name;
}

int game::numberOfPlayers() const
{
	return m_players.count();
}

int game::numberOfSpectators() const
{
	return m_spectators.count();
}

int game::maxNumberOfPlayers() const
{
	return m_maxPlayers;
}

bool game::hasStarted() const
{
	return m_started;
}

bool game::allowsSpectators() const
{
	return m_allowSpectators;
}

QString game::comment() const
{
	return m_comment;
}

QString game::language() const
{
	return m_dict->language();
}

QMap<QString, QString> game::toMap(const QString &tag) const
{
	QMap<QString, QString> result;
	result[tag + "id"] = QString::number(m_id);
	result[tag + "name"] = m_name;
	result[tag + "numberOfPlayers"] = QString::number(numberOfPlayers());
	result[tag + "numberOfSpectators"] = QString::number(numberOfSpectators());
	result[tag + "maxPlayers"] = QString::number(m_maxPlayers);
	result[tag + "started"] = QString::number(m_started);
	result[tag + "allowSpectators"] = QString::number(m_allowSpectators);
	result[tag + "comment"] = m_comment;
	result[tag + "language"] = language();
	return result;
}

void game::checkCanStartGame(player *p)
{
	int players = m_players.count();
	if (m_maxPlayers != players) return;
	bool allReady = true;
	for (int i = 0; i < players; i++)
	{
		allReady = allReady && m_players[i]->ready();
	}

	if (allReady)
	{
		// all the clients are ready, start the game
		m_board.reset();
		
		tileFactory tf;
		tf.readTileSet(m_dict->language());
		m_tileBox.newGame(tf.tiles());
		QList<tile*> tiles;
		for (int i = 0; i < players; i++)
		{
			tiles = m_tileBox.getTiles(7);
			aiPlayer *pp = dynamic_cast<aiPlayer *>(m_players[i]);
			if (pp)
				pp->startGame(tiles);
			else
				m_players[i]->startGame(tiles);
		}
		for (int i = 0; i < m_spectators.count(); i++)
			m_spectators[i]->startGame();

		m_started = true;

		// general rule says the player with the "latest" letter is the first
		const QList< QPair<QString, int> > &tileStrings = tf.tileStringList();
		QList< QPair<QString, int> >::const_iterator firstTile = tileStrings.begin();
		// This relies on the XML files beign correctly ordered, make this more prominent
		// somewhere, empty tile should be at the beginning as it beats all tiles
		bool found = false;
		while (!found)
		{
			for (int i = 0; i < players; i++)
			{
				if (m_players[i]->hasTile((*firstTile).first))
				{
					found = true;
					m_currentPlayer = i;
				}
			}
			if (!found) ++firstTile;
		}
		
		// we substract one because nextTurn is going to increase it by one
		m_currentPlayer -= 1;
		nextTurn();
	}
	else
	{
		// some clients still are not ready, sending the changed
		// status for the player *p
		::sendStatus(m_players, p);
		::sendStatus(m_spectators, p);
	}
}

void game::gotTiles(player *p, const tilePositionList &tiles)
{
	// CHECK p is m_players[m_currentPlayer] if not CHEATER !!!
	kDebug() << "tiles count: "<< tiles.count() << endl;
	int error = 0;
	int score = 0;
	if (p->freeTiles() == 0) score += 50;
	m_board.setPlayedTiles(tiles);
	const wordsAndTilePositions &wordsAndTiles = m_board.checkValidTilesPositions(score, error);
	if (!error)
	{
		m_incorrectWordAskingTimeouted = false;
		m_scoreLastWord = score;
		
		bool wordOk, wordsOk, usesExclusiveTiles;
		wordsAndTilePositions::const_iterator wordsIt, wordsItEnd;
		wordsIt = wordsAndTiles.begin();
		wordsItEnd = wordsAndTiles.end();
		wordsOk = true;
		usesExclusiveTiles = false;
		for ( ; wordsOk && wordsIt != wordsItEnd; ++wordsIt)
		{
			const QString &word = (*wordsIt).first;
			QStringList::const_iterator possibleWordsIt, possibleWordsItEnd;
			QStringList possibleWords = m_rh.possibleWords(word);
			possibleWordsIt = possibleWords.constBegin();
			possibleWordsItEnd = possibleWords.constEnd();
			wordOk = false;
			while(!wordOk && possibleWordsIt != possibleWordsItEnd)
			{
				wordOk = m_dict->isCorrect(*possibleWordsIt);
				++possibleWordsIt;
			}
			
			if (wordOk)
			{
				QStringList::const_iterator exclusiveTilesIt, exclusiveTilesItEnd;
				exclusiveTilesIt = m_exclusiveTileList.constBegin();
				exclusiveTilesItEnd = m_exclusiveTileList.constEnd();
				for ( ; wordOk && exclusiveTilesIt != exclusiveTilesItEnd; ++exclusiveTilesIt)
				{
					int count = word.count(*exclusiveTilesIt);
					if (count)
					{
						foreach(const tilePosition &tp, (*wordsIt).second)
							if (tp.ourTileC()->string() == *exclusiveTilesIt) count--;
					}
					if (count)
					{
						usesExclusiveTiles = true;
						wordsOk = false;
					}
				}
			}
			else wordsOk = false;
		}
		
		if (!usesExclusiveTiles)
		{
			if (wordsOk) m_players[m_currentPlayer]->addToScore(score);
			::sendTiles(m_players, p, tiles, wordsOk, score);
			::sendTiles(m_spectators, p, tiles, wordsOk, score);
		}
		else
		{
			for (int i = 0; i < m_players.count(); i++)
			{
				aiPlayer *pp = dynamic_cast<aiPlayer *>(m_players[i]);
				if (pp)
					pp->usesExclusiveTiles(p->nick());
				else
					m_players[i]->usesExclusiveTiles(p->nick());
			}
		}
		
		if (wordsOk)
		{
			kDebug() << "game::gotTiles(), words are OK !!!" << endl;
			int num = p->freeTiles();
			QList<tile*> newTiles = m_tileBox.getTiles(num);
			if (newTiles.count() == 0 && num == 7) gameEnded(p);
			else
			{
				aiPlayer *pp = dynamic_cast<aiPlayer *>(p);
				if (pp)
				{
					pp->sendNewTiles(newTiles);
				}
				else
				{
					p->sendNewTiles(newTiles);
				}
				kDebug() << "game::gotTiles()" << endl;
				m_board.finalizePlayedTiles(ACCEPT);
				nextTurn();
//				m_board.dumpBoard();
			}
		}
		else if (usesExclusiveTiles)
		{
			nextTurn();
			m_board.finalizePlayedTiles(DELETE);
		}
		else
		{
			// the word is incorrect
			// TODO that may change on the client
			QTimer::singleShot(15000 + 3000, this, SLOT(incorrectWordTimeout()));
		}
		
	}
	// else cheater kill him!!!
}


void game::nextTurn()
{
	int players = m_players.count();
	m_currentPlayer = (m_currentPlayer + 1) % players;
	QString nick = m_players[m_currentPlayer]->nick();
	::nextTurn(m_players, nick);
	::nextTurn(m_spectators, nick);
}

void game::gameEnded(player *finisher)
{
	// TODO Remove to each player the score amounting the tiles he has at the deck at end
	// TODO Add to finished the sum of the above scores
	// TODO In case after that there's a tie between to players wins the one that had more scores before the above calculation
	// TODO if still tied no idea :-D
	m_started = false;
	player *winner;
	winner = m_players[0];
	foreach (player *p, m_players)
	{
		if (p->score() > winner->score()) winner = p;
	}
	::gameEnded(m_players, finisher->nick(), winner->nick());
	::gameEnded(m_spectators, finisher->nick(), winner->nick());

	m_started = false;
}

void game::checkVotingResult()
{
	int clients = m_players.count();
	int votesAccept = 0;
	int votesReject = 0;
	for (int i = 0; i < clients; i++)
	{
		switch(m_players[i]->votingState())
		{
			case player::VOTING_UNKNOWN:
			break;
			case player::VOTING_ACCEPTED:
				votesAccept++;
			break;
			case player::VOTING_REJECTED:
				votesReject++;
			break;
		}
	}
	
	// don't count the player that started the voting
	if (votesAccept + votesReject == clients - 1 || m_incorrectWordAskingTimeouted)
	{
		// the actual number of good votes has to be at least:
		// good > 50% + 1
		if (votesAccept >= floor( (clients - 1) / 2.0) + 1)
		{
			m_players[m_currentPlayer]->addToScore(m_scoreLastWord);
			// most of the players voted ok for it
			for (int i = 0; i < clients; i++)
			{
				aiPlayer *pp = dynamic_cast<aiPlayer *>(m_players[i]);
				if (pp)
					pp->correctWord(m_scoreLastWord);
				else
					m_players[i]->correctWord(m_scoreLastWord);
				m_players[i]->prepareForNewVoting();
			}
			for (int i = 0; i < m_spectators.count(); i++)
			{
				m_spectators[i]->correctWord(m_scoreLastWord);
			}
			// TODO see how to merge this with the gotTiles part that does the same :-D
			// basically find a name for the suitable function
			int num = m_players[m_currentPlayer]->freeTiles();
			QList<tile*> newTiles = m_tileBox.getTiles(num);
			if (newTiles.count() == 0 && num == 7) gameEnded(m_players[m_currentPlayer]);
			else
			{
				aiPlayer *pp = dynamic_cast<aiPlayer *>(m_players[m_currentPlayer]);
				if (pp)
					pp->sendNewTiles(newTiles);
				else
					m_players[m_currentPlayer]->sendNewTiles(newTiles);
				nextTurn();
				m_board.finalizePlayedTiles(ACCEPT);
			}
		}
		else
		{
			// mostly rejected
			for (int i = 0; i < clients; i++)
			{
				aiPlayer *pp = dynamic_cast<aiPlayer *>(m_players[i]);
				if (pp)
					pp->rejectVotedTiles(m_players[m_currentPlayer]->nick());
				else
					m_players[i]->rejectVotedTiles(m_players[m_currentPlayer]->nick());
				m_players[i]->prepareForNewVoting();
			}
			for (int i = 0; i < m_spectators.count(); i++)
			{
				m_spectators[i]->rejectVotedTiles(m_players[m_currentPlayer]->nick());
			}
			nextTurn();
			m_board.finalizePlayedTiles(DELETE);
		}
		
	}
}

void game::incorrectWordTimeout()
{
	m_incorrectWordAskingTimeouted = true;
	//TODO
//	checkVotingResult();
}

void game::passTurn(player *p)
{
	if (p == m_players.at(m_currentPlayer))
	{
		::passedTurn(m_players, p->nick());
		::passedTurn(m_spectators, p->nick());
		nextTurn();
	}
	// else cheater kill him !!!
}

void game::exchangeTiles(player *p, const QList<tile*> &tiles)
{
	if (p == m_players.at(m_currentPlayer))
	{
		int nTiles = tiles.count();
		QList<tile*> newTiles = m_tileBox.getTiles(nTiles);
		int nNewTiles = newTiles.count();
		m_tileBox.addTiles(tiles);
		if (nTiles != nNewTiles)
			newTiles += m_tileBox.getTiles(nTiles - nNewTiles);
		p->sendExchangedTiles(newTiles);
		::exchangedTiles(m_players, p->nick(), nTiles);
		::exchangedTiles(m_spectators, p->nick(), nTiles);
		nextTurn();
	}
	// else cheater kill him !!!
}

#include "game.moc"
