/***************************************************************************
 *   Copyright (C) 2005 by Albert Astals Cid <tsdgeos@terra.es>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "tilefactory.h"
#include "tile.h"

#include <QDomDocument>
#include <QFile>
#include <QStringList>

#include <klocale.h>
#include <kstandarddirs.h>

static QString basepath("kombination/tiles/%1.xml");

tileFactory::tileFactory()
{
}

bool tileFactory::haveTilesFor(const QString &langcode)
{
	QString path = KStandardDirs::locate("data", basepath.arg(langcode));
	return !path.isEmpty();
}

bool tileFactory::readTileSet(const QString &langcode)
{
	// look for the tileset of the specified language
	QString path = KStandardDirs::locate("data", basepath.arg(langcode));
	if (path.isEmpty())
	{
		// not found, so search better (if we can)
		QString language;
		QString country;
		QString charset;
		QString modifier;

		KLocale::splitLocale(langcode, language, country, modifier, charset);

		if (!country.isEmpty())
			// if we are here, then we have a language like en_GB,
			// and we have no tileset for that language. so, strip
			// the country part and look for the "base" language
			path = KStandardDirs::locate("data", basepath.arg(language));
		if (path.isEmpty())
			// all the attempts above failed, so the last hope for
			// us is to use the english tiles
			path = KStandardDirs::locate("data", basepath.arg("en"));
	}
	QFile file(path);
	if (!file.open(QIODevice::ReadOnly)) return false;
	QDomDocument doc("tileset");
	if (!doc.setContent(&file))
	{
		file.close();
		return false;
	}
	file.close();

	m_tileList.clear();

	bool found, atEnd;
	QDomNode n = doc.documentElement().firstChild();
	found = false;
	atEnd = false;
	// find the tiles child
	while (!found && !atEnd)
	{
		if (n.isNull()) atEnd = true;
		else if (!n.toElement().isNull() && n.toElement().tagName() == "tiles") found = true;
		else n = n.nextSibling();
	}
	if (found) readTiles(n.firstChild());
	else return false;
	
	n = doc.documentElement().firstChild();
	found = false;
	atEnd = false;
	// find the exceptions child
	while (!found && !atEnd)
	{
		if (n.isNull()) atEnd = true;
		else if (!n.toElement().isNull() && n.toElement().tagName() == "exceptions") found = true;
		else n = n.nextSibling();
	}
	if (found)
	{
		QDomNode n2;
		n2 = n.firstChild();
		found = false;
		atEnd = false;
		// find the replacements child
		while (!found && !atEnd)
		{
			if (n2.isNull()) atEnd = true;
			else if (!n2.toElement().isNull() && n2.toElement().tagName() == "replacements") found = true;
			else n2 = n2.nextSibling();
		}
		if (found) readReplacements(n2.firstChild());
		
		n2 = n.firstChild();
		found = false;
		atEnd = false;
		// find the exclusiveTiles child
		while (!found && !atEnd)
		{
			if (n2.isNull()) atEnd = true;
			else if (!n2.toElement().isNull() && n2.toElement().tagName() == "exclusiveTiles") found = true;
			else n2 = n2.nextSibling();
		}
		if (found) readExclusiveTiles(n2.firstChild());
	}
	
	return true;
}

QList<tile> tileFactory::tiles() const
{
	return m_tiles;
}

QList< QPair<QString, int> > tileFactory::tileStringList() const
{
	return m_tileList;
}

QStringList tileFactory::exclusiveTileList() const
{
	return m_exclusiveTileList;
}

replacementsHandler tileFactory::getReplacementsHandler() const
{
	return m_replacements;
}

QString tileFactory::simpleXMLSerialization(const tile *t)
{
	QDomDocument doc;
	QDomElement root = doc.createElement("tile");
	root.setAttribute("string", t->string());
	root.setAttribute("score", QString::number(t->score()));
	root.setAttribute("blank", QString::number(t->isBlank()));
	doc.appendChild(root);
	return doc.toString().remove(QChar('\n'));
}

tile* tileFactory::simpleXMLDeserialization(const QString &s)
{
	QDomDocument doc;
	if (!doc.setContent(s)) return 0;
	QDomNodeList tiles = doc.elementsByTagName("tile");
	if (tiles.count() != 1) return 0;
	QDomElement elem = tiles.item(0).toElement();
	QString str = elem.attribute("string");
	QString strscore = elem.attribute("score");
	QString strblank = elem.attribute("blank");
	bool ok = true;
	int score = strscore.toInt(&ok);
	if (!ok) return 0;
	bool blank = strblank.toInt(&ok);
	if (!ok) return 0;
	tile *t = new tile(str);
	t->setScore(score);
	t->setBlank(blank);
	return t;
}

void tileFactory::readTiles(const QDomNode &n)
{
	m_tiles.clear();
	QDomNode aux = n;
	while(!aux.isNull())
	{
		QDomElement e = aux.toElement(); // try to convert the node to an element.
		if(!e.isNull() && e.tagName() == "tile")
		{
			QString s;
			int q = 0;
			uint score = 0;
			bool hadLetter = false, hadQuantity = false, hadScore = false;
			
			QDomNode n2 = e.firstChild();
			while(!n2.isNull())
			{
				QDomElement e2 = n2.toElement(); // try to convert the node to an element.
				if (e2.tagName() == "letter")
				{
					s = e2.text();
					hadLetter = true;
				}
				else if (e2.tagName() == "quantity")
				{
					q = e2.text().toInt();
					hadQuantity = true;
				}
				else if (e2.tagName() == "score")
				{
					score = e2.text().toInt();
					hadScore = true;
				}
				
				n2 = n2.nextSibling();
			}
			if (hadLetter && hadQuantity && hadScore)
			{
				QPair<QString, int> p(s, score);
				if (m_tileList.indexOf(p) < 0)
					m_tileList << p;
				
				for ( ; q > 0; q--)
				{
					tile t;
					t.setString(s);
					t.setScore(score);
					m_tiles.append(t);
				}
			}
		}
		aux = aux.nextSibling();
	}
}

void tileFactory::readReplacements(const QDomNode &n)
{
	QDomNode aux = n;
	while(!aux.isNull())
	{
		QDomElement e = aux.toElement(); // try to convert the node to an element.
		if(!e.isNull() && e.tagName() == "replacement")
		{
			QString s, r;
			bool hadLetter = false, hadReplacement = false;
			
			QDomNode n2 = e.firstChild();
			while(!n2.isNull())
			{
				QDomElement e2 = n2.toElement(); // try to convert the node to an element.
				if (e2.tagName() == "letter")
				{
					s = e2.text();
					hadLetter = true;
				}
				else if (e2.tagName() == "replaces")
				{
					r = e2.text();
					hadReplacement = true;
				}
				n2 = n2.nextSibling();
			}
			if (hadLetter && hadReplacement)
			{
				m_replacements.addReplacement(s, r);
			}
		}
		aux = aux.nextSibling();
	}
}

void tileFactory::readExclusiveTiles(const QDomNode &n)
{
	QDomNode aux = n;
	while(!aux.isNull())
	{
		QDomElement e = aux.toElement(); // try to convert the node to an element.
		if(!e.isNull() && e.tagName() == "exclusiveTile")
		{
			QString s = e.text();
			if (m_exclusiveTileList.indexOf(s) < 0)
				m_exclusiveTileList << s;
		}
		aux = aux.nextSibling();
	}
}
