/****************************************************************************
 *   Copyright (C) 2005 by Albert Astals Cid <tsdgeos@terra.es>             *
 *   Copyright (C) 2005 by Pino Toscano      <toscano.pino@tiscali.it>      *
 * Copyright (C) 2007-2008 by Michal Sciubidlo <michal.sciubidlo@gmail.com> *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 2 of the License, or      *
 *   (at your option) any later version.                                    *
 ****************************************************************************/
#include "tilewidget.h"
#include "tilewidget.moc"

#include <QApplication>
#include <QDrag>
#include <QDragMoveEvent>
#include <QDropEvent>
#include <QLayout>
#include <QMouseEvent>
#include <QPainter>
#include <QSizePolicy>
#include <QStringList>
#include <QPointF>
#include <QGraphicsSceneMouseEvent>
#include <kdebug.h>

#include "tile.h"
#include "tilefactory.h"
#include "boardwidget.h"
#include "utils.h"
#include "renderer.h"

static QString basepath("kombination/graphics/%1.svg");

tileWidget::tileWidget(int nTiles) : QObject()
{
	m_tiles.reserve(nTiles);
	for (int i = 0; i < m_tiles.capacity(); i++)
		m_tiles.append(0L);
	m_startdrag = QPoint(-1, -1);
	m_tileSize = 32;

	//exchangeRack[i] = addPixmap(*(pixmapTable[6]));
	//exchangeRack[i]->setPos(8*32 + i*32, 15*32);

	m_oldDragIndex = -1;
	setAcceptDrops(true);
	// fake value, just to start layouting and updating
	//changeTileSize(15.0);
}

tileWidget::~tileWidget()
{
	qDeleteAll(m_tiles);
}

void tileWidget::setTile(tile *t, int index)
{
	if (!Utils::checkRange(index, 0, m_tiles.size())) return;
	m_tiles[index] = t;
	update();
}

void tileWidget::removeDroppedTile(tile* t)
{
	delete m_tiles[m_oldDragIndex];
	setTile(t, m_oldDragIndex);
	m_oldDragIndex = -1;
	update();
	emit numTilesChanged(freeSpaces());
}

int tileWidget::freeSpaces() const
{
	return m_tiles.count(0);
}

bool tileWidget::isFull() const
{
	return freeSpaces() == 0;
}

const QVector<tile*> &tileWidget::tiles() const
{
 	return m_tiles;
}

// void tileWidget::setTiles(const QVector<tile*> &tiles)
// {
// 	int count = tiles.count();
// 	for (int i = 0; i < count; ++i)
// 	{
// 		tile *t = new tile(*(tiles.at(i)));
// 		setTile(t, i);
// 	}
// }

void tileWidget::removeTiles(const QVector<tile*> &tiles)
{
	int count = m_tiles.size();
	foreach(tile *t, tiles)
	{
		for (int i = 0; i < count; ++i)
		{
			tile *t2 = m_tiles.at(i);
			if (t2 && *t == *t2)
			{
				delete t2;
				m_tiles[i] = 0;
			}
		}
	}
	emit numTilesChanged(freeSpaces());
	update();
}

// void tileWidget::changeTileSize(double size)
// {
// 	m_tileSize = size;
// 
// 	setFixedSize((int)(m_tileSize * m_tiles.size() + m_tiles.size() - 1), (int)m_tileSize + 2);
// 
// 	update();
// 	parentWidget()->layout()->update();
// }

void tileWidget::addTile(tile *t)
{
	int id = m_tiles.indexOf(0);
	if (id > -1)
	{
		m_tiles[id] = t;
		update();
	}
}

void tileWidget::insertAndShift(tile *t, int pos, int oldpos)
{
	if (!(Utils::checkRange(pos, 0, m_tiles.size()) || Utils::checkRange(oldpos, 0, m_tiles.size()))) return;
	if (m_tiles.at(pos))
	{
		m_tiles.remove(oldpos);
		if (pos == m_tiles.capacity())
		{
			kDebug() << "append t" << endl;
			m_tiles.append(t);
		}
		else
		{
			kDebug() << "insert at:" << pos << endl;
			m_tiles.insert(pos, t);
		}
	}
	else
	{
		kDebug() << "o" << endl;
		m_tiles[pos] = t;
		m_tiles[oldpos] = 0;
	}
}

void tileWidget::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
	Q_UNUSED(option);
	Q_UNUSED(widget);
	for (int i = 0; i < m_tiles.size(); i++)
	{
		painter->drawPixmap(i*m_tileSize, 0, renderer::self()->rackBackground());
		if (m_tiles.at(i))
		{
			painter->drawPixmap(i*m_tileSize, 0, renderer::self()->tileBackground( m_tiles.at(i)->isBlank() ));
			painter->setRenderHint(QPainter::Antialiasing, true);
			if (!m_tiles.at(i)->string().isEmpty())
			{
				//if (boardWidget::tileFontSize != -1)
				//{
					//QFont f;
					//f.setPointSizeF(boardWidget::tileFontSize);
					//p->setFont(f);
				//}
				painter->drawText(i*m_tileSize, 0, m_tileSize, m_tileSize, Qt::AlignCenter, m_tiles.at(i)->string());
			}
		}
	}
	
}

QRectF tileWidget::boundingRect() const
{
	return QRectF(0, 0, m_tileSize*7, m_tileSize);
}

void tileWidget::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
	if (event->button() == Qt::LeftButton)
	{
		m_startdrag = event->pos().toPoint();
	}
}

void tileWidget::mouseMoveEvent(QGraphicsSceneMouseEvent *event)
{
	if (!(event->buttons() & Qt::LeftButton)) return;

	if ((event->pos().toPoint() - m_startdrag).manhattanLength() < QApplication::startDragDistance()) return;

	//int tmp = Utils::calcMousePositionInt(m_startdrag.x(), m_tileSize);
	int tmp = m_startdrag.x() / m_tileSize;
	if (!Utils::checkRange(tmp, 0, m_tiles.size())) return;
	
	tile *t = m_tiles.at(tmp);
	//TODO: send tileBack or blankBack
	QDrag *drag = Utils::createDragForTile(event->widget(), (QObject*)this, renderer::self()->tileBackground( t->isBlank() ), t, m_tileSize);
	if (drag)
	{
		m_oldDragIndex = tmp;
		Qt::DropAction dropAction = drag->start(Qt::MoveAction);
		Q_UNUSED(dropAction);
		m_startdrag = event->pos().toPoint();
	}
	else
	{
		m_startdrag = QPoint(-1, -1);
	}
}

void tileWidget::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{
	int tmp = m_startdrag.x() / m_tileSize;
	if (!Utils::checkRange(tmp, 0, m_tiles.size())) return;
	tile *t = m_tiles.at(tmp);
	if (event->button() == Qt::LeftButton && t)
	{
		//TODO: tileBack i blankBack
		QDrag *drag = Utils::createDragForTile(event->widget(), (QObject*)this, renderer::self()->tileBackground( t->isBlank() ), t, m_tileSize);
		m_oldDragIndex = tmp;
		Qt::DropAction dropAction = drag->start(Qt::MoveAction);
		Q_UNUSED(dropAction);
		m_startdrag = event->pos().toPoint();
	}
}

void tileWidget::dragEnterEvent(QGraphicsSceneDragDropEvent * event)
{
	// always accept the drop, as will be the dragMoveEvent to take care of checking where we are
	event->acceptProposedAction();
}

void tileWidget::dragMoveEvent(QGraphicsSceneDragDropEvent * event)
{
	int tmp = m_startdrag.x() / m_tileSize;

	const Utils::sMimeData *myData = qobject_cast<const Utils::sMimeData *>(event->mimeData());

	if (event->mimeData()->hasText() && Utils::checkRange(tmp, 0, m_tiles.size()) && (((m_tiles.at(tmp) == 0)) || (myData)))
		event->acceptProposedAction();
	else
		event->ignore();
}

void tileWidget::dropEvent(QGraphicsSceneDragDropEvent * event)
{
	boardWidget *bw;
	tileWidget *tw;
	int tmp = event->pos().x() / m_tileSize;

	const Utils::sMimeData *myData =  qobject_cast<const Utils::sMimeData *>(event->mimeData());

	tile *t = tileFactory::simpleXMLDeserialization(event->mimeData()->text());
	if (t)
	{
		QObject *source = myData->senderItem(); //event->source();
		if ((void*)source == (void*)this)
		{
			insertAndShift(t, tmp, m_oldDragIndex);
		}
		else if ((bw = dynamic_cast<boardWidget*>(source)))
		{
			bw->removeDroppedTile(m_tiles[tmp]);
			m_tiles[tmp] = t;
			emit numTilesChanged(freeSpaces());
		}
		else if ((tw = dynamic_cast<tileWidget*>(source)))
		{
			tw->removeDroppedTile(m_tiles[tmp]);
			m_tiles[tmp] = t;	
			emit numTilesChanged(freeSpaces());
		}
                else
                        delete t;
	}

	m_startdrag = QPoint(-1, -1);
	m_oldDragIndex = -1;

	event->acceptProposedAction();

	update();
}

void tileWidget::resize(double size)
{
	prepareGeometryChange();
	m_tileSize = size;
}

void tileWidget::reset()
{
	for (int i = 0; i < m_tiles.capacity(); i++)
		m_tiles[i] = 0;
}
