/*
 *  Koware - The ancient African game, also known as Mancala, for KDE
 *  Copyright (C) 2005  Michael Collins  <michaelcollins@ivorycity.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Steet, Fifth Floor, Boston, MA  02110-1301  USA
 */


#include "game.h"

char* illegalmoves[] =
{ "You cannot play an empty cup.\n",
  "You cannot leave the opponent with 0 beads by capture.\n",
  "You must sow the opponent if possible.\n"};

static int  need_endgame_score ( struct game* g);
static int  endgame_score      ( struct game* g, int begincup, int ndist );
static int  cycle_score        ( struct game* g );
static int  score              ( struct game* g, int endcup );
static void restore            ( int distcup, int cupval, int board[]);

void
oware_init(struct game* g)
{
  int i;
  g->player1score=0;
  g->player2score=0;
  g->gameover    =0;
  for(i=0;i<12;i++) g->cups[i]=4;
}

int
oware_move(struct game* g, int move)
{
  int cupval     = g->cups[move];
  int cupvalsave = cupval;
  int curcup     = move+1;
  int s;

  if(g->cups[move]==0) return ERR_MOVE0;

  g->cups[move]=0;

  for(;;)
  {
    if(curcup==12)  { curcup=0;           }
    if(curcup==move){ curcup++; continue; }

    g->cups[curcup]++;

    if(--cupval==0) break;
    else            curcup++;
  }

  s = score( g, curcup );

  if( need_endgame_score(g) ){
    if( s>0 ) return ERR_MOVE1;                           /* Player left opponent with no beads by taking all of them */
    else      return endgame_score(g, move, cupvalsave);  /* Either gameover or illegal move */
  }

  if( (g->player1score+g->player2score)==46 ) if( cycle_score(g) ) return 1;  /* Check for cycles */

  if(g->turn==PLAYER1){ g->player1score+=s; g->turn=PLAYER2; }
  else                { g->player2score+=s; g->turn=PLAYER1; }

  return s;
}

void
oware_dump( struct game* g )
{
  int i;
  for( i=5; i>=0; i--) printf("[%d]\t",g->cups[i]); printf("\n");
  for( i=6; i<12; i++) printf("[%d]\t",g->cups[i]);
  printf("\nScore P1 %d P2 %d\n", g->player1score, g->player2score);
}

/* Check to see if opponent has 0 beads, if so it could be endgame*/
static int
need_endgame_score(struct game* g)
{
  int i;
  if(g->turn==1) for( i=0;i<6;i++)  { if( g->cups[i] > 0) return 0; }/* Look at opponents beads (player 2) */
  else           for( i=6;i<12;i++) { if( g->cups[i] > 0) return 0; }/* Ditto (player 1) */
  return 1;
}

/* score the endgame or identify an illegal move */
static int
endgame_score( struct game* g, int begincup, int ndist )
{
  int score = 0;
  int i;

  restore( begincup, ndist, g->cups );

  if( g->turn == PLAYER1 ) for( i=6;i<12;i++ )
  {
    int maxbeads = 11-i;                        /* Most beads we could have in this cup to not populate opponent */
    int beads    = g->cups[i];
    if( beads > maxbeads )   return ERR_MOVE2;  /* Player could have made a legel move */
    else                     score += beads;
  }

  else for( i=0; i<6; i++)
  {
    int maxbeads = 5-i;                         /* Most beads we can distribute legally */
    int beads    = g->cups[i];
    if( beads > maxbeads )  return ERR_MOVE2;   /* Player could have made a legel move */
    else                    score += beads;
  }

  if(g->turn==PLAYER1){ g->player1score+=score; g->turn=PLAYER2; }
  else                { g->player2score+=score; g->turn=PLAYER1; }

  g->gameover = 1;
  for(i=0;i<12;i++) g->cups[i]=0;
  return score;
}

/* score the endgame or identify an illegal move */
static int
cycle_score( struct game* g )
{
  int i;
  int beadstart=-1;
  for(i=0;i<12;i++)
  {
    if( g->cups[i]==1 )
    {
      if(beadstart==-1) beadstart=i;
      else if(i-beadstart==6)
      {
        g->player1score++;
        g->player2score++;
        g->cups[beadstart]=0;
        g->cups[i]=0;
        g->gameover=1;
        return 1;
      }
    }
  }
  return 0;
}

/* undo a move */
static void
restore(int distcup, int cupval, int board[])
{
  int curcup = distcup+1;
  board[distcup]=cupval;

  while(cupval)
  {
    if(curcup==12)     { curcup=0;           }
    if(curcup==distcup){ curcup++; continue; }
    board[curcup++]--;
    cupval--;
  }
}

static int
score( struct game* g, int endcup  )
{
  int score=0;

  if( g->turn == PLAYER1 )
  {
    if( endcup > 5 ) return 0; /* ended on her own side (cup 6-11), no points */
    for(;;)
    {
      if( g->cups[endcup]==2 || g->cups[endcup]==3 ){ score+=g->cups[endcup]; g->cups[endcup]=0; }
      else break;
      if(--endcup==-1) break;
    }
  }

  else /*(player == 2)*/
  {
    if( endcup < 6 ) return 0;  /* ended on her own side (cup 0-5), no points */
    for(;;)
    {
      if( g->cups[endcup]==2 || g->cups[endcup]==3 ){ score+=g->cups[endcup]; g->cups[endcup]=0; }
      else break;
      if(--endcup==5) break;
    }
  }

  return score;
}



