/**************************************************************************
*   Copyright (C) 2005 by Lukas Tinkl                                     *
*   lukas@kde.org                                                         *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.             *
***************************************************************************/

#ifndef _GAMEMANAGER_H_
#define _GAMEMANAGER_H_

#include "launch.h"

#include <qvaluelist.h>
#include <qcombobox.h>
#include <qdom.h>

class QString;
class Server;
class Game;

/**
 * \typedef A value-based list of games
 */
typedef QValueList<Game> GameList;


/**
 * A pool that handles all games known and available to kqf. This
 * class holds the list of games in m_gameList
 * @short Game pool
 */
class GameManager: public QObject
{
    Q_OBJECT
public:
    /**
     * CTOR
     */
    GameManager();

    ~GameManager();

    /**
     * @return a list of all the games available in the system
     */
    GameList availableGames() const;

    /**
     * @return a list of all the games known to qstat
     */
    GameList allGames() const;

    /** @see Game::serverOption */
    QString serverOption( const QString & gameID ) const;
    /** @see Game::masterServerOption */
    QString masterServerOption( const QString & gameID ) const;
    /** @see Game::masterServer */
    QString masterServer( const QString & gameID ) const;

    /**
     * @return a small pixmap of the game (icon)
     */
    QPixmap gameIcon( const QString & gameID ) const;

    /**
     * @return pixmap representing a server being password-protected
     */
    static QPixmap lockedIcon();

    /**
     * @return pixmap representing a server having Punkbuster on
     */
    static QPixmap pbIcon();

    /**
     * @return pixmap representing a server having both Punkbuster and
     * password effective
     */
    static QPixmap pbLockedIcon();

    /**
     * @param color name of the icon color (can be red, black or yellow)
     * @return pixmap representing the status of player count
     */
    static QPixmap manIcon( const QString & color );

    /**
     * @return pixmap for a @p country
     */
    static QPixmap flagIcon( const QString & country );

    /**
     * @return the rule name that is used for analyzing the server password option
     * (whether the server is locked or not)
     *
     * First we try specific games, then we look up for the whole "family" of servers
     */
    QString passwordRule( const QString & gameID );

    /**
     * @return the rule name that is used for analyzing the availability of Punkbuster
     * or another cheating counter measure
     */
    QString pbRule( const QString & gameID );

    /**
     * @return the list of rule names that are used for getting the name of the game mod
     * @see Server::gameName()
     */
    QStringList gameNameRule( const QString & gameID );

    /**
     * @return the list of rule names that are used for getting the type of the game
     * @see Server::gameType()
     */
    QStringList gameTypeRule( const QString & gameID );

    /**
     * Executes the client game (binary); decides about
     * game type and command like parameters
     * @return whether the game was actually started
     */
    bool execClient( Server * server );

    /**
     * @return true if ther exists a flag for a game
     * @param gameID ID of the game
     * @param name name of the flag
     */
    bool hasFlag( const QString & gameID, const QString & name ) const;

    /**
     * @return user visible name of the game by its ID
     */
    QString gameName( const QString & gameID ) const;

signals:
    /**
     * Emitted when the user stops playing the game (ie when the game process finishes)
     * (propagated from the Launch class up to the view)
     */
    void gameFinished();

private:
    /**
     * Helper function to parse the game definitions
     * @param xmlFile path to the XML file to parse
     * @return whether the parsing succeeded
     */
    bool parseXML( const QString & xmlFile );

    /**
     * @return a specific game by its @p gameID
     */
    Game * gameByID( const QString & gameID ) const;

    /**
     * @return the base game (similar, parent), if any
     */
    Game * baseGame( const QString & serverType ) const;

    /**
     * Try to find the type of the game client to execute
     */
    QString execClientType( const QString & gameID );

    /// internal list holding all the games
    GameList m_gameList;

    /// game launcher helper class
    Launch * m_launch;
};


/**
 * Combobox with games, available or all of the supported ones.
 * The combo displays a game's icon together with its name, gameID() returns
 * ID of the game that the user selected
 *
 * @short Game selection combobox
 */
class GameComboBox: public QComboBox
{
    Q_OBJECT
public:
    /**
     * CTOR
     * @param manager GameManager handling the games
     */
    GameComboBox( GameManager * manager, QWidget * parent = 0, const char * name = 0 );

    /**
     * @return game ID of the currently selected item
     */
    QString gameID() const;

    /**
     * Set the game ID (changes the current combo item)
     */
    void setGameID( const QString & gameID );

private:
    QMap<int, QString> m_gameMap;
};

#endif
