/**************************************************************************
*   Copyright (C) 2005 by Lukas Tinkl                                     *
*   lukas@kde.org                                                         *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.             *
***************************************************************************/

#include <stdlib.h>

#include "kqf.h"
#include "settings.h"
#include "prefs_games.h"
#include "prefs_servers.h"
#include "prefs_im.h"

#include <qdragobject.h>

#include <kconfigdialog.h>
#include <kglobal.h>
#include <klocale.h>
#include <kiconloader.h>
#include <kdeversion.h>
#include <kmenubar.h>
#include <kstatusbar.h>
#include <kkeydialog.h>
#include <kstdaccel.h>
#include <kaction.h>
#include <kstdaction.h>
#include <kactionclasses.h>
#include <kpopupmenu.h>
#include <kstringhandler.h>
#include <kstandarddirs.h>
#include <kmessagebox.h>

kqf::kqf()
    : KMainWindow( 0, "kqf" ), m_view( new kqfView( this ) )
{
    // accept dnd
    setAcceptDrops( true );

    // tell the KMainWindow that this is indeed the main widget
    setCentralWidget( m_view );

    // and a status bar
    statusBar()->show();

    // allow the view to change the statusbar and caption
    connect( m_view, SIGNAL( signalChangeStatusbar( const QString& ) ),
             statusBar(), SLOT( message( const QString& ) ) );
    connect( m_view, SIGNAL( signalChangeCaption( const QString& ) ),
             this, SLOT( setCaption( const QString& ) ) );

    connect( m_view, SIGNAL( signalEnableUpdateActions( bool ) ),
             this, SLOT( slotEnableUpdateActions( bool ) ) );
    connect( m_view, SIGNAL( signalEnableServerActions( bool, bool ) ),
             this, SLOT( slotEnableServerActions( bool, bool ) ) );

    connect( m_view->lvServers, SIGNAL( contextMenuRequested ( QListViewItem *, const QPoint &, int ) ),
             this, SLOT( slotShowServerRMB( QListViewItem *, const QPoint &, int ) ) );

    actionCollection()->setHighlightingEnabled( true );
    connect( actionCollection(), SIGNAL( actionStatusText( const QString & ) ),
             statusBar(), SLOT( message( const QString & ) ) );
    connect( actionCollection(), SIGNAL( clearStatusText() ),
             statusBar(), SLOT( clear() ) );

    // then, setup our actions
    setupActions();

    slotEnableUpdateActions( true );

    setCaption( i18n( "Game Server Browser" ) );

    if ( !checkQStat() )
    {
        KMessageBox::error( this, "<qt>" + i18n( "QStat was not found in the system; kqf will not run without it.<br>"
                                                 "You can get the latest version (2.8 at least) "
                                                 "from the <a href=\"http://sourceforge.net/project/showfiles.php?group_id=56603&package_id=51888\">qstat</a> "
                                                 "download page." ), i18n( "QStat Not Found" ), KMessageBox::Notify | KMessageBox::AllowLink );
        exit( EXIT_FAILURE );
    }
}

kqf::~kqf()
{
}

void kqf::setupActions()
{
    // File menu
    KStdAction::quit( kapp, SLOT( quit() ), actionCollection() );

    // Edit menu
    KStdAction::copy( m_view, SLOT( slotCopy() ), actionCollection() );
    action( KStdAction::name( KStdAction::Copy ) )->setToolTip( i18n( "Copy server address" ) );
    action( KStdAction::name( KStdAction::Copy ) )->
        setWhatsThis( i18n( "Copies the currently selected server's address to the clipboard" ) );

    // KDE 4 icon name: edit-find
    new KAction( i18n( "Find Player" ), "find", KStdAccel::shortcut( KStdAccel::Find ),
                 m_view, SLOT( slotFindPlayer() ), actionCollection(), "find_player" );

    // View menu
    // KDE 4 icon name: view-refresh
    new KAction( i18n( "Refresh View" ), "reload", "Ctrl+R", m_view, SLOT( slotRefreshView() ),
                 actionCollection(), "refresh_view" );

    new KAction( i18n( "Refresh Selected" ), 0, "Ctrl+S", m_view, SLOT( slotRefreshSelected() ),
                 actionCollection(), "refresh_selected" );
    action( "refresh_selected" )->setToolTip( i18n( "Refresh the server info" ) );
    action( "refresh_selected" )->setWhatsThis( i18n( "Refreshes the currently selected server's information." ) );

    // KDE 4 icon name: run-build-rebuild?
    new KAction( i18n( "Update From Master" ), "rebuild", "Ctrl+U", m_view, SLOT( slotUpdateView() ),
                 actionCollection(), "update_view" );
    action( "update_view" )->setToolTip( i18n( "Update the server list from master" ) );
    action( "update_view" )->setWhatsThis( i18n( "Rebuilds the whole server list by querying the master server." ) );

    // KDE 4 icon name: process-stop
    new KAction( i18n( "Stop" ), "stop", KShortcut(), m_view, SLOT( slotStopUpdating() ),
                 actionCollection(), "stop" );
    action( "stop" )->setToolTip( i18n( "Stop updating data" ) );
    action( "stop" )->setWhatsThis( i18n( "Stops the current process of updating data from the server." ) );

    new KToggleAction( i18n( "Resolve Hostnames" ), 0, KShortcut(),
                       this, SLOT( slotToggleResolving() ), actionCollection(), "resolve_hostnames" );
    static_cast<KToggleAction *>( action( "resolve_hostnames" ) )->setChecked( Settings::resolveHostnames() );
    action( "resolve_hostnames" )->setToolTip( i18n( "Show server hostname" ) );
    action( "resolve_hostnames" )->setWhatsThis( i18n( "Toggles showing server's hostname instead of its IP address." ) );


    // Server menu
    // KDE 4 icon name: (not known at the time of writing)
    new KAction( i18n( "Connect" ), "connect_creating", KShortcut(), m_view, SLOT( slotConnect() ),
                 actionCollection(), "connect" );
    action( "connect" )->setToolTip( i18n( "Connect to the server" ) );
    action( "connect" )->setWhatsThis( i18n( "Start the game and connect to the selected server." ) );

    new KAction( i18n( "DNS Lookup" ), 0, "Ctrl+L", m_view, SLOT( slotDNSLookup() ),
                 actionCollection(), "dns_lookup" );

    // KDE 4 icon name: document-new
    new KAction( i18n( "Add New Server" ), "filenew", "Ctrl+N", m_view, SLOT( slotAddNewServer() ),
                 actionCollection(), "add_new_server" );
    action( "add_new_server" )->setToolTip( i18n( "Add a new server to favorites" ) );
    action( "add_new_server" )->setWhatsThis( i18n( "Manually adds a server to favorites." ) );

    // KDE 4 icon name: bookmark-new, or maybe list-add
    new KAction( i18n( "Add to Favorites" ), "bookmark_add", "Ctrl+B", m_view, SLOT( slotAddToFavorites() ),
                 actionCollection(), "add_to_favorites" );
    action( "add_to_favorites" )->setToolTip( i18n( "Add server to favorites" ) );
    action( "add_to_favorites" )->setWhatsThis( i18n( "Adds the currently selected server to favorites." ) );

    // KDE 4 icon name: edit-delete, or maybe list-remove
    new KAction( i18n( "Remove from Favorites" ), "editdelete", KShortcut(), m_view, SLOT( slotRemoveFromFavorites() ),
                 actionCollection(), "remove_from_favorites" );
    action( "remove_from_favorites" )->setToolTip( i18n( "Remove server from favorites" ) );
    action( "remove_from_favorites" )->setWhatsThis( i18n( "Removes the currently selected server from favorites." ) );

    // KDE 4 icon name: document-properties
    new KAction( i18n( "Properties" ), "edit", KShortcut(), m_view, SLOT( slotProperties() ),
                 actionCollection(), "properties" );
    action( "properties" )->setToolTip( i18n( "Display server properties" ) );
    action( "properties" )->setWhatsThis( i18n( "Show a dialog with the selected server's properties." ) );

    KStdAction::preferences( this, SLOT( optionsPreferences() ), actionCollection() );

    setupGUI( ToolBar | Keys | StatusBar | Save | Create );
}

void kqf::optionsPreferences()
{
    if ( KConfigDialog::showDialog( "settings" ) )
        return;

    KConfigDialog * dialog = new KConfigDialog( this, "settings", Settings::self() );
    dialog->addPage( new PrefsGames(), i18n( "Games" ), "package_games", i18n( "Configure games" ) );
    dialog->addPage( new PrefsServers(), i18n( "Servers" ), "package_settings", i18n( "Configure servers" ) );
    dialog->addPage( new PrefsIM(), i18n( "IM presence" ), "presence_away", i18n( "Configure IM presence" ) );
    connect( dialog, SIGNAL( settingsChanged() ), m_view, SLOT( slotLoadSettings() ) );
    dialog->show();
}

bool kqf::queryExit()
{
    m_view->saveServerList();
    m_view->slotStopUpdating();
    Settings::writeConfig();    // necessary for non-dialog settings to get saved
    m_view->saveConfig();
    return true;
}

void kqf::slotToggleResolving()
{
    if ( static_cast<KToggleAction *>( action( "resolve_hostnames" ) )->isChecked() )
        Settings::setResolveHostnames( true );
    else
        Settings::setResolveHostnames( false );
}

void kqf::slotEnableUpdateActions( bool enable )
{
    action( "stop" )->setEnabled( !enable );
    action( "connect" )->setEnabled( enable );
    action( "dns_lookup" )->setEnabled( enable );
    action( "refresh_view" )->setEnabled( enable );
    action( "update_view" )->setEnabled( enable );
    action( "add_new_server" )->setEnabled( enable );
    action( "properties" )->setEnabled( enable );
    action( "refresh_selected" )->setEnabled( enable && m_view->lvServers->selectedItem() );
}

void kqf::slotEnableServerActions( bool add, bool remove )
{
    action( "add_to_favorites" )->setEnabled( add );
    action( "remove_from_favorites" )->setEnabled( remove );
}

void kqf::slotShowServerRMB( QListViewItem * item, const QPoint & pos, int /*col*/ )
{
    Server * server = static_cast<Server *>( item );

    if ( server )
    {
        KPopupMenu * menu = new KPopupMenu( this, "server_popup" );
        menu->insertTitle( m_view->gameIcon( server->gameID() ),
                           KStringHandler::rsqueeze( server->strippedName(), 30 ) );
        action( "connect" )->plug( menu );
        menu->insertSeparator();
        action( "add_to_favorites" )->plug( menu );
        action( "remove_from_favorites" )->plug( menu );
        action( KStdAction::name( KStdAction::Copy ) )->plug( menu );
        menu->insertSeparator();
        action( "properties" )->plug( menu );
        menu->popup( pos );
    }
}

bool kqf::checkQStat()
{
    return !KStandardDirs::findExe( "qstat" ).isEmpty();
}

#include "kqf.moc"
