/****************************************************************
Copyright 2008  Matt Williams matt@milliams.com
Copyright 2008  Parker Coates parker.coates@gmail.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) version 3, or any
later version accepted by the membership of KDE e.V. (or its
successor approved by the membership of KDE e.V.), which shall
act as a proxy defined in Section 6 of version 3 of the license.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************/

#include "kscoremanager.h"

#include <KDebug>
#include <KGlobal>
#include <KHighscore>
#include <KLocale>
#include <KUser>

#include <QtCore/QDate>
#include <QtCore/QStringList>
#include <QtCore/QVariant>

class KScoreManagerPrivate
{
    public:
        KScoreManagerPrivate();
        ~KScoreManagerPrivate();

        KScoreManager instance;

        QMap<QString, KScoreManager::ScoreList > scores; ///<Maps config group name to GroupScores

        bool loaded;
        bool setupHasBeenFinalized;

        int scoreBoardLength;

        QString groupOfLatestAddition;
        int rankOfLatestAddition;
        bool waitingForName;

        KHighscore * highscoreObject;
        QMap<QString, QString> translatedGroupNames; ///<List of the translated group names.
        QMap<KScoreManager::Field, QString> keys; ///<Keys for fields. Maps field index to a string
        QMap<KScoreManager::Field, QString> headers; ///<Header for fields. Maps field index to a string
        QList<KScoreManager::Field> visibleFields;

        KScoreManager::Options options;

        QString lastEnteredName;

        //Functions
        void loadScores(const QString& groupKey);
        void loadAllScores();
        void saveScores(const QString& groupKey);
        void finalizeSetup();
};

KScoreManagerPrivate::KScoreManagerPrivate()
  : loaded(false),
    setupHasBeenFinalized(false),
    scoreBoardLength(10),
    groupOfLatestAddition("BADGROUP"),
    rankOfLatestAddition(0),
    waitingForName(false),
    highscoreObject(new KHighscore())
{
}

KScoreManagerPrivate::~KScoreManagerPrivate()
{
    delete highscoreObject;
}

void KScoreManagerPrivate::loadScores(const QString& groupKey)
{
    kDebug() << "Loading scores for" << groupKey;
    scores[groupKey].clear();

    highscoreObject->setHighscoreGroup(groupKey);

    for (int i = 1; i <= scoreBoardLength; ++i)
    {
        KScoreManager::ScoreEntry score;
        foreach(KScoreManager::Field field, keys.keys())
        {
            score[field] = highscoreObject->readEntry(i, keys[field], QString("-"));
        }
        scores[groupKey].append(score);
    }
}

void KScoreManagerPrivate::loadAllScores()
{
    scores.clear();

    highscoreObject->setHighscoreGroup("LastPlayer");
    lastEnteredName = highscoreObject->readEntry(0, "LastPlayer", QString());

    QList<QString> groupKeyList; //This will be a list of all the groups in the config fie
    foreach(const QString & group, highscoreObject->groupList())
    {
        if (group != "LastPlayer")
        {
            loadScores(group);

            if( scores[group].first().value(KScoreManager::Score) == '-' )
            {
                scores.remove(group);
            }
        }
    }
    loaded = true;
}

void KScoreManagerPrivate::saveScores(const QString& groupKey)
{
    kDebug() << "Saving scores for" << groupKey;

    if (highscoreObject->lockForWriting())
    {
        highscoreObject->setHighscoreGroup("LastPlayer");
        highscoreObject->writeEntry(0, "LastPlayer", lastEnteredName);

        highscoreObject->setHighscoreGroup(groupKey);

        for (int i = 1; i <= scoreBoardLength; ++i)
        {
            KScoreManager::ScoreEntry score = scores[groupKey].at(i-1);
            foreach(KScoreManager::Field field, keys.keys())
            {
                highscoreObject->writeEntry(i, keys[field], score[field]);
            }
        }

        highscoreObject->writeAndUnlock();
    }
}

void KScoreManagerPrivate::finalizeSetup()
{
    if (!keys.contains(KScoreManager::Name))
    {
        instance.addField(KScoreManager::Name);
    }
    if (!keys.contains(KScoreManager::Score))
    {
        instance.addField(KScoreManager::Score);
    }
    if (options.testFlag(KScoreManager::AutomaticTimeStamp) && !keys.contains(KScoreManager::Date))
    {
        instance.addField(KScoreManager::Date);
    }

    setupHasBeenFinalized = true;
}

K_GLOBAL_STATIC( KScoreManagerPrivate, p )

KScoreManager* KScoreManager::self()
{
    return &p->instance;
}

KScoreManager::KScoreManager()
  : QObject()
{
}

KScoreManager::~KScoreManager()
{
}

void KScoreManager::addScoreGroupName(const QString& groupKey, const QString& displayName)
{
    if (!p->translatedGroupNames.contains(groupKey))
    {
        p->translatedGroupNames[groupKey] = displayName;
    }
}

void KScoreManager::addScoreGroupNames(const QMap<QString, QString>& groups)
{
    foreach(const QString& groupKey, groups.keys())
    {
        addScoreGroupName(groupKey, groups[groupKey]);
    }
}

void KScoreManager::addCustomField(const Field& field, const QString& key, const QString& header, bool visible)
{
    if ( !p->setupHasBeenFinalized )
    {
        p->keys[field] = key;
        p->headers[field] = header;

        if ( p->visibleFields.contains(field) )
        {
            p->visibleFields.removeOne(field);
        }
        if ( visible )
        {
            p->visibleFields << field;
        }
    }
    else
    {
        kWarning() << "Setup has already been finalized. Cannot add field.";
    }
}

void KScoreManager::addField(const Field& field, bool visible)
{
    switch(field)
    {
    case Name :
        addCustomField(field, "Name", i18n("Name"), visible);
        break;
    case Level :
        addCustomField(field, "Level", i18n("Level"), visible);
        break;
    case Date :
        addCustomField(field, "Date", i18n("Date"), visible);
        break;
    case Time :
        addCustomField(field, "Time", i18n("Time"), visible);
        break;
    case Score :
        addCustomField(field, "Score", i18n("Score"), visible);
        break;
    case Custom1 :
        addCustomField(field, "Custom1", i18n("Custom 1"), visible);
        break;
    case Custom2 :
        addCustomField(field, "Custom2", i18n("Custom 2"), visible);
        break;
    case Custom3 :
        addCustomField(field, "Custom3", i18n("Custom 3"), visible);
        break;
    case Custom4 :
        addCustomField(field, "Custom4", i18n("Custom 4"), visible);
        break;
    case Custom5 :
        addCustomField(field, "Custom5", i18n("Custom 5"), visible);
        break;
    }
}

void KScoreManager::setScoreBoardLength(int numberOfScores)
{
    if ( !p->setupHasBeenFinalized )
    {
        if ( numberOfScores > 0 )
        {
            p->scoreBoardLength = numberOfScores;
        }
    }
    else
    {
        kWarning() << "Setup has already been finalized. Cannot change score board length.";
    }
}

int KScoreManager::scoreBoardLength() const
{
    return p->scoreBoardLength;
}

void KScoreManager::setOptions(const KScoreManager::Options& options)
{
    if ( !p->setupHasBeenFinalized )
    {
        p->options = options;
    }
    else
    {
        kWarning() << "Setup has already been finalized. Cannot change score board length.";
    }
}

KScoreManager::Options KScoreManager::options() const
{
    return p->options;
}

int KScoreManager::submitScore(const ScoreEntry& score, const QString& groupKey)
{
    if (!p->setupHasBeenFinalized)
    {
        p->finalizeSetup();
    }

    if (!groupKey.isEmpty() && !p->translatedGroupNames.contains(groupKey))
    {
        kWarning() << "Attempting to add a score to a group that has not been translated.";
        p->translatedGroupNames[groupKey] = groupKey;
    }

    if (!p->loaded)
    {
        p->loadAllScores();
    }

    if (!p->scores.contains(groupKey))
    {
        p->loadScores(groupKey);
    }

    if (!score.contains(Score))
    {
        kWarning() << "Attempting to add a score entry without the 'Score' property set.";
        return 0;
    }

    bool lowerScoresAreBetter = p->options.testFlag(LowerScoresAreBetter);
    int submittedScore = score.value(Score).toInt();

    for (int i = 0; i < p->scoreBoardLength; ++i)
    {
        bool ok; //will be false if there isn't any score yet in position i
        int scoreToCompare = p->scores.value(groupKey).at(i).value(Score).toInt(&ok); //test if the stored score is a number

        if ( (!lowerScoresAreBetter && (submittedScore > scoreToCompare))
             || (lowerScoresAreBetter && (submittedScore < scoreToCompare))
             || !ok
           )
        {
            p->scores[groupKey].insert(i, score);
            p->scores[groupKey].removeAt(p->scoreBoardLength);

            if (score.contains(Name))
            {
                p->lastEnteredName = score.value(Name).toString();
            }
            else
            {
                kDebug() << "How'd we get in here?";
                // If the name is empty, use the last entered name.
                QString name = p->lastEnteredName;
                if(name.isEmpty())
                {
                    // If the name is still empty, use the current user's full name.
                    KUser user;
                    name = user.property(KUser::FullName).toString();
                    if (name.isEmpty())
                    {
                        // If the name is still empty, use the current user's user name.
                        name = user.loginName();
                    }
                }
                p->scores[groupKey][i][Name] = name;
            }

            if (p->options.testFlag(AutomaticTimeStamp))
            {
                QString timeStamp = KGlobal::locale()->formatDate(QDate::currentDate(), KLocale::ShortDate);
                p->scores[groupKey][i][Date] = timeStamp;
            }

            p->groupOfLatestAddition = groupKey;
            p->rankOfLatestAddition = i + 1;
            p->waitingForName = ! p->options.testFlag(DoNotAskName);

            p->saveScores(groupKey);

            emit newScore(p->groupOfLatestAddition, p->rankOfLatestAddition);

            return i + 1;
        }
    }
    return 0;
}

int KScoreManager::submitScore(int score, const QString& groupKey)
{
    KScoreManager::ScoreEntry scoreObject;
    scoreObject[Score] = score;
    return submitScore(scoreObject, groupKey);
}

bool KScoreManager::nameHasBeenRequested() const
{
    return p->waitingForName;
}

void KScoreManager::submitName(const QString & newName)
{
    if ( p->waitingForName )
    {
        kDebug() << "Name submitted:" << newName;
        p->lastEnteredName = newName;

        ScoreEntry & score = p->scores[p->groupOfLatestAddition][p->rankOfLatestAddition-1];
        if (newName != score.value(Name))
        {
            score[Name] = newName;
            p->saveScores(p->groupOfLatestAddition);
        }
        p->waitingForName = false;
    }
}

const QMap<QString, QString> KScoreManager::groupNames() const
{
    return p->translatedGroupNames;
}

const QList<KScoreManager::Field> KScoreManager::visibleFields() const
{
    if (!p->setupHasBeenFinalized)
    {
        p->finalizeSetup();
    }

    return p->visibleFields;
}

const QMap<KScoreManager::Field, QString> KScoreManager::fieldHeaders() const
{
    if (!p->setupHasBeenFinalized)
    {
        p->finalizeSetup();
    }

    return p->headers;
}

const QList<QString> KScoreManager::groups() const
{
    if (!p->setupHasBeenFinalized)
    {
        p->finalizeSetup();
    }

    if (!p->loaded)
    {
        p->loadAllScores();
    }

    return p->scores.keys();
}

const KScoreManager::ScoreList KScoreManager::scores(const QString& groupKey) const
{
    kDebug() << "Score list requested for" << groupKey;

    if (!p->setupHasBeenFinalized)
    {
        p->finalizeSetup();
    }

    if (!p->loaded)
    {
        p->loadAllScores();
    }

    if (!p->scores.contains(groupKey))
    {
        p->loadScores(groupKey);
    }

    return p->scores[groupKey];
}

QPair<QString, int> KScoreManager::newestScore() const
{
    return qMakePair(p->groupOfLatestAddition, p->rankOfLatestAddition);
}

#include "kscoremanager.moc"
