/****************************************************************
Copyright 2008  Matt Williams matt@milliams.com
Copyright 2008  Parker Coates parker.coates@gmail.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) version 3, or any
later version accepted by the membership of KDE e.V. (or its
successor approved by the membership of KDE e.V.), which shall
act as a proxy defined in Section 6 of version 3 of the license.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************/

#ifndef KSCOREMANGER_H
#define KSCOREMANGER_H

#include <libkdegames_export.h>

#include <QtCore/QString>
#include <QtCore/QMap>
#include <QtCore/QPair>
class QVariant;

// TODO
// - clear highscores (Shouldn't this be KHighscore's responsibility?)
// - provide a way to seed data for time

/**
 * \class KScoreManager kscoremanager.h <KScoreManager>
 *
 * \brief A simple high score implementation
 *
 * This class acts as a highscore manager for your game. It provides function
 * for adding scores to various groups as well as displaying the resultant
 * highscore table.
 *
 * <h4>Basic Operation</h4>
 *
 * It is a singleton and so requires no setup for the basic tasks. At the
 * simplest, you can show the highscore table with
 * \code
 * KScoreBoardDialog board();
 * board.show();
 * \endcode
 * This will default to showing the Name and Score for the default score group.
 * To add a score, it is as simple as calling
 * \code
 * KScoreManager::self()->submitScore(playersScore);
 * \endcode
 * This will (by default) prompt the player to enter their name. If you want to
 * display the table to the player to show them how they did you can call
 * \code
 * KScoreBoardDialog board();
 * board.show(KScoreDialog::HighlightLastScore | KScoreDialog::DisplayMessage);
 * \endcode
 * which will display the table with the most recently added score highlighted
 * and a message congratulating/commiserating the player for their performance.
 *
 * <h4>Passing Detailed Information</h4>
 *
 * You can pass in more information than just the player's score by passing
 * addScore() a ScoreInfo object. You can build up the ScoreInfo object with
 * the relevant information and then submit it:
 * \code
 * KScoreManager::ScoreInfo scoreInfo;
 * scoreInfo[KScoreDialog::Name] = playerName; //the player's name string
 * scoreInfo[KScoreDialog::Score] = playerScore; //the player's score (int)
 * KScoreManager::self()->submitScore(scoreInfo);
 * \endcode
 *
 * If you want to display more than just name and score (e.g. level or time
 * taken) you can do the following:
 * \code
 * KScoreManager::self()->setDisplayFields(KScoreDialog::Name |
 *                                         KScoreDialog::Score |
 *                                         KScoreDialog::Level);
 *
 * KScoreManager::ScoreInfo scoreInfo;
 * scoreInfo[KScoreDialog::Name] = playerName;
 * scoreInfo[KScoreDialog::Score] = playerScore;
 * scoreInfo[KScoreDialog::Level] = playerLevel; //the level the player got to
 * KScoreManager::self()->addScore(scoreInfo);
 *
 * KScoreBoardDialog board();
 * board.show(KScoreDialog::HighlightLastScore | KScoreDialog::DisplayMessage);
 * \endcode
 *
 * <h4>Custom Fields</h4>
 *
 * If you wish to define your own custom fields, you can do so with the
 * addField() function. You add it into one of the custom fields and set the UI
 * string. Note that you must both call addField() and set it to be displayed
 * with setDisplayFields() if you want the field to show up in the dialog.
 * \code
 * //This defines the new field
 * KScoreManager::self()->addField(KScoreDialog::Custom1, i18n("Moves Taken"));
 *
 * KScoreManager::self()->setDisplayFields(KScoreDialog::Name |
 *                                         KScoreDialog::Score |
 *                                         KScoreDialog::Custom1);
 *
 * KScoreManager::ScoreInfo scoreInfo;
 * scoreInfo[KScoreDialog::Name] = playerName;
 * scoreInfo[KScoreDialog::Score] = playerScore;
 * scoreInfo[KScoreDialog::Custom1] = numberOfMovesTaken;
 * KScoreManager::self()->addScore(scoreInfo);
 *
 * KScoreBoardDialog board();
 * board.show(KScoreDialog::HighlightLastScore | KScoreDialog::DisplayMessage);
 * \endcode
 *
 * <h4>Grouping Scores</h4>
 *
 * So far all the examples have been adding the score to the default group. You
 * can group your scores (e.g. based on the difficulty of the level). Note that
 * you must always addScoreGroupName() for all groups that could exist in your
 * highscore file so that when reading the highscores from file, the
 * translations are available.
 * \code
 * KScoreManager::self()->addScoreGroupName("easy", i18n("Easy"));
 * KScoreManager::self()->addScoreGroupName("medium", i18n("Medium"));
 * KScoreManager::self()->addScoreGroupName("hard", i18n("Hard"));
 *
 * //Using the simple addScore() function that takes an int
 * KScoreManager::self()->addScore(playerScore, "easy");
 *
 * KScoreBoardDialog board();
 * board.show(KScoreDialog::HighlightLastScore | KScoreDialog::DisplayMessage);
 * \endcode
 *
 * <h4>Specific Settings</h4>
 *
 * There are two functions which globally set properties:
 * \code
 * KScoreManager::self()->setOption(KScoreManager::DoNotAskName |
 *                                  KScoreManager::LowerScoreIsBetter);
 * \endcode
 * This will make the UI never ask the user for their name. Make sure you pass
 * a name via a ScoreInfo call to addScore() or the default name from KUser
 * will be used. The second flag will make lower scores be placed higher on the
 * scoreboard than higher scores.
 *
 * <h4>Outcome Messages</h4>
 *
 * Finally, you can set custom messages to be displayed when the dialog is
 * shown. There are three possible situations define by OutcomeFlag and you can
 * set a custom message for any of them. All the outcomes have default messages,
 * this is only for overriding them.
 * \code
 * KScoreDialog::self()->setMessage(KScoreDialog::TopScore,
 *                                  i18n("You got the highest score ever!"));
 * KScoreDialog::self()->setMessage(KScoreDialog::NoHighscore,
 *                                  QString()); //No message will be displayed
 *                                              //for this outcome
 * \endcode
 *
 *
 * \since KDE 4.3
 * \author Matt Williams <matt@milliams.com>
 * \author Parker Coates <parker.coates@gmail.com>
 */
class KDEGAMES_EXPORT KScoreManager : public QObject
{
    Q_OBJECT
    public:
        /**
         * \enum Option
         *
         * \brief General option flags
         *
         * This provides the values for the \a Options flag set
         *
         * \see setOptions()
         */
        enum Option
        {
            DoNotAskName = 0x1, ///<If set, the user will not be able to change the name after the score has been submitted.
            LowerScoresAreBetter = 0x2, ///<If set, the scores will be sorted in ascending order instead of the default descending order.
            AutomaticTimeStamp = 0x4
        };
        Q_DECLARE_FLAGS(Options, Option)

        /**
         * \enum Field
         *
         * \brief Highscore fields
         *
         * Perhaps instead of enums here which each match to a key string which
         * also match a translated string, I could just use key and translation
         * mapping. This would allow for almost completely free key additions.
         *
         * This provides the values for the \a Fields flag set
         *
         * \see addField()
         */
        enum Field
        {
            Score, ///<The player's score
            Name, ///<Name of the player
            Date, ///<The date (and time?) when the score was recorded
            Level, ///<The level of the game achieved
            Time, ///<The time taken to complete the game or level
            //TODO: replace the Custom{1-5} fields by some UserField (similar to Qt::UserRole or QGraphicsItem::UserType)
            Custom1, ///<Field for custom information
            Custom2, ///<Field for custom information
            Custom3, ///<Field for custom information
            Custom4, ///<Field for custom information
            Custom5  ///<Field for custom information
        };

        /**
         * \typedef ScoreEntry
         *
         * A map of key and values for the score
         *
         * \see submitScore()
         */
        typedef QMap<Field, QVariant> ScoreEntry;

        /**
         * \typedef ScoreList
         *
         * A list of score objects
         */
        typedef QList<ScoreEntry> ScoreList;

        /**
         * \brief The entry point for the class.
         */
        static KScoreManager* self();

        /**
         * \name Settings and Translations
         */
        //@{
        /**
         * \brief Add a translated group name
         *
         * Add a score group name. If you wish to use any score grouping, all
         * groups must be added using either this function or addScoreGroupNames().
         * Translations must be added for each existing score group before the
         * first call to addScore() or scores(). If a new group is created at runtime,
         * be sure to add its translation before adding any scores to it.
         *
         * \param groupKey a unique key used to define the group. This must not
         *                 be a translated string.
         * \param groupTitle the i18n'd string used for the UI
         */
        void addScoreGroupName(const QString& groupKey, const QString& displayName);
        /**
         * \brief Add a list of translated group names
         *
         * Add a list of score groups. A convenience function.
         *
         * \param groups a QMap of group keys to group titles
         *
         * \see addScoreGroupName
         */
        void addScoreGroupNames(const QMap<QString, QString>& groups);

        /**
         * \brief Add a standard field
         *
         * Add one of the standard fields using a standard heading and key.
         * All added fields are stored in the config file, but only those that
         * are visible are shown in the UI. Fields will be displayed in the
         * order that they are added. The Name and Score fields are mandatory
         * and will be added automatically should you fail to do so. All fields
         * must be added before the first call to addScore() or scores().
         *
         * \param field the field to add
         * \param visible the visibility of the field in the UI
         *
         * \see addCustomField
         */
        void addField(const Field& field, bool visible=true);
        /**
         * \brief Add a custom field
         *
         * Add a custom field. Like addField, but allows customization of the
         * header and key.
         *
         * Generally, the standard fields should be added using addField and fields
         * Custom1 to Custom5 should be added with addCustomField.
         *
         * \param field the field to add
         * \param header the translated string used to label the field in the UI
         * \param key the key used to store the field in the config file
         * \param visible the visibility of the field in the UI
         *
         * \see addField
         */
        void addCustomField(const Field& field, const QString& header, const QString& key, bool visible=true);

        /**
         * \brief Set various options
         *
         * The class defaults to asking the player for their name and making
         * higher scores worth more. You can change these by respectively
         * passing KScoreManager::DoNotAskName and KScoreManager::LowerScoresAreBetter.
         * Any call to this function overrides any previous call and so calling
         * setOption(KScoreManager::Option()) at any time will reset the options
         * to the default values.
         *
         * \param options a bitwise OR of the Option flags you wish to set
         */
        void setOptions(const Options & options);
        /**
         * \note Perhaps this is not needed. It was only used for checking
         * AskName but if nameHasBeenRequested() is true then maybe we can assume
         * this is true.
         * \return the currently set options
         */
        Options options() const;
        /**
         * \brief Set the number of scores to save and display
         *
         * \param numberOfScores the number of scores to save and display on the board
         */
        void setScoreBoardLength(int numberOfScores);
        /**
         * \brief Retrieve the number of scores to save and display
         *
         * \return the number of scores being saved and displayed on the board
         */
        int scoreBoardLength() const;
        //@}

        /**
         * \name Score Manipulation
         * Functions for adding and removing scores
         */
        //@{
        /**
         * \brief Submit a new score to the manager
         *
         * If no group is specified, the score will be added to the default
         * group. If the KScoreManager::Name field is set in the score and the
         * player is promted for their name, the  contents of the
         * KScoreManager::Name field will be used to prefill the name entry text
         * box
         *
         * \param score a ScoreEntry detailing the information about the new score
         * \param groupKey the group to add the new score to. This must match
         *                 one of the keys you passed to addScoreGroupName() or
         *                 addScoreGroupNames()
         *
         * \return the highscore position if the score was good enough to make
         *         it into the list (1 being topscore) or 0 otherwise.
         */
        int submitScore(const ScoreEntry& score, const QString& groupKey = QString());
        /**
         * \brief Submit a new score to the mananger
         *
         * If no group is specified, the score will be added to the default
         * group
         *
         * \param score the score the player achieved. This will be added to
         *              the KScoreManager::Score field
         * \param groupKey the group to add the new score to. This must match
         *                 one of the keys you passed to addScoreGroupName() or
         *                 addScoreGroupNames()
         *
         * \return the highscore position if the score was good enough to make
         *         it into the list (1 being topscore) or 0 otherwise.
         */
        int submitScore(int score, const QString& groupKey = QString());
        //@}

    Q_SIGNALS:
        void newScore(QString groupKey, int rank);

    private:
        /**
         * \name Functions for use by the UI
         */
        //@{
        /**
         * \brief Translations for the group names
         */
        const QMap<QString, QString> groupNames() const;
        /**
         * \brief The visible fields in the order they are to be displayed in the UI
         */
        const QList<Field> visibleFields() const;
        /**
         * \brief Translations for the field headers
         */
        const QMap<Field, QString> fieldHeaders() const;
        const QList<QString> groups() const;
        /**
         * \brief Maps the group key the list of scores for that group
         */
        const ScoreList scores(const QString& groupKey = QString()) const;
        /**
         * \brief Group and index of the most recently added score
         */
        QPair<QString, int> newestScore() const;
        /**
         * \return whether the score specified by newestScore() is expecting a name to be entered
         */
        bool nameHasBeenRequested() const;
        /**
         * This has no effect if nameHasBeenRequested() is false
         * \param newName the name to be put in the slot specified by newestscore()
         * \post nameHasBeenRequested() will be set to false
         */
        void submitName(const QString& newName);

        friend class KScoreBoardBase;
        //@}

        KScoreManager(); ///<Private constructor since it's a singleton
        ~KScoreManager(); ///<Private destructor since it's a singleton

        friend class KScoreManagerPrivate;

        Q_DISABLE_COPY(KScoreManager)
};

Q_DECLARE_OPERATORS_FOR_FLAGS(KScoreManager::Options)

#endif //KSCOREMANGER_H
