/*
Copyright (c) 2006, 2007, Nicolas Roffet, <nicolas-kde@roffet.com>

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU Library General Public License along with this program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
*/

#ifndef KSIMILICARD_H
#define KSIMILICARD_H



#include <QByteArray>
class QMouseEvent;
#include <QtSvg/QSvgWidget>
class QWidget;


class KSimiliBoard;



/**
 * @brief Widget that represents a card.
 *
 * Each card is made of a generated SVG file. 
 *    - http://www.w3.org/Graphics/SVG/
 *    - http://en.wikipedia.org/wiki/Scalable_Vector_Graphics
 * 
 * There are differents possible states:
 *    - normal
 *    - selected
 *    - highlighted because in a combination
 *    - highlighted because not in a combination
 */
class KSimiliCard : public QSvgWidget
{
	Q_OBJECT

	public:
		/**
		 * @enum cardForm Form of the card.
		 */
		enum cardForm { ellipse, rectangle, losange, specialForm };

		/**
		 * @enum cardFilling Filling type of the card.
		 */
		enum cardFilling { opaque, shading, transparent, specialFilling };

		/**
		 * @brief First constructor of a card with the card properties.
		 * 
		 * @param form The form of the card symbol(s) (0, 1 or 2).
		 * 
		 * @param color The color of the card symbol(s) (0, 1 or 2).
		 * 
		 * @param number The number of the card symbols (1, 2 or 3).
		 * 
		 * @param filling The filling art of the card symbol(s) (0, 1 or 2).
		 */
		KSimiliCard(QWidget *parent, KSimiliBoard *board, const cardForm form, const int color, const  int number, const cardFilling filling, const int position);

		/**
		 * @brief Second constructor of a card with the card ID.
		 * 
		 */
		KSimiliCard(QWidget *parent, KSimiliBoard *board, const int cardId, const int position);

		/**
		 * @brief Set the unique identifier of the card 
		 * 
		 * @param newCardId Identifier of the card: value between 0 and 80 of the Card.
		 * There are indeed 81 different cards in the game, because there are 4 properties and 3 values for each property: 
		 * @f${{property}_{value}}^{{property}_{number}} = 3^4 = 81@f$.
		 *
		 * This function can be used to reuse the widget "card" with another card, without to destroy it and to create a new one.
		 */
		void setNewCardId(const int newCardId);


		/**
		 * @brief Get the unique identifier of the card 
		 * @see setNewCardId
		 */
		int getCardId();


		
		cardForm form;
		// 0, 1 or 2
		int color;
		// 1, 2 or 3
		int number;
		// 0, 1 or 2
		cardFilling filling;

		/**
		 * @brief Position on the board. 
		 * 
		 * Needed by destroing the card to know where there are some free positions.
		 */
		int position;

	signals:
		/**
		 * @brief The state of the card changed.
		 * @param selected If the card is selected, then true. If not: false.
		 */
		void selectionChanged(bool selected);
	
	public slots:
		/**
		 * @brief Change the state of the card.
		 * 
		 * @param selected
		 */
		void selectCard(bool selected);

		/**
		 * @brief Tell the card to display itself as "highligted" (found in a combination) or not.
		 *
		 * In this case, the card is not selected anymore.
		 */
		void highlightOk(bool highlighted);

		/**
		 * @brief Tell the card to display itself as "highligted for error" (not found in a combination) or not.
		 *
		 * In this case, the card is not selected anymore.
		 */
		void highlightNotOk(bool highlighted);

		/**
		 * @brief Tell the card to display itself as "highligted as hint" (found in a combination) or not.
		 *
		 * In this case, the card is not selected anymore.
		 */
		void highlightHint(bool highlighted);


	private:
		void createSvg(const cardForm form, const int color, const  int number, const cardFilling filling);
		void mousePressEvent ( QMouseEvent * event );
		void updateSvg();

		QByteArray getXmlDefs(const QByteArray& color, cardFilling filling);
		QByteArray getSvgStyleFill(const QByteArray& color, cardFilling filling);
		QByteArray getXmlForm(cardForm form, const QByteArray& color, cardFilling filling, int offsetX);

		/**
		 * m_cardId=-1 if nothing on the card (special case)
		 */
		int m_cardId;
		bool m_selected;
		bool m_highlightedOk;
		bool m_highlightedNotOk;
		bool m_highlightedHint;
		KSimiliBoard *m_board;


		QByteArray m_cardSvg;
		QByteArray m_selectedSvg;
		QByteArray m_partOfKSimiliSvg;
		QByteArray m_notPartOfKSimiliSvg;
		QByteArray m_hintSvg;
};

#endif //KSIMILICARD_H
