//Magazynier, a warehouse assistant game.
//Copyright (C) 2008, 2009 Tadeusz Andrzej Kadlubowski yess@hell.org.pl
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License along
//with this program; if not, write to the Free Software Foundation, Inc.,
//51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

#include "window.h"

#include <math.h>

#include <KAction>
#include <KActionCollection>
#include <KConfig>
#include <KConfigDialog>
#include <KConfigGroup>
#include <KDebug>
#include <KGameClock>
#include <KGamePopupItem>
#include <KGameThemeSelector>
#include <KGlobal>
#include <KInputDialog>
#include <KLocale>
#include <KMenu>
#include <KMenuBar>
#include <KMessageBox>
#include <KScoreDialog>
#include <KStandardDirs>
#include <KStandardGameAction>
#include <KStatusBar>
#include <KUndoStack>
#include <QActionGroup>
#include <QGraphicsItem>
#include <QGraphicsView>
#include <QMutex>

#include "commands.h"
#include "levelModel.h"
#include "levelPack.h"
#include "renderer.h"
#include "settings.h"

MagazynierWindow::MagazynierWindow(QWidget *parent) : //{{{
	KXmlGuiWindow(parent),
	m_levelPack(NULL),
	m_levelModel(NULL),
	m_gameClockRunning(false)
{	
	m_gamePopupItem = new KGamePopupItem();
	m_gamePopupItem->setMessageTimeout(5000);

	m_view = new QGraphicsView();
	m_view->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
	m_view->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);

	m_undoStack = new KUndoStack(this);
	setupActions();

	setCentralWidget(m_view);

	statusBar()->insertPermanentItem("", 1); //LevelPack: 42/69
	statusBar()->insertPermanentItem("", 2); //Moves: 1337
	statusBar()->insertPermanentItem("", 3); //Shifts: 7331
	statusBar()->insertPermanentItem("", 4); //Time: 13:37

	m_gameClock = new KGameClock(this, KGameClock::MinSecOnly);
	connect(m_gameClock, SIGNAL(timeChanged(const QString&)), SLOT(updateTimer(void)));

	setupGUI();
	
	Renderer::instance()->setViewSize(m_view->height(), m_view->width());

	findLevelPacks();
	setupLastPlayedLevelPack();
	setupLevelPackMenu();	

} //}}}

void MagazynierWindow::setupLastPlayedLevelPack(void) //{{{
{
	KConfig config;
	KConfigGroup general = config.group("General");
	QString lastLevelPackName = general.readEntry("LastLevelPack");
	if (lastLevelPackName.isNull()) {
		lastLevelPackName = QString("Sasquatch");
	}

	LevelPack *newLevelPack;
	
	if (m_levelPacks[lastLevelPackName] && !m_levelPacks[lastLevelPackName]->getName().isNull()) {
		newLevelPack = m_levelPacks[lastLevelPackName];
	} else {
		KMessageBox::error(this, i18n("Could not find level collection \"%1\". Please check your installation.", lastLevelPackName));
		exit(EXIT_FAILURE);
	
	}
	
	levelPackChanged(newLevelPack);
	return;
} //}}}

void MagazynierWindow::findLevelPacks(void) //{{{
{
	QStringList levelPackNames;
	KGlobal::dirs()->findAllResources("data", QString("*/1.txt"),
			KStandardDirs::Recursive|KStandardDirs::NoDuplicates, levelPackNames);

	if (levelPackNames.size() == 0) {
		kDebug() << "No level collections.";
		KMessageBox::error(this, i18n("No level collections found. Please check your installation."));
		exit(EXIT_FAILURE);
	}

	for (int i = 0; i < levelPackNames.size(); ++i) {
		LevelPack *lp = new LevelPack(levelPackNames[i]);
		m_levelPacks[lp->getName()] = lp;
		connect(lp, SIGNAL(changeLevelPack(LevelPack *)),
			this, SLOT(levelPackChanged(LevelPack *)));
	}
	
	return;
} //}}}

void MagazynierWindow::setupLevelPackMenu(void) //{{{
{
	QActionGroup *m_levelPackActionGroup = new QActionGroup(this);
	QMap<QString, LevelPack *>::const_iterator i;

	for (i = m_levelPacks.constBegin(); i != m_levelPacks.constEnd(); ++i) {
		KAction *levelPackAction = new KAction(m_levelPackActionGroup);
		levelPackAction->setCheckable(true);
		levelPackAction->setText(i.key());
		if (i.value() == m_levelPack) {
			levelPackAction->setChecked(true);
		}
		connect(levelPackAction, SIGNAL(triggered(void)), 
			i.value(), SLOT(thisLevelPackSelected(void)));

		m_levelPackActionGroup->addAction(levelPackAction);
	}
	plugActionList("level-pack-actions", m_levelPackActionGroup->actions());
}
//}}}

void MagazynierWindow::setupActions(void) //{{{
{
	KStandardGameAction::quit(this, SLOT(close()), actionCollection());

	KStandardGameAction::highscores(this, SLOT(showHighscore()),
		actionCollection());

	m_undoStack->createUndoAction(actionCollection(), "");
	m_undoStack->createRedoAction(actionCollection(), "");

	KAction* prevLevelAction = new KAction(this);
	prevLevelAction->setText(i18n("Previous Level"));
	prevLevelAction->setIcon(KIcon("arrow-left"));
	prevLevelAction->setShortcut(QKeySequence(Qt::ALT | Qt::Key_Left));
	actionCollection()->addAction("prev-level", prevLevelAction);

	KAction* nextLevelAction = new KAction(this);
	nextLevelAction->setText(i18n("Next Level"));
	nextLevelAction->setIcon(KIcon("arrow-right"));
	nextLevelAction->setShortcut(QKeySequence(Qt::ALT | Qt::Key_Right));
	actionCollection()->addAction("next-level", nextLevelAction);

	KAction* reloadLevelAction = new KAction(this);
	reloadLevelAction->setText(i18n("Reload Level"));
	reloadLevelAction->setIcon(KIcon("view-refresh"));
	reloadLevelAction->setShortcut(QKeySequence(Qt::ALT | Qt::Key_R));
	actionCollection()->addAction("reload-level", reloadLevelAction);

	KAction* changeLevelAction = new KAction(this);
	changeLevelAction->setText(i18n("Change level..."));
	connect(changeLevelAction, SIGNAL(triggered()), this, SLOT(changeLevel()));
	actionCollection()->addAction("change-level", changeLevelAction);

	KAction *moveUpAction = new KAction(this);
	moveUpAction->setText(i18n("Move up"));
	moveUpAction->setShortcut(QKeySequence(Qt::Key_Up));
	actionCollection()->addAction("move-up", moveUpAction);
	
	KAction *moveDownAction = new KAction(this);
	moveDownAction->setText(i18n("Move down"));
	moveDownAction->setShortcut(QKeySequence(Qt::Key_Down));
	actionCollection()->addAction("move-down", moveDownAction);

	KAction *moveLeftAction = new KAction(this);
	moveLeftAction->setText(i18n("Move left"));
	moveLeftAction->setShortcut(QKeySequence(Qt::Key_Left));
	actionCollection()->addAction("move-left", moveLeftAction);

	KAction *moveRightAction = new KAction(this);
	moveRightAction->setText(i18n("Move right"));
	moveRightAction->setShortcut(QKeySequence(Qt::Key_Right));
	actionCollection()->addAction("move-right", moveRightAction);

	KStandardAction::preferences(this, SLOT(configureSettings()), actionCollection());

	return;
} //}}}

void MagazynierWindow::updateLevelChangeActions(void)//{{{
{
	QAction *nextLevelAction = actionCollection()->action("next-level");
	nextLevelAction->setEnabled(m_levelPack->hasNextAvailableLevels());
	
	QAction *prevLevelAction = actionCollection()->action("prev-level");
	prevLevelAction->setEnabled(m_levelPack->hasPrevLevels());
}//}}}

void MagazynierWindow::connectMoveActions(void) //{{{
{
	QAction *moveRightAction = actionCollection()->action("move-right");
	connect(moveRightAction, SIGNAL(triggered()),
		m_levelModel, SLOT(moveRight()));

	QAction *moveLeftAction = actionCollection()->action("move-left");
	connect(moveLeftAction, SIGNAL(triggered()),
		m_levelModel, SLOT(moveLeft()));
		
	QAction *moveUpAction = actionCollection()->action("move-up");
	connect(moveUpAction, SIGNAL(triggered()),
		m_levelModel, SLOT(moveUp()));

	QAction *moveDownAction = actionCollection()->action("move-down");
	connect(moveDownAction, SIGNAL(triggered()),
		m_levelModel, SLOT(moveDown()));
} //}}}

void MagazynierWindow::disconnectMoveActions(void) //{{{
{
	QAction *moveRightAction = actionCollection()->action("move-right");
	disconnect(moveRightAction, SIGNAL(triggered()),
		m_levelModel, SLOT(moveRight()));

	QAction *moveLeftAction = actionCollection()->action("move-left");
	disconnect(moveLeftAction, SIGNAL(triggered()),
		m_levelModel, SLOT(moveLeft()));
		
	QAction *moveUpAction = actionCollection()->action("move-up");
	disconnect(moveUpAction, SIGNAL(triggered()),
		m_levelModel, SLOT(moveUp()));

	QAction *moveDownAction = actionCollection()->action("move-down");
	disconnect(moveDownAction, SIGNAL(triggered()),
		m_levelModel, SLOT(moveDown()));
} //}}}

void MagazynierWindow::levelChanged(LevelModel *newLevelModel) //{{{
{
	kDebug() << "New level is loaded";

	m_gamePopupItem->forceHide();

	LevelModel *oldLevelModel = m_levelModel;
	m_levelModel = newLevelModel;

	m_view->setScene(m_levelModel);

	m_levelModel->addItem(m_gamePopupItem);

	redrawNeeded();

	updateLevelChangeActions();
	connectMoveActions();

	m_undoStack->clear();
	connect(newLevelModel, SIGNAL(addNewUndoCommand(MagazynierCommand *)),
		this, SLOT(newUndoCommandCreated(MagazynierCommand *)));

	connect(newLevelModel, SIGNAL(levelCompleted(void)),
		m_levelPack, SLOT(markCurrentLevelAsSolved(void)));
	connect(newLevelModel, SIGNAL(levelCompleted(void)),
		this, SLOT(updateLevelChangeActions(void)));
	connect(newLevelModel, SIGNAL(levelCompleted(void)),
		m_gameClock, SLOT(pause(void)));
	connect(newLevelModel, SIGNAL(levelCompleted(void)),
		this, SLOT(addHighscore(void)));
	connect(newLevelModel, SIGNAL(levelCompleted(void)),
		this, SLOT(disconnectMoveActions(void)));
	connect(newLevelModel, SIGNAL(levelCompleted(void)),
		this, SLOT(showLevelCompletedPopup(void)));

	statusBar()->changeItem(
		i18n(
			"%1 Level: %2/%3",
			m_levelPack->getName(),
			m_levelPack->getCurrentLevelNumber(),
			m_levelPack->getLevelCount()
		), 1);
	statusBar()->changeItem(i18n("Moves: %1", 0), 2);
	statusBar()->changeItem(i18n("Shifts: %1", 0), 3);

	m_gameClock->setTime(0);
	m_gameClock->pause();
	m_gameClockRunning = false;
	updateTimer();

	connect(newLevelModel, SIGNAL(moveCountChanged(int)),
		this, SLOT(updateMoveCounter(int)));
	connect(newLevelModel, SIGNAL(shiftCountChanged(int)),
		this, SLOT(updateShiftCounter(int)));

	delete oldLevelModel;

	return;
} //}}}

void MagazynierWindow::updateMoveCounter(int i) //{{{
{
	statusBar()->changeItem(i18n("Moves: %1", i), 2);
} //}}}

void MagazynierWindow::updateTimer(void) //{{{
{
	statusBar()->changeItem(i18n("Time: %1", m_gameClock->timeString()), 4);
} //}}}

void MagazynierWindow::updateShiftCounter(int i) //{{{
{
	statusBar()->changeItem(i18n("Shifts: %1", i), 3);
} //}}}

void MagazynierWindow::newUndoCommandCreated(MagazynierCommand *command) //{{{
{
	kDebug() << "New command pushed onto undo stack.";
	m_undoStack->push(command);

	if (not m_gameClockRunning) {
		m_gameClockRunning = true;
		m_gameClock->restart();
		m_gameClock->setTime(1);
	}
	return;
} //}}}

void MagazynierWindow::redrawNeeded(void) //{{{
{
	Renderer::instance()->setViewSize(m_view->height(), m_view->width());
	
	m_levelModel->populateScene();
	
	return;
} //}}}

void MagazynierWindow::levelPackChanged(LevelPack *newLevelPack) //{{{
{
	Q_ASSERT(newLevelPack != NULL);
	LevelPack *oldLevelPack = m_levelPack;

	m_levelPack = newLevelPack;
	
	if (oldLevelPack != NULL) {
		disconnect(oldLevelPack, SIGNAL(levelLoaded(LevelModel *)),
			   this, SLOT(levelChanged(LevelModel *)));
	}
	connect(m_levelPack, SIGNAL(levelLoaded(LevelModel *)),
		this, SLOT(levelChanged(LevelModel *)));

	QAction *prevLevelAction = actionCollection()->action("prev-level");
	if (oldLevelPack != NULL) {
		disconnect(prevLevelAction, SIGNAL(triggered()),
			   oldLevelPack, SLOT(loadPrevLevel()));
	}
	connect(prevLevelAction, SIGNAL(triggered()),
			 m_levelPack, SLOT(loadPrevLevel()));

	QAction *nextLevelAction = actionCollection()->action("next-level");
	if (oldLevelPack != NULL) {
		disconnect(nextLevelAction, SIGNAL(triggered()),
			   oldLevelPack, SLOT(loadNextLevel()));
	}
	connect(nextLevelAction, SIGNAL(triggered()),
		m_levelPack, SLOT(loadNextLevel()));

	QAction *reloadLevelAction = actionCollection()->action("reload-level");
	if (oldLevelPack != NULL) {
		disconnect(reloadLevelAction, SIGNAL(triggered()),
			   oldLevelPack, SLOT(reloadCurrentLevel()));
	}
	connect(reloadLevelAction, SIGNAL(triggered()),
			 m_levelPack, SLOT(reloadCurrentLevel()));
	
	m_levelPack->loadFirstUnsolvedLevel();
	return;
} //}}}

void MagazynierWindow::resizeEvent(QResizeEvent *event) //{{{
{
	if (m_levelModel != NULL) {
		redrawNeeded();
	}
	KXmlGuiWindow::resizeEvent(event);
} //}}}

void MagazynierWindow::configureSettings(void) //{{{
{
	KConfigDialog *dialog = new KConfigDialog(this, "settings", Settings::self());
	dialog->addPage(new KGameThemeSelector(dialog, Settings::self(),
			KGameThemeSelector::NewStuffDisableDownload), i18n("Theme"), "games-config-theme");

	connect(dialog, SIGNAL( settingsChanged(const QString&)), this,
			SLOT(reloadTheme()));

	dialog->show();

	return;
} //}}}

void MagazynierWindow::reloadTheme() //{{{
{
	Renderer::instance()->loadTheme();
	if(m_levelModel != NULL) {
		m_levelModel->invalidate();
		redrawNeeded();
	}
} //}}}

void MagazynierWindow::showHighscore(void) //{{{
{
	KScoreDialog scoreDialog(KScoreDialog::Name | KScoreDialog::Time | KScoreDialog::Level, this);
	scoreDialog.addField(KScoreDialog::Custom1, i18n("Number of moves"), "moves");
	scoreDialog.addField(KScoreDialog::Custom2, i18n("Number of shifts"), "shifts");
	scoreDialog.addField(KScoreDialog::Custom3, i18n("Level pack"), "level-pack");
	scoreDialog.exec();
} //}}}

void MagazynierWindow::addHighscore(void) //{{{
{
	KScoreDialog::FieldInfo scoreInfo;
	scoreInfo[KScoreDialog::Time] = m_gameClock->timeString();
	scoreInfo[KScoreDialog::Level].setNum(m_levelPack->getCurrentLevelNumber());
	scoreInfo[KScoreDialog::Custom1].setNum(m_levelModel->getMoveCount());
	scoreInfo[KScoreDialog::Custom2].setNum(m_levelModel->getShiftCount());
	scoreInfo[KScoreDialog::Custom3] = m_levelPack->getName();
	scoreInfo[KScoreDialog::Score].setNum(12); //TODO: Think of something that makes sense as a score

	KScoreDialog scoreDialog(KScoreDialog::Name | KScoreDialog::Time | KScoreDialog::Level, this);
	scoreDialog.addField(KScoreDialog::Custom1, i18n("Number of moves"), "moves");
	scoreDialog.addField(KScoreDialog::Custom2, i18n("Number of shifts"), "shifts");
	scoreDialog.addField(KScoreDialog::Custom3, i18n("Level pack"), "level-pack");

	if(scoreDialog.addScore(scoreInfo) != 0) {
		scoreDialog.exec();
	}
} //}}}

void MagazynierWindow::showLevelCompletedPopup(void) //{{{
{
	m_gamePopupItem->showMessage(
			i18n("You have completed this level! Click on 'Next Level' to continue."),
			KGamePopupItem::Center);
} //}}}

void MagazynierWindow::changeLevel(void) //{{{
{
	bool ok;
	int newLevelNum = KInputDialog::getInteger(i18n("Change level"),
			i18n("Enter the number of the level you wish to change to"),
			m_levelPack->getCurrentLevelNumber(),
			1,
			m_levelPack->getSolvedLevelCount() + 1,
			1,
			&ok,
			this);

	if(ok) {
		m_levelPack->loadLevelByNumber(newLevelNum);
	}
} //}}}

// vim: foldmethod=marker

