/***************************************************************************
 *   Copyright (C) 2008 Copyright 2008  Tom Vollerthun <vollerthun@gmx.de>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "data.h"
#include <KDebug>
#include <Qt>

Data::Data(QObject *parent) : QObject(parent)
{
    m_currentColor = Qt::black;
    m_backGround = Qt::white;

    rowList = new QList< QList<HintEntry * > * > ();
    colList = new QList< QList<HintEntry * > * > ();
    m_matrix = new QList< QList<MatrixEntry * > * > ();


}

void Data::togglePosition(int x, int y)
{
    kDebug() << "Toggling Position [" << x << "/" << y << "]";
    MatrixEntry * m = entry(x, y);

    if (m->box_status == BOX_UNDEFINED || (
            (m->box_color != m_currentColor) && (m->box_color != m_backGround)
            ) ) {

        setOn(x, y);
    }
    else if (m->box_status == BOX_ON) {

        setOff(x, y);
    } else {

        setUndefined(x, y);
    }
}

void Data::toggle_UndefinedOn(int x, int y)
{
    MatrixEntry * m = entry(x, y);

    if (m->box_status == BOX_ON && m->box_color == m_currentColor) {
        setUndefined(x, y);
    }
    else {
        setOn(x, y);
    }
}

void Data::toggle_UndefinedOff(int x, int y)
{
    MatrixEntry * m = entry(x, y);

    if (m->box_status == BOX_OFF) {
        setUndefined(x, y);
    }
    else {
        setOff(x, y);
    }
}

void Data::continueLastChange(int x, int y)
{
    if (lastChange == BOX_ON) {
        setOn(x, y);
    } else if (lastChange == BOX_OFF) {
        setOff(x, y);
    } else if (lastChange == BOX_UNDEFINED) {
        setUndefined(x, y);
    }
}


void Data::setOn(int x, int y)
{
    MatrixEntry * m = entry(x, y);
    m->box_status = BOX_ON;
    m->box_color = m_currentColor;

    lastChange = BOX_ON;

    emit positionChanged(x, y);
    emit positionChanged();
}

void Data::setOff(int x, int y)
{
    MatrixEntry * m = entry(x, y);

    m->box_status = BOX_OFF;
    m->box_color = m_backGround;

    lastChange = BOX_OFF;

    emit positionChanged(x, y);
    emit positionChanged();
}

void Data::setUndefined(int x, int y)
{
    MatrixEntry * m = entry(x, y);

    m->box_status = BOX_UNDEFINED;
    m->box_color = m_backGround;

    lastChange = BOX_UNDEFINED;

    emit positionChanged(x, y);
    emit positionChanged();

}


void Data::resetGame()
{

    //reset header-hints (active/inactive)
    for (int x = 0; x < rowList->size(); ++x){
        for(int y = 0; y < rowList->at(x)->size(); y++){

            HintEntry * h = rowList->at(x)->at(y);
            h->active = true;
        }
    }

    for (int x = 0; x < colList->size(); ++x){
        for(int y = 0; y < colList->at(x)->size(); y++){

            HintEntry * h = colList->at(x)->at(y);
            h->active = true;
        }
    }

    //reset grid-boxes
    for (int x = 0; x < getColumnCount() ; ++x) {
        for (int y = 0; y < getRowCount() ; ++y) {

            setUndefined(x, y);
        }
    }



    kDebug() << "Game reset!";
    emit positionChanged();
}

void Data::setRowHints(int row, QList<HintEntry *> * values)
{
    rowList->replace(row, values);
}

void Data::setRowHints(QList< QList<Data::HintEntry *> *> * values)
{
    rowList = values;
}

void Data::setColumnHints(int column, QList<HintEntry *> * values)
{
    colList->replace(column, values);
}

void Data::setColumnHints(QList< QList<Data::HintEntry *> *> * values)
{
    colList = values;
}


QList<QColor> * Data::availableColors()
{
    return m_availableColors;
}

void Data::setAvailableColors(QList<QColor> * colors)
{
    m_availableColors = colors;
    emit colorsChanged(this);
}

void Data::update()
{

    /*remove empty rows and columns at the beginning and the end*/
    bool dropDone = false;
    for (int y = 0; y < rowList->size() && !dropDone; y++) {
        Data::HintEntry * hint = rowList->first()->first();
        if (hint->amount == 0) {
            rowList->removeFirst();
        } else {
            dropDone = true;
        }
    }

    for (int i = rowList->size() -1; i >=0; i--) {
        Data::HintEntry * hint = rowList->last()->first();
        if (hint->amount == 0) {
            rowList->removeLast();
        } else {
            dropDone = true;
        }
    }

    for (int x = 0; x < colList->size(); x++) {
        Data::HintEntry * hint = colList->first()->first();
        if (hint->amount == 0) {
            colList->removeFirst();
        } else {
            dropDone = true;
        }
    }

    for (int i = rowList->size() -1; i >=0; i--) {
        Data::HintEntry * hint = colList->last()->first();
        if (hint->amount == 0) {
            colList->removeLast();
        } else {
            dropDone = true;
        }
    }

    /*find all available colors*/
    QList<QColor> * availableColors = new QList<QColor>();
    for (int y = 0; y < rowList->size(); y++) {

        QList<HintEntry * > * row = rowList->at(y);
        for (int x = 0; x < row->size(); x++) {

            HintEntry * h = row->at(x);

            /*color of current pixel*/
            QColor color = h->hintColor;

            if (color != Qt::white && !availableColors->contains(color)){
                availableColors->append(color);
            }
        }
    }
    setAvailableColors(availableColors);

    m_matrix = new QList< QList<MatrixEntry * > * > ();

    /*fill the grid's content with undefined MatrixEntries.*/
    for (int x = 0; x<rowList->size(); x++) {

        QList<MatrixEntry*> * aList = new QList<MatrixEntry *> () ;
        for (int y = 0; y<colList->size(); y++) {

            MatrixEntry * m = new MatrixEntry();
            m->box_status = BOX_UNDEFINED;
            m->box_color = m_backGround;
            aList->append(m);
        }
        m_matrix->append(aList);
    }

}

QList<Data::HintEntry *> * Data::rowHints(int row)
{
    QList<HintEntry *> * retval = rowList->at(row);
    return retval;
}

QList<Data::HintEntry *> * Data::columnHints(int column)
{
    QList<HintEntry *> * retval = colList->at(column);
    return retval;
}

int Data::getRowCount()
{
    if (rowList == NULL) {
        return 0;
    }

    return rowList->size();
}

int Data::getColumnCount()
{
    if (colList == NULL) {
        return 0;
    }

    return colList->size();
}

QList< QList<Data::HintEntry *> *> * Data::rowHints()
{
    return rowList;
}

QList< QList<Data::HintEntry *> *> * Data::columnHints()
{
    return colList;
}


QList< QList<Data::MatrixEntry *> *> * Data::matrix()
{
    return m_matrix;
}

QString Data::toString(QList<HintEntry *> * entries)
{

    QString entries_String = "|";

    for (int i = 0; i < entries->size(); ++i) {

        HintEntry * entry = entries->at(i);

        if (entry->amount>0) {
            entries_String.append(QString::number(entry->amount) + ','
                + entry->hintColor.name() + '|');
        }
    }

    return entries_String;
}

QString Data::toString(QList< QList<HintEntry *> *> * entries)
{
    QString entries_String = "-";

    for (int i = 0; i < entries->size(); ++i) {

        QList<HintEntry *> * entry = entries->at(i);
        entries_String.append(toString(entry));
        entries_String.append("-");
    }

    return entries_String;
}

QString Data::toString(QList< QList<MatrixEntry *> *> * entries)
{
    QString entries_String = "-";

    for (int i = 0; i < entries->size(); ++i) {

        QList<MatrixEntry *> * aRow = entries->at(i);
        entries_String.append('|');

        for (int j = 0; j < aRow->size(); ++j) {
            MatrixEntry * matrixEntry = aRow->at(j);

            entries_String.append(QString::number(matrixEntry->box_status) + ','
                    + matrixEntry->box_color.name() + '|');
        }

        entries_String.append('-');
    }

    return entries_String;
}

QList< QList<Data::HintEntry *> *> * Data::fromString(const QString & hintString)
{

    QList< QList<Data::HintEntry *> *> *retval = new QList< QList<Data::HintEntry *> *>();

    QStringList hint_list_list = hintString.split('-', QString::SkipEmptyParts);
    for (int i = 0; i < hint_list_list.size(); ++i) {
        QString hint_list_string = hint_list_list.at(i);

        QList<Data::HintEntry*> * aRow = new QList<Data::HintEntry*>();

        QStringList hint_list = hint_list_string.split('|', QString::SkipEmptyParts);

        if (hint_list.size() != 0) {
            //hintlist contains at least one hint

            for (int j = 0; j < hint_list.size(); ++j) {

                QString hint_string = hint_list.at(j);

                if (!hint_string.isEmpty() && !hint_string.isNull()) {

                    QStringList hintEntryList = hint_string.split(',', QString::KeepEmptyParts);

                    Data::HintEntry * hint = new Data::HintEntry();
                    hint->amount = hintEntryList.at(0).toInt();
                    hint->hintColor = QColor(hintEntryList.at(1));
                    hint->active = true;
                    aRow->append(hint);
                }
            }
        } else {
            //hintlist doesn't contain any hint. Default to a "0" so there's at least something :)

            Data::HintEntry * hint = new Data::HintEntry();
            hint->amount = 0;
            hint->hintColor = Qt::black;
            hint->active = true;

            aRow->append(hint);
        }

        retval->append(aRow);
    }


    return retval;
}

/**
 * Get the Data::MatrixEntry at position x/y
 * @param x the column (horizontal - left to right)
 * @param y the row (vertical - top to bottom)
 */
Data::MatrixEntry * Data::entry(int x, int y) const
{
    if ( y >= m_matrix->size()
        || x >= m_matrix->at(0)->size() ) {

        kDebug() << "Accessing at [" << x << "/" << y << "]";
        kDebug() << "but matrix size is only : [" << m_matrix->at(0)->size() << "/" << m_matrix->size() << "]";
        kDebug() << "My name is " << m_description->name();
    }

    MatrixEntry * m = m_matrix->at(y)->at(x);
    return m;
}

void Data::setDescription(Description * description)
{
    m_description = description;
}

Description * Data::description()
{
    return m_description;
}

void Data::setActiveColor(const QColor & c)
{
    if (m_currentColor != c) {
        m_currentColor = c;
    }
}


void Data::connectNotify(const char* signal) {
QObject::connectNotify(signal);
    kDebug() << "Connected " << description()->name();
;}


void Data::disconnectNotify(const char* signal) {
QObject::disconnectNotify(signal);
    kDebug() << "Disconnected " << description()->name();
}

bool Data::event(QEvent* e) {
    bool retval = QObject::event(e);
    kDebug() << "Event for " << description()->name();
    return retval;
}

