/***************************************************************************
 *   Copyright (C) 2008 Copyright 2008  Tom Vollerthun <vollerthun@gmx.de>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "a_grid.h"
#include <KDebug>
#include <qevent.h>
#include <QColor>
#include <QTimer>
#include "ds/config.h"

#include <KComponentData>
#include <KConfig>
#include <KConfigGroup>
#include <KGlobal>

int Abstract_Grid::HORIZONTAL_LEFT_OFFSET=1;

int Abstract_Grid::VERTICAL_UPPER_OFFSET=1;

Abstract_Grid::Abstract_Grid(QWidget *parent) :
    QSvgWidget(parent)
{
    m_data = NULL;

    /*get mouse move events even if no button is pressed
     * This is for the highlighting of the active cell*/
    setMouseTracking(true);

    connect(this, SIGNAL(settingsChanged()), SLOT(readConfig()));

    selectedCell = new QPoint(-1, -1);
}

void Abstract_Grid::paintEvent(QPaintEvent * /*event*/)
{
    QPainter painter(this);

    calculatePositions();

    if (doPaintHorizontal()) {
        paintHorizontal(&painter);
    }

    if (doPaintVertical()) {
        paintVertical(&painter);
    }
}

void Abstract_Grid::mouseMoveEvent(QMouseEvent * event)
{
    QPoint * p = getGridPositions(event);

    if (! doPaintHorizontal() ) {
        p->setY(-1);
    }

    if (! doPaintVertical() ) {
        p->setX(-1);
    }

    /*if nothing has changed, there's no need to get all fussy.*/
    if ( selectedCell->x() == p->x() && selectedCell->y() == p->y() )
    {
        return;
    }

    kDebug() << "Moved to: [" << p->x() << "/" << p->y() << "]";
    emit cellSelected(p);

    if (autoselectClickColor) {
        QTimer * t = new  QTimer(this);
        t->setSingleShot(true);
        connect(t, SIGNAL(timeout()), this, SLOT(autoselectColor()));
        connect(this, SIGNAL(cellSelected(QPoint*)), t, SLOT(stop()));
        connect(this, SIGNAL(settingsChanged()), t, SLOT(stop()));
        t->start(waitBeforeAutoselect);
    }

}

void Abstract_Grid::autoselectColor() {
    kDebug() << "Checking if color autoselection is possible";
    bool onlyOneColor = true;
    QColor noColor = QColor(0,0,0,0) ;
    QColor selectThisColor = noColor;

    if (selectedCell->y() >=0 && selectedCell->y() < m_rows) {
        QList<Data::HintEntry *> * rowHints = m_data->rowHints(selectedCell->y());
        for (int i = 0; onlyOneColor && i < rowHints->size(); i++) {

            QColor currentHintColor = rowHints->at(i)->hintColor;
            if (selectThisColor == noColor) {
                selectThisColor = currentHintColor;
            }

            onlyOneColor = currentHintColor==selectThisColor;

        }
    }

    if (onlyOneColor && selectedCell->x() >=0 && selectedCell->x() < m_cols) {
        QList<Data::HintEntry *> * colHints = m_data->columnHints(selectedCell->x());
        for (int i = 0; onlyOneColor && i < colHints->size(); i++) {

            QColor currentHintColor = colHints->at(i)->hintColor;
            if (selectThisColor == noColor) {
                selectThisColor = currentHintColor;
            }

            onlyOneColor = currentHintColor==selectThisColor;

        }
    }

    if (onlyOneColor) {
        kDebug() << "automatically selecting color [" <<selectThisColor << "]";
        emit wantColor(&selectThisColor);
    }
}


void Abstract_Grid::selectCell(QPoint* newlySelectedCell) {

    if ( newlySelectedCell->x() >= m_cols
        || newlySelectedCell->y() >= m_rows
        || newlySelectedCell->x() < -1
        || newlySelectedCell->y() < -1) {

        return;
    }

    selectedCell = newlySelectedCell;
    update();
}


void Abstract_Grid::paintHorizontal(QPainter * painter)
{
    QPen pen = painter->pen();


    /*draw horizontal lines and the boxes*/
    for (int y=0; y<m_rows; y++) {
        int pos_y = y*m_v_spacing;

        /*paint additional horizontal stuff at pos_y*/
        paintAdditionalHorizontal(painter, pos_y, y);

        if (y % 5 == 0) {
            pen.setWidth(2);
        }
        else {
            pen.setWidth(0);
        }

        if ( y == selectedCell->y() ||
                y == selectedCell->y()+1
        ) {
            pen.setColor(highlightColor);
        } else {
            pen.setColor(gridColor);
        }

        painter->setPen(pen);
        painter->drawLine(HORIZONTAL_LEFT_OFFSET, pos_y+VERTICAL_UPPER_OFFSET,
            widthOfLastLine(), pos_y+VERTICAL_UPPER_OFFSET);
    }

    pen.setWidth(2);
    if (m_rows == selectedCell->y()+1 ) {
        pen.setColor(highlightColor);
    } else {
        pen.setColor(gridColor);
    }
    painter->setPen(pen);
    painter->drawLine(HORIZONTAL_LEFT_OFFSET, heightOfLastLine(), widthOfLastLine(),
        heightOfLastLine());

}

void Abstract_Grid::paintVertical(QPainter * painter)
{
    QPen pen = painter->pen();

    for (int x=0; x<m_cols; x++) {
        int pos_x = x*m_h_spacing;

        /*paint additional vertical stuff at pos_x*/
        paintAdditionalVertical(painter, pos_x, x);

        if (x % 5 == 0) {
            pen.setWidth(2);
        }
        else {
            pen.setWidth(0);
        }

        if ( x == selectedCell->x() ||
                x == selectedCell->x()+1 ) {
            pen.setColor(highlightColor);
        } else {
            pen.setColor(gridColor);
        }

        painter->setPen(pen);
        painter->drawLine(pos_x+HORIZONTAL_LEFT_OFFSET, VERTICAL_UPPER_OFFSET,
            pos_x+HORIZONTAL_LEFT_OFFSET, heightOfLastLine());
    }

    pen.setWidth(2);
    if (m_cols == selectedCell->x()+1 ) {
        pen.setColor(highlightColor);
    } else {
        pen.setColor(gridColor);
    }
    painter->setPen(pen);
    painter->drawLine(widthOfLastLine(), VERTICAL_UPPER_OFFSET, widthOfLastLine(), heightOfLastLine());
}

Data * Abstract_Grid::data()
{
    return m_data;
}

void Abstract_Grid::setData(Data * data)
{
    if (m_data != NULL) {
        kDebug() << "Disconnecting myself from old " << m_data->description()->name();
        disconnect(m_data);
    }

    m_data = data;
    connect(m_data, SIGNAL(positionChanged()), SLOT(update()));

    calculatePositions();

    update();

    kDebug() << "Changed data: ["<<m_cols<<" cols / "<< m_rows<<" rows] ==> " << data->description()->name() ;
}

void Abstract_Grid::calculatePositions()
{
    if (m_data != NULL) {
        m_cols = m_data->getColumnCount();
        m_rows = m_data->getRowCount();

        int height = this->height() - VERTICAL_UPPER_OFFSET;
        int width = this->width() - HORIZONTAL_LEFT_OFFSET;

        m_maxHeight = height - (height % m_rows );
        m_maxWidth = width - (width % m_cols );

        m_v_spacing = m_maxHeight / m_rows;
        m_h_spacing = m_maxWidth / m_cols;

        m_maxHeight+=VERTICAL_UPPER_OFFSET;
        m_maxWidth+=HORIZONTAL_LEFT_OFFSET;
    } else {
        m_cols = 0;
        m_rows = 0;
        m_maxHeight = 1;
        m_maxWidth = 1;
        m_v_spacing = 1;
        m_h_spacing = 1;
    }
}

QPoint * Abstract_Grid::getGridPositions (QMouseEvent * event)
{

    int x_world = event->x() - HORIZONTAL_LEFT_OFFSET;
    int y_world = event->y() - VERTICAL_UPPER_OFFSET;

    int x_grid = x_world / m_h_spacing;
    int y_grid = y_world / m_v_spacing;

    QPoint * p = new QPoint(x_grid, y_grid);

    return p;
}


void Abstract_Grid::readConfig() {
    KConfigGroup colors = KGlobal::config()->group(Config::CONFIG_GROUP_COLORS);
    gridColor = colors.readEntry( Config::CONFIG_COLOR_GRID, Config::DEFAULT_COLOR_GRID );
    highlightColor = colors.readEntry( Config::CONFIG_COLOR_HIGHLIGHT_GRID, Config::DEFAULT_COLOR_HIGHLIGHT_GRID );

    KConfigGroup intelligence = KGlobal::config()->group(Config::CONFIG_GROUP_INTELLIGENCE);
    autoselectClickColor =  intelligence.readEntry( Config::CONFIG_AUTOSELECT_CLICK_COLOR,
        Config::DEFAULT_AUTOSELECT_CLICK_COLOR );

    waitBeforeAutoselect = intelligence.readEntry( Config::CONFIG_WAIT_BEFORE_AUTOSELECT_CLICK_COLOR,
        Config::DEFAULT_WAIT_BEFORE_AUTOSELECT_CLICK_COLOR );


    update();
}


