/***************************************************************************
 *   Copyright (C) 2008 Copyright 2008  Tom Vollerthun <vollerthun@gmx.de>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "grid.h"

#include <KDebug>

#include <QPainter>
#include <QGridLayout>

#include "grid_header.h"
#include "grid_center.h"
#include "colorselector.h"

#include "ds/data.h"
#include "ds/config.h"

#include <kaction.h>
#ifndef KXMLGUIWINDOW_H
    #include <kxmlguiwindow.h>
#endif
#include <klocalizedstring.h>
#include <KActionCollection>

#include <KComponentData>
#include <KConfig>
#include <KConfigGroup>
#include <KGlobal>
#include "../ds/data.h"
#include <ds/solver.h>
#include <threadweaver/ThreadWeaver.h>

Grid::Grid(KXmlGuiWindow *parent) :
    QWidget(parent)
{

    setAutoFillBackground(true);

    /*create parts of the gui*/
    colors = new ColorSelector(parent);
    picturePreview = new OriginalSizePicture(this);
    central = new GridCenter(this);
    column_header = new GridHeader(this, false);
    line_header = new GridHeader(this, true);


    /*connect repaint signals*/
    connectGridParts(central);
    connectGridParts(column_header);
    connectGridParts(line_header);

    connectGridPartsToMoreSignals();
    connect(parent, SIGNAL(settingsChanged()), SIGNAL(settingsChanged()));

    int h_max = 6;
    int v_max = 6;
    QGridLayout *gridLayout = new QGridLayout;

    //upper left: preview and colorselection
    gridLayout->addWidget(picturePreview, 0, 0, 1, 1);
    gridLayout->addWidget(colors, 1, 0, 1, 1);

    //remainder: horizontal and vertical headings
    // and the actual grid.
    gridLayout->addWidget(line_header, 2, 0, h_max, 1);
    gridLayout->addWidget(column_header, 0, 1, 2, v_max);
    gridLayout->addWidget(central, 2, 1, h_max, v_max);

    gridLayout->setHorizontalSpacing(0);
    gridLayout->setVerticalSpacing(0);

    setLayout(gridLayout);

    setupActions(parent);

}

void Grid::setupActions(KXmlGuiWindow *parent) {

    KAction* upAction = new KAction(this);
    upAction->setText(i18n("Up"));
    upAction->setIcon(KIcon("up"));
    upAction->setShortcut(Qt::Key_Up);
    parent->actionCollection()->addAction("up", upAction);
    connect(upAction, SIGNAL(triggered(bool)), central, SLOT(up()));

    KAction* downAction = new KAction(this);
    downAction->setText(i18n("Down"));
    downAction->setIcon(KIcon("down"));
    downAction->setShortcut(Qt::Key_Down);
    parent->actionCollection()->addAction("down", downAction);
    connect(downAction, SIGNAL(triggered(bool)), central, SLOT(down()));

    KAction* leftAction = new KAction(this);
    leftAction->setText(i18n("Left"));
    leftAction->setIcon(KIcon("left"));
    leftAction->setShortcut(Qt::Key_Left);
    parent->actionCollection()->addAction("left", leftAction);
    connect(leftAction, SIGNAL(triggered(bool)), central, SLOT(left()));

    KAction* rightAction = new KAction(this);
    rightAction->setText(i18n("Right"));
    rightAction->setIcon(KIcon("right"));
    rightAction->setShortcut(Qt::Key_Right);
    parent->actionCollection()->addAction("right", rightAction);
    connect(rightAction, SIGNAL(triggered(bool)), central, SLOT(right()));

    KAction* toggleAction = new KAction(this);
    toggleAction->setText(i18n("Toggle state"));
    toggleAction->setIcon(KIcon("toggle"));
    toggleAction->setShortcut(Qt::Key_Space);
    parent->actionCollection()->addAction("toggle", toggleAction);
    connect(toggleAction, SIGNAL(triggered(bool)), central, SLOT(toggleCurrent()));


    KAction* solveAction = new KAction(this);
    solveAction->setText(i18n("Solve"));
    solveAction->setIcon(KIcon("solve"));
    solveAction->setShortcut(Qt::Key_S);
    parent->actionCollection()->addAction("solve", solveAction);
    connect(solveAction, SIGNAL(triggered(bool)), this, SLOT(findSolution()));

}

void Grid::findSolution() {
    Solver * solver = new Solver();
    solver->setData(data());
    ThreadWeaver::Weaver::instance()->enqueue( solver );
}


void Grid::setData(Data * data)
{

    /*First try to disconnect from whichever data was connected before*/
    Data* oldData = central->data();
    if (oldData != NULL){
        kDebug() << "Disconnecting everything from old" << oldData->description()->name();
         oldData->disconnect();
    }
    kDebug() << "Disconnecting everything from " << data->description()->name();
    data->disconnect();

    kDebug() << "Setting Data to " << data->description()->name();

    emit dataSet(data);

    kDebug() << "Finished setting Data";
}

Data * Grid::data()
{
    return central->data();
}


void Grid::connectGridParts(Abstract_Grid * source)
{
    connect(source, SIGNAL(cellSelected(QPoint*)), central, SLOT(selectCell(QPoint*)));
    connect(source, SIGNAL(cellSelected(QPoint*)), line_header, SLOT(selectCell(QPoint*)));
    connect(source, SIGNAL(cellSelected(QPoint*)), column_header, SLOT(selectCell(QPoint*)));

    connect(source, SIGNAL(wantColor(QColor*)), colors, SLOT(selectColor(QColor*)));

    connect(this, SIGNAL(settingsChanged()), source, SIGNAL(settingsChanged()));
}

void Grid::connectGridPartsToMoreSignals() {
    connect(this, SIGNAL(dataSet(Data*)), picturePreview, SLOT(setData(Data*)));
    connect(this, SIGNAL(dataSet(Data*)), colors, SLOT(setData(Data*)));

    connect(this, SIGNAL(dataSet(Data*)), line_header, SLOT(setData(Data*)));
    connect(this, SIGNAL(dataSet(Data*)), column_header, SLOT(setData(Data*)));
    connect(this, SIGNAL(dataSet(Data*)), central, SLOT(setData(Data*)));

    connect(this, SIGNAL(settingsChanged()), SLOT(readConfig()));
    connect(this, SIGNAL(settingsChanged()), picturePreview, SLOT(readConfig()));
}



void Grid::readConfig() {
    KConfigGroup config = KGlobal::config()->group(Config::CONFIG_GROUP_COLORS);
    background = config.readEntry( Config::CONFIG_COLOR_BACKGROUND, Config::DEFAULT_COLOR_BACKGROUND );
    setPalette(QPalette(background));
    update();
}


