/***************************************************************************
 *   Copyright (C) 2008 Copyright 2008  Tom Vollerthun <vollerthun@gmx.de>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include <stdio.h>

#include <QtTest/QtTest>
#include <qtest_kde.h>

#include <KDebug>
#include <KLocale>
#include <KGlobal>

#include "ds/data.h"
#include "ds/description.h"

#include "io/io_factory.h"
#include "io/nonogram_io.h"
#include "io/nog_out.h"
#include "io/nop_format.h"

class CombinedTest : public QObject
{
    Q_OBJECT
    private slots:
        void testReadXbmImage();
        void testDataToString();
        void testDataToggles();
        void testStringToData();
        void loadMulticoloredImage();

        void testNogOut();
        void testNogIn();

        void testNopIn();

        void printLongString(const QString & str);
};

/**
 * Test the ability to properly read an image file into a Data object.
 * This test check only xbm since this is available.
 *
 * Test the ability to serialize data's row- and columnhints to String.
 *
 */
void CombinedTest::testReadXbmImage()
{
    kDebug() << "Starting test";
    NonogramIo *in = Io_Factory::createIoForType("xbm");
    Data * data = in->readFile("data/mouse_white.xbm");

    QCOMPARE(Data::toString(data->rowHints(6)), QString("|1,#000000|1,#000000|5,#000000|1,#000000|"));
    QCOMPARE(Data::toString(data->columnHints(data->getColumnCount()-1)), QString("|4,#000000|"));

    QCOMPARE(Data::toString(data->rowHints()), QString("-|-|-|-|2,#000000|-|2,#000000|-|2,#000000|2,#000000|-|1,#000000|1,#000000|5,#000000|1,#000000|-|1,#000000|1,#000000|1,#000000|2,#000000|2,#000000|-|4,#000000|2,#000000|1,#000000|-|2,#000000|2,#000000|1,#000000|1,#000000|-|1,#000000|1,#000000|1,#000000|1,#000000|1,#000000|-|1,#000000|1,#000000|1,#000000|1,#000000|-|2,#000000|3,#000000|3,#000000|1,#000000|-|14,#000000|-|-|-"));
    QCOMPARE(Data::toString(data->columnHints()), QString("-|2,#000000|-|4,#000000|-|5,#000000|1,#000000|-|1,#000000|3,#000000|2,#000000|-|3,#000000|2,#000000|-|1,#000000|3,#000000|-|1,#000000|1,#000000|-|1,#000000|2,#000000|-|1,#000000|4,#000000|-|1,#000000|1,#000000|2,#000000|-|1,#000000|1,#000000|1,#000000|-|1,#000000|2,#000000|1,#000000|-|2,#000000|2,#000000|1,#000000|-|2,#000000|3,#000000|1,#000000|-|3,#000000|2,#000000|-|4,#000000|-"));
}

/**
 * Test the ability to serialize data's content-matrix to String.
 */
void CombinedTest::testDataToString()
{
    kDebug() << "Starting test";
    NonogramIo *in = Io_Factory::createIoForType("xbm");
    Data * data = in->readFile("data/mouse_white.xbm");

    /*Hints to string are already tested above.
     * It's enough to test the grid-content toString.*/
    QString expected = QString("-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-");
    QString actual = Data::toString(data->matrix());

    printLongString(expected);
    printLongString(actual);

    QCOMPARE(actual, expected);

}

/**
 * Test if the Data's different toggles work properly
 */
void CombinedTest::testDataToggles()
{
    kDebug() << "Starting test";
    NonogramIo *in = Io_Factory::createIoForType("xbm");
    Data * data = in->readFile("data/mouse_white.xbm");

    /*--------- General Toggles (left mouse button) ---------*/

    /*First toggle: Undefined->On*/
    kDebug() << "First general toggle";
    data->togglePosition(0, 0);
    QCOMPARE( Data::toString(data->matrix()), QString("-|0,#000000|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-"));

    /*Second toggle: On->Off*/
    kDebug() << "Second general toggle";
    data->togglePosition(0, 0);
    QCOMPARE( Data::toString(data->matrix()), QString("-|1,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-"));

    /*Third toggle: Off->Undefined*/
    kDebug() << "Third general toggle";
    data->togglePosition(0, 0);
    QCOMPARE( Data::toString(data->matrix()), QString("-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-"));

    /*--------- ON/UNDEFINED Toggles (middle mouse button) ---------*/

    /*First toggle: UNDEFINED->ON*/
    kDebug() << "First ON toggle";
    data->toggle_UndefinedOn(0, 0);
    QCOMPARE( Data::toString(data->matrix()), QString("-|0,#000000|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-"));

    /*Second toggle: OFF->ON (next row)*/
    kDebug() << "Second ON toggle";
    //prepare
    data->togglePosition(1, 0);
    QCOMPARE( Data::toString(data->matrix()), QString("-|0,#000000|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|0,#000000|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-"));
    data->togglePosition(1, 0);
    QCOMPARE( Data::toString(data->matrix()), QString("-|0,#000000|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|1,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-"));

    //this is the toggle I'm interested in
    data->toggle_UndefinedOn(1, 0);
    QCOMPARE( Data::toString(data->matrix()), QString("-|0,#000000|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|0,#000000|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-"));

    /*Third toggle: ON->UNDEFINED (both positions)*/
    kDebug() << "Third ON toggle";
    data->toggle_UndefinedOn(0, 0);
    data->toggle_UndefinedOn(1, 0);
    QCOMPARE( Data::toString(data->matrix()), QString("-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-"));

    /*--------- OFF/UNDEFINED Toggles (right mouse button) ---------*/
    /*First toggle: UNDEFINED->OFF*/
    kDebug() << "First OFF toggle";
    data->toggle_UndefinedOff(0, 0); //was undefined, should now be off
    QCOMPARE( Data::toString(data->matrix()), QString("-|1,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-"));

    /*Second toggle: ON->OFF (next row)*/
    kDebug() << "Second OFF toggle";
    //prepare
    data->toggle_UndefinedOn(1, 0); //was undefined, should now be on
    QCOMPARE( Data::toString(data->matrix()), QString("-|1,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|0,#000000|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-"));
    //this is the toggle I'm interested in
    data->toggle_UndefinedOff(1, 0); //was on, should now be off
    QCOMPARE( Data::toString(data->matrix()), QString("-|1,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|1,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-"));

    /*Third toggle: ON->UNDEFINED (both positions)*/
    kDebug() << "Third OFF toggle";
    data->toggle_UndefinedOff(0, 0); //was off, should now be undefined
    data->toggle_UndefinedOff(1, 0); //same
    QCOMPARE( Data::toString(data->matrix()), QString("-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|2,#ffffff|-"));

}

/**
 * Test if the serialized hints can be deserialized back again
 */
void CombinedTest::testStringToData()
{
    kDebug() << "Starting test";
    NonogramIo *in = Io_Factory::createIoForType("xbm");
    Data * data = in->readFile("data/mouse_white.xbm");

    QCOMPARE(Data::toString(data->rowHints()), QString("-|-|-|-|2,#000000|-|2,#000000|-|2,#000000|2,#000000|-|1,#000000|1,#000000|5,#000000|1,#000000|-|1,#000000|1,#000000|1,#000000|2,#000000|2,#000000|-|4,#000000|2,#000000|1,#000000|-|2,#000000|2,#000000|1,#000000|1,#000000|-|1,#000000|1,#000000|1,#000000|1,#000000|1,#000000|-|1,#000000|1,#000000|1,#000000|1,#000000|-|2,#000000|3,#000000|3,#000000|1,#000000|-|14,#000000|-|-|-"));
    QCOMPARE(Data::toString(data->columnHints()), QString("-|2,#000000|-|4,#000000|-|5,#000000|1,#000000|-|1,#000000|3,#000000|2,#000000|-|3,#000000|2,#000000|-|1,#000000|3,#000000|-|1,#000000|1,#000000|-|1,#000000|2,#000000|-|1,#000000|4,#000000|-|1,#000000|1,#000000|2,#000000|-|1,#000000|1,#000000|1,#000000|-|1,#000000|2,#000000|1,#000000|-|2,#000000|2,#000000|1,#000000|-|2,#000000|3,#000000|1,#000000|-|3,#000000|2,#000000|-|4,#000000|-"));

    QCOMPARE(Data::toString(Data::fromString(Data::toString(data->rowHints()))), Data::toString(data->rowHints()));
    QCOMPARE(Data::toString(Data::fromString(Data::toString(data->columnHints()))), Data::toString(data->columnHints()));
}

/**
 * Test the ability to properly load a multicolored game in an image file
 */
void CombinedTest::loadMulticoloredImage()
{
    kDebug() << "Starting test";
    NonogramIo *in = Io_Factory::createIoForType("xpm");
    Data * data = in->readFile("data/tongue.xpm");

    QCOMPARE(Data::toString(data->rowHints()), QString("-|-|-|3,#ff0000|3,#ff0000|-|9,#ff0000|-|3,#ff0000|5,#000000|3,#ff0000|-|2,#ff0000|9,#000000|2,#ff0000|-|2,#ff0000|11,#000000|2,#ff0000|-|1,#ff0000|13,#000000|1,#ff0000|-|2,#ff0000|3,#000000|3,#ff0000|1,#000000|2,#ff0000|2,#000000|2,#ff0000|-|2,#ff0000|1,#000000|3,#ff0000|1,#000000|4,#ff0000|1,#000000|1,#ff0000|-|4,#ff0000|1,#000000|6,#ff0000|-|3,#ff0000|1,#000000|5,#ff0000|-|3,#ff0000|1,#000000|4,#ff0000|-|3,#ff0000|1,#000000|4,#ff0000|-|6,#ff0000|-|4,#ff0000|-"));
    QCOMPARE(Data::toString(data->columnHints()), QString("-|3,#ff0000|-|2,#ff0000|1,#000000|2,#ff0000|-|2,#ff0000|3,#000000|2,#ff0000|-|2,#ff0000|5,#000000|4,#ff0000|-|3,#ff0000|4,#000000|6,#ff0000|-|2,#ff0000|4,#000000|8,#ff0000|-|2,#ff0000|4,#000000|2,#ff0000|4,#000000|2,#ff0000|-|1,#ff0000|4,#000000|1,#ff0000|1,#000000|6,#ff0000|-|2,#ff0000|5,#000000|7,#ff0000|-|2,#ff0000|4,#000000|7,#ff0000|-|3,#ff0000|3,#000000|6,#ff0000|-|2,#ff0000|4,#000000|3,#ff0000|-|2,#ff0000|4,#000000|1,#ff0000|-|2,#ff0000|1,#000000|2,#ff0000|-|3,#ff0000|-|-"));
}

/**
 * Test the ability to serialize a complete Data object into the NOG fileformat
 */
void CombinedTest::testNogOut()
{
    kDebug() << "Starting test";
    NonogramIo *in = Io_Factory::createIoForType("xbm");
    Data * data = in->readFile("data/island.xbm");

    QString expected = "<nonogram_game>\n";
    expected.append("\t<name>Island</name>\n");
    expected.append("\t<hint></hint>\n");
    expected.append("\t<description></description>\n");
    expected.append("\t<rowHints>-|-|-|3,#000000|-|2,#000000|-|5,#000000|2,#000000|-|2,#000000|6,#000000|-|1,#000000|5,#000000|2,#000000|-|6,#000000|1,#000000|-|3,#000000|3,#000000|1,#000000|-|2,#000000|2,#000000|-|1,#000000|2,#000000|-|1,#000000|2,#000000|-|2,#000000|-|2,#000000|-|4,#000000|-|16,#000000|-</rowHints>\n");
    expected.append("\t<columnHints>-|1,#000000|-|1,#000000|-|1,#000000|-|1,#000000|-|1,#000000|-|2,#000000|1,#000000|-|2,#000000|3,#000000|1,#000000|-|1,#000000|6,#000000|1,#000000|-|1,#000000|3,#000000|1,#000000|-|4,#000000|1,#000000|-|6,#000000|1,#000000|-|2,#000000|5,#000000|1,#000000|-|1,#000000|2,#000000|4,#000000|2,#000000|-|1,#000000|3,#000000|6,#000000|-|4,#000000|4,#000000|-|2,#000000|-</columnHints>\n");
    expected.append("</nonogram_game>");

    QCOMPARE(NogOut::createOutput(data, false), expected);
}

/**
 * Test the ability to deserialize a NOG file back into a Data again
 */
void CombinedTest::testNogIn()
{
    NonogramIo *inXbm = Io_Factory::createIoForType("xbm");
    Data * dataXbm = inXbm->readFile("data/island.xbm");

    NonogramIo *inNog = Io_Factory::createIoForType("nog");
    Data * dataNog = inNog->readFile("data/island.nog");

    QCOMPARE(NogOut::createOutput(dataXbm, false), NogOut::createOutput(dataNog, false));

}

/**
 * Test if multiple games in one package (NOP file format) can be read back.
 */
void CombinedTest::testNopIn()
{
    NopFormat * nopIn = new NopFormat();
    Package * p = nopIn->readPackageFile("data/testpackage.nop");

    QCOMPARE(p->size(), 8);

    Data * d = p->at(5);

//    printf("%s\n", NogOut::createOutput(d, false).toAscii().constData());

    QString rowExpect = "-|16,#000000|-|16,#000000|-|2,#000000|7,#000000|3,#000000|-|2,#000000|1,#000000|5,#000000|1,#000000|3,#000000|-|2,#000000|2,#000000|3,#000000|2,#000000|3,#000000|-|2,#000000|3,#000000|3,#000000|3,#000000|-|16,#000000|-|4,#000000|3,#000000|5,#000000|-|2,#000000|9,#000000|3,#000000|-|3,#000000|2,#000000|2,#000000|4,#000000|-|1,#000000|2,#000000|2,#000000|2,#000000|-|3,#000000|3,#000000|3,#000000|4,#000000|-|2,#000000|3,#000000|1,#000000|3,#000000|3,#000000|-|4,#000000|3,#000000|5,#000000|-|5,#000000|6,#000000|-|16,#000000|-";
    QString colExpect = "-|16,#000000|-|10,#000000|5,#000000|-|2,#000000|2,#000000|1,#000000|1,#000000|3,#000000|-|2,#000000|6,#000000|4,#000000|-|3,#000000|3,#000000|2,#000000|2,#000000|2,#000000|-|4,#000000|2,#000000|5,#000000|1,#000000|-|5,#000000|3,#000000|2,#000000|1,#000000|1,#000000|-|5,#000000|3,#000000|2,#000000|1,#000000|-|5,#000000|3,#000000|2,#000000|1,#000000|1,#000000|-|4,#000000|2,#000000|5,#000000|1,#000000|-|3,#000000|3,#000000|2,#000000|2,#000000|2,#000000|-|2,#000000|6,#000000|4,#000000|-|2,#000000|2,#000000|1,#000000|1,#000000|3,#000000|-|10,#000000|5,#000000|-|16,#000000|-|16,#000000|-";

    QCOMPARE(Data::toString(d->rowHints()), rowExpect);
    QCOMPARE(Data::toString(d->columnHints()), colExpect);
    QCOMPARE(d->description()->name(), QString("Kitten"));
}

void CombinedTest::printLongString(const QString & str)
{
    printf("%s\n", str.toAscii().constData());
}

        QTEST_KDEMAIN(CombinedTest, GUI)
#include "combinedTest.moc"
