/*
    Copyright 2008 Gábor Lehel <illissius@gmail.com>

    This file is part of Superpong.

    Superpong is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Superpong is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Superpong (files COPYING and COPYING.LESSER).
    If not, see <http://www.gnu.org/licenses/>.
*/

#include <QDebug>
#include <QGraphicsEllipseItem>
#include <QGraphicsRectItem>
#include <QKeyEvent>
#include <QTime>
#include <QTimerEvent>

#include <KLocale>

#include "pongscene.h"


static const qreal fieldh  = 800;
static const qreal fieldw  = fieldh * 4 / 3;
static const qreal margin  = 50;
static const qreal totalh  = fieldh + margin * 2;
static const qreal totalw  = fieldw + margin * 2;
static const qreal paddleh = fieldh / 4;
static const qreal paddlew = 2;
static const qreal balld   = fieldh / 16;
static const qreal hzone   = fieldw / 5;

static const qreal speed   = fieldw / 5;
static const qreal vaccel  = fieldw * 4;
static const qreal haccel  = fieldw * 2;
static const qreal resist  = 0.1;

static const int   fps     = 60;

qreal randf()
{
    return qreal( qrand() ) / RAND_MAX;
}


class BallItem;

class PongItem
{
    public:
    virtual void update( qreal delta ) = 0;
    virtual bool collidesWith( BallItem * ) const = 0;
    virtual void collideWith( BallItem * ) = 0;
    bool colliding;
    PongItem(): colliding( false ) { }
    virtual ~PongItem() { }
};

class BallItem: public QGraphicsEllipseItem
{
    typedef QGraphicsEllipseItem super;

    public:
    BallItem( const QPointF &pos );
    virtual void update( qreal delta, QList<PongItem*> obstacles );
    void launch();

    qreal hmoving, vmoving;
};

class PaddleItem: public QGraphicsRectItem, public PongItem
{
    typedef QGraphicsRectItem super;

    public:
    PaddleItem( const QPointF &pos );
    virtual void update( qreal delta );
    virtual bool collidesWith( BallItem *b ) const;
    virtual void collideWith( BallItem *b );
    QRectF territory;
    int hinput, vinput;

private:
    qreal m_hmoving, m_vmoving;
};

class SpeedStripItem: public QGraphicsRectItem, public PongItem
{
    typedef QGraphicsRectItem super;

    public:
    SpeedStripItem( const QPointF &pos, qreal speed = 0 );
    virtual void update( qreal ) { }
    virtual bool collidesWith( BallItem *b ) const;
    virtual void collideWith( BallItem *b );
    void setSpeed( qreal speed );

    private:
    qreal m_speed;
};

class TeleportItem: public QGraphicsEllipseItem, public PongItem
{
    typedef QGraphicsEllipseItem super;

    public:
    TeleportItem( const QPointF &pos, TeleportItem *exit = 0 );
    virtual void update( qreal ) { }
    virtual bool collidesWith( BallItem *b ) const;
    virtual void collideWith( BallItem *b );
    void setExit( TeleportItem *other );

    private:
    TeleportItem *m_exit;
    bool m_incoming;
};

PaddleItem::PaddleItem( const QPointF &pos )
    : super( QRectF( pos, QSizeF( paddlew, paddleh ) ) ), hinput( 0 ), vinput( 0 ), m_hmoving( 0 ), m_vmoving( 0 )
{
    setBrush( QBrush( Qt::green ) );
}

BallItem::BallItem( const QPointF &pos )
    : super( QRectF( pos, QSizeF( balld, balld ) ) ), hmoving( 0 ), vmoving( 0 )
{
    setPen( QPen( Qt::yellow ) );
}

SpeedStripItem::SpeedStripItem( const QPointF &pos, qreal speed )
    : super( QRectF( pos, QSizeF( 75, 25 ) ) )
{
    setSpeed( speed );
}

void SpeedStripItem::setSpeed( qreal speed )
{
    m_speed = speed;
    if( speed > 1 ) setBrush( QBrush( Qt::green ) );
    if( speed < 1 ) setBrush( QBrush( Qt::red ) );
    if( speed == 1 ) setBrush( QBrush( Qt::magenta ) ); //wtf
}

TeleportItem::TeleportItem( const QPointF &pos, TeleportItem *exit )
    : super( QRectF( pos, QSizeF( 50, 50 ) ) ), m_incoming( false )
{
    setBrush( QBrush( Qt::blue ) );
    setExit( exit );
}

void TeleportItem::setExit( TeleportItem *other )
{
    m_exit = other;
}



class PongScene::Private
{
    public:
    QList<PongItem*> items;
    PaddleItem *left, *right;
    BallItem *ball;
    QGraphicsTextItem *score, *message;
    QTime time;
    int timer;
    bool started;
    int sr, sl;
    bool paused, autopaused;
};

PongScene *PongScene::s_instance = 0;

PongScene *PongScene::instance()
{
    if( !s_instance )
        s_instance = new PongScene;
    return s_instance;
}

PongScene::PongScene()
    : super( 0, 0, totalw, totalh ), d( new Private )
{
    newGame();
}

PongScene::~PongScene()
{
    delete d;
    s_instance = 0;
}

void PongScene::newGame()
{
    foreach( QGraphicsItem *i, items() )
        delete i;

    d->items.clear();

    d->sr = d->sl = 0;

    QGraphicsLineItem *lz = new QGraphicsLineItem( hzone + 1, 0, hzone + 1, fieldh );
    lz->setPen( QPen( Qt::gray ) );
    QGraphicsLineItem *rz = new QGraphicsLineItem( fieldw - hzone - 1, 0, fieldw - hzone - 1, fieldh );
    rz->setPen( QPen( Qt::gray ) );
    QGraphicsLineItem *uw = new QGraphicsLineItem( -1, -1, fieldw + 1, -1 );
    uw->setPen( QPen( Qt::white ) );
    QGraphicsLineItem *dw = new QGraphicsLineItem( -1, fieldh + 1, fieldw + 1, fieldh + 1 );
    dw->setPen( QPen( Qt::white ) );
    QGraphicsLineItem *lw = new QGraphicsLineItem( -1, -1, -1, fieldh + 1 );
    lw->setPen( QPen( Qt::red ) );
    QGraphicsLineItem *rw = new QGraphicsLineItem( fieldw + 1, -1, fieldw + 1, fieldh + 1 );
    rw->setPen( QPen( Qt::red ) );

    foreach( QGraphicsItem *i, QList<QGraphicsItem*>() << lz << rz << lw << rw << uw << dw )
        i->setZValue( -1 );

    d->score = new QGraphicsTextItem;
    d->score->setFont( QFont( "Monospace", 10 ) );
    d->score->setDefaultTextColor( Qt::white );

    d->message = new QGraphicsTextItem;
    d->message->setFont( QFont( "Sans", 36 ) );
    d->message->setDefaultTextColor( Qt::white );

    d->left = new PaddleItem( QPointF( 0, ( fieldh - paddleh ) / 2 ) );
    d->right = new PaddleItem( QPointF( fieldw - paddlew, ( fieldh - paddleh ) / 2 ) );
    d->ball = new BallItem( QPointF( ( fieldw - balld ) / 2, ( fieldh - balld ) / 2 ) );

    d->left->territory = QRectF( 0, 0, hzone, fieldh );
    d->right->territory = QRectF( fieldw - hzone, 0, hzone, fieldh );

    SpeedStripItem *fast = new SpeedStripItem( QPointF( hzone + randf() * ( fieldw - hzone * 2 - 100 ), randf() * ( fieldh - 50 ) ), 1.5 );
    SpeedStripItem *slow = new SpeedStripItem( QPointF( hzone + randf() * ( fieldw - hzone * 2 - 100 ), randf() * ( fieldh - 50 ) ), 0.5 );
    TeleportItem *port1 = new TeleportItem( QPointF( hzone + randf() * ( fieldw - hzone * 2 - 50 ), randf() * ( fieldh - 50 ) ) );
    TeleportItem *port2 = new TeleportItem( QPointF( hzone + randf() * ( fieldw - hzone * 2 - 50 ), randf() * ( fieldh - 50 ) ) );
    port1->setExit( port2 );
    port2->setExit( port1 );

    d->items << d->left << d->right << fast << slow << port1 << port2;


    foreach( QGraphicsItem *i, QList<QGraphicsItem*>() << d->ball << lz << rz << lw << rw << uw << dw << d->score << d->message << d->left << d->right << fast << slow << port1 << port2 )
    {
        i->translate( margin, margin );
        addItem( i );
    }

    showScore();
    d->started = false;
    d->paused = false;
    d->autopaused = false;
    d->ball->hide();
    showMessage( i18n( "Press any key to continue..." ) );
}

void PongScene::start()
{
    d->started = true;
    hideMessage();
    d->ball->show();
    d->ball->launch();
    d->timer = startTimer( 1000 / fps );
    d->time.start();
}

void PongScene::showScore()
{
    d->score->setPlainText( i18n( "%1 - %2", d->sl, d->sr ) );
    d->score->setPos( ( fieldw - d->score->boundingRect().width() ) / 2, fieldh + ( margin - d->score->boundingRect().height() ) / 2 );
}

void PongScene::showMessage( QString s )
{
    d->message->setPlainText( s );
    d->message->setPos( ( fieldw - d->message->boundingRect().width() ) / 2, ( fieldh - d->message->boundingRect().height() ) / 2 );
    d->message->show();
}

void PongScene::hideMessage()
{
    d->message->hide();
}

void PongScene::goal( int who )
{
    if( who > 0 ) d->sl++;
    else if( who < 0 ) d->sr++;

    showScore();
    d->ball->launch();
}

void PongScene::pause( bool whether )
{
    if( d->started )
    {
        if( whether )
        {
            killTimer( d->timer );
            showMessage( i18n( "Paused" ) );
        }
        else
        {
            hideMessage();
            d->timer = startTimer( 1000 / fps );
            d->time.start();
        }
        d->paused = whether;
    }
}

void PongScene::keyPressEvent( QKeyEvent *e )
{
    if( e->isAutoRepeat() )
    {
        e->ignore();
        return;
    }

    if( !d->started ) start();

    switch( e->key() )
    {
        case Qt::Key_Up:    d->right->vinput +=  1;  break;
        case Qt::Key_Down:  d->right->vinput += -1;  break;
        case Qt::Key_Right: d->right->hinput +=  1;  break;
        case Qt::Key_Left:  d->right->hinput += -1;  break;

        case Qt::Key_W:     d->left->vinput  +=  1;  break;
        case Qt::Key_S:     d->left->vinput  += -1;  break;
        case Qt::Key_D:     d->left->hinput  +=  1;  break;
        case Qt::Key_A:     d->left->hinput  += -1;  break;

        case Qt::Key_P:     pause( !d->paused ); break;

        default:            e->ignore(); return;
    }

    e->accept();
}

void PongScene::keyReleaseEvent( QKeyEvent *e )
{
    if( e->isAutoRepeat() )
    {
        e->ignore();
        return;
    }

    switch( e->key() )
    {
        case Qt::Key_Up:    d->right->vinput -=  1;  break;
        case Qt::Key_Down:  d->right->vinput -= -1;  break;
        case Qt::Key_Right: d->right->hinput -=  1;  break;
        case Qt::Key_Left:  d->right->hinput -= -1;  break;

        case Qt::Key_W:     d->left->vinput -=  1;  break;
        case Qt::Key_S:     d->left->vinput -= -1;  break;
        case Qt::Key_D:     d->left->hinput -=  1;  break;
        case Qt::Key_A:     d->left->hinput -= -1;  break;

        default:            e->ignore(); return;
    }

    e->accept();
}

void PongScene::focusInEvent( QFocusEvent *e )
{
    if( d->autopaused )
    {
        pause( false );
        d->autopaused = false;
    }
    e->accept();
}

void PongScene::focusOutEvent( QFocusEvent *e )
{
    if( !d->paused )
    {
        pause( true );
        d->autopaused = true;
    }
    e->accept();
}

void PongScene::timerEvent( QTimerEvent *e )
{
    if( e->timerId() != d->timer )
    {
        e->ignore();
        return;
    }
    e->accept();
    animate();
}

void PongScene::animate()
{
    const int e = d->time.elapsed();
    d->time.start();
    const qreal delta = qreal( e ) / 1000;

    foreach( PongItem *i, d->items )
        i->update( delta );

    d->ball->update( delta, d->items );
}

void PaddleItem::update( qreal delta )
{
    QRectF r = rect(), t = territory;

    m_vmoving += -delta * vaccel * vinput - resist * m_vmoving;
    r.moveTop( r.top() + delta * m_vmoving );
    if( r.bottom() >= t.bottom() || r.top() <= t.top() ) m_vmoving = 0;
    if( r.bottom() > t.bottom() ) r.moveBottom( t.bottom() );
    if( r.top() < t.top() ) r.moveTop( t.top() );

    m_hmoving += delta * haccel * hinput - resist * m_hmoving;
    r.moveLeft( r.left() + delta * m_hmoving );
    if( r.left() <= t.left() || r.right() >= t.right() ) m_hmoving = 0;
    if( r.left() < t.left() ) r.moveLeft( t.left() );
    if( r.right() > t.right()  ) r.moveRight( t.right() );

    setRect( r );
}

bool PaddleItem::collidesWith( BallItem *b ) const
{
    return b->rect().intersects( rect() );
}

void PaddleItem::collideWith( BallItem *b )
{
    QRectF r = b->rect();
    if( r.left() < fieldw / 2 )
        r.moveLeft( rect().right() );
    else
        r.moveRight( rect().left() );
    b->hmoving = -b->hmoving + m_hmoving;
    b->vmoving += m_vmoving / 2;
    b->setRect( r );
}

bool SpeedStripItem::collidesWith( BallItem *b ) const
{
    return b->rect().intersects( rect() );
}

void SpeedStripItem::collideWith( BallItem *b )
{
    b->hmoving *= m_speed;
    b->vmoving *= m_speed;
}

bool TeleportItem::collidesWith( BallItem *b ) const
{
    return b->rect().intersects( rect() );
}

void TeleportItem::collideWith( BallItem *b )
{
    if( m_exit && m_exit != this && !m_incoming )
    {
        b->setRect( QRectF( m_exit->rect().topLeft() + ( b->rect().topLeft() - rect().topLeft() ), b->rect().size() ) );
        m_exit->m_incoming = true;
    }
    m_incoming = false;
}

void BallItem::update( qreal delta, QList<PongItem*> obstacles )
{
    QRectF r = rect();

    r.moveTop( r.top() + delta * vmoving );
    if( r.top() < 0 || r.bottom() > fieldh ) vmoving = -vmoving;
    if( r.top() < 0 ) r.moveTop( 0 );
    if( r.bottom() > fieldh ) r.moveBottom( fieldh );

    bool out = false;
    r.moveRight( r.right() + delta * hmoving );
    if( r.right() > fieldw || r.left() < 0 ) out = true;
    if( r.right() > fieldw ) r.moveRight( fieldw );
    if( r.left() < 0 ) r.moveLeft( 0 );

    setRect( r );

    bool colliding = false;

    foreach( PongItem *i, obstacles )
    {
        if( i->collidesWith( this ) )
        {
            if( !i->colliding )
                i->collideWith( this );
            i->colliding = true;
        }
        else i->colliding = false;

        if( i->colliding ) colliding = true;
    }
 
    if( out && !colliding )
        PongScene::instance()->goal( int( r.left() - fieldw / 2 ) / abs( int( r.left() - fieldw / 2 ) ) ); //-1 or +1
}

void BallItem::launch()
{
    setRect( ( fieldw - balld ) / 2, ( fieldh - balld ) / 2, balld, balld );
    hmoving = speed * ( ( qrand() % 2 ) * 2 - 1 ); //randomly: -1, 1
    vmoving = ( qrand() % 100 ) * ( ( qrand() % 2 ) * 2 - 1 );
}
